#ifndef MATHTOOLKIT_GENERIC_AFFINEMAPPING_H
#define MATHTOOLKIT_GENERIC_AFFINEMAPPING_H

#include "ReportException.h"

#include "NumericalMatrix.h"
#include "NumericalVector.h"
#include "ArithmeticOperations.h"


namespace MathToolkit
{
	namespace Generic
	{
		template <typename T>
		class AffineMapping
		{
			using value_type = T;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			AffineMapping();
			explicit AffineMapping(const NumericalMatrix<value_type>&);
			explicit AffineMapping(NumericalMatrix<value_type>&&);
			AffineMapping(const NumericalMatrix<value_type>&, const NumericalVector<value_type>&);
			AffineMapping(NumericalMatrix<value_type>&&, NumericalVector<value_type>&&);

			AffineMapping(const AffineMapping&) = default;
			AffineMapping(AffineMapping&&) noexcept = default;
			AffineMapping& operator=(const AffineMapping&) = default;
			AffineMapping& operator=(AffineMapping&&) noexcept = default;

			virtual ~AffineMapping() = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Unary operators

			AffineMapping& operator*=(const NumericalMatrix<value_type>&);
			AffineMapping& operator*=(const AffineMapping&);

		// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Mathematical methods

			AffineMapping& transform(const NumericalMatrix<value_type>&);
			AffineMapping& transform(const AffineMapping&);
			AffineMapping& transform(const value_type scalingMultiplier, const NumericalMatrix<value_type>&);
			AffineMapping& transform(const value_type scalingMultiplier, const AffineMapping&);
			AffineMapping& multiply(const value_type scalingMultiplier, const NumericalMatrix<value_type>&);
			AffineMapping& multiply(const value_type scalingMultiplier, const AffineMapping&);

		// Mathematical methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			const NumericalMatrix<value_type>& matrixPart() const;
			const NumericalVector<value_type>& columnPart() const;

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			NumericalVector<value_type> apply(const NumericalVector<value_type>&) const;
			bool isAvailable() const;

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			NumericalMatrix<value_type> m_matrixPart;
			NumericalVector<value_type> m_columnPart;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>::AffineMapping()
	: m_matrixPart{}
	, m_columnPart{}
{
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>::AffineMapping(const NumericalMatrix<T>& matrixPart)
	: m_matrixPart{ matrixPart }
	, m_columnPart{}
{
	if (m_matrixPart.isEmpty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", "\"matrixPart\" is empty."));
	}

	else
		m_columnPart = NumericalVector<T>(m_matrixPart.columnSize());
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>::AffineMapping(NumericalMatrix<T>&& matrixPart)
	: m_matrixPart{ std::move(matrixPart) }
	, m_columnPart{}
{
	if (m_matrixPart.isEmpty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", "\"matrixPart\" is empty."));
	}

	else
		m_columnPart = NumericalVector<T>(m_matrixPart.columnSize());
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>::AffineMapping(const NumericalMatrix<T>& matrixPart, const NumericalVector<T>& columnPart)
	: m_matrixPart{ matrixPart }
	, m_columnPart{ columnPart }
{
	if (m_matrixPart.isEmpty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", "\"matrixPart\" is empty."));
	}

	else
	{
		if (m_matrixPart.columnSize() != m_columnPart.size())
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "constructor", "The column size of \"matrixPart\" is not equal to the size of \"columnPart\"."));
		}
	}
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>::AffineMapping(NumericalMatrix<T>&& matrixPart, NumericalVector<T>&& columnPart)
	: m_matrixPart{ std::move(matrixPart) }
	, m_columnPart{ std::move(columnPart) }
{
	if (m_matrixPart.isEmpty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", "\"matrixPart\" is empty."));
	}

	else
	{
		if (m_matrixPart.columnSize() != m_columnPart.size())
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "constructor", "The column size of \"matrixPart\" is not equal to the size of \"translation\"."));
		}
	}
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Unary operators

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>& MathToolkit::Generic::AffineMapping<T>::operator*=(const NumericalMatrix<value_type>& nm)
{
	m_matrixPart *= nm;
	return *this;
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>& MathToolkit::Generic::AffineMapping<T>::operator*=(const AffineMapping& am)
{
	m_matrixPart *= am.matrixPart;
	m_columnPart.add(m_matrixPart, am.columnPart());
	return *this;
}

// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Mathematical methods

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>& MathToolkit::Generic::AffineMapping<T>::transform(const NumericalMatrix<value_type>& nm)
{
	m_matrixPart.transform(nm);
	m_columnPart.transform(nm);
	return *this;
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>& MathToolkit::Generic::AffineMapping<T>::transform(const AffineMapping& am)
{
	m_matrixPart.transform(am.matrixPart);
	m_columnPart.transform(am.matrixPart);
	m_columnPart += am.columnPart();
	return *this;
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>& MathToolkit::Generic::AffineMapping<T>::transform(const value_type scalingMultiplier, const NumericalMatrix<value_type>& nm)
{
	m_matrixPart.transform(scalingMultiplier, nm);
	m_columnPart.transform(nm);
	m_columnPart *= scalingMultiplier;
	return *this;
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>& MathToolkit::Generic::AffineMapping<T>::transform(const value_type scalingMultiplier, const AffineMapping& am)
{
	m_matrixPart.transform(scalingMultiplier, am.matrixPart);
	m_columnPart.transform(am.matrixPart);
	m_columnPart += am.columnPart();
	m_columnPart *= scalingMultiplier;
	return *this;
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>& MathToolkit::Generic::AffineMapping<T>::multiply(const value_type scalingMultiplier, const NumericalMatrix<value_type>& nm)
{
	m_matrixPart.multiply(scalingMultiplier, nm.matrixPart);
	return *this;
}

template <typename T>
inline MathToolkit::Generic::AffineMapping<T>& MathToolkit::Generic::AffineMapping<T>::multiply(const value_type scalingMultiplier, const AffineMapping& am)
{
	m_matrixPart.multiply(scalingMultiplier, am.matrixPart);
	m_columnPart += scalingMultiplier * (m_matrixPart * am.columnPart());
	return *this;
}

// Mathematical methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

template <typename T>
inline const MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::AffineMapping<T>::matrixPart() const
{
	return m_matrixPart;
}

template <typename T>
inline const MathToolkit::Generic::NumericalVector<T>& MathToolkit::Generic::AffineMapping<T>::columnPart() const
{
	return m_columnPart;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

template <typename T>
inline MathToolkit::Generic::NumericalVector<T> MathToolkit::Generic::AffineMapping<T>::apply(const NumericalVector<T>& nv) const
{
	NumericalVector<T> newNumericalVector{ m_columnPart };
	newNumericalVector.add(m_matrixPart, nv);
	return newNumericalVector;
}

template <typename T>
inline bool MathToolkit::Generic::AffineMapping<T>::isAvailable() const
{
	return !(m_matrixPart.isEmpty());
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHTOOLKIT_GENERIC_AFFINEMAPPING_H
