#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_ANIONICCONSTRAINT_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_ANIONICCONSTRAINT_H

#include "ReportException.h"

#include "ConstraintOwner.h"
#include "MinimumDistanceConstraint.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			class AnionicConstraint :public Internal::MinimumDistanceConstraint
			{
				using AnionicConstraintOwner = ConstraintOwner<AnionicConstraint>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors and destructor

			private:
				AnionicConstraint(ModelAtom&, const Ligand&);
				AnionicConstraint(ModelAtom&, Ligand&&);
				AnionicConstraint(const Ligand&, const Ligand&);
				explicit AnionicConstraint(const std::shared_ptr<AnionicConstraint>& reversePtr);

			public:
				virtual ~AnionicConstraint() = default;

			// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static AnionicConstraintOwner create(const ConstraintKey&);
				static bool shouldCreate(const ModelAtom&, const Ligand&, const double cutoffRatio);
				static bool shouldCreate(const ConstraintKey&, const double cutoffRatio);

				static double defaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio(const double cutoffRatio);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				static bool canCreate(const ModelAtom&, const Ligand&);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				static double m_defaultCutoffRatio;


			private:
				AnionicConstraint() = delete;
				AnionicConstraint(const AnionicConstraint&) = delete;
				AnionicConstraint(AnionicConstraint&&) noexcept = delete;
				AnionicConstraint& operator=(const AnionicConstraint&) = delete;
				AnionicConstraint& operator=(AnionicConstraint&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint::AnionicConstraintOwner MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint::create(const ConstraintKey& constraintKey)
{
	ModelAtom& modelAtom = constraintKey.modelAtomRef();
	const Ligand& ligand = constraintKey.ligand();

	if (canCreate(modelAtom, ligand))
	{
		std::shared_ptr<AnionicConstraint> anionicConstraintPtr{ new AnionicConstraint(modelAtom, ligand) };
		std::shared_ptr<AnionicConstraint> reverseAnionicConstraintPtr{ new AnionicConstraint(anionicConstraintPtr) };
		anionicConstraintPtr->setReversePtr(std::dynamic_pointer_cast<InteratomicConstraintBase>(reverseAnionicConstraintPtr));

		modelAtom.track(anionicConstraintPtr);
		ligand.get().track(reverseAnionicConstraintPtr);

		return AnionicConstraintOwner{ std::move(anionicConstraintPtr), std::move(reverseAnionicConstraintPtr) };
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException("create of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint", "Could not create anionic constraint."));
	}
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint::shouldCreate(const ModelAtom& modelAtom, const Ligand& ligand, const double cutoffRatio)
{
	if (canCreate(modelAtom, ligand))
	{
		NumericalVector displacement = ligand.coordinate() - modelAtom.coordinate();
		double cutoffLength = cutoffRatio * (modelAtom.coordinationFeature().repulsiveIonicRadius() + ligand.get().coordinationFeature().repulsiveIonicRadius());

		if (displacement.normSquare() < (cutoffLength * cutoffLength))
			return true;
		else
			return false;
	}

	else
		return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint::shouldCreate(const ConstraintKey& constraintKey, const double cutoffRatio)
{
	return shouldCreate(constraintKey.modelAtomRef(), constraintKey.ligand(), cutoffRatio);
}

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint::defaultCutoffRatio() noexcept
{
	return m_defaultCutoffRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint::setDefaultCutoffRatio() noexcept
{
	m_defaultCutoffRatio = 2.0;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint::setDefaultCutoffRatio(const double val)
{
	if (0.0 < val)
		m_defaultCutoffRatio = val;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("setDefaultCutoffRatio of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint", "\"defaultCutoffRatio\" is not more than zero."));
	}
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint::canCreate(const ModelAtom& modelAtom, const Ligand& ligand)
{
	if (modelAtom.coordinationFeature().atomCategory().isAnion() && ligand.get().coordinationFeature().atomCategory().isAnion())
		return true;
	else
		return false;
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_ANIONICCONSTRAINT_H
