#ifndef CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_ATOMSITE_H
#define CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_ATOMSITE_H

#include <string>
#include <stdexcept>

#include "LinearAlgebra.h"
#include "ReportException.h"


namespace ChemToolkit
{
	namespace Crystallography
	{
		namespace SpaceGroups
		{
			class AtomSite final
			{
				using size_type = std::size_t;
				using NumericalVector = MathToolkit::Generic::NumericalVector<double>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				AtomSite();
				~AtomSite() = default;

				AtomSite(const AtomSite&) = default;
				AtomSite(AtomSite&&) noexcept = default;
				AtomSite& operator=(const AtomSite&) = default;
				AtomSite& operator=(AtomSite&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				const std::string& siteLabel() const;
				const std::string& elementSymbol() const;
				double occupancy() const;
				const NumericalVector& fractionalCoordinate() const;
				const std::string& siteSymmetrySymbol() const;
				const std::string& wyckoffSymbol() const;

				void setSiteLabel(const std::string&);
				void setElementSymbol(const std::string&);
				void setOccupancy(const double);
				void setFractionalCoordinate(const NumericalVector& fractionalCoordinate);
				void setSiteSymmetrySymbol(const std::string&);
				void setWyckoffSymbol(const std::string&);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public utility

				void clear();
				bool isValid() const;

			// Public utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::string m_siteLabel;
				std::string m_elementSymbol;
				double m_occupancy;
				NumericalVector m_fractionalCoordinate;

				std::string m_siteSymmetrySymbol;
				std::string m_wyckoffSymbol;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline const std::string& ChemToolkit::Crystallography::SpaceGroups::AtomSite::siteLabel() const
{
	return m_siteLabel;
}

inline const std::string& ChemToolkit::Crystallography::SpaceGroups::AtomSite::elementSymbol() const
{
	return m_elementSymbol;
}

inline double ChemToolkit::Crystallography::SpaceGroups::AtomSite::occupancy() const
{
	return m_occupancy;
}

inline const MathToolkit::Generic::NumericalVector<double>& ChemToolkit::Crystallography::SpaceGroups::AtomSite::fractionalCoordinate() const
{
	return m_fractionalCoordinate;
}

inline const std::string& ChemToolkit::Crystallography::SpaceGroups::AtomSite::siteSymmetrySymbol() const
{
	return m_siteSymmetrySymbol;
}

inline const std::string& ChemToolkit::Crystallography::SpaceGroups::AtomSite::wyckoffSymbol() const
{
	return m_wyckoffSymbol;
}

inline void ChemToolkit::Crystallography::SpaceGroups::AtomSite::setSiteLabel(const std::string& siteLabel)
{
	if (siteLabel.empty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setSiteLabel", "\"siteLabel\" is empty."));
	}

	else
		m_siteLabel = siteLabel;
}

inline void ChemToolkit::Crystallography::SpaceGroups::AtomSite::setElementSymbol(const std::string& symbol)
{
	if (symbol.empty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setElementSymbol", "\"elementSymbol\" is empty."));
	}

	else
		m_elementSymbol = symbol;
}

inline void ChemToolkit::Crystallography::SpaceGroups::AtomSite::setOccupancy(const double value)
{
	if (0.0 < value)
	{
		if (1.0 < value)
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "setOccupancy", "\"occupancy\" is invalid."));
		}

		else
			m_occupancy = value;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setOccupancy", "\"occupancy\" is invalid."));
	}
}

inline void ChemToolkit::Crystallography::SpaceGroups::AtomSite::setFractionalCoordinate(const NumericalVector& coordinate)
{
	if (coordinate.size() == 3)
		m_fractionalCoordinate = coordinate;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setFractionalCoordinate", "The size of \"fractionalCoordinate\" is not three."));
	}
}

inline void ChemToolkit::Crystallography::SpaceGroups::AtomSite::setSiteSymmetrySymbol(const std::string& symbol)
{
	if (symbol.empty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setSiteSymmetrySymbol", "\"siteSymmetrySymbol\" is empty."));
	}

	else
		m_siteSymmetrySymbol = symbol;
}

inline void ChemToolkit::Crystallography::SpaceGroups::AtomSite::setWyckoffSymbol(const std::string& symbol)
{
	if (symbol.empty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setWyckoffSymbol", "\"wyckoffSymbol\" is empty."));
	}

	else
		m_wyckoffSymbol = symbol;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public utility

inline void ChemToolkit::Crystallography::SpaceGroups::AtomSite::clear()
{
	m_siteLabel.clear();
	m_elementSymbol.clear();
	m_occupancy = 1.0;
	m_fractionalCoordinate.clear();
	m_siteSymmetrySymbol.clear();
	m_wyckoffSymbol.clear();
}

inline bool ChemToolkit::Crystallography::SpaceGroups::AtomSite::isValid() const
{
	if (m_siteLabel.empty())
		return false;
	if (m_elementSymbol.empty())
		return false;
	if (m_fractionalCoordinate.isEmpty())
		return false;

	return true;
}

// Public utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_ATOMSITE_H
