#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_ATOMICENVIRONMENT_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_ATOMICENVIRONMENT_H

#include <string>
#include <vector>
#include <utility>

#include "ModelAtom.h"
#include "PeriodicReferenceWrapper.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			class AtomicEnvironment
			{
				using size_type = std::size_t;
				using Ligand = ChemToolkit::Crystallography::PeriodicReferenceWrapper<ModelAtom, ChemToolkit::Crystallography::UnitCell>;

			public:
				struct Hasher
				{
					std::size_t operator()(const AtomicEnvironment& atomicEnvironment) const { return atomicEnvironment.getHashValue(); }
				};

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			private:
				AtomicEnvironment();

			public:
				virtual ~AtomicEnvironment() = default;

				AtomicEnvironment(const AtomicEnvironment&) = default;
				AtomicEnvironment(AtomicEnvironment&&) noexcept = default;
				AtomicEnvironment& operator=(const AtomicEnvironment&) = default;
				AtomicEnvironment& operator=(AtomicEnvironment&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				std::string elementSymbol() const noexcept;
				std::string atomCategorySymbol() const noexcept;
				std::string siteSymmetrySymbol() const noexcept;
				std::string wyckoffSymbol() const noexcept;
				size_type siteIndex() const noexcept;

				size_type getHashValue() const;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				std::string getReportText() const;

				std::string getFingerprint() const;
				std::string getPlainFingerprint() const;
				
			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				std::string writeConstraintsFingerprint() const;
				std::string getChemicalBondReportText(const std::vector<std::string>&) const;
				std::string getCationicConstraintReportText() const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private access methods

				void setElementSymbol(const std::string&);
				void setAtomCategorySymbol(const std::string&);
				void setSiteSymmetrySymbol(const std::string&);
				void setWyckoffSymbol(const std::string&);
				void setSiteIndex(const size_type);

				void setIonicBondFingerprint(const std::vector<Ligand>&);
				void setMetallicBondFingerprint(const std::vector<Ligand>&);
				void setCationicConstraintFingerprint(const std::vector<std::pair<Ligand, std::vector<Ligand>>>&);

			// Private access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::string m_elementSymbol;
				std::string m_atomCategorySymbol;
				std::string m_siteSymmetrySymbol;
				std::string m_wyckoffSymbol;
				size_type m_siteIndex;

				std::vector<std::string> m_ionicBondFingerprint;
				std::vector<std::string> m_metallicBondFingerprint;
				std::vector<std::pair<std::string, std::vector<std::string>>> m_cationicConstraintFingerprint;


				friend ModelAtom;
				friend bool operator<(const AtomicEnvironment& aea, const AtomicEnvironment& aeb);
				friend bool operator>(const AtomicEnvironment& aea, const AtomicEnvironment& aeb);
				friend bool operator==(const AtomicEnvironment& aea, const AtomicEnvironment& aeb);
				friend bool operator!=(const AtomicEnvironment& aea, const AtomicEnvironment& aeb);
			};


			inline bool operator<(const AtomicEnvironment& aea, const AtomicEnvironment& aeb)
			{
				if (aea.m_atomCategorySymbol < aeb.m_atomCategorySymbol)
					return true;
				else if (aeb.m_atomCategorySymbol < aea.m_atomCategorySymbol)
					return false;
				else
				{
					if (aea.m_ionicBondFingerprint < aeb.m_ionicBondFingerprint)
						return true;
					else if (aeb.m_ionicBondFingerprint < aea.m_ionicBondFingerprint)
						return false;
					else
					{
						if (aea.m_metallicBondFingerprint < aeb.m_metallicBondFingerprint)
							return true;
						else if (aeb.m_metallicBondFingerprint < aea.m_metallicBondFingerprint)
							return false;
						else
						{
							if (aea.m_cationicConstraintFingerprint < aeb.m_cationicConstraintFingerprint)
								return true;
							else if (aeb.m_cationicConstraintFingerprint < aea.m_cationicConstraintFingerprint)
								return false;
							else
							{
								if (aea.m_siteSymmetrySymbol < aeb.m_siteSymmetrySymbol)
									return true;
								else
									return false;
							}
						}
					}
				}
			}

			inline bool operator>(const AtomicEnvironment& aea, const AtomicEnvironment& aeb)
			{
				return (aeb < aea);
			}

			inline bool operator==(const AtomicEnvironment& aea, const AtomicEnvironment& aeb)
			{
				if (aea.m_atomCategorySymbol == aeb.m_atomCategorySymbol)
				{
					if (aea.m_siteSymmetrySymbol == aeb.m_siteSymmetrySymbol)
					{
						if (aea.m_ionicBondFingerprint == aeb.m_ionicBondFingerprint)
						{
							if (aea.m_metallicBondFingerprint == aeb.m_metallicBondFingerprint)
							{
								if (aea.m_cationicConstraintFingerprint == aeb.m_cationicConstraintFingerprint)
									return true;
								else
									return false;
							}
							else
								return false;
						}
						else
							return false;
					}
					else
						return false;
				}
				else
					return false;
			}

			inline bool operator!=(const AtomicEnvironment& aea, const AtomicEnvironment& aeb)
			{
				return !(aea == aeb);
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline std::string MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::elementSymbol() const noexcept
{
	return m_elementSymbol;
}

inline std::string MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::atomCategorySymbol() const noexcept
{
	return m_atomCategorySymbol;
}

inline std::string MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::siteSymmetrySymbol() const noexcept
{
	return m_siteSymmetrySymbol;
}

inline std::string MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::wyckoffSymbol() const noexcept
{
	return m_wyckoffSymbol;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::siteIndex() const noexcept
{
	return m_siteIndex;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::setElementSymbol(const std::string& symbol)
{
	m_elementSymbol = symbol;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::setAtomCategorySymbol(const std::string& symbol)
{
	m_atomCategorySymbol = symbol;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::setSiteSymmetrySymbol(const std::string& symbol)
{
	m_siteSymmetrySymbol = symbol;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::setWyckoffSymbol(const std::string& symbol)
{
	m_wyckoffSymbol = symbol;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment::setSiteIndex(const size_type index)
{
	m_siteIndex = index;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_ATOMICENVIRONMENT_H
