#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_CATIONICCONSTRAINT_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_CATIONICCONSTRAINT_H

#include <vector>

#include "ReportException.h"

#include "ConstraintOwner.h"
#include "MinimumDistanceConstraint.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			class CationicConstraint :public Internal::MinimumDistanceConstraint
			{
				using CationicConstraintOwner = ConstraintOwner<CationicConstraint>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors and destructor

			private:
				CationicConstraint(ModelAtom&, const Ligand&);
				CationicConstraint(ModelAtom&, Ligand&&);
				CationicConstraint(const Ligand&, const Ligand&);
				explicit CationicConstraint(const std::shared_ptr<CationicConstraint>& reversePtr);

			public:
				virtual ~CationicConstraint() = default;

			// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				const std::vector<Ligand>& commonBridgingAnions() const;

				void addCommonBridgingAnions(const Ligand&);
				void addCommonBridgingAnions(Ligand&&);
				void clearCommonBridgingAnions();

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static CationicConstraintOwner create(const ConstraintKey&);
				static bool shouldCreate(const ModelAtom&, const Ligand&);
				static bool shouldCreate(const ConstraintKey&);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::vector<Ligand> m_commonBridgingAnions;


			private:
				CationicConstraint() = delete;
				CationicConstraint(const CationicConstraint&) = delete;
				CationicConstraint(CationicConstraint&&) noexcept = delete;
				CationicConstraint& operator=(const CationicConstraint&) = delete;
				CationicConstraint& operator=(CationicConstraint&&) noexcept = delete;
			};		
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline const std::vector<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::Ligand>& MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::commonBridgingAnions() const
{
	return m_commonBridgingAnions;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::addCommonBridgingAnions(const Ligand& ligand)
{
	Ligand reverseLigand = ligand;
	reverseLigand.reverseLatticePoint();

	m_commonBridgingAnions.push_back(ligand);

	std::shared_ptr<CationicConstraint> reverseCationicConstraintPtr = std::dynamic_pointer_cast<CationicConstraint>(InteratomicConstraintBase::reversePtr());
	reverseCationicConstraintPtr->m_commonBridgingAnions.push_back(std::move(reverseLigand));
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::addCommonBridgingAnions(Ligand&& ligand)
{
	Ligand reverseLigand = ligand;
	reverseLigand.reverseLatticePoint();
	
	m_commonBridgingAnions.push_back(std::move(ligand));

	std::shared_ptr<CationicConstraint> reverseCationicConstraintPtr = std::dynamic_pointer_cast<CationicConstraint>(InteratomicConstraintBase::reversePtr());
	reverseCationicConstraintPtr->m_commonBridgingAnions.push_back(std::move(reverseLigand));
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::clearCommonBridgingAnions()
{
	m_commonBridgingAnions.clear();

	std::shared_ptr<CationicConstraint> reverseCationicConstraintPtr = std::dynamic_pointer_cast<CationicConstraint>(InteratomicConstraintBase::reversePtr());
	reverseCationicConstraintPtr->m_commonBridgingAnions.clear();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::CationicConstraintOwner MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::create(const ConstraintKey& constraintKey)
{
	ModelAtom& modelAtom = constraintKey.modelAtomRef();
	const Ligand& ligand = constraintKey.ligand();

	if (shouldCreate(modelAtom, ligand))
	{
		std::shared_ptr<CationicConstraint> cationicConstraintPtr{ new CationicConstraint(modelAtom, ligand) };
		std::shared_ptr<CationicConstraint> reverseCationicConstraintPtr{ new CationicConstraint(cationicConstraintPtr) };
		cationicConstraintPtr->setReversePtr(std::dynamic_pointer_cast<InteratomicConstraintBase>(reverseCationicConstraintPtr));

		modelAtom.track(cationicConstraintPtr);
		ligand.get().track(reverseCationicConstraintPtr);

		return CationicConstraintOwner{ std::move(cationicConstraintPtr), std::move(reverseCationicConstraintPtr) };
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException("create of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint", "Could not create cationic constraint."));
	}
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::shouldCreate(const ModelAtom& modelAtom, const Ligand& ligand)
{
	if (modelAtom.coordinationFeature().atomCategory().isCation() && ligand.get().coordinationFeature().atomCategory().isCation())
		return true;
	else
		return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint::shouldCreate(const ConstraintKey& constraintKey)
{
	return shouldCreate(constraintKey.modelAtomRef(), constraintKey.ligand());
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_CATIONICCONSTRAINT_H
