#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_INTERNAL_CHEMICALBONDRECORDER_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_INTERNAL_CHEMICALBONDRECORDER_H

#include <stdexcept>
#include <string>
#include <filesystem>

#include "ReportException.h"
#include "FileWriter.h"

#include "ModelCrystalStructure.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Recording
		{
			class StructureDesignRecorder;


			namespace Internal
			{
				class ChemicalBondRecorder
				{
					friend StructureDesignRecorder;

					using size_type = std::size_t;
					using FileWriter = OperatingSystem::IO::FileWriter;

					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				private:
					ChemicalBondRecorder();

				public:
					virtual ~ChemicalBondRecorder() = default;

					ChemicalBondRecorder(const ChemicalBondRecorder&) = default;
					ChemicalBondRecorder(ChemicalBondRecorder&&) noexcept = default;
					ChemicalBondRecorder& operator=(const ChemicalBondRecorder&) = default;
					ChemicalBondRecorder& operator=(ChemicalBondRecorder&&) noexcept = default;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

					void initialize();

					bool needRecord() const;
					size_type recordFrequency() const noexcept;

					void setRecordFrequency(const size_type recordFrequency = 0);

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void record(const ModelCrystalStructure&, const double feasibleConstraintErrorRate) const;

					void createDataFiles(const std::filesystem::path& workDirectoryPath);
					void createGplFiles(const std::filesystem::path& workDirectoryPath);
					void moveFiles(const std::filesystem::path& newDirectoryPath);
					void releaseFiles();

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					void recordFeasibleAtomRate(const double rate) const;
					void recordFeasibleChemicalBondRate(const double rate) const;
					void recordAverageChemicalBondError(const double error) const;

					void createFeasibleAtomRatesGplFile(const std::filesystem::path& workDirectoryPath);
					void createFeasibleChemicalBondRatesGplFile(const std::filesystem::path& workDirectoryPath);
					void createAverageChemicalBondErrorsGplFile(const std::filesystem::path& workDirectoryPath);

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private utility

					std::string writeCommonGplHeaderText(const std::string& fileStem) const;

				// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					size_type m_recordFrequency;

					FileWriter m_feasibleAtomRatesWriter;
					FileWriter m_feasibleChemicalBondRatesWriter;
					FileWriter m_averageChemicalBondErrorsWriter;

					FileWriter m_feasibleAtomRatesGplFileWriter;
					FileWriter m_feasibleChemicalBondRatesGplFileWriter;
					FileWriter m_averageChemicalBondErrorsGplFileWriter;


					mutable size_type m_recordTimer{ 0 };
					mutable size_type m_structureOptimizing{ 0 };

					static std::string m_feasibleAtomRatesFileStem;
					static std::string m_feasibleChemicalBondRatesFileStem;
					static std::string m_averageChemicalBondErrorsFileStem;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::initialize()
{
	m_recordFrequency = 0;
	
	m_feasibleAtomRatesWriter.releaseFile();
	m_feasibleChemicalBondRatesWriter.releaseFile();
	m_averageChemicalBondErrorsWriter.releaseFile();

	m_feasibleAtomRatesGplFileWriter.releaseFile();
	m_feasibleChemicalBondRatesGplFileWriter.releaseFile();
	m_averageChemicalBondErrorsGplFileWriter.releaseFile();

	m_recordTimer = 0;
	m_structureOptimizing = 0;
}

inline bool MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::needRecord() const
{
	return static_cast<bool>(m_recordFrequency);
}

inline std::size_t MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::recordFrequency() const noexcept
{
	return m_recordFrequency;
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::setRecordFrequency(const size_type recordFrequency)
{
	m_recordFrequency = recordFrequency;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::createDataFiles(const std::filesystem::path& workDirectoryPath)
{
	std::filesystem::path feasibleAtomRatesFilePath = workDirectoryPath;
	{
		std::string filename = m_feasibleAtomRatesFileStem;
		filename += ".txt";
		feasibleAtomRatesFilePath /= filename;
	}
	m_feasibleAtomRatesWriter = FileWriter{ feasibleAtomRatesFilePath };
	m_feasibleAtomRatesWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing);

	std::filesystem::path feasibleChemicalBondRatesFilePath = workDirectoryPath;
	{
		std::string filename = m_feasibleChemicalBondRatesFileStem;
		filename += ".txt";
		feasibleChemicalBondRatesFilePath /= filename;
	}
	m_feasibleChemicalBondRatesWriter = FileWriter{ feasibleChemicalBondRatesFilePath };
	m_feasibleChemicalBondRatesWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing);

	std::filesystem::path averageChemicalBondErrorsFilePath = workDirectoryPath;
	{
		std::string filename = m_averageChemicalBondErrorsFileStem;
		filename += ".txt";
		averageChemicalBondErrorsFilePath /= filename;
	}
	m_averageChemicalBondErrorsWriter = FileWriter{ averageChemicalBondErrorsFilePath };
	m_averageChemicalBondErrorsWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing);
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::createGplFiles(const std::filesystem::path& workDirectoryPath)
{
	createFeasibleAtomRatesGplFile(workDirectoryPath);
	createFeasibleChemicalBondRatesGplFile(workDirectoryPath);
	createAverageChemicalBondErrorsGplFile(workDirectoryPath);
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::moveFiles(const std::filesystem::path& newDirectoryPath)
{
	if (m_feasibleAtomRatesWriter.isValid())
	{
		m_feasibleAtomRatesWriter.moveDirectory(newDirectoryPath, OperatingSystem::IO::FileRenameOptions::overwrite_existing);
		m_feasibleAtomRatesGplFileWriter.moveDirectory(newDirectoryPath, OperatingSystem::IO::FileRenameOptions::overwrite_existing);
	}

	if (m_feasibleChemicalBondRatesWriter.isValid())
	{
		m_feasibleChemicalBondRatesWriter.moveDirectory(newDirectoryPath, OperatingSystem::IO::FileRenameOptions::overwrite_existing);
		m_feasibleChemicalBondRatesGplFileWriter.moveDirectory(newDirectoryPath, OperatingSystem::IO::FileRenameOptions::overwrite_existing);
	}	

	if (m_averageChemicalBondErrorsWriter.isValid())
	{
		m_averageChemicalBondErrorsWriter.moveDirectory(newDirectoryPath, OperatingSystem::IO::FileRenameOptions::overwrite_existing);
		m_averageChemicalBondErrorsGplFileWriter.moveDirectory(newDirectoryPath, OperatingSystem::IO::FileRenameOptions::overwrite_existing);
	}
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::releaseFiles()
{
	m_feasibleAtomRatesWriter.releaseFile();
	m_feasibleChemicalBondRatesWriter.releaseFile();
	m_averageChemicalBondErrorsWriter.releaseFile();
	m_recordTimer = 0;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::recordFeasibleAtomRate(const double rate) const
{
	std::string outputText;
	{
		outputText += std::to_string(m_structureOptimizing);
		outputText += "   ";
		outputText += std::to_string(rate);
		outputText += "\n";
	}

	m_feasibleAtomRatesWriter.write(outputText);
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::recordFeasibleChemicalBondRate(const double rate) const
{
	std::string outputText;
	{
		outputText += std::to_string(m_structureOptimizing);
		outputText += "   ";
		outputText += std::to_string(rate);
		outputText += "\n";
	}

	m_feasibleChemicalBondRatesWriter.write(outputText);
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::recordAverageChemicalBondError(const double error) const
{
	std::string outputText;
	{
		outputText += std::to_string(m_structureOptimizing);
		outputText += "   ";
		outputText += std::to_string(error);
		outputText += "\n";
	}

	m_averageChemicalBondErrorsWriter.write(outputText);
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

inline std::string MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder::writeCommonGplHeaderText(const std::string& fileStem) const
{
	std::string outputText;
	{
		outputText += "set terminal postscript enhanced eps color size 7,3\n";
		outputText += "set output \"";
		{
			outputText += fileStem;
			outputText += ".eps\"\n\n";
		}

		outputText += "set xrange [0:";
		{
			outputText += std::to_string(m_structureOptimizing + (m_structureOptimizing / 100));
			outputText += "]\n";
		}
		outputText += "set yrange [0.0:1.2]\n";

		outputText += "set bmargin 4\n";
		outputText += "set ylabel offset 0.5,0.5\n";
		outputText += "set tics font \"Times-Roman,18\"\n";
		outputText += "set key right top font \"Times-Roman,18\" spacing 1.1 maxcols 1 samplen 1\n\n";
	}

	return outputText;
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_INTERNAL_CHEMICALBONDRECORDER_H
