#ifndef CHEMTOOLKIT_GENERIC_CHEMICALCOMPOSITION_H
#define CHEMTOOLKIT_GENERIC_CHEMICALCOMPOSITION_H

#include <string>
#include <map>


namespace ChemToolkit
{
	namespace Generic
	{
		class ChemicalComposition
		{
			using size_type = std::size_t;
			using iterator = std::map<std::string, size_type>::iterator;
			using const_iterator = std::map<std::string, size_type>::const_iterator;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			ChemicalComposition() = default;
			virtual ~ChemicalComposition() = default;

			ChemicalComposition(const ChemicalComposition&) = default;
			ChemicalComposition(ChemicalComposition&&) noexcept = default;
			ChemicalComposition& operator=(const ChemicalComposition&) = default;
			ChemicalComposition& operator=(ChemicalComposition&&) noexcept = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			iterator begin();
			const_iterator begin() const;
			iterator end();
			const_iterator end() const;

			size_type countElementSpecies() const;
			bool contains(const std::string& elementSymbol) const;

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			std::string name() const;
			std::string getChemicalFormula() const;
			std::string getChemicalSystemName() const;

			ChemicalComposition getReducedChemicalComposition() const;

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Customize methods

			void add(const std::string& elementSymbol, const size_type numAtoms = 1);
			void erase(const std::string& elementSymbol);
			void join(const ChemicalComposition&);

		// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			bool isEmpty() const;
			void clear();
			size_type getHashValue() const;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::map<std::string, size_type> m_dictionary;


			friend bool operator<(const ChemicalComposition&, const ChemicalComposition&);
			friend bool operator>(const ChemicalComposition&, const ChemicalComposition&);
			friend bool operator==(const ChemicalComposition&, const ChemicalComposition&);
			friend bool operator!=(const ChemicalComposition&, const ChemicalComposition&);
		};


		inline bool operator<(const ChemicalComposition& cca, const ChemicalComposition& ccb)
		{
			return (cca.m_dictionary < ccb.m_dictionary);
		}

		inline bool operator>(const ChemicalComposition& cca, const ChemicalComposition& ccb)
		{
			return (cca.m_dictionary > ccb.m_dictionary);
		}

		inline bool operator==(const ChemicalComposition& cca, const ChemicalComposition& ccb)
		{
			return (cca.m_dictionary == ccb.m_dictionary);
		}

		inline bool operator!=(const ChemicalComposition& cca, const ChemicalComposition& ccb)
		{
			return (cca.m_dictionary != ccb.m_dictionary);
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline ChemToolkit::Generic::ChemicalComposition::iterator ChemToolkit::Generic::ChemicalComposition::begin()
{
	return m_dictionary.begin();
}

inline ChemToolkit::Generic::ChemicalComposition::const_iterator ChemToolkit::Generic::ChemicalComposition::begin() const
{
	return m_dictionary.begin();
}

inline ChemToolkit::Generic::ChemicalComposition::iterator ChemToolkit::Generic::ChemicalComposition::end()
{
	return m_dictionary.end();
}

inline ChemToolkit::Generic::ChemicalComposition::const_iterator ChemToolkit::Generic::ChemicalComposition::end() const
{
	return m_dictionary.end();
}

inline ChemToolkit::Generic::ChemicalComposition::size_type ChemToolkit::Generic::ChemicalComposition::countElementSpecies() const
{
	return m_dictionary.size();
}

inline bool ChemToolkit::Generic::ChemicalComposition::contains(const std::string& symbol) const
{
	return !(m_dictionary.find(symbol) == m_dictionary.end());
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

inline void ChemToolkit::Generic::ChemicalComposition::add(const std::string& symbol, const size_type numAtoms)
{
	iterator iter = m_dictionary.find(symbol);

	if (iter == m_dictionary.end())
		m_dictionary.emplace(symbol, numAtoms);
	else
		iter->second += numAtoms;
}

inline void ChemToolkit::Generic::ChemicalComposition::erase(const std::string& symbol)
{
	m_dictionary.erase(symbol);
}

inline void ChemToolkit::Generic::ChemicalComposition::join(const ChemicalComposition& chemicalComposition)
{
	for (const auto& symbolAndNum : chemicalComposition)
		add(symbolAndNum.first, symbolAndNum.second);
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline bool ChemToolkit::Generic::ChemicalComposition::isEmpty() const
{
	return m_dictionary.empty();
}

inline void ChemToolkit::Generic::ChemicalComposition::clear()
{
	m_dictionary.clear();
}

inline std::size_t ChemToolkit::Generic::ChemicalComposition::getHashValue() const
{
	size_type hashValue = 0;
	{
		for (const auto& symbolAndNum : m_dictionary)
		{
			hashValue ^= std::hash<std::string>{}(symbolAndNum.first);
			hashValue ^= std::hash<size_type>{}(symbolAndNum.second);
		}
	}

	return hashValue;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CHEMTOOLKIT_GENERIC_CHEMICALCOMPOSITION_H
