#ifndef CHEMTOOLKIT_CRYSTALLOGRAPHY_FILEIO_CIFFILEWRITER_H
#define CHEMTOOLKIT_CRYSTALLOGRAPHY_FILEIO_CIFFILEWRITER_H

#include "FileWriter.h"
#include "LinearAlgebra.h"

#include "UnitCell.h"
#include "CrystalStructure.h"


namespace ChemToolkit
{
	namespace Crystallography
	{
		namespace FileIO
		{
			class CifFileWriter :public OperatingSystem::IO::FileWriter
			{
				using NumericalVector = MathToolkit::Generic::NumericalVector<double>;
				using NumericalMatrix = MathToolkit::Generic::NumericalMatrix<double>;
				using FileCreateOptions = OperatingSystem::IO::FileCreateOptions;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				CifFileWriter() = default;
				explicit CifFileWriter(const std::filesystem::path& filePath);
				explicit CifFileWriter(std::filesystem::path&& filePath);

				CifFileWriter(const CifFileWriter&) = default;
				CifFileWriter(CifFileWriter&&) noexcept = default;
				CifFileWriter& operator=(const CifFileWriter&) = default;
				CifFileWriter& operator=(CifFileWriter&&) noexcept = default;

				virtual ~CifFileWriter() = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static std::string correctExtension();
				static bool isCorrectExtension(const std::filesystem::path filePath);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				template <typename A>
				void createFile(const ChemToolkit::Crystallography::CrystalStructure<A>&, FileCreateOptions fileCreateOptions = FileCreateOptions::overwrite_existing) const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				std::string writeHeaderMessage(const std::string& materialName) const;
				std::string writeCrystallographicInfo(const ChemToolkit::Generic::ChemicalComposition&, const double primitiveCellVolume) const;

				std::string writeUnitCellParameters(const ChemToolkit::Crystallography::UnitCell&) const;

				template <typename A>
				std::string writeAtomSites(const ChemToolkit::Crystallography::CrystalStructure<A>&) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline std::string ChemToolkit::Crystallography::FileIO::CifFileWriter::correctExtension()
{
	return std::string{ ".cif" };
}

inline bool ChemToolkit::Crystallography::FileIO::CifFileWriter::isCorrectExtension(const std::filesystem::path filePath)
{
	if (filePath.extension().string() == correctExtension())
		return true;
	else
		return false;
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

template <typename A>
inline void ChemToolkit::Crystallography::FileIO::CifFileWriter::createFile(const ChemToolkit::Crystallography::CrystalStructure<A>& crystalStructure, FileCreateOptions fileCreateOptions) const
{
	ChemToolkit::Generic::ChemicalComposition chemicalFormulaSum = crystalStructure.getChemicalComposition();

	std::string outputString;
	outputString += writeHeaderMessage(chemicalFormulaSum.name());
	outputString += writeUnitCellParameters(crystalStructure.unitCell());
	outputString += writeCrystallographicInfo(chemicalFormulaSum, crystalStructure.unitCell().getCellVolume());
	outputString += writeAtomSites(crystalStructure);

	
	OperatingSystem::IO::FileWriter::createFile(fileCreateOptions, outputString);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename A>
inline std::string ChemToolkit::Crystallography::FileIO::CifFileWriter::writeAtomSites(const ChemToolkit::Crystallography::CrystalStructure<A>& crystalStructure) const
{
	std::string outputText;
	{
		outputText += " _atom_site_type_symbol\n";
		outputText += " _atom_site_label\n";
		outputText += " _atom_site_symmetry_multiplicity\n";
		outputText += " _atom_site_fract_x\n";
		outputText += " _atom_site_fract_y\n";
		outputText += " _atom_site_fract_z\n";
		outputText += " _atom_site_occupancy\n";
	}

	std::stringstream outputStream;
	{
		outputStream.precision(12);
		outputStream.flags(std::ios_base::fixed);

		std::map<std::string, std::size_t> composition;

		for (const auto& atom : crystalStructure.atomicArrangement())
		{
			std::map<std::string, std::size_t>::iterator iter = composition.find(atom.elementSymbol());
			{
				if (iter == composition.end())
					composition.emplace(atom.elementSymbol(), 1);
				else
					iter->second += 1;
			}

			NumericalVector fractionalCoordinate = crystalStructure.unitCell().inverseBasisVectors() * atom.coordinate();
			outputStream << "  " << atom.elementSymbol();
			outputStream << "  " << atom.elementSymbol() << composition[atom.elementSymbol()] << "  1  ";
			outputStream << fractionalCoordinate[0] << "  " << fractionalCoordinate[1] << "  " << fractionalCoordinate[2];
			outputStream << "  1\n";
		}

		outputText += outputStream.str();
	}

	return outputText;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CHEMTOOLKIT_CRYSTALLOGRAPHY_FILEIO_CIFFILEWRITER_H
