#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_FEATURES_CONSTRAINTCOMPOSITION_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_FEATURES_CONSTRAINTCOMPOSITION_H

#include <stdexcept>
#include <string>
#include <map>


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Features
		{
			class ConstraintComposition
			{
				using size_type = std::size_t;
				using iterator = std::map<std::string, size_type>::iterator;
				using const_iterator = std::map<std::string, size_type>::const_iterator;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				ConstraintComposition();
				explicit ConstraintComposition(const std::string& name);
				virtual ~ConstraintComposition() = default;

				ConstraintComposition(const ConstraintComposition&) = default;
				ConstraintComposition(ConstraintComposition&&) noexcept = default;
				ConstraintComposition& operator=(const ConstraintComposition&) = default;
				ConstraintComposition& operator=(ConstraintComposition&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				iterator begin();
				const_iterator begin() const;
				iterator end();
				const_iterator end() const;

				bool isUnconstrained() const;
				bool isInfeasible() const;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				bool contains(const std::string& atomCategorySymbol) const;

				std::string name() const;
				std::string getSystemName() const;
				size_type getHashValue() const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Customize methods

				void initialize();
				void initialize(const std::string& name);

				void enableConstraints();
				void disableConstraints();

				void add(const std::string& atomCategorySymbol, const size_type numAtoms = 1);
				void erase(const std::string& atomCategorySymbol);
				void join(const ConstraintComposition&);

			// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				size_type getFeasibleCoordinationNumber() const;

				static std::map<std::string, size_type> getSymbolComposition(const std::string&);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				bool m_isUnconstrained;
				std::map<std::string, size_type> m_symbolComposition;


				friend bool operator<(const ConstraintComposition& cca, const ConstraintComposition& ccb);
				friend bool operator>(const ConstraintComposition& cca, const ConstraintComposition& ccb);
			};


			inline bool operator<(const ConstraintComposition& cca, const ConstraintComposition& ccb)
			{
				if (cca.m_isUnconstrained)
					return false;
				else
				{
					if (ccb.m_isUnconstrained)
						return true;
					else
					{
						if (cca.getFeasibleCoordinationNumber() < ccb.getFeasibleCoordinationNumber())
							return true;
						else if (ccb.getFeasibleCoordinationNumber() < cca.getFeasibleCoordinationNumber())
							return true;
						else
						{
							if (cca.m_symbolComposition.size() < ccb.m_symbolComposition.size())
								return true;
							else
								return false;
						}
					}
				}
			}

			inline bool operator>(const ConstraintComposition& cca, const ConstraintComposition& ccb)
			{
				return (ccb < cca);
			}

			inline bool operator==(const ConstraintComposition& cca, const ConstraintComposition& ccb)
			{
				return (cca.name() == ccb.name());
			}

			inline bool operator!=(const ConstraintComposition& cca, const ConstraintComposition& ccb)
			{
				return (cca.name() != ccb.name());
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::iterator MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::begin()
{
	return m_symbolComposition.begin();
}

inline MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::const_iterator MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::begin() const
{
	return m_symbolComposition.begin();
}

inline MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::iterator MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::end()
{
	return m_symbolComposition.end();
}

inline MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::const_iterator MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::end() const
{
	return m_symbolComposition.end();
}

inline bool MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::isUnconstrained() const
{
	return m_isUnconstrained;
}

inline bool MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::isInfeasible() const
{
	if (m_symbolComposition.empty())
		return !(m_isUnconstrained);
	else
		return false;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::contains(const std::string& atomCategorySymbol) const
{
	if (m_symbolComposition.find(atomCategorySymbol) == m_symbolComposition.end())
		return false;
	else
		return true;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::getHashValue() const
{
	size_type hashValue = std::hash<bool>{}(m_isUnconstrained);
	{
		for (const auto& symbolAndNum : m_symbolComposition)
		{
			hashValue ^= std::hash<std::string>{}(symbolAndNum.first);
			hashValue ^= std::hash<size_type>{}(symbolAndNum.second);
		}
	}

	return hashValue;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

inline void MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::initialize()
{
	m_isUnconstrained = false;
	m_symbolComposition.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::initialize(const std::string& name)
{
	m_symbolComposition = getSymbolComposition(name);

	if (name == "*")
		m_isUnconstrained = true;
	else
		m_isUnconstrained = false;
}

inline void MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::enableConstraints()
{
	m_isUnconstrained = false;
	m_symbolComposition.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::disableConstraints()
{
	m_isUnconstrained = true;
	m_symbolComposition.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::add(const std::string& symbol, const size_type numAtoms)
{
	m_isUnconstrained = false;
	iterator iter = m_symbolComposition.find(symbol);

	if (iter == m_symbolComposition.end())
		m_symbolComposition.emplace(symbol, numAtoms);
	else
		iter->second += numAtoms;
}

inline void MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::erase(const std::string& symbol)
{
	m_symbolComposition.erase(symbol);
}

inline void MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::join(const ConstraintComposition& chemicalComposition)
{
	for (const auto& symbolAndNum : chemicalComposition)
		add(symbolAndNum.first, symbolAndNum.second);
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::size_t MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition::getFeasibleCoordinationNumber() const
{
	size_type feasibleCoordinationNumber = 0;
	{
		for (const auto& symbolAndNum : m_symbolComposition)
			feasibleCoordinationNumber += symbolAndNum.second;
	}

	return feasibleCoordinationNumber;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_FEATURES_CONSTRAINTCOMPOSITION_H
