#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_CONSTRAINTKEY_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_CONSTRAINTKEY_H

#include "PeriodicReferenceWrapper.h"
#include "UnitCell.h"

#include "ModelAtom.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			class ConstraintKey
			{
				using size_type = std::size_t;
				using ModelAtom = MathematicalCrystalChemistry::StructureModel::Components::ModelAtom;
				using Ligand = ChemToolkit::Crystallography::PeriodicReferenceWrapper<ModelAtom, ChemToolkit::Crystallography::UnitCell>;

			public:
				struct Hasher
				{
					std::size_t operator()(const ConstraintKey& constraintKey) const { return constraintKey.hashValue(); }
				};

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			private:
				ConstraintKey(ModelAtom&, const Ligand&);
				ConstraintKey(ModelAtom&, Ligand&&);

			public:
				virtual ~ConstraintKey() = default;

				ConstraintKey(const ConstraintKey&);
				ConstraintKey(ConstraintKey&&) noexcept;
				ConstraintKey& operator=(const ConstraintKey&);
				ConstraintKey& operator=(ConstraintKey&&) noexcept;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static ConstraintKey createConstraintKey(ModelAtom&, const Ligand&);
				static ConstraintKey createConstraintKey(const Ligand&, const Ligand&);

				static Ligand selectModelAtomKey(const Ligand&, const Ligand&);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				ModelAtom& modelAtomRef() const;
				const Ligand& ligand() const;
				size_type hashValue() const;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Utility

			private:
				static std::string toLatticePointString(const Ligand::LatticePoint&) noexcept;

			// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::reference_wrapper<ModelAtom> m_modelAtomRef;
				Ligand m_ligand;
				size_type m_hashValue;
			};


			inline bool operator==(const ConstraintKey& cka, const ConstraintKey& ckb)
			{
				if (&(cka.modelAtomRef()) != &(ckb.modelAtomRef()))
					return false;
				else
					return (cka.ligand() == ckb.ligand());
			}

			inline bool operator!=(const ConstraintKey& cka, const ConstraintKey& ckb)
			{
				return !(cka == ckb);
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::createConstraintKey(ModelAtom& atom, const Ligand& inputLigand)
{
	if (&atom < &(inputLigand.get()))
		return ConstraintKey{ atom, inputLigand };

	else if (&(inputLigand.get()) < &atom)
	{
		Ligand tempLigand{ atom, inputLigand };
		tempLigand.reverseLatticePoint();
		return ConstraintKey{ inputLigand.get(), std::move(tempLigand) };
	}

	else
	{
		Ligand::LatticePoint modelAtomLatticePoint{ 0,0,0 };
		Ligand::LatticePoint ligandLatticePoint = inputLigand.getLatticePoint();


		if (modelAtomLatticePoint < ligandLatticePoint)
			return ConstraintKey{ atom, inputLigand };

		else if (ligandLatticePoint < modelAtomLatticePoint)
		{
			Ligand tempLigand{ atom, inputLigand };
			tempLigand.reverseLatticePoint();
			return ConstraintKey{ inputLigand.get(), std::move(tempLigand) };
		}

		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::logic_error(reportException("MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::createConstraintKey", "Could not create constraint key since \"atom\" and \"ligand\" point the same atom."));
		}
	}
}

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::createConstraintKey(const Ligand& ligandA, const Ligand& ligandB)
{
	if (&(ligandA.get()) < &(ligandB.get()))
		return ConstraintKey{ ligandA.get(), Ligand(ligandB.get(), ligandB.latticeSystemRef(), ligandA.getRelativeLatticePointTo(ligandB)) };

	else if (&(ligandB.get()) < &(ligandA.get()))
		return ConstraintKey{ ligandB.get(), Ligand(ligandA.get(), ligandA.latticeSystemRef(), ligandB.getRelativeLatticePointTo(ligandA)) };

	else
	{
		Ligand::LatticePoint ligandLatticePointA = ligandA.getLatticePoint();
		Ligand::LatticePoint ligandLatticePointB = ligandB.getLatticePoint();


		if (ligandLatticePointA < ligandLatticePointB)
			return ConstraintKey{ ligandA.get(), Ligand(ligandB.get(), ligandB.latticeSystemRef(), ligandA.getRelativeLatticePointTo(ligandB)) };

		else if (ligandLatticePointB < ligandLatticePointA)
			return ConstraintKey{ ligandB.get(), Ligand(ligandA.get(), ligandA.latticeSystemRef(), ligandB.getRelativeLatticePointTo(ligandA)) };

		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::logic_error(reportException("MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::createConstraintKey", "Could not create constraint key since \"ligandA\" and \"ligandB\" point the same atom."));
		}
	}
}

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::Ligand MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::selectModelAtomKey(const Ligand& ligandA, const Ligand& ligandB)
{
	if (&(ligandA.get()) < &(ligandB.get()))
		return ligandA;

	else if (&(ligandB.get()) < &(ligandA.get()))
		return ligandB;

	else
	{
		Ligand::LatticePoint ligandLatticePointA = ligandA.getLatticePoint();
		Ligand::LatticePoint ligandLatticePointB = ligandB.getLatticePoint();


		if (ligandLatticePointA < ligandLatticePointB)
			return ligandA;

		else if (ligandLatticePointB < ligandLatticePointA)
			return ligandB;

		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::logic_error(reportException("MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::createConstraintKey", "Could not create constraint key since \"ligandA\" and \"ligandB\" point the same atom."));
		}
	}
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline MathematicalCrystalChemistry::StructureModel::Components::ModelAtom& MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::modelAtomRef() const
{
	return m_modelAtomRef.get();
}

inline const ChemToolkit::Crystallography::PeriodicReferenceWrapper<MathematicalCrystalChemistry::StructureModel::Components::ModelAtom, ChemToolkit::Crystallography::UnitCell>& MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::ligand() const
{
	return m_ligand;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::hashValue() const
{
	return m_hashValue;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::string MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::toLatticePointString(const Ligand::LatticePoint& latticePoint) noexcept
{
	std::string latticePointText;
	{
		for (const auto& num : latticePoint)
		{
			latticePointText += std::to_string(num);
			latticePointText += ",";
		}

		latticePointText.pop_back();
	}

	return latticePointText;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_CONSTRAINTKEY_H
