#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_CONSTRAINTMANAGER_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_CONSTRAINTMANAGER_H

#include <stdexcept>
#include <vector>
#include <unordered_map>
#include <memory>
#include <random>

#include "ConstraintKey.h"
#include "ConstraintOwner.h"

#include "IonicBond.h"
#include "MetallicBond.h"
#include "AnionicConstraint.h"
#include "CationicConstraint.h"
#include "NonIonicBond.h"
#include "NonMetallicBond.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			namespace Internal
			{
				class ConstraintManager
				{
					using size_type = std::size_t;
					using ModelAtom = MathematicalCrystalChemistry::StructureModel::Components::ModelAtom;
					using ConstraintKey = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey;

					using IonicBondOwner = GeometricalConstraints::ConstraintOwner<GeometricalConstraints::IonicBond>;
					using MetallicBondOwner = GeometricalConstraints::ConstraintOwner<GeometricalConstraints::MetallicBond>;
					using AnionicConstraintOwner = GeometricalConstraints::ConstraintOwner<GeometricalConstraints::AnionicConstraint>;
					using CationicConstraintOwner = GeometricalConstraints::ConstraintOwner<GeometricalConstraints::CationicConstraint>;
					using NonIonicBondOwner = GeometricalConstraints::ConstraintOwner<GeometricalConstraints::NonIonicBond>;
					using NonMetallicBondOwner = GeometricalConstraints::ConstraintOwner<GeometricalConstraints::NonMetallicBond>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operator

				protected:
					ConstraintManager();

				public:
					virtual ~ConstraintManager() = default;

					ConstraintManager(const ConstraintManager&);
					ConstraintManager(ConstraintManager&&) noexcept;
					ConstraintManager& operator=(const ConstraintManager&);
					ConstraintManager& operator=(ConstraintManager&&) noexcept;

				// Constructors, destructor, and operator
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					const std::unordered_map<ConstraintKey, IonicBondOwner, ConstraintKey::Hasher>& ionicBondOwners() const noexcept;
					const std::unordered_map<ConstraintKey, MetallicBondOwner, ConstraintKey::Hasher>& metallicBondOwners() const noexcept;
					const std::unordered_map<ConstraintKey, AnionicConstraintOwner, ConstraintKey::Hasher>& anionicConstraintOwners() const noexcept;
					const std::unordered_map<ConstraintKey, CationicConstraintOwner, ConstraintKey::Hasher>& cationicConstraintOwners() const noexcept;
					const std::unordered_map<ConstraintKey, NonIonicBondOwner, ConstraintKey::Hasher>& nonIonicBondOwners() const noexcept;
					const std::unordered_map<ConstraintKey, NonMetallicBondOwner, ConstraintKey::Hasher>& nonMetallicBondOwners() const noexcept;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Protected methods

				protected:
					void createGeometricalConstraintOwners(const std::vector<ConstraintKey>& trackingAtomPairs, std::vector<ModelAtom>&);
					void createGeometricalConstraintOwners(const std::vector<ConstraintKey>& trackingAtomPairs, std::vector<ModelAtom>&, const double minimumConstraintCutoffRatio, const double nonBondingRangeRatio, const double ionicBondCutoffRatio, const double metallicBondCutoffRatio);
					void createChemicalBondOwners(std::vector<ModelAtom>&);
					void createChemicalBondOwners(std::vector<ModelAtom>&, const double ionicBondCutoffRatio, const double metallicBondCutoffRatio);
					void createCationicConstraintOwners(std::vector<ModelAtom>&);
					void substituteIonicBondOwners(std::vector<ModelAtom>&, const double feasibleConstraintErrorRate);

					void clearGeometricalConstraintOwners();
					void clearIonicBondOwners();
					void clearMetallicBondOwners();
					void clearCationicConstraintOwners();

					bool hasInfeasibleGeometricalConstraints(const double feasibleConstraintErrorRate) const;

				// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					std::vector<ModelAtom*> getInfeasibleAtomPtrs(std::vector<ModelAtom>&, const double feasibleConstraintErrorRate) const;

					void createBasicMinimumConstraintOwners(const std::vector<ConstraintKey>& trackingAtomPairs);
					void createBasicMinimumConstraintOwners(const std::vector<ConstraintKey>& trackingAtomPairs, const double minimumConstraintCutoffRatio, const double nonBondingRangeRatio);
					void createIonicBondOwners(std::vector<ModelAtom>&);
					void createIonicBondOwners(std::vector<ModelAtom>&, const double ionicBondCutoffRatio);
					void createMetallicBondOwners(std::vector<ModelAtom>&);
					void createMetallicBondOwners(std::vector<ModelAtom>&, const double metallicBondCutoffRatio);

					template <typename T>
					bool hasInfeasibleConstraints(const double errorRate, const std::unordered_map<ConstraintKey, T, ConstraintKey::Hasher>&) const;

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					size_type m_randomValueGenerating;
					std::mt19937 m_engine;

					std::unordered_map<ConstraintKey, IonicBondOwner, ConstraintKey::Hasher> m_ionicBondOwners;
					std::unordered_map<ConstraintKey, MetallicBondOwner, ConstraintKey::Hasher> m_metallicBondOwners;

					std::unordered_map<ConstraintKey, AnionicConstraintOwner, ConstraintKey::Hasher> m_anionicConstraintOwners;
					std::unordered_map<ConstraintKey, CationicConstraintOwner, ConstraintKey::Hasher> m_cationicConstraintOwners;
					std::unordered_map<ConstraintKey, NonIonicBondOwner, ConstraintKey::Hasher> m_nonIonicBondOwners;
					std::unordered_map<ConstraintKey, NonMetallicBondOwner, ConstraintKey::Hasher> m_nonMetallicBondOwners;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline const std::unordered_map<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey, MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::IonicBondOwner, MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::Hasher>& MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::ionicBondOwners() const noexcept
{
	return m_ionicBondOwners;
}

inline const std::unordered_map<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey, MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::MetallicBondOwner, MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::Hasher>& MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::metallicBondOwners() const noexcept
{
	return m_metallicBondOwners;
}

inline const std::unordered_map<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey, MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::AnionicConstraintOwner, MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::Hasher>& MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::anionicConstraintOwners() const noexcept
{
	return m_anionicConstraintOwners;
}

inline const std::unordered_map<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey, MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::CationicConstraintOwner, MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::Hasher>& MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::cationicConstraintOwners() const noexcept
{
	return m_cationicConstraintOwners;
}

inline const std::unordered_map<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey, MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::NonIonicBondOwner, MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::Hasher>& MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::nonIonicBondOwners() const noexcept
{
	return m_nonIonicBondOwners;
}

inline const std::unordered_map<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey, MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::NonMetallicBondOwner, MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey::Hasher>& MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::nonMetallicBondOwners() const noexcept
{
	return m_nonMetallicBondOwners;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::createGeometricalConstraintOwners(const std::vector<ConstraintKey>& trackingAtomPairs, std::vector<ModelAtom>& atomicArrangement)
{
	createBasicMinimumConstraintOwners(trackingAtomPairs);
	createIonicBondOwners(atomicArrangement);
	createMetallicBondOwners(atomicArrangement);
	createCationicConstraintOwners(atomicArrangement);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::createGeometricalConstraintOwners(const std::vector<ConstraintKey>& trackingAtomPairs, std::vector<ModelAtom>& atomicArrangement, const double minimumConstraintCutoffRatio, const double nonBondingRangeRatio, const double ionicBondCutoffRatio, const double metallicBondCutoffRatio)
{
	createBasicMinimumConstraintOwners(trackingAtomPairs, minimumConstraintCutoffRatio, nonBondingRangeRatio);
	createIonicBondOwners(atomicArrangement, ionicBondCutoffRatio);
	createMetallicBondOwners(atomicArrangement, metallicBondCutoffRatio);
	createCationicConstraintOwners(atomicArrangement);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::createChemicalBondOwners(std::vector<ModelAtom>& atomicArrangement)
{
	createIonicBondOwners(atomicArrangement);
	createMetallicBondOwners(atomicArrangement);
	createCationicConstraintOwners(atomicArrangement);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::createChemicalBondOwners(std::vector<ModelAtom>& atomicArrangement, const double ionicBondCutoffRatio, const double metallicBondCutoffRatio)
{
	createIonicBondOwners(atomicArrangement, ionicBondCutoffRatio);
	createMetallicBondOwners(atomicArrangement, metallicBondCutoffRatio);
	createCationicConstraintOwners(atomicArrangement);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::clearGeometricalConstraintOwners()
{
	m_ionicBondOwners.clear();
	m_metallicBondOwners.clear();
	m_anionicConstraintOwners.clear();
	m_cationicConstraintOwners.clear();
	m_nonIonicBondOwners.clear();
	m_nonMetallicBondOwners.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::clearIonicBondOwners()
{
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond;

	for (const auto& ionicBondOwner : m_ionicBondOwners)
	{
		if (m_nonIonicBondOwners.find(ionicBondOwner.first) == m_nonIonicBondOwners.end())
			m_nonIonicBondOwners.emplace(ionicBondOwner.first, NonIonicBondOwner{ NonIonicBond::create(ionicBondOwner.first) });
		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::logic_error(reportException(typeid(*this), "clearIonicBondOwners", "Could not translate ionic bond into non-ionic bond."));
		}
	}

	m_ionicBondOwners.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::clearMetallicBondOwners()
{
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond;

	for (const auto& metallicBondOwner : m_metallicBondOwners)
	{
		if (m_nonMetallicBondOwners.find(metallicBondOwner.first) == m_nonMetallicBondOwners.end())
			m_nonMetallicBondOwners.emplace(metallicBondOwner.first, NonMetallicBondOwner{ NonMetallicBond::create(metallicBondOwner.first) });
		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::logic_error(reportException(typeid(*this), "clearMetallicBondOwners", "Could not translate metallic bond into non-metallic bond."));
		}		
	}

	m_metallicBondOwners.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::clearCationicConstraintOwners()
{
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond;

	for (const auto& cationicConstraintOwner : m_cationicConstraintOwners)
	{
		if (m_nonMetallicBondOwners.find(cationicConstraintOwner.first) == m_nonMetallicBondOwners.end())
			m_nonMetallicBondOwners.emplace(cationicConstraintOwner.first, NonMetallicBondOwner{ NonMetallicBond::create(cationicConstraintOwner.first) });
		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::logic_error(reportException(typeid(*this), "clearCationicConstraintOwners", "Could not translate cationic constraints into non-metallic bond."));
		}
	}

	m_cationicConstraintOwners.clear();
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::hasInfeasibleGeometricalConstraints(const double feasibleConstraintErrorRate) const
{
	if (hasInfeasibleConstraints(feasibleConstraintErrorRate, m_ionicBondOwners))
		return true;
	if (hasInfeasibleConstraints(feasibleConstraintErrorRate, m_metallicBondOwners))
		return true;
	if (hasInfeasibleConstraints(feasibleConstraintErrorRate, m_anionicConstraintOwners))
		return true;
	if (hasInfeasibleConstraints(feasibleConstraintErrorRate, m_cationicConstraintOwners))
		return true;
	if (hasInfeasibleConstraints(feasibleConstraintErrorRate, m_nonIonicBondOwners))
		return true;
	if (hasInfeasibleConstraints(feasibleConstraintErrorRate, m_nonMetallicBondOwners))
		return true;

	return false;
}

// Protected method
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename T>
inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::ConstraintManager::hasInfeasibleConstraints(const double errorRate, const std::unordered_map<ConstraintKey, T, ConstraintKey::Hasher>& constraintOwners) const
{
	for (const auto& keyAndConstraintOwner : constraintOwners)
	{
		if (!(keyAndConstraintOwner.second.constraintPtr()->isFeasible(errorRate)))
			return true;
	}

	return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_CONSTRAINTMANAGER_H
