#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_COORDINATIONCENTER_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_COORDINATIONCENTER_H

#include "LinearAlgebra.h"

#include "AtomBase.h"
#include "UnitCell.h"
#include "PeriodicReferenceWrapper.h"

#include "CoordinationFeature.h"
#include "CoordinationFeatureSymbol.h"

#include "CoordinationCenterBase.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			namespace Internal
			{
				class CoordinationCenter :public CoordinationCenterBase, public ChemToolkit::Generic::AtomBase
				{
					using NumericalSymmetricMatrix = MathToolkit::Generic::NumericalSymmetricMatrix;
					using Ligand = ChemToolkit::Crystallography::PeriodicReferenceWrapper<CoordinationCenter, ChemToolkit::Crystallography::UnitCell>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				protected:
					explicit CoordinationCenter(const std::string& elementSymbol);
					CoordinationCenter(const std::string& elementSymbol, const CoordinationFeatureSymbol&);
					CoordinationCenter(const std::string& elementSymbol, const NumericalVector& cartesianCoordinate);
					CoordinationCenter(const std::string& elementSymbol, const CoordinationFeatureSymbol&, const NumericalVector& cartesianCoordinate);

				public:
					virtual ~CoordinationCenter() = default;

					CoordinationCenter(const CoordinationCenter&);
					CoordinationCenter(CoordinationCenter&&) noexcept;
					CoordinationCenter& operator=(const CoordinationCenter&);
					CoordinationCenter& operator=(CoordinationCenter&&) noexcept;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void optimizeCoordinationPolyhedra();

					bool hasAnySymmetry(const std::vector<AffineMapping>& symmetryOperations) const;
					bool isHomoleptic() const;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Protected methods

				protected:
					void recognizeIonicLigands(std::vector<Ligand>&&);
					void recognizeMetallicLigands(std::vector<Ligand>&&);
					void clearRecognizedLigands();

				// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					NumericalMatrix getRefinedLocalCoordinateSystem() const;
					NumericalSymmetricMatrix getInertiaTensor() const;

					bool hasSymmetry(AffineMapping symmetryOperation, const NumericalMatrix& localCoordinateSystem, const NumericalMatrix& inverseLocalCoordinateSystem) const;
					bool hasSameAtom(const std::string& elementSymbol, const NumericalVector& coordinate) const;

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					NumericalMatrix m_localCoordinateSystem;
					std::vector<Ligand> m_ionicLigands;
					std::vector<Ligand> m_metallicLigands;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::optimizeCoordinationPolyhedra()
{
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::hasAnySymmetry(const std::vector<AffineMapping>& symmetryOperations) const
{
	NumericalMatrix localCoordinateSystem = getRefinedLocalCoordinateSystem();
	NumericalMatrix inverseLocalCoordinateSystem = localCoordinateSystem.getInverseMatrix();

	for (const auto& symmetryOperation : symmetryOperations)
	{
		if (hasSymmetry(symmetryOperation, localCoordinateSystem, inverseLocalCoordinateSystem))
			return true;
	}

	return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::isHomoleptic() const
{
	std::string atomCategorySymbol;

	for (const auto& ionicLigand : m_ionicLigands)
	{
		if (atomCategorySymbol.empty())
			atomCategorySymbol = ionicLigand.get().coordinationFeature().atomCategory().symbol();

		else
		{
			if (atomCategorySymbol != ionicLigand.get().coordinationFeature().atomCategory().symbol())
				return false;
		}
	}

	for (const auto& metallicLigand : m_metallicLigands)
	{
		if (atomCategorySymbol.empty())
			atomCategorySymbol = metallicLigand.get().coordinationFeature().atomCategory().symbol();

		else
		{
			if (atomCategorySymbol != metallicLigand.get().coordinationFeature().atomCategory().symbol())
				return false;
		}
	}

	if (atomCategorySymbol.empty())
		return false;
	else
		return true;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::recognizeIonicLigands(std::vector<Ligand>&& ligands)
{
	m_ionicLigands = std::move(ligands);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::recognizeMetallicLigands(std::vector<Ligand>&& ligands)
{
	m_metallicLigands = std::move(ligands);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::clearRecognizedLigands()
{
	m_ionicLigands.clear();
	m_metallicLigands.clear();
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline MathToolkit::Generic::NumericalMatrix<double> MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::getRefinedLocalCoordinateSystem() const
{
	std::vector<NumericalVector> eigenVectors = getInertiaTensor().getEigenVectors();

	if (eigenVectors.size() == 3)
	{
		double normA = eigenVectors[0].norm();
		double normB = eigenVectors[1].norm();
		double normC = eigenVectors[2].norm();
		eigenVectors[0] /= normA;
		eigenVectors[1] /= normB;
		eigenVectors[2] /= normC;
		{
			double coordinateSign = 0.0;
			{
				coordinateSign += eigenVectors[0][0] * ((eigenVectors[1][1] * eigenVectors[2][2]) - (eigenVectors[1][2] * eigenVectors[2][1]));
				coordinateSign += eigenVectors[0][1] * ((eigenVectors[1][2] * eigenVectors[2][0]) - (eigenVectors[1][0] * eigenVectors[2][2]));
				coordinateSign += eigenVectors[0][2] * ((eigenVectors[1][0] * eigenVectors[2][1]) - (eigenVectors[1][1] * eigenVectors[2][0]));
			}

			if (coordinateSign < 0.0)
			{
				eigenVectors[0] *= (-1.0);
				eigenVectors[1] *= (-1.0);
				eigenVectors[2] *= (-1.0);
			}
		}

		return NumericalMatrix(eigenVectors);
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "getRefinedLocalCoordinateSystem", "Could not create three dimensional coordinate system."));
	}
}

inline MathToolkit::Generic::NumericalSymmetricMatrix MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::getInertiaTensor() const
{
	NumericalSymmetricMatrix inertiaTensor(3);
	{
		for (const auto& ligand : m_ionicLigands)
		{
			NumericalSymmetricMatrix thisInertiaTensor(3);
			{
				NumericalVector thisCoordinate = ligand.get().coordinate() - AtomBase::coordinate();

				thisInertiaTensor(0, 0) = ((thisCoordinate[1] * thisCoordinate[1]) + (thisCoordinate[2] * thisCoordinate[2]));
				thisInertiaTensor(1, 0) = ((-1.0) * thisCoordinate[0] * thisCoordinate[1]);
				thisInertiaTensor(2, 0) = ((-1.0) * thisCoordinate[0] * thisCoordinate[2]);
				thisInertiaTensor(1, 1) = ((thisCoordinate[2] * thisCoordinate[2]) + (thisCoordinate[0] * thisCoordinate[0]));
				thisInertiaTensor(2, 1) = ((-1.0) * thisCoordinate[1] * thisCoordinate[2]);
				thisInertiaTensor(2, 2) = ((thisCoordinate[0] * thisCoordinate[0]) + (thisCoordinate[1] * thisCoordinate[1]));
			}

			inertiaTensor += thisInertiaTensor;
		}

		for (const auto& ligand : m_metallicLigands)
		{
			NumericalSymmetricMatrix thisInertiaTensor(3);
			{
				NumericalVector thisCoordinate = ligand.get().coordinate() - AtomBase::coordinate();

				thisInertiaTensor(0, 0) = ((thisCoordinate[1] * thisCoordinate[1]) + (thisCoordinate[2] * thisCoordinate[2]));
				thisInertiaTensor(1, 0) = ((-1.0) * thisCoordinate[0] * thisCoordinate[1]);
				thisInertiaTensor(2, 0) = ((-1.0) * thisCoordinate[0] * thisCoordinate[2]);
				thisInertiaTensor(1, 1) = ((thisCoordinate[2] * thisCoordinate[2]) + (thisCoordinate[0] * thisCoordinate[0]));
				thisInertiaTensor(2, 1) = ((-1.0) * thisCoordinate[1] * thisCoordinate[2]);
				thisInertiaTensor(2, 2) = ((thisCoordinate[0] * thisCoordinate[0]) + (thisCoordinate[1] * thisCoordinate[1]));
			}

			inertiaTensor += thisInertiaTensor;
		}
	}

	return inertiaTensor;
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::hasSymmetry(AffineMapping symmetryOperation, const NumericalMatrix& localCoordinateSystem, const NumericalMatrix& inverseLocalCoordinateSystem) const
{
	symmetryOperation.transform(localCoordinateSystem);
	symmetryOperation *= inverseLocalCoordinateSystem;

	for (const auto& ligand : m_ionicLigands)
	{
		const std::string elementSymbol = ligand.get().elementSymbol();
		NumericalVector coordinate = ligand.coordinate() - AtomBase::coordinate();
		{
			coordinate.transform(symmetryOperation.matrixPart());
			coordinate += symmetryOperation.columnPart();
		}

		if (!(hasSameAtom(elementSymbol, coordinate)))
			return false;
	}

	for (const auto& ligand : m_metallicLigands)
	{
		const std::string elementSymbol = ligand.get().elementSymbol();
		NumericalVector coordinate = ligand.coordinate() - AtomBase::coordinate();
		{
			coordinate.transform(symmetryOperation.matrixPart());
			coordinate += symmetryOperation.columnPart();
		}

		if (!(hasSameAtom(elementSymbol, coordinate)))
			return false;
	}

	return true;
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenter::hasSameAtom(const std::string& elementSymbol, const NumericalVector& coordinate) const
{
	for (const auto& ionicLigand : m_ionicLigands)
	{
		if (elementSymbol == ionicLigand.get().elementSymbol())
		{
			NumericalVector difference = ionicLigand.coordinate() - AtomBase::coordinate() - coordinate;

			if (difference.normSquare() < (pointGroupPrecision() * pointGroupPrecision()))
				return true;
		}
	}

	for (const auto& metallicLigand : m_metallicLigands)
	{
		if (elementSymbol == metallicLigand.get().elementSymbol())
		{
			NumericalVector difference = metallicLigand.coordinate() - AtomBase::coordinate() - coordinate;

			if (difference.normSquare() < (pointGroupPrecision() * pointGroupPrecision()))
				return true;
		}
	}

	return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_COORDINATIONCENTER_H
