#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_FEATURES_COORDINATIONFEATURE_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_FEATURES_COORDINATIONFEATURE_H

#include <stdexcept>
#include <string>
#include <utility>

#include "ConstraintComposition.h"

#include "AtomCategory.h"
#include "CoordinationFeatureSymbol.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Features
		{
			class CoordinationFeature
			{
				using size_type = std::size_t;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				CoordinationFeature();
				explicit CoordinationFeature(const CoordinationFeatureSymbol&);
				virtual ~CoordinationFeature() = default;

				CoordinationFeature(const CoordinationFeature&) = default;
				CoordinationFeature(CoordinationFeature&&) noexcept = default;
				CoordinationFeature& operator=(const CoordinationFeature&) = default;
				CoordinationFeature& operator=(CoordinationFeature&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				const CoordinationFeatureSymbol& symbol() const noexcept;

				const AtomCategory& atomCategory() const noexcept;
				double upperIonicBondRadius() const noexcept;
				double lowerIonicBondRadius() const noexcept;
				double repulsiveIonicRadius() const noexcept;
				double upperMetallicBondRadius() const noexcept;
				double lowerMetallicBondRadius() const noexcept;
				const ConstraintComposition& feasibleIonicBondComposition() const noexcept;
				const ConstraintComposition& feasibleMetallicBondComposition() const noexcept;

				void setCoordinationFeatureSymbol(const CoordinationFeatureSymbol&);
				void setIonicBondRadii(const std::string& upperRadiusSymbol, const std::string& lowerRadiusSymbol);
				void setRepulsiveIonicRadius(const std::string& radiusSymbol);
				void setMetallicBondRadii(const std::string& upperRadiusSymbol, const std::string& lowerRadiusSymbol);
				void setFeasibleIonicBondComposition(const ConstraintComposition&);
				void setFeasibleMetallicBondComposition(const ConstraintComposition&);

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept;
				void initializeFromElementSymbol(const std::string& elementSymbol, const int oxidationState);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static utility

				static double getDefinedRadius(const std::string& radiusSymbol);
				static double getDefinedRadius(const size_type radiusNumber);

				static std::string getUppercaseRadiusSymbol(const size_type radiusNumber);
				static std::string getLowercaseRadiusSymbol(const size_type radiusNumber);

				static size_type getMaxRadiusNumber() noexcept;
				static size_type getRadiusNumber(const std::string& radiusSymbol);
				static size_type getFloorRadiusNumber(const double radius);

				static size_type getMinimumFeasibleRadiusNumber(const size_type coordinationNumber);

			// Static utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				CoordinationFeatureSymbol m_coordinationFeatureSymbol;

				AtomCategory m_atomCategory;
				double m_upperIonicBondRadius;
				double m_lowerIonicBondRadius;
				double m_repulsiveIonicRadius;
				double m_upperMetallicBondRadius;
				double m_lowerMetallicBondRadius;

				ConstraintComposition m_feasibleIonicBondComposition;
				ConstraintComposition m_feasibleMetallicBondComposition;
			};


			inline bool operator<(const CoordinationFeature& cfa, const CoordinationFeature& cfb)
			{
				if (!(cfa.symbol().isValid()))
					return false;
				else
				{
					if (!(cfb.symbol().isValid()))
						return true;
					else
					{
						if (cfa.atomCategory() < cfb.atomCategory())
							return true;
						else if (cfb.atomCategory() < cfa.atomCategory())
							return false;
						else
						{
							if (std::make_pair(cfa.upperIonicBondRadius(), cfa.lowerIonicBondRadius()) < std::make_pair(cfb.upperIonicBondRadius(), cfb.lowerIonicBondRadius()))
								return true;
							else if (std::make_pair(cfb.upperIonicBondRadius(), cfb.lowerIonicBondRadius()) < std::make_pair(cfa.upperIonicBondRadius(), cfa.lowerIonicBondRadius()))
								return false;

							else
							{
								if (cfa.repulsiveIonicRadius() < cfb.repulsiveIonicRadius())
									return true;
								else if (cfb.repulsiveIonicRadius() < cfa.repulsiveIonicRadius())
									return false;
								else
								{
									if (std::make_pair(cfa.upperMetallicBondRadius(), cfa.lowerMetallicBondRadius()) < std::make_pair(cfb.upperMetallicBondRadius(), cfb.lowerMetallicBondRadius()))
										return true;
									else if (std::make_pair(cfb.upperMetallicBondRadius(), cfb.lowerMetallicBondRadius()) < std::make_pair(cfa.upperMetallicBondRadius(), cfa.lowerMetallicBondRadius()))
										return false;
									else
									{
										if (cfa.feasibleIonicBondComposition() < cfb.feasibleIonicBondComposition())
											return true;
										else if (cfb.feasibleIonicBondComposition() < cfa.feasibleIonicBondComposition())
											return false;
										else
										{
											if (cfa.feasibleMetallicBondComposition() < cfb.feasibleMetallicBondComposition())
												return true;
											else
												return false;
										}
									}
								}
							}
						}
					}
				}
			}

			inline bool operator>(const CoordinationFeature& cfa, const CoordinationFeature& cfb)
			{
				return (cfb < cfa);
			}

			inline bool operator==(const CoordinationFeature& cfa, const CoordinationFeature& cfb)
			{
				return (cfa.symbol() == cfb.symbol());
			}

			inline bool operator!=(const CoordinationFeature& cfa, const CoordinationFeature& cfb)
			{
				return (cfa.symbol() != cfb.symbol());
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeatureSymbol& MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::symbol() const noexcept
{
	return m_coordinationFeatureSymbol;
}

inline const MathematicalCrystalChemistry::StructureModel::Features::AtomCategory& MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::atomCategory() const noexcept
{
	return m_atomCategory;
}

inline double MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::upperIonicBondRadius() const noexcept
{
	return m_upperIonicBondRadius;
}

inline double MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::lowerIonicBondRadius() const noexcept
{
	return m_lowerIonicBondRadius;
}

inline double MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::repulsiveIonicRadius() const noexcept
{
	return m_repulsiveIonicRadius;
}

inline double MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::upperMetallicBondRadius() const noexcept
{
	return m_upperMetallicBondRadius;
}

inline double MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::lowerMetallicBondRadius() const noexcept
{
	return m_lowerMetallicBondRadius;
}

inline const MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition& MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::feasibleIonicBondComposition() const noexcept
{
	return m_feasibleIonicBondComposition;
}

inline const MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition& MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::feasibleMetallicBondComposition() const noexcept
{
	return m_feasibleMetallicBondComposition;
}

inline void MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::setIonicBondRadii(const std::string& upperRadiusSymbol, const std::string& lowerRadiusSymbol)
{
	m_upperIonicBondRadius = getDefinedRadius(upperRadiusSymbol);
	m_lowerIonicBondRadius = getDefinedRadius(lowerRadiusSymbol);


	if (m_upperIonicBondRadius < m_lowerIonicBondRadius)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "setIonicBondRadii", "Upper ionic radius is less than lower ionic radius."));
	}

	else
	{
		m_coordinationFeatureSymbol.setUpperIonicBondRadiusSymbol(upperRadiusSymbol);
		m_coordinationFeatureSymbol.setLowerIonicBondRadiusSymbol(lowerRadiusSymbol);
	}
}

inline void MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::setRepulsiveIonicRadius(const std::string& radiusSymbol)
{
	m_repulsiveIonicRadius = getDefinedRadius(radiusSymbol);
	m_coordinationFeatureSymbol.setRepulsiveIonicRadiusSymbol(radiusSymbol);
}

inline void MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::setMetallicBondRadii(const std::string& upperRadiusSymbol, const std::string& lowerRadiusSymbol)
{
	m_upperMetallicBondRadius = getDefinedRadius(upperRadiusSymbol);
	m_lowerMetallicBondRadius = getDefinedRadius(lowerRadiusSymbol);


	if (m_upperMetallicBondRadius < m_lowerMetallicBondRadius)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "setMetallicBondRadii", "Upper ionic radius is less than lower ionic radius."));
	}

	else
	{
		m_coordinationFeatureSymbol.setUpperMetallicBondRadiusSymbol(upperRadiusSymbol);
		m_coordinationFeatureSymbol.setLowerMetallicBondRadiusSymbol(lowerRadiusSymbol);
	}
}

inline void MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::setFeasibleIonicBondComposition(const ConstraintComposition& constraintComposition)
{
	m_feasibleIonicBondComposition = constraintComposition;
	m_coordinationFeatureSymbol.setIonicBondCompositionSymbol(m_feasibleIonicBondComposition.name());
}

inline void MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::setFeasibleMetallicBondComposition(const ConstraintComposition& constraintComposition)
{
	m_feasibleMetallicBondComposition = constraintComposition;
	m_coordinationFeatureSymbol.setMetallicBondCompositionSymbol(m_feasibleMetallicBondComposition.name());
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline double MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getDefinedRadius(const size_type radiusNumber)
{
	std::string uppercaseRadiusSymbol = getUppercaseRadiusSymbol(radiusNumber);

	if (uppercaseRadiusSymbol.empty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("getDefinedRadius of MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature", "\"radiusNumber\" is out of definitions."));
	}

	else
		return getDefinedRadius(uppercaseRadiusSymbol);
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber() noexcept
{
	return 22;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getRadiusNumber(const std::string& radiusSymbol)
{
	if (radiusSymbol == "Z" || radiusSymbol == "z")
		return 0;

	if (radiusSymbol == "T" || radiusSymbol == "t")
		return 1;
	if (radiusSymbol == "Q" || radiusSymbol == "q")
		return 2;
	if (radiusSymbol == "H" || radiusSymbol == "h")
		return 3;
	if (radiusSymbol == "S" || radiusSymbol == "s")
		return 4;
	if (radiusSymbol == "E" || radiusSymbol == "e")
		return 5;
	if (radiusSymbol == "N" || radiusSymbol == "n")
		return 6;
	if (radiusSymbol == "D" || radiusSymbol == "d")
		return 7;
	if (radiusSymbol == "WD" || radiusSymbol == "wd")
		return 8;

	if (radiusSymbol == "O" || radiusSymbol == "o")
		return 9;

	if (radiusSymbol == "TD" || radiusSymbol == "td")
		return 10;
	if (radiusSymbol == "QD" || radiusSymbol == "qd")
		return 11;
	if (radiusSymbol == "PD" || radiusSymbol == "pd")
		return 12;
	if (radiusSymbol == "HD" || radiusSymbol == "hd")
		return 13;
	if (radiusSymbol == "SD" || radiusSymbol == "sd")
		return 14;
	if (radiusSymbol == "ED" || radiusSymbol == "ed")
		return 15;
	if (radiusSymbol == "ND" || radiusSymbol == "nd")
		return 16;
	if (radiusSymbol == "V" || radiusSymbol == "v")
		return 17;

	if (radiusSymbol == "UV" || radiusSymbol == "uv")
		return 18;
	if (radiusSymbol == "WV" || radiusSymbol == "wv")
		return 19;
	if (radiusSymbol == "TV" || radiusSymbol == "tv")
		return 20;
	if (radiusSymbol == "QV" || radiusSymbol == "qv")
		return 21;
	if (radiusSymbol == "PV" || radiusSymbol == "pv")
		return 22;


	OperatingSystem::Configuration::ReportException reportException;
	throw std::invalid_argument(reportException("getRadiusNumber of MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature", "\"radiusSymbol\" is out of definitions."));
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_FEATURES_COORDINATIONFEATURE_H
