#ifndef MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_COORDINATIONFEATUREASSIGNER_H
#define MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_COORDINATIONFEATUREASSIGNER_H

#include <string>
#include <filesystem>
#include <stdexcept>

#include "ReportException.h"

#include "ModelCrystalStructure.h"
#include "CoordinationFeatureSpecifyOptions.h"


namespace MathematicalCrystalChemistry
{
	namespace Runtime
	{
		namespace Analysis
		{
			namespace Internal
			{
				class CoordinationFeatureAssigner
				{
					using size_type = std::size_t;
					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				public:
					CoordinationFeatureAssigner() = default;
					virtual ~CoordinationFeatureAssigner() = default;

					CoordinationFeatureAssigner(const CoordinationFeatureAssigner&) = default;
					CoordinationFeatureAssigner(CoordinationFeatureAssigner&&) noexcept = default;
					CoordinationFeatureAssigner& operator=(const CoordinationFeatureAssigner&) = default;
					CoordinationFeatureAssigner& operator=(CoordinationFeatureAssigner&&) noexcept = default;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void initializeRadiusNumbers();
					void initializeRadiusNumbersRange() noexcept;
					bool advanceRadiusNumbers();
					void setNumbersRanges(const ModelCrystalStructure&, size_type upperIonicBondRadiusNumberExtensionWidth, size_type repulsiveIonicRadiusNumberExtensionWidth);

					ModelCrystalStructure getModelCrystalStructure(const ModelCrystalStructure& inputModelCrystalStructure) const;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					void setDefaultMinimumIonicBondRadiusNumbersRange(const ModelCrystalStructure&);
					void setDefaultMaximumIonicBondRadiusNumbersRange(const ModelCrystalStructure&);

					void setIonicBondRadiusNumbersRange(const ModelCrystalStructure&, const size_type upperIonicBondRadiusNumberExtensionWidth);
					void setRepulsiveIonicRadiusNumbersRange(const ModelCrystalStructure&, const size_type repulsiveIonicRadiusNumberExtensionWidth);

					void initializeRadiusNumbers(const std::map<size_type, std::pair<size_type, size_type>>& numberRanges, std::map<size_type, size_type>&);
					bool advanceRadiusNumbers(const std::map<size_type, std::pair<size_type, size_type>>& numberRanges, std::map<size_type, size_type>&);

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private utility

					size_type countRadiusNumberPossibles() const;

				// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					std::map<size_type, size_type> m_upperIonicBondRadiusNumbers;
					std::map<size_type, size_type> m_lowerIonicBondRadiusNumbers;
					std::map<size_type, size_type> m_repulsiveIonicRadiusNumbers;
					std::map<size_type, std::pair<size_type, size_type>> m_upperIonicBondRadiusNumbersRange;
					std::map<size_type, std::pair<size_type, size_type>> m_lowerIonicBondRadiusNumbersRange;
					std::map<size_type, std::pair<size_type, size_type>> m_repulsiveIonicRadiusNumbersRange;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::CoordinationFeatureAssigner::initializeRadiusNumbers()
{
	initializeRadiusNumbers(m_upperIonicBondRadiusNumbersRange, m_upperIonicBondRadiusNumbers);
	initializeRadiusNumbers(m_lowerIonicBondRadiusNumbersRange, m_lowerIonicBondRadiusNumbers);
	initializeRadiusNumbers(m_repulsiveIonicRadiusNumbersRange, m_repulsiveIonicRadiusNumbers);
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::CoordinationFeatureAssigner::initializeRadiusNumbersRange() noexcept
{
	m_upperIonicBondRadiusNumbersRange.clear();
	m_lowerIonicBondRadiusNumbersRange.clear();
	m_repulsiveIonicRadiusNumbersRange.clear();
}

inline bool MathematicalCrystalChemistry::Runtime::Analysis::Internal::CoordinationFeatureAssigner::advanceRadiusNumbers()
{
	if (advanceRadiusNumbers(m_repulsiveIonicRadiusNumbersRange, m_repulsiveIonicRadiusNumbers))
		return true;

	else
	{
		while (true)
		{
			if (advanceRadiusNumbers(m_lowerIonicBondRadiusNumbersRange, m_lowerIonicBondRadiusNumbers))
			{
				bool validRadii = true;
				{
					for (const auto& indexAndNumber : m_upperIonicBondRadiusNumbers)
					{
						if (indexAndNumber.second < m_lowerIonicBondRadiusNumbers.at(indexAndNumber.first))
						{
							validRadii = false;
							break;
						}
					}
				}

				if (validRadii)
					return true;
			}

			else
			{
				if (advanceRadiusNumbers(m_upperIonicBondRadiusNumbersRange, m_upperIonicBondRadiusNumbers))
				{
					bool validRadii = true;
					{
						for (const auto& indexAndNumber : m_upperIonicBondRadiusNumbers)
						{
							if (indexAndNumber.second < m_lowerIonicBondRadiusNumbers.at(indexAndNumber.first))
							{
								validRadii = false;
								break;
							}
						}
					}

					if (validRadii)
						return true;
				}

				else
					return false;
			}
		}
	}
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::CoordinationFeatureAssigner::setNumbersRanges(const ModelCrystalStructure& modelCrystalStructure, size_type upperIonicBondRadiusNumberExtensionWidth, size_type repulsiveIonicRadiusNumberExtensionWidth)
{
	setIonicBondRadiusNumbersRange(modelCrystalStructure, upperIonicBondRadiusNumberExtensionWidth);
	setRepulsiveIonicRadiusNumbersRange(modelCrystalStructure, repulsiveIonicRadiusNumberExtensionWidth);


	while ((0 < upperIonicBondRadiusNumberExtensionWidth) || (0 < repulsiveIonicRadiusNumberExtensionWidth))
	{
		if (100000 < countRadiusNumberPossibles())
		{
			if (0 < repulsiveIonicRadiusNumberExtensionWidth)
			{
				--repulsiveIonicRadiusNumberExtensionWidth;

				setIonicBondRadiusNumbersRange(modelCrystalStructure, upperIonicBondRadiusNumberExtensionWidth);
				setRepulsiveIonicRadiusNumbersRange(modelCrystalStructure, repulsiveIonicRadiusNumberExtensionWidth);
			}

			else
			{
				if (0 < upperIonicBondRadiusNumberExtensionWidth)
				{
					--upperIonicBondRadiusNumberExtensionWidth;

					setIonicBondRadiusNumbersRange(modelCrystalStructure, upperIonicBondRadiusNumberExtensionWidth);
					setRepulsiveIonicRadiusNumbersRange(modelCrystalStructure, repulsiveIonicRadiusNumberExtensionWidth);
				}
			}
		}

		else
			break;
	}


	if (100000 < countRadiusNumberPossibles())
	{
		setDefaultMinimumIonicBondRadiusNumbersRange(modelCrystalStructure);

		if (100000 < countRadiusNumberPossibles())
		{
			setDefaultMaximumIonicBondRadiusNumbersRange(modelCrystalStructure);

			if (100000 < countRadiusNumberPossibles())
			{
				OperatingSystem::Configuration::ReportException reportException;
				throw std::runtime_error(reportException(typeid(*this), "setNumbersRanges", "There are too possible sets of radius numbers."));
			}
		}
	}
}

inline MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure MathematicalCrystalChemistry::Runtime::Analysis::Internal::CoordinationFeatureAssigner::getModelCrystalStructure(const ModelCrystalStructure& inputCrystalStructure) const
{
	ModelCrystalStructure modelCrystalStructure = inputCrystalStructure;
	{
		for (auto& modelAtom : modelCrystalStructure.atomicArrangement())
		{
			using MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature;
			size_type siteIndex = modelAtom.siteIndex();

			std::string upperIonicBondRadiusSymbol = CoordinationFeature::getUppercaseRadiusSymbol(m_upperIonicBondRadiusNumbers.at(siteIndex));
			std::string lowerIonicBondRadiusSymbol = CoordinationFeature::getLowercaseRadiusSymbol(m_lowerIonicBondRadiusNumbers.at(siteIndex));
			std::string repulsiveIonicRadiusSymbol = CoordinationFeature::getUppercaseRadiusSymbol(m_repulsiveIonicRadiusNumbers.at(siteIndex));

			modelAtom.coordinationFeature().setIonicBondRadii(upperIonicBondRadiusSymbol, lowerIonicBondRadiusSymbol);
			modelAtom.coordinationFeature().setRepulsiveIonicRadius(repulsiveIonicRadiusSymbol);
		}
	}

	return modelCrystalStructure;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::CoordinationFeatureAssigner::initializeRadiusNumbers(const std::map<size_type, std::pair<size_type, size_type>>& numberRanges, std::map<std::size_t, std::size_t>& radiusNumbers)
{
	radiusNumbers.clear();

	for (const auto& indexAndRange : numberRanges)
		radiusNumbers.emplace(indexAndRange.first, indexAndRange.second.first);
}

inline bool MathematicalCrystalChemistry::Runtime::Analysis::Internal::CoordinationFeatureAssigner::advanceRadiusNumbers(const std::map<size_type, std::pair<size_type, size_type>>& numberRanges, std::map<size_type, size_type>& radiusNumbers)
{
	for (auto& indexAndNumber : radiusNumbers)
	{
		const std::pair<size_type, size_type>& range = numberRanges.at(indexAndNumber.first);

		if (indexAndNumber.second < range.second)
		{
			++indexAndNumber.second;
			return true;
		}

		else
			indexAndNumber.second = range.first;
	}

	return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

inline std::size_t MathematicalCrystalChemistry::Runtime::Analysis::Internal::CoordinationFeatureAssigner::countRadiusNumberPossibles() const
{
	size_type numPossibles = 1;
	{
		for (const auto& upperIonicBondIndexAndRange : m_upperIonicBondRadiusNumbersRange)
			numPossibles *= (upperIonicBondIndexAndRange.second.second + 1 - upperIonicBondIndexAndRange.second.first);

		for (const auto& lowerIonicBondIndexAndRange : m_lowerIonicBondRadiusNumbersRange)
			numPossibles *= (lowerIonicBondIndexAndRange.second.second + 1 - lowerIonicBondIndexAndRange.second.first);

		for (const auto& repulsiveIndexAndRange : m_repulsiveIonicRadiusNumbersRange)
			numPossibles *= (repulsiveIndexAndRange.second.second + 1 - repulsiveIndexAndRange.second.first);
	}

	return numPossibles;
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_COORDINATIONFEATUREASSIGNER_H
