#ifndef MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_COORDINATIONFEATURESPECIFYOPTIONS_H
#define MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_COORDINATIONFEATURESPECIFYOPTIONS_H

#include <stdexcept>
#include <string>
#include <filesystem>
#include <vector>

#include "ChemicalParameters.h"
#include "GeometricalConstraintParameters.h"
#include "StructuralOptimizationParameters.h"
#include "StructureDesignRecordOptions.h"


namespace MathematicalCrystalChemistry
{
	namespace Runtime
	{
		namespace Analysis
		{
			class CoordinationFeatureSpecifyOptions
			{
				using size_type = std::size_t;

				using ChemicalParameters = ChemToolkit::Generic::ChemicalParameters;
				using GeometricalConstraintParameters = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::GeometricalConstraintParameters;
				using StructuralOptimizationParameters = MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters;
				using StructureDesignRecordOptions = MathematicalCrystalChemistry::Design::Recording::StructureDesignRecordOptions;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				CoordinationFeatureSpecifyOptions();
				virtual ~CoordinationFeatureSpecifyOptions() = default;

				CoordinationFeatureSpecifyOptions(const CoordinationFeatureSpecifyOptions&) = default;
				CoordinationFeatureSpecifyOptions(CoordinationFeatureSpecifyOptions&&) noexcept = default;
				CoordinationFeatureSpecifyOptions& operator=(const CoordinationFeatureSpecifyOptions&) = default;
				CoordinationFeatureSpecifyOptions& operator=(CoordinationFeatureSpecifyOptions&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept;
				void initializeFromFiles(const std::vector<std::filesystem::path>& filePaths);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				const std::filesystem::path& cifFilesDirectoryPath() const noexcept;
				const std::filesystem::path& workDirectoryPath() const noexcept;
				const std::filesystem::path& outputDirectoryPath() const noexcept;

				bool needInfeasibleResult() const noexcept;
				size_type upperIonicBondRadiusNumberExtensionWidth() const noexcept;
				size_type repulsiveIonicRadiusNumberExtensionWidth() const noexcept;
				size_type maxReoptimizing() const noexcept;
				double maxAtomFractuationRatio() const noexcept;
				double initialMinimumConstraintCutoffRatio() const noexcept;
				double initialIonicBondCutoffRatio() const noexcept;
				double initialMetallicBondCutoffRatio() const noexcept;

				const ChemicalParameters& chemicalParameters() const noexcept;
				const GeometricalConstraintParameters& geometricalConstraintParameters() const noexcept;
				const StructuralOptimizationParameters& structuralOptimizationParameters() const noexcept;
				const StructureDesignRecordOptions& structureDesignRecordOptions() const noexcept;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void readInfeasibleResultNecessity(const std::string& inputText);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::filesystem::path m_cifFilesDirectoryPath;
				std::filesystem::path m_workDirectoryPath;
				std::filesystem::path m_outputDirectoryPath;

				bool m_needInfeasibleResult;
				size_type m_upperIonicBondRadiusNumberExtensionWidth;
				size_type m_repulsiveIonicRadiusNumberExtensionWidth;
				size_type m_maxReoptimizing;
				double m_maxAtomFractuationRatio;
				double m_initialMinimumConstraintCutoffRatio;
				double m_initialIonicBondCutoffRatio;
				double m_initialMetallicBondCutoffRatio;

				ChemicalParameters m_chemicalParameters;
				GeometricalConstraintParameters m_geometricalConstraintParameters;
				StructuralOptimizationParameters m_structuralOptimizationParameters;
				StructureDesignRecordOptions m_structureDesignRecordOptions;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const std::filesystem::path& MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::cifFilesDirectoryPath() const noexcept
{
	return m_cifFilesDirectoryPath;
}

inline const std::filesystem::path& MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::workDirectoryPath() const noexcept
{
	return m_workDirectoryPath;
}

inline const std::filesystem::path& MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::outputDirectoryPath() const noexcept
{
	return m_outputDirectoryPath;
}

inline bool MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::needInfeasibleResult() const noexcept
{
	return m_needInfeasibleResult;
}

inline std::size_t MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::upperIonicBondRadiusNumberExtensionWidth() const noexcept
{
	return m_upperIonicBondRadiusNumberExtensionWidth;
}

inline std::size_t MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::repulsiveIonicRadiusNumberExtensionWidth() const noexcept
{
	return m_repulsiveIonicRadiusNumberExtensionWidth;
}

inline std::size_t MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::maxReoptimizing() const noexcept
{
	return m_maxReoptimizing;
}

inline double MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::maxAtomFractuationRatio() const noexcept
{
	return m_maxAtomFractuationRatio;
}

inline double MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::initialMinimumConstraintCutoffRatio() const noexcept
{
	return m_initialMinimumConstraintCutoffRatio;
}

inline double MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::initialIonicBondCutoffRatio() const noexcept
{
	return m_initialIonicBondCutoffRatio;
}

inline double MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::initialMetallicBondCutoffRatio() const noexcept
{
	return m_initialMetallicBondCutoffRatio;
}

inline const ChemToolkit::Generic::ChemicalParameters& MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::chemicalParameters() const noexcept
{
	return m_chemicalParameters;
}

inline const MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::GeometricalConstraintParameters& MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::geometricalConstraintParameters() const noexcept
{
	return m_geometricalConstraintParameters;
}

inline const MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters& MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::structuralOptimizationParameters() const noexcept
{
	return m_structuralOptimizationParameters;
}

inline const MathematicalCrystalChemistry::Design::Recording::StructureDesignRecordOptions& MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::structureDesignRecordOptions() const noexcept
{
	return m_structureDesignRecordOptions;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void MathematicalCrystalChemistry::Runtime::Analysis::CoordinationFeatureSpecifyOptions::readInfeasibleResultNecessity(const std::string& inputText)
{
	if ((inputText == "ON") || (inputText == "On") || (inputText == "on"))
		m_needInfeasibleResult = true;

	else if ((inputText == "OFF") || (inputText == "Off") || (inputText == "off"))
		m_needInfeasibleResult = false;

	else
	{
		std::string errorMessage;
		{
			errorMessage += inputText;
			errorMessage += " is out of definitions.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "readInfeasibleResultNecessity", errorMessage));
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_COORDINATIONFEATURESPECIFYOPTIONS_H
