#ifndef MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_COORDINATIONPOLYHEDRACONSTRAINTS_H
#define MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_COORDINATIONPOLYHEDRACONSTRAINTS_H

#include <stdexcept>
#include <string>

#include "LinearAlgebra.h"
#include "ReportException.h"

#include "AtomCategory.h"
#include "ConstraintComposition.h"
#include "CoordinationFeatureSymbol.h"


namespace MathematicalCrystalChemistry
{
	namespace Analysis
	{
		class CoordinationPolyhedraConstraints
		{
			using size_type = std::size_t;
			using AffineMapping = MathToolkit::Generic::AffineMapping<double>;

			using AtomCategory = MathematicalCrystalChemistry::StructureModel::Features::AtomCategory;
			using ConstraintComposition = MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition;
			using CoordinationFeatureSymbol = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeatureSymbol;
			
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			CoordinationPolyhedraConstraints();
			virtual ~CoordinationPolyhedraConstraints() = default;

			CoordinationPolyhedraConstraints(const CoordinationPolyhedraConstraints&) = default;
			CoordinationPolyhedraConstraints(CoordinationPolyhedraConstraints&&) noexcept = default;
			CoordinationPolyhedraConstraints& operator=(const CoordinationPolyhedraConstraints&) = default;
			CoordinationPolyhedraConstraints& operator=(CoordinationPolyhedraConstraints&&) noexcept = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			void initialize();
			void initializeFromInputFileTexts(const std::string& inputFileTexts);

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			std::string centralElementSymbol() const;
			const CoordinationFeatureSymbol& centralCoordiantionFeatureSymbol() const;
			const std::vector<AtomCategory>& feasibleCentralAtomCategories() const;

			size_type coordinationNumber() const;
			size_type ionicCoordinationNumber() const;
			size_type metallicCoordinationNumber() const;
			const ConstraintComposition& chemicalBondComposition() const;
			const ConstraintComposition& ionicBondComposition() const;
			const ConstraintComposition& metallicBondComposition() const;

			double pointGroupPrecision() const;
			const std::vector<AffineMapping>& symmetryConstraints() const;

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			void readCentralAtomCategories(const std::string& inputTexts);
			void readConstraintComposition(const std::string& inputTexts, const std::string& flag, ConstraintComposition&);
			void readSymmetryConstraints(const std::string& inputTexts);

		// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::string m_centralElementSymbol;
			CoordinationFeatureSymbol m_centralCoordinationFeatureSymbol;
			std::vector<AtomCategory> m_feasibleCentralAtomCategories;

			size_type m_coordinationNumber;
			size_type m_ionicCoordinationNumber;
			size_type m_metallicCoordinationNumber;
			ConstraintComposition m_chemicalBondComposition;
			ConstraintComposition m_ionicBondComposition;
			ConstraintComposition m_metallicBondComposition;

			double m_pointGroupPrecision;
			std::vector<AffineMapping> m_symmetryConstraints;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline std::string MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::centralElementSymbol() const
{
	return m_centralElementSymbol;
}

inline const MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeatureSymbol& MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::centralCoordiantionFeatureSymbol() const
{
	return m_centralCoordinationFeatureSymbol;
}

inline const std::vector<MathematicalCrystalChemistry::StructureModel::Features::AtomCategory>& MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::feasibleCentralAtomCategories() const
{
	return m_feasibleCentralAtomCategories;
}

inline std::size_t MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::coordinationNumber() const
{
	return m_coordinationNumber;
}

inline std::size_t MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::ionicCoordinationNumber() const
{
	return m_ionicCoordinationNumber;
}

inline std::size_t MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::metallicCoordinationNumber() const
{
	return m_metallicCoordinationNumber;
}

inline const MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition& MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::chemicalBondComposition() const
{
	return m_chemicalBondComposition;
}

inline const MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition& MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::ionicBondComposition() const
{
	return m_ionicBondComposition;
}

inline const MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition& MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::metallicBondComposition() const
{
	return m_metallicBondComposition;
}

inline double MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::pointGroupPrecision() const
{
	return m_pointGroupPrecision;
}

inline const std::vector<MathToolkit::Generic::AffineMapping<double>>& MathematicalCrystalChemistry::Analysis::CoordinationPolyhedraConstraints::symmetryConstraints() const
{
	return m_symmetryConstraints;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_COORDINATIONPOLYHEDRACONSTRAINTS_H
