#ifndef MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_INTERNAL_CREATESTRUCTURES_H
#define MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_INTERNAL_CREATESTRUCTURES_H

#include <stdexcept>
#include <string>
#include <filesystem>
#include <memory>
#include <mutex>

#include "ModelAtomCollection.h"
#include "ModelCrystalStructure.h"

#include "InitialStructureGenerator.h"
#include "StructuralOptimizer.h"

#include "StructureCreationReportOptions.h"


namespace MathematicalCrystalChemistry
{
	namespace Runtime
	{
		namespace Creation
		{
			namespace Internal
			{
				class CreateStructures
				{
					using size_type = std::size_t;
					using ModelAtomCollection = MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection;
					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

					using InitialStructureGenerationOptions = MathematicalCrystalChemistry::Design::Generation::InitialStructureGenerationOptions;
					using InitialStructureGenerator = MathematicalCrystalChemistry::Design::Generation::InitialStructureGenerator;
					using StructuralOptimizer = MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor and operators

				public:
					CreateStructures();
					virtual ~CreateStructures() = default;

					CreateStructures(const CreateStructures&) = default;
					CreateStructures(CreateStructures&&) noexcept = default;
					CreateStructures& operator=(const CreateStructures&) = default;
					CreateStructures& operator=(CreateStructures&&) noexcept = default;

				// Constructors, destructor and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void operator()(ModelAtomCollection);

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

					void setStructureCreationReportOptions(const StructureCreationReportOptions&);

					void setInitialStructureGenerator(const InitialStructureGenerator&);
					void setStructuralOptimizer(const StructuralOptimizer& structuralOptimizer);
					void setStructuralOptimizer(StructuralOptimizer&& structuralOptimizer);

					static void initializeStructureDesigning();
					static void setMaxStructureDesigning(const size_type);

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					void createWithRecord(const ModelAtomCollection&) const;
					void createWithoutRecord(const ModelAtomCollection&) const;

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private utility

					bool shouldDesign() const;
					void initializeDesignDirectoryPath() const;

				// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					StructureCreationReportOptions m_structureCreationReportOptions;
					InitialStructureGenerator m_initialStructureGenerator;
					StructuralOptimizer m_structuralOptimizer;

					mutable size_type m_sampleIndex{ 0 };
					mutable std::filesystem::path m_designDirectoryPath{};

					static size_type m_structureDesigning;
					static size_type m_maxStructureDesigning;
					static std::mutex m_designMutex;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::CreateStructures::operator()(ModelAtomCollection modelAtomCollection)
{
	if (m_structureCreationReportOptions.structureDesignRecordOptions().needMdRecord() || m_structureCreationReportOptions.structureDesignRecordOptions().needChemicalBondsRecord())
		createWithRecord(modelAtomCollection);
	else
		createWithoutRecord(modelAtomCollection);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::CreateStructures::setStructureCreationReportOptions(const StructureCreationReportOptions& structureCreationReportOptions)
{
	m_structureCreationReportOptions = structureCreationReportOptions;
}

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::CreateStructures::setInitialStructureGenerator(const InitialStructureGenerator& initialStructureGenerator)
{
	m_initialStructureGenerator = initialStructureGenerator;
}

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::CreateStructures::setStructuralOptimizer(const StructuralOptimizer& structuralOptimizer)
{
	m_structuralOptimizer = structuralOptimizer;
}

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::CreateStructures::setStructuralOptimizer(StructuralOptimizer&& structuralOptimizer)
{
	m_structuralOptimizer = std::move(structuralOptimizer);
}

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::CreateStructures::initializeStructureDesigning()
{
	m_structureDesigning = 0;
}

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::CreateStructures::setMaxStructureDesigning(const size_type num)
{
	m_maxStructureDesigning = num;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

inline bool MathematicalCrystalChemistry::Runtime::Creation::Internal::CreateStructures::shouldDesign() const
{
	std::lock_guard<std::mutex> guard{ m_designMutex };

	if (m_structureDesigning < m_maxStructureDesigning)
	{
		m_sampleIndex = m_structureDesigning;
		++m_structureDesigning;
		return true;
	}

	else
		return false;
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_INTERNAL_CREATESTRUCTURES_H
