#ifndef CHEMTOOLKIT_CRYSTALLOGRAPHY_CRYSTALSTRUCTURE_H
#define CHEMTOOLKIT_CRYSTALLOGRAPHY_CRYSTALSTRUCTURE_H

#include <vector>
#include <algorithm>

#include "LinearAlgebra.h"

#include "ChemicalComposition.h"
#include "UnitCell.h"
#include "SpglibWrapper.h"


namespace ChemToolkit
{
	namespace Crystallography
	{
		template <typename A>
		class CrystalStructure
		{
		public:
			using size_type = std::size_t;
			using Atom = A;
			using AtomicArrangement = std::vector<A>;

			using NumericalVector = MathToolkit::Generic::NumericalVector<double>;
			using NumericalMatrix = MathToolkit::Generic::NumericalMatrix<double>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			CrystalStructure();
			CrystalStructure(const UnitCell& unitCell, const AtomicArrangement& atomicArrangement = AtomicArrangement{});
			CrystalStructure(const UnitCell& unitCell, AtomicArrangement&& atomicArrangement);

			template <typename B>
			CrystalStructure(const CrystalStructure<B>&);

			virtual ~CrystalStructure() = default;

			CrystalStructure(const CrystalStructure&) = default;
			CrystalStructure(CrystalStructure&&) noexcept = default;
			CrystalStructure& operator=(const CrystalStructure&) = default;
			CrystalStructure& operator=(CrystalStructure&&) noexcept = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			const UnitCell& unitCell() const noexcept;
			UnitCell& unitCell() noexcept;
			const AtomicArrangement& atomicArrangement() const noexcept;
			AtomicArrangement& atomicArrangement() noexcept;

			size_type spaceGroupNumber() const noexcept;
			virtual void clearSpaceGroupInfo();

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			virtual void conventionalizeStructure();
			virtual void refineStructure();
			virtual void refineStructure(const double spaceGroupPrecision);
			void updateSpaceGroupInfo();

			virtual void normalizeFractionalCoordinates();

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			virtual bool isValid() const;
			virtual void clear();
			virtual void sort();
			virtual void eraseAtoms(const std::string& elementSymbol);

			ChemToolkit::Generic::ChemicalComposition getChemicalComposition() const;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			UnitCell m_unitCell;
			AtomicArrangement m_atomicArrangement;

			size_type m_spaceGroupNumber;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

template <typename A>
inline ChemToolkit::Crystallography::CrystalStructure<A>::CrystalStructure()
	: m_unitCell{}
	, m_atomicArrangement{}
	, m_spaceGroupNumber{ 1 }
{
}

template <typename A>
inline ChemToolkit::Crystallography::CrystalStructure<A>::CrystalStructure(const UnitCell& unitCell, const AtomicArrangement& atomicArrangement)
	: m_unitCell{ unitCell }
	, m_atomicArrangement{ atomicArrangement }
	, m_spaceGroupNumber{ 1 }
{
}

template <typename A>
inline ChemToolkit::Crystallography::CrystalStructure<A>::CrystalStructure(const UnitCell& unitCell, AtomicArrangement&& atomicArrangement)
	: m_unitCell{ unitCell }
	, m_atomicArrangement{ std::move(atomicArrangement) }
	, m_spaceGroupNumber{ 1 }
{
}

template <typename A>
template <typename B>
inline ChemToolkit::Crystallography::CrystalStructure<A>::CrystalStructure(const CrystalStructure<B>& crystalStructure)
	: m_unitCell{ crystalStructure.unitCell() }
	, m_atomicArrangement{}
	, m_spaceGroupNumber{ crystalStructure.spaceGroupNumber() }
{
	if (crystalStructure.isValid())
	{
		for (const auto& inputAtom : crystalStructure.atomicArrangement())
			m_atomicArrangement.push_back(Atom{ inputAtom.atomicNumber(), inputAtom.coordinate() });
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException(typeid(*this), "constructor", "\"crystalStructure\" is invalid."));
	}
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

template <typename A>
inline const ChemToolkit::Crystallography::UnitCell& ChemToolkit::Crystallography::CrystalStructure<A>::unitCell() const noexcept
{
	return m_unitCell;
}

template <typename A>
inline ChemToolkit::Crystallography::UnitCell& ChemToolkit::Crystallography::CrystalStructure<A>::unitCell() noexcept
{
	return m_unitCell;
}

template <typename A>
inline const typename ChemToolkit::Crystallography::CrystalStructure<A>::AtomicArrangement& ChemToolkit::Crystallography::CrystalStructure<A>::atomicArrangement() const noexcept
{
	return m_atomicArrangement;
}

template <typename A>
inline typename ChemToolkit::Crystallography::CrystalStructure<A>::AtomicArrangement& ChemToolkit::Crystallography::CrystalStructure<A>::atomicArrangement() noexcept
{
	return m_atomicArrangement;
}

template <typename A>
inline std::size_t ChemToolkit::Crystallography::CrystalStructure<A>::spaceGroupNumber() const noexcept
{
	return m_spaceGroupNumber;
}

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::clearSpaceGroupInfo()
{
	m_spaceGroupNumber = 1;

	for (auto& atom : atomicArrangement())
		atom.setSiteSymmetrySymbol("");
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::conventionalizeStructure()
{
	ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper spglibWrapper;
	m_spaceGroupNumber = spglibWrapper.findConventionalStructure(m_unitCell, m_atomicArrangement);
}

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::refineStructure()
{
	ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper spglibWrapper;
	spglibWrapper.findPrimitiveStructure(m_unitCell, m_atomicArrangement);
}

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::refineStructure(const double spaceGroupPrecision)
{
	ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper spglibWrapper;
	spglibWrapper.findPrimitiveStructure(m_unitCell, m_atomicArrangement, spaceGroupPrecision);
}

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::updateSpaceGroupInfo()
{
	ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper spglibWrapper;
	m_spaceGroupNumber = spglibWrapper.getSpaceGroupInformation(m_unitCell, m_atomicArrangement);
}

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::normalizeFractionalCoordinates()
{
	for (auto& atom : m_atomicArrangement)
	{
		NumericalVector fractionalCoordinate = m_unitCell.inverseBasisVectors() * atom.coordinate();
		int floorValueA = static_cast<int>(std::floor(fractionalCoordinate[0]));
		int floorValueB = static_cast<int>(std::floor(fractionalCoordinate[1]));
		int floorValueC = static_cast<int>(std::floor(fractionalCoordinate[2]));

		if (floorValueA != 0)
			fractionalCoordinate[0] -= static_cast<double>(floorValueA);
		if (floorValueB != 0)
			fractionalCoordinate[1] -= static_cast<double>(floorValueB);
		if (floorValueC != 0)
			fractionalCoordinate[2] -= static_cast<double>(floorValueC);

		atom.coordinate() = m_unitCell.basisVectors() * fractionalCoordinate;
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

template <typename A>
inline bool ChemToolkit::Crystallography::CrystalStructure<A>::isValid() const
{
	if (!(m_unitCell.isValid()))
		return false;
	if (m_atomicArrangement.empty())
		return false;

	return true;
}

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::clear()
{
	m_unitCell.clear();
	m_atomicArrangement.clear();
}

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::sort()
{
	std::sort(m_atomicArrangement.begin(), m_atomicArrangement.end());
}

template <typename A>
inline void ChemToolkit::Crystallography::CrystalStructure<A>::eraseAtoms(const std::string& elementSymbol)
{
	for (auto iter = m_atomicArrangement.begin(); iter != m_atomicArrangement.end();)
	{
		if (iter->elementSymbol() == elementSymbol)
			iter = m_atomicArrangement.erase(iter);
		else
			++iter;
	}
}

template <typename A>
inline ChemToolkit::Generic::ChemicalComposition ChemToolkit::Crystallography::CrystalStructure<A>::getChemicalComposition() const
{
	ChemToolkit::Generic::ChemicalComposition chemicalComposition;
	{
		for (const auto& atom : m_atomicArrangement)
			chemicalComposition.add(atom.elementSymbol(), 1);
	}

	return chemicalComposition;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CHEMTOOLKIT_CRYSTALLOGRAPHY_CRYSTALSTRUCTURE_H
