#ifndef CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_CRYSTALLOGRAPHICINFORMATION_H
#define CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_CRYSTALLOGRAPHICINFORMATION_H

#include <utility>
#include <vector>

#include "UnitCell.h"
#include "CrystalStructure.h"

#include "AtomSite.h"


namespace ChemToolkit
{
	namespace Crystallography
	{
		namespace SpaceGroups
		{
			class CrystallographicInformation final
			{
				using size_type = std::size_t;
				using NumericalVector = MathToolkit::Generic::NumericalVector<double>;
				using AffineMapping = MathToolkit::Generic::AffineMapping<double>;
				using UnitCell = ChemToolkit::Crystallography::UnitCell;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				CrystallographicInformation();
				~CrystallographicInformation() = default;

				CrystallographicInformation(const CrystallographicInformation&) = default;
				CrystallographicInformation(CrystallographicInformation&&) noexcept = default;
				CrystallographicInformation& operator=(const CrystallographicInformation&) = default;
				CrystallographicInformation& operator=(CrystallographicInformation&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				const UnitCell& unitCell() const;
				const std::vector<AtomSite>& atomSites() const;
				const std::vector<AffineMapping>& symmetryOperations() const;

				const std::string& structureType() const;
				size_type spaceGroupNumber() const;

				void setUnitCell(const UnitCell&);
				void setAtomSites(const std::vector<AtomSite>&);
				void setAtomSites(std::vector<AtomSite>&&);
				void setSymmetryOperations(const std::vector<AffineMapping>&);
				void setSymmetryOperations(std::vector<AffineMapping>&&);

				void setStructureType(const std::string&);
				void setSpaceGroupNumber(const size_type);

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				bool hasPartiallyFilledAtomSite() const;
				std::string getSpaceGroupShortSymbol() const;

				bool isValid() const;
				void clear();

				template <typename A>
				void toCrystalStructure(CrystalStructure<A>&) const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				bool isSamePosition(const NumericalVector&, const NumericalVector&) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				UnitCell m_unitCell;
				std::vector<AtomSite> m_atomSites;
				std::vector<AffineMapping> m_symmetryOperations;

				std::string m_structureType;
				size_type m_spaceGroupNumber;

				double m_positionPrecision;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const ChemToolkit::Crystallography::UnitCell& ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::unitCell() const
{
	return m_unitCell;
}

inline const std::vector<ChemToolkit::Crystallography::SpaceGroups::AtomSite>& ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::atomSites() const
{
	return m_atomSites;
}

inline const std::vector<MathToolkit::Generic::AffineMapping<double>>& ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::symmetryOperations() const
{
	return m_symmetryOperations;
}

inline const std::string& ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::structureType() const
{
	return m_structureType;
}

inline std::size_t ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::spaceGroupNumber() const
{
	return m_spaceGroupNumber;
}

inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::setUnitCell(const UnitCell& unitCell)
{
	m_unitCell = unitCell;
}

inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::setAtomSites(const std::vector<AtomSite>& atomSites)
{
	m_atomSites = atomSites;
}

inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::setAtomSites(std::vector<AtomSite>&& atomSites)
{
	m_atomSites = std::move(atomSites);
}

inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::setSymmetryOperations(const std::vector<AffineMapping>& symmetryOperations)
{
	for (const auto& symmetryOperation : symmetryOperations)
	{
		if (symmetryOperation.matrixPart().columnSize() != 3 || symmetryOperation.matrixPart().rowSize() != 3)
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "setSymmetryOperations", "\"symmetryOperations\" contains an invalid linear mapping."));
		}
	}

	m_symmetryOperations = symmetryOperations;
}

inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::setSymmetryOperations(std::vector<AffineMapping>&& symmetryOperations)
{
	for (const auto& symmetryOperation : symmetryOperations)
	{
		if (symmetryOperation.matrixPart().columnSize() != 3 || symmetryOperation.matrixPart().rowSize() != 3)
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "setSymmetryOperations", "\"symmetryOperations\" contains an invalid linear mapping."));
		}
	}

	m_symmetryOperations = std::move(symmetryOperations);
}

inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::setStructureType(const std::string& name)
{
	if (name.empty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setStructureType", "\"structureType\" is empty."));
	}

	else
		m_structureType = name;
}

inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::setSpaceGroupNumber(const size_type number)
{
	if ((0 < number) && (number < 231))
		m_spaceGroupNumber = number;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setSpaceGroupNumber", "Space group number is out of definitions."));
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::hasPartiallyFilledAtomSite() const
{
	for (const auto& atomSite : m_atomSites)
	{
		if (atomSite.occupancy() < 1.0)
			return true;
	}

	return false;
}

inline bool ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::isValid() const
{
	if (!(m_unitCell.isValid()))
		return false;
	if (m_atomSites.empty())
		return false;
	if (m_symmetryOperations.empty())
		return false;

	return true;
}

inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::clear()
{
	m_unitCell.clear();
	m_atomSites.clear();
	m_symmetryOperations.clear();

	m_structureType.clear();
	m_spaceGroupNumber = 0;
}

template <typename A>
inline void ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::toCrystalStructure(CrystalStructure<A>& crystalStructure) const
{
	using Atom = A;
	crystalStructure.unitCell() = m_unitCell;
	crystalStructure.atomicArrangement().clear();
	{
		for (const auto& atomSite : m_atomSites)
		{
			std::vector<NumericalVector> registeredCartesianCoordinates;
			{
				for (const auto& symmetryOperation : m_symmetryOperations)
				{
					NumericalVector cartesianCoordinate;
					{
						NumericalVector fractionalCoordinate = symmetryOperation.apply(atomSite.fractionalCoordinate());
						fractionalCoordinate[0] -= std::floor(fractionalCoordinate[0]);
						fractionalCoordinate[1] -= std::floor(fractionalCoordinate[1]);
						fractionalCoordinate[2] -= std::floor(fractionalCoordinate[2]);

						cartesianCoordinate = m_unitCell.basisVectors() * fractionalCoordinate;
					}

					bool isRegisteredCoordinate = false;
					{
						for (const auto& registeredCartesianCoordinate : registeredCartesianCoordinates)
						{
							if (isSamePosition(registeredCartesianCoordinate, cartesianCoordinate))
							{
								isRegisteredCoordinate = true;
								break;
							}
						}
					}

					if (!isRegisteredCoordinate)
						registeredCartesianCoordinates.push_back(std::move(cartesianCoordinate));
				}
			}

			for (const auto& cartesianCoordinate : registeredCartesianCoordinates)
				crystalStructure.atomicArrangement().push_back(Atom{ atomSite.elementSymbol(), cartesianCoordinate });
		}
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation::isSamePosition(const NumericalVector& nva, const NumericalVector& nvb) const
{
	if ((nva - nvb).normSquare() < (m_positionPrecision * m_positionPrecision))
		return true;
	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_CRYSTALLOGRAPHICINFORMATION_H
