#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_INTERNAL_DYNAMICSRECORDER_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_INTERNAL_DYNAMICSRECORDER_H

#include <stdexcept>
#include <string>
#include <filesystem>

#include "ReportException.h"

#include "MdFileWriter.h"
#include "ModelCrystalStructure.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Recording
		{
			class StructureDesignRecorder;


			namespace Internal
			{
				class DynamicsRecorder
				{
					friend StructureDesignRecorder;

					using size_type = std::size_t;
					using MdFileWriter = OpenMX::FileIO::MdFileWriter;
					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				private:
					DynamicsRecorder();

				public:
					virtual ~DynamicsRecorder() = default;

					DynamicsRecorder(const DynamicsRecorder&) = default;
					DynamicsRecorder(DynamicsRecorder&&) noexcept = default;
					DynamicsRecorder& operator=(const DynamicsRecorder&) = default;
					DynamicsRecorder& operator=(DynamicsRecorder&&) noexcept = default;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

					void initialize();

					bool needRecord() const;
					size_type recordFrequency() const noexcept;

					void setRecordFrequency(const size_type recordFrequency = 0);

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void record(const ModelCrystalStructure&);

					void createFile(const std::filesystem::path& workDirectoryPath, const std::string& crystalName);
					void moveFile(const std::filesystem::path& newDirectoryPath);
					void releaseFile();

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					size_type m_recordFrequency;
					MdFileWriter m_mdFileWriter;

					mutable size_type m_recordTimer{ 0 };
					mutable size_type m_structureOptimizing{ 0 };
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

inline MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::DynamicsRecorder()
	: m_recordFrequency{ 0 }
	, m_mdFileWriter{}
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::initialize()
{
	m_recordFrequency = 0;
	m_mdFileWriter.releaseFile();

	m_recordTimer = 0;
	m_structureOptimizing = 0;
}

inline bool MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::needRecord() const
{
	return static_cast<bool>(m_recordFrequency);
}

inline std::size_t MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::recordFrequency() const noexcept
{
	return m_recordFrequency;
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::setRecordFrequency(const size_type mdRecordFrequency)
{
	m_recordFrequency = mdRecordFrequency;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::record(const ModelCrystalStructure& modelCrystalStructure)
{
	++m_recordTimer;
	++m_structureOptimizing;

	if (m_recordTimer == m_recordFrequency)
	{
		m_mdFileWriter.drawFrame(modelCrystalStructure, m_structureOptimizing);
		m_recordTimer = 0;
	}
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::createFile(const std::filesystem::path& workDirectoryPath, const std::string& crystalName)
{
	std::filesystem::path mdFilePath = workDirectoryPath;
	{
		std::string filename = crystalName;
		filename += ".md";
		mdFilePath /= filename;
	}

	m_mdFileWriter = OpenMX::FileIO::MdFileWriter{ mdFilePath };
	m_mdFileWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing);
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::moveFile(const std::filesystem::path& newDirectoryPath)
{
	if (m_mdFileWriter.isValid())
		m_mdFileWriter.moveDirectory(newDirectoryPath, OperatingSystem::IO::FileRenameOptions::overwrite_existing);
}

inline void MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder::releaseFile()
{
	m_mdFileWriter.releaseFile();
	m_recordTimer = 0;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_INTERNAL_DYNAMICSRECORDER_H
