#ifndef OPERATINGSYSTEM_IO_FILEBASE_H
#define OPERATINGSYSTEM_IO_FILEBASE_H

#include <string>
#include <filesystem>
#include <mutex>

#include "ReportException.h"
#include "ReportException.h"


namespace OperatingSystem
{
	namespace IO
	{
		enum class FileCopyOptions
		{
			none,
			skip_existing,
			overwrite_existing,
			update_existing
		};

		enum class FileRenameOptions
		{
			none,
			skip_existing,
			overwrite_existing,
		};


		class FileBase
		{
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		protected:
			FileBase() = default;
			explicit FileBase(const std::filesystem::path& filePath);
			explicit FileBase(std::filesystem::path&& filePath);

		public:
			FileBase(const FileBase&) = default;
			FileBase(FileBase&&) noexcept = default;
			FileBase& operator=(const FileBase&) = default;
			FileBase& operator=(FileBase&&) noexcept = default;

			virtual ~FileBase() = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			std::string filename() const;
			const std::filesystem::path& filePath() const;
			virtual void releaseFile();

			virtual bool isValid() const;
			explicit operator bool() const;

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods on file system

			void copyFile(const std::filesystem::path& copyFilePath, FileCopyOptions) const;
			void renameFile(const std::filesystem::path& newFilePath, FileRenameOptions);
			void moveDirectory(const std::filesystem::path& directoryPath, FileRenameOptions);
			void removeFile() const;

		// Public methods on file system
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Protected methods

		protected:
			bool isValidFilePath(const std::filesystem::path& filePath) const;

		// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::filesystem::path m_filePath;

			static std::mutex m_filesystemMutex;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline std::string OperatingSystem::IO::FileBase::filename() const
{
	return m_filePath.filename().string();
}

inline const std::filesystem::path& OperatingSystem::IO::FileBase::filePath() const
{
	return m_filePath;
}

inline void OperatingSystem::IO::FileBase::releaseFile()
{
	m_filePath.clear();
}

inline bool OperatingSystem::IO::FileBase::isValid() const
{
	return std::filesystem::is_regular_file(m_filePath);
}

inline OperatingSystem::IO::FileBase::operator bool() const
{
	return isValid();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods on file system

inline void OperatingSystem::IO::FileBase::moveDirectory(const std::filesystem::path& directoryPath, FileRenameOptions option)
{
	if (std::filesystem::is_directory(directoryPath))
	{
		std::filesystem::path newFilePath = directoryPath;
		newFilePath /= m_filePath.filename();
		renameFile(newFilePath, option);
	}

	else
	{
		std::string errorMessage;
		{
			errorMessage += "\"directoryPath\" of \"";
			errorMessage += directoryPath.generic_string();
			errorMessage += "\" is not a path to a directory.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "moveDirectory", errorMessage));
	}
}

inline void OperatingSystem::IO::FileBase::removeFile() const
{
	std::lock_guard<std::mutex> guard{ m_filesystemMutex };
	std::filesystem::remove_all(m_filePath);
}

// Public methods on file system
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline bool OperatingSystem::IO::FileBase::isValidFilePath(const std::filesystem::path& filePath) const
{
	if (filePath.has_filename())
		return std::filesystem::is_directory(filePath.parent_path());
	else
		return false;
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !OPERATINGSYSTEM_IO_FILEBASE_H
