#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_INTERNAL_GEOMETRICALBARRIERHANDLER_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_INTERNAL_GEOMETRICALBARRIERHANDLER_H

#include <array>
#include <vector>

#include "ModelAtom.h"
#include "ModelCrystalStructure.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Optimization
		{
			namespace Internal
			{
				class GeometricalBarrierHandler
				{
					using size_type = std::size_t;
					using LatticePoint = std::array<int, 3>;
					using ConstraintKey = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey;

					using NumericalVector = MathToolkit::Generic::NumericalVector<double>;
					using ModelAtom = MathematicalCrystalChemistry::StructureModel::Components::ModelAtom;
					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;


					struct SimpleMinimumBarrierKey
					{
						size_type atomIndex;
						size_type ligandIndex;
						double minimumDistance;
					};

					struct MinimumBarrierKey
					{
						size_type atomIndex;
						size_type ligandIndex;
						LatticePoint latticePoint;
						double minimumDistance;
					};

					struct SimpleMinMaxBarrierKey
					{
						size_type atomIndex;
						size_type ligandIndex;
						double minimumDistance;
						double maximumDistance;
					};

					struct MinMaxBarrierKey
					{
						size_type atomIndex;
						size_type ligandIndex;
						LatticePoint latticePoint;
						double minimumDistance;
						double maximumDistance;
					};

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				protected:
					GeometricalBarrierHandler();

				public:
					virtual ~GeometricalBarrierHandler() = default;

					GeometricalBarrierHandler(const GeometricalBarrierHandler&) = default;
					GeometricalBarrierHandler(GeometricalBarrierHandler&&) noexcept = default;
					GeometricalBarrierHandler& operator=(const GeometricalBarrierHandler&) = default;
					GeometricalBarrierHandler& operator=(GeometricalBarrierHandler&&) noexcept = default;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

					const std::vector<SimpleMinimumBarrierKey>& simpleMinimumBarrierKeys() const noexcept;
					const std::vector<MinimumBarrierKey>& minimumBarrierKeys() const noexcept;
					const std::vector<SimpleMinMaxBarrierKey>& simpleMinMaxBarrierKeys() const noexcept;
					const std::vector<MinMaxBarrierKey>& minMaxBarrierKeys() const noexcept;

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					virtual void initialize() noexcept;
					virtual void initialize(const ModelCrystalStructure&);

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					template <typename T>
					void setMinimumBarrierKey(const std::unordered_map<ConstraintKey, T, ConstraintKey::Hasher>&);

					template <typename T>
					void setMinMaxBarrierKey(const std::unordered_map<ConstraintKey, T, ConstraintKey::Hasher>&);

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					std::vector<SimpleMinimumBarrierKey> m_simpleMinimumBarrierKeys;
					std::vector<MinimumBarrierKey> m_minimumBarrierKeys;
					std::vector<SimpleMinMaxBarrierKey> m_simpleMinMaxBarrierKeys;
					std::vector<MinMaxBarrierKey> m_minMaxBarrierKeys;

					std::unordered_map<const ModelAtom*, size_type> m_modelAtomIndicesDictionary;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const std::vector<MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::SimpleMinimumBarrierKey>& MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::simpleMinimumBarrierKeys() const noexcept
{
	return m_simpleMinimumBarrierKeys;
}

inline const std::vector<MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::MinimumBarrierKey>& MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::minimumBarrierKeys() const noexcept
{
	return m_minimumBarrierKeys;
}

inline const std::vector<MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::SimpleMinMaxBarrierKey>& MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::simpleMinMaxBarrierKeys() const noexcept
{
	return m_simpleMinMaxBarrierKeys;
}

inline const std::vector<MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::MinMaxBarrierKey>& MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::minMaxBarrierKeys() const noexcept
{
	return m_minMaxBarrierKeys;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::initialize() noexcept
{
	m_simpleMinimumBarrierKeys.clear();
	m_minimumBarrierKeys.clear();
	m_simpleMinMaxBarrierKeys.clear();
	m_minMaxBarrierKeys.clear();

	m_modelAtomIndicesDictionary.clear();
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::initialize(const ModelCrystalStructure& modelCrystalStructure)
{
	initialize();

	for (size_type index = 0; index < modelCrystalStructure.atomicArrangement().size(); ++index)
		m_modelAtomIndicesDictionary.emplace(&(modelCrystalStructure.atomicArrangement()[index]), index);

	setMinMaxBarrierKey(modelCrystalStructure.ionicBondOwners());
	setMinMaxBarrierKey(modelCrystalStructure.metallicBondOwners());

	setMinimumBarrierKey(modelCrystalStructure.anionicConstraintOwners());
	setMinimumBarrierKey(modelCrystalStructure.cationicConstraintOwners());
	setMinimumBarrierKey(modelCrystalStructure.nonIonicBondOwners());
	setMinimumBarrierKey(modelCrystalStructure.nonMetallicBondOwners());
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename T>
inline void MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::setMinimumBarrierKey(const std::unordered_map<ConstraintKey, T, ConstraintKey::Hasher>& constraintOwners)
{
	for (const auto& keyAndConstraintOwner : constraintOwners)
	{
		size_type atomIndex = m_modelAtomIndicesDictionary.at(&(keyAndConstraintOwner.first.modelAtomRef()));
		size_type ligandIndex = m_modelAtomIndicesDictionary.at(&(keyAndConstraintOwner.first.ligand().get()));
		double minimumDistance = keyAndConstraintOwner.second.constraintPtr()->minimumDistance();


		if (keyAndConstraintOwner.first.ligand().hasOriginalLatticePoint())
			m_simpleMinimumBarrierKeys.push_back(SimpleMinimumBarrierKey{ atomIndex, ligandIndex, minimumDistance });

		else
		{
			LatticePoint latticePoint;
			{
				auto ligandLatticePoint = keyAndConstraintOwner.first.ligand().getLatticePoint();

				latticePoint[0] = ligandLatticePoint[0];
				latticePoint[1] = ligandLatticePoint[1];
				latticePoint[2] = ligandLatticePoint[2];
			}

			m_minimumBarrierKeys.push_back(MinimumBarrierKey{ atomIndex, ligandIndex, latticePoint, minimumDistance });
		}
	}
}

template <typename T>
inline void MathematicalCrystalChemistry::Design::Optimization::Internal::GeometricalBarrierHandler::setMinMaxBarrierKey(const std::unordered_map<ConstraintKey, T, ConstraintKey::Hasher>& constraintOwners)
{
	for (const auto& keyAndConstraintOwner : constraintOwners)
	{
		size_type atomIndex = m_modelAtomIndicesDictionary.at(&(keyAndConstraintOwner.first.modelAtomRef()));
		size_type ligandIndex = m_modelAtomIndicesDictionary.at(&(keyAndConstraintOwner.first.ligand().get()));
		double minimumDistance = keyAndConstraintOwner.second.constraintPtr()->minimumDistance();
		double maximumDistance = keyAndConstraintOwner.second.constraintPtr()->maximumDistance();


		if (keyAndConstraintOwner.first.ligand().hasOriginalLatticePoint())
			m_simpleMinMaxBarrierKeys.push_back(SimpleMinMaxBarrierKey{ atomIndex, ligandIndex, minimumDistance, maximumDistance });

		else
		{
			LatticePoint latticePoint;
			{
				auto ligandLatticePoint = keyAndConstraintOwner.first.ligand().getLatticePoint();

				latticePoint[0] = ligandLatticePoint[0];
				latticePoint[1] = ligandLatticePoint[1];
				latticePoint[2] = ligandLatticePoint[2];
			}

			m_minMaxBarrierKeys.push_back(MinMaxBarrierKey{ atomIndex, ligandIndex, latticePoint, minimumDistance, maximumDistance });
		}
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_INTERNAL_GEOMETRICALBARRIERHANDLER_H
