#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_IONICBOND_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_IONICBOND_H

#include "ReportException.h"

#include "ConstraintOwner.h"
#include "MinMaxDistanceConstraint.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			class IonicBond :public Internal::MinMaxDistanceConstraint
			{
				using IonicBondOwner = ConstraintOwner<IonicBond>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private constructors and public destructor

			private:
				IonicBond(ModelAtom&, const Ligand&);
				IonicBond(ModelAtom&, Ligand&&);
				IonicBond(const Ligand&, const Ligand&);
				explicit IonicBond(const std::shared_ptr<IonicBond>& reversePtr);

			public:
				virtual ~IonicBond() = default;

			// Private constructors and public destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static IonicBondOwner create(const ConstraintKey&);
				static bool shouldCreate(const ModelAtom&, const Ligand&, const double cutoffRatio);
				static bool shouldCreate(const ConstraintKey&, const double cutoffRatio);

				static double defaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio(const double cutoffRatio);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				static bool canCreate(const ModelAtom&, const Ligand&);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				static double m_defaultCutoffRatio;


			private:
				IonicBond() = delete;
				IonicBond(const IonicBond&) = delete;
				IonicBond(IonicBond&&) noexcept = delete;
				IonicBond& operator=(const IonicBond&) = delete;
				IonicBond& operator=(IonicBond&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond::IonicBondOwner MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond::create(const ConstraintKey& constraintKey)
{
	ModelAtom& modelAtom = constraintKey.modelAtomRef();
	const Ligand& ligand = constraintKey.ligand();

	if (canCreate(modelAtom, ligand))
	{
		std::shared_ptr<IonicBond> ionicBondPtr{ new IonicBond(modelAtom, ligand) };
		std::shared_ptr<IonicBond> reverseIonicBondPtr{ new IonicBond(ionicBondPtr) };
		ionicBondPtr->setReversePtr(std::static_pointer_cast<InteratomicConstraintBase>(reverseIonicBondPtr));

		modelAtom.track(ionicBondPtr);
		ligand.get().track(reverseIonicBondPtr);

		return IonicBondOwner{ std::move(ionicBondPtr), std::move(reverseIonicBondPtr) };
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException("create of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond", "Could not create ionic bond."));
	}
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond::shouldCreate(const ModelAtom& modelAtom, const Ligand& ligand, const double cutoffRatio)
{
	if (canCreate(modelAtom, ligand))
	{
		NumericalVector displacement = ligand.coordinate() - modelAtom.coordinate();
		double cutoffLength = cutoffRatio * (modelAtom.coordinationFeature().upperIonicBondRadius() + ligand.get().coordinationFeature().upperIonicBondRadius());

		if (displacement.normSquare() < (cutoffLength * cutoffLength))
			return true;
		else
			return false;
	}

	else
		return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond::shouldCreate(const ConstraintKey& constraintKey, const double cutoffRatio)
{
	return shouldCreate(constraintKey.modelAtomRef(), constraintKey.ligand(), cutoffRatio);
}

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond::defaultCutoffRatio() noexcept
{
	return m_defaultCutoffRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond::setDefaultCutoffRatio() noexcept
{
	m_defaultCutoffRatio = 2.0;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond::setDefaultCutoffRatio(const double val)
{
	if (0.0 < val)
		m_defaultCutoffRatio = val;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("setDefaultCutoffRatio of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond", "\"defaultCutoffRatio\" is not more than zero"));
	}
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond::canCreate(const ModelAtom& modelAtom, const Ligand& ligand)
{
	if ((modelAtom.coordinationFeature().atomCategory().isAnion() && ligand.get().coordinationFeature().atomCategory().isCation()) || (modelAtom.coordinationFeature().atomCategory().isCation() && ligand.get().coordinationFeature().atomCategory().isAnion()))
		return true;
	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_IONICBOND_H
