#ifndef MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_INTERNAL_ISFEASIBLECHEMICALCOMPOSITION_H
#define MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_INTERNAL_ISFEASIBLECHEMICALCOMPOSITION_H

#include "ChemicalCompositionConstraints.h"

#include "ModelCrystalStructure.h"


namespace MathematicalCrystalChemistry
{
	namespace Analysis
	{
		class CrystalAnalyzer;


		namespace Internal
		{
			class IsFeasibleChemicalComposition
			{
				friend CrystalAnalyzer;

				using size_type = std::size_t;
				using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			private:
				IsFeasibleChemicalComposition() = default;

			public:
				virtual ~IsFeasibleChemicalComposition() = default;

				IsFeasibleChemicalComposition(const IsFeasibleChemicalComposition&) = default;
				IsFeasibleChemicalComposition(IsFeasibleChemicalComposition&&) noexcept = default;
				IsFeasibleChemicalComposition& operator=(const IsFeasibleChemicalComposition&) = default;
				IsFeasibleChemicalComposition& operator=(IsFeasibleChemicalComposition&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				bool operator()(const ModelCrystalStructure&) const;

				void setChemicalCompositionConstraints(const ChemicalCompositionConstraints&);
				void setChemicalCompositionConstraints(ChemicalCompositionConstraints&&);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				bool hasFeasibleNumAtoms(const ModelCrystalStructure&) const;
				bool belongFeasibleCrystalCategory(const ModelCrystalStructure&) const;
				bool hasNecessaryChemicalElements(const ModelCrystalStructure&) const;

				bool hasInfeasibleChemicalElements(const ModelCrystalStructure&) const;
				bool hasInfeasibleAtomCategories(const ModelCrystalStructure&) const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				ChemicalCompositionConstraints m_chemicalCompositionConstraints;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleChemicalComposition::operator()(const ModelCrystalStructure& modelCrystalStructure) const
{
	if(!(hasFeasibleNumAtoms(modelCrystalStructure)))
		return false;
	if (!(belongFeasibleCrystalCategory(modelCrystalStructure)))
		return false;
	if (!(hasNecessaryChemicalElements(modelCrystalStructure)))
		return false;

	if (hasInfeasibleChemicalElements(modelCrystalStructure))
		return false;
	if (hasInfeasibleAtomCategories(modelCrystalStructure))
		return false;

	return true;
}

inline void MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleChemicalComposition::setChemicalCompositionConstraints(const ChemicalCompositionConstraints& chemicalCompositionConstraints)
{
	m_chemicalCompositionConstraints = chemicalCompositionConstraints;
}

inline void MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleChemicalComposition::setChemicalCompositionConstraints(ChemicalCompositionConstraints&& chemicalCompositionConstraints)
{
	m_chemicalCompositionConstraints = std::move(chemicalCompositionConstraints);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleChemicalComposition::hasFeasibleNumAtoms(const ModelCrystalStructure& modelCrystalStructure) const
{
	if (modelCrystalStructure.atomicArrangement().size() < m_chemicalCompositionConstraints.minNumAtoms())
		return false;
	else
	{
		if (m_chemicalCompositionConstraints.maxNumAtoms() < modelCrystalStructure.atomicArrangement().size())
			return false;
		else
			return true;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleChemicalComposition::belongFeasibleCrystalCategory(const ModelCrystalStructure& modelCrystalStructure) const
{
	if (m_chemicalCompositionConstraints.feasibleCrystalCategories().empty())
		return true;
	else
	{
		MathematicalCrystalChemistry::StructureModel::Features::CrystalCategory crystalCategory = modelCrystalStructure.getCrystalCategory();

		for (const auto& feasibleCrystalCategory : m_chemicalCompositionConstraints.feasibleCrystalCategories())
		{
			if (crystalCategory == feasibleCrystalCategory)
				return true;
		}

		return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleChemicalComposition::hasNecessaryChemicalElements(const ModelCrystalStructure& modelCrystalStructure) const
{
	if (m_chemicalCompositionConstraints.necessaryElementSymbols().empty())
		return true;
	else
	{
		for (const auto& necessaryElementSymbol : m_chemicalCompositionConstraints.necessaryElementSymbols())
		{
			bool hasNecessaryElementSymbol = false;
			{
				for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
				{
					if (modelAtom.elementSymbol() == necessaryElementSymbol)
					{
						hasNecessaryElementSymbol = true;
						break;
					}
				}
			}

			if (!hasNecessaryElementSymbol)
				return false;
		}

		return true;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleChemicalComposition::hasInfeasibleChemicalElements(const ModelCrystalStructure& modelCrystalStructure) const
{
	for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
	{
		for (const auto& infeasibleElementSymbol : m_chemicalCompositionConstraints.infeasibleElementSymbols())
		{
			if (modelAtom.elementSymbol() == infeasibleElementSymbol)
				return true;
		}
	}

	return false;
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleChemicalComposition::hasInfeasibleAtomCategories(const ModelCrystalStructure& modelCrystalStructure) const
{
	if (m_chemicalCompositionConstraints.feasibleAtomCategories().empty())
		return false;
	else
	{
		for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
		{
			bool belongFeasibleAtomCategory = false;
			{
				for (const auto& feasibleAtomCategory : m_chemicalCompositionConstraints.feasibleAtomCategories())
				{
					if (modelAtom.coordinationFeature().atomCategory() == feasibleAtomCategory)
					{
						belongFeasibleAtomCategory = true;
						break;
					}
				}
			}

			if (!belongFeasibleAtomCategory)
				return true;
		}

		return false;
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_INTERNAL_ISFEASIBLECHEMICALCOMPOSITION_H
