#ifndef MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_INTERNAL_ISFEASIBLECOORDINATIONPOLYHEDRA_H
#define MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_INTERNAL_ISFEASIBLECOORDINATIONPOLYHEDRA_H

#include <limits>

#include "ModelCrystalStructure.h"

#include "CoordinationPolyhedraConstraints.h"


namespace MathematicalCrystalChemistry
{
	namespace Analysis
	{
		class CrystalAnalyzer;


		namespace Internal
		{
			class IsFeasibleCoordinationPolyhedra
			{
				friend CrystalAnalyzer;

				using size_type = std::size_t;
				using AffineMapping = MathToolkit::Generic::AffineMapping<double>;
				using ModelAtom = MathematicalCrystalChemistry::StructureModel::Components::ModelAtom;
				using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			private:
				IsFeasibleCoordinationPolyhedra() = default;

			public:
				virtual ~IsFeasibleCoordinationPolyhedra() = default;

				IsFeasibleCoordinationPolyhedra(const IsFeasibleCoordinationPolyhedra&) = default;
				IsFeasibleCoordinationPolyhedra(IsFeasibleCoordinationPolyhedra&&) noexcept = default;
				IsFeasibleCoordinationPolyhedra& operator=(const IsFeasibleCoordinationPolyhedra&) = default;
				IsFeasibleCoordinationPolyhedra& operator=(IsFeasibleCoordinationPolyhedra&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				bool operator()(const ModelCrystalStructure&) const;

				void setCoordinationPolyhedraConstraints(const CoordinationPolyhedraConstraints&);
				void setCoordinationPolyhedraConstraints(CoordinationPolyhedraConstraints&&);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				bool meetCentralAtomSymbol(const ModelAtom&) const;
				bool meetCoordinationFeatureSymbol(const ModelAtom&) const;
				bool isFeasibleCentralAtomCategory(const ModelAtom&) const;

				bool isFeasibleCoordinationNumber(const ModelAtom&) const;
				bool isFeasibleIonicCoordinationNumber(const ModelAtom&) const;
				bool isFeasibleMetallicCoordinationNumber(const ModelAtom&) const;

				bool isFeasibleChemicalBondComposition(const ModelAtom&) const;
				bool isFeasibleIonicBondComposition(const ModelAtom&) const;
				bool isFeasibleMetallicBondComposition(const ModelAtom&) const;

				bool satisfySymmetryConstraint(const ModelAtom&) const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				CoordinationPolyhedraConstraints m_coordinationPolyhedraConstraints;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::operator()(const ModelCrystalStructure& modelCrystalStructure) const
{
	for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
	{
		if (!(meetCentralAtomSymbol(modelAtom)))
			continue;
		if (!(meetCoordinationFeatureSymbol(modelAtom)))
			continue;
		if (!(isFeasibleCentralAtomCategory(modelAtom)))
			continue;

		if (!(isFeasibleCoordinationNumber(modelAtom)))
			continue;
		if (!(isFeasibleIonicCoordinationNumber(modelAtom)))
			continue;
		if (!(isFeasibleMetallicCoordinationNumber(modelAtom)))
			continue;

		if (!(isFeasibleChemicalBondComposition(modelAtom)))
			continue;
		if (!(isFeasibleIonicBondComposition(modelAtom)))
			continue;
		if (!(isFeasibleMetallicBondComposition(modelAtom)))
			continue;

		if (!(satisfySymmetryConstraint(modelAtom)))
			continue;

		return true;
	}

	return false;
}

inline void MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::setCoordinationPolyhedraConstraints(const CoordinationPolyhedraConstraints& coordinationPolyhedraConstraints)
{
	m_coordinationPolyhedraConstraints = coordinationPolyhedraConstraints;
}

inline void MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::setCoordinationPolyhedraConstraints(CoordinationPolyhedraConstraints&& coordinationPolyhedraConstraints)
{
	m_coordinationPolyhedraConstraints = std::move(coordinationPolyhedraConstraints);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::meetCentralAtomSymbol(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.centralElementSymbol().empty())
		return true;
	else
	{
		if (modelAtom.elementSymbol() == m_coordinationPolyhedraConstraints.centralElementSymbol())
			return true;
		else
			return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::meetCoordinationFeatureSymbol(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.centralCoordiantionFeatureSymbol().isValid())
	{
		if (modelAtom.coordinationFeature().symbol() == m_coordinationPolyhedraConstraints.centralCoordiantionFeatureSymbol())
			return true;
		else
			return false;
	}

	else
		return true;
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::isFeasibleCentralAtomCategory(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.feasibleCentralAtomCategories().empty())
		return true;
	else
	{
		for (const auto& atomCategory : m_coordinationPolyhedraConstraints.feasibleCentralAtomCategories())
		{
			if (modelAtom.coordinationFeature().atomCategory() == atomCategory)
				return true;
		}

		return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::isFeasibleCoordinationNumber(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.coordinationNumber() == std::numeric_limits<size_type>::max())
		return true;
	else
	{
		if (modelAtom.getCoordinationNumber() == m_coordinationPolyhedraConstraints.coordinationNumber())
			return true;
		else
			return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::isFeasibleIonicCoordinationNumber(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.ionicCoordinationNumber() == std::numeric_limits<size_type>::max())
		return true;
	else
	{
		if (modelAtom.getIonicCoordinationNumber() == m_coordinationPolyhedraConstraints.ionicCoordinationNumber())
			return true;
		else
			return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::isFeasibleMetallicCoordinationNumber(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.metallicCoordinationNumber() == std::numeric_limits<size_type>::max())
		return true;
	else
	{
		if (modelAtom.getMetallicCoordinationNumber() == m_coordinationPolyhedraConstraints.metallicCoordinationNumber())
			return true;
		else
			return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::isFeasibleChemicalBondComposition(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.chemicalBondComposition().isUnconstrained())
		return true;
	else
	{
		if (modelAtom.getChemicalBondComposition() == m_coordinationPolyhedraConstraints.chemicalBondComposition())
			return true;
		else
			return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::isFeasibleIonicBondComposition(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.ionicBondComposition().isUnconstrained())
		return true;
	else
	{
		if (modelAtom.getIonicBondComposition() == m_coordinationPolyhedraConstraints.ionicBondComposition())
			return true;
		else
			return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::isFeasibleMetallicBondComposition(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.metallicBondComposition().isUnconstrained())
		return true;
	else
	{
		if (modelAtom.getMetallicBondComposition() == m_coordinationPolyhedraConstraints.metallicBondComposition())
			return true;
		else
			return false;
	}
}

inline bool MathematicalCrystalChemistry::Analysis::Internal::IsFeasibleCoordinationPolyhedra::satisfySymmetryConstraint(const ModelAtom& modelAtom) const
{
	if (m_coordinationPolyhedraConstraints.symmetryConstraints().empty())
		return true;
	else
		return modelAtom.hasAnySymmetry(m_coordinationPolyhedraConstraints.symmetryConstraints());
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_ANALYSIS_INTERNAL_ISFEASIBLECOORDINATIONPOLYHEDRA_H
