#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_LIGANDTRACKER_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_LIGANDTRACKER_H

#include <type_traits>
#include <vector>
#include <list>
#include <utility>
#include <memory>
#include <cmath>
#include <limits>

#include "ConstraintComposition.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			class IonicBond;
			class MetallicBond;

			class AnionicConstraint;
			class CationicConstraint;
			class NonIonicBond;
			class NonMetallicBond;
		}


		namespace Components
		{
			namespace Internal
			{
				class LigandTracker
				{
					using size_type = std::size_t;
					using ConstraintComposition = MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition;

					using IonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond;
					using MetallicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond;
					using AnionicConstraint = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint;
					using CationicConstraint = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint;
					using NonIonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond;
					using NonMetallicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				protected:
					LigandTracker() = default;

				public:
					virtual ~LigandTracker() = default;

					LigandTracker(const LigandTracker&);
					LigandTracker(LigandTracker&&) noexcept;
					LigandTracker& operator=(const LigandTracker&);
					LigandTracker& operator=(LigandTracker&&) noexcept;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

					template <typename T>
					void track(const std::shared_ptr<T>&);

					std::vector<std::shared_ptr<IonicBond>> getIonicBondPtrs() const;
					std::vector<std::shared_ptr<MetallicBond>> getMetallicBondPtrs() const;
					std::vector<std::shared_ptr<AnionicConstraint>> getAnionicConstraintPtrs() const;
					std::vector<std::shared_ptr<CationicConstraint>> getCationicConstraintPtrs() const;
					std::vector<std::shared_ptr<NonIonicBond>> getNonIonicBondPtrs() const;
					std::vector<std::shared_ptr<NonMetallicBond>> getNonMetallicBondPtrs() const;

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					size_type getCoordinationNumber() const;
					size_type getIonicCoordinationNumber() const;
					size_type getMetallicCoordinationNumber() const;

					ConstraintComposition getChemicalBondComposition() const;
					ConstraintComposition getIonicBondComposition() const;
					ConstraintComposition getMetallicBondComposition() const;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Protected methods

				protected:
					void eraseInactiveConstraintPtrs();
					void clearConstraintPtrs();
					void clearChemicalBondPtrs();
					void clearCationicConstraintPtrs();

					std::vector<std::shared_ptr<NonIonicBond>> selectIonicLigandCandidates() const;
					std::vector<std::shared_ptr<NonIonicBond>> selectIonicLigandCandidates(const double ionicBondCutoffRatio) const;
					std::vector<std::shared_ptr<NonIonicBond>> selectIonicLigandCandidates(const ConstraintComposition&) const;
					std::vector<std::shared_ptr<NonIonicBond>> selectIonicLigandCandidates(const ConstraintComposition&, const double ionicBondCutoffRatio) const;
					std::vector<std::shared_ptr<NonMetallicBond>> selectMetallicLigandCandidates() const;
					std::vector<std::shared_ptr<NonMetallicBond>> selectMetallicLigandCandidates(const double metallicBondCutoffRatio) const;
					std::vector<std::shared_ptr<NonMetallicBond>> selectMetallicLigandCandidates(const ConstraintComposition&) const;
					std::vector<std::shared_ptr<NonMetallicBond>> selectMetallicLigandCandidates(const ConstraintComposition&, const double metallicBondCutoffRatio) const;

				// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					template <typename T>
					std::vector<std::shared_ptr<T>> getConstraintPtrs(const std::list<std::weak_ptr<T>>&) const;

					template <typename T>
					void eraseInactiveConstraintPtrs(std::list<std::weak_ptr<T>>&);

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					std::list<std::weak_ptr<IonicBond>> m_ionicBondPtrs;
					std::list<std::weak_ptr<MetallicBond>> m_metallicBondPtrs;
					
					std::list<std::weak_ptr<AnionicConstraint>> m_anionicConstraintPtrs;
					std::list<std::weak_ptr<CationicConstraint>> m_cationicConstraintPtrs;
					std::list<std::weak_ptr<NonIonicBond>> m_nonIonicBondPtrs;
					std::list<std::weak_ptr<NonMetallicBond>> m_nonMetallicBondPtrs;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

template <>
inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::track(const std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond>& constraintPtr)
{
	m_ionicBondPtrs.push_back(std::weak_ptr<IonicBond>{ constraintPtr });
}

template <>
inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::track(const std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond>& constraintPtr)
{
	m_metallicBondPtrs.push_back(std::weak_ptr<MetallicBond>{ constraintPtr });
}

template <>
inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::track(const std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint>& constraintPtr)
{
	m_anionicConstraintPtrs.push_back(std::weak_ptr<AnionicConstraint>{ constraintPtr });
}

template <>
inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::track(const std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint>& constraintPtr)
{
	m_cationicConstraintPtrs.push_back(std::weak_ptr<CationicConstraint>{ constraintPtr });
}

template <>
inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::track(const std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond>& constraintPtr)
{
	m_nonIonicBondPtrs.push_back(std::weak_ptr<NonIonicBond>{ constraintPtr });
}

template <>
inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::track(const std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond>& constraintPtr)
{
	m_nonMetallicBondPtrs.push_back(std::weak_ptr<NonMetallicBond>{ constraintPtr });
}

inline std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond>> MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getIonicBondPtrs() const
{
	return getConstraintPtrs(m_ionicBondPtrs);
}

inline std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond>> MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getMetallicBondPtrs() const
{
	return getConstraintPtrs(m_metallicBondPtrs);
}

inline std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint>> MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getAnionicConstraintPtrs() const
{
	return getConstraintPtrs(m_anionicConstraintPtrs);
}

inline std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint>> MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getCationicConstraintPtrs() const
{
	return getConstraintPtrs(m_cationicConstraintPtrs);
}

inline std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond>> MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getNonIonicBondPtrs() const
{
	return getConstraintPtrs(m_nonIonicBondPtrs);
}

inline std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond>> MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getNonMetallicBondPtrs() const
{
	return getConstraintPtrs(m_nonMetallicBondPtrs);
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline std::size_t MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getCoordinationNumber() const
{
	size_type coordinationNumber = 0;
	{
		coordinationNumber += getIonicCoordinationNumber();
		coordinationNumber += getMetallicCoordinationNumber();
	}

	return coordinationNumber;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getIonicCoordinationNumber() const
{
	size_type ionicCoordinationNumber = 0;
	{
		for (const auto& ionicBondWeakPtr : m_ionicBondPtrs)
		{
			if (std::shared_ptr<IonicBond> ionicBondPtr = ionicBondWeakPtr.lock())
				ionicCoordinationNumber += 1;
		}
	}

	return ionicCoordinationNumber;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getMetallicCoordinationNumber() const
{
	size_type metallicCoordinationNumber = 0;
	{
		for (const auto& metallicBondWeakPtr : m_metallicBondPtrs)
		{
			if (std::shared_ptr<MetallicBond> metallicBondPtr = metallicBondWeakPtr.lock())
				metallicCoordinationNumber += 1;
		}
	}

	return metallicCoordinationNumber;
}

inline MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getChemicalBondComposition() const
{
	ConstraintComposition bondComposition = getIonicBondComposition();
	bondComposition.join(getMetallicBondComposition());

	return bondComposition;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::eraseInactiveConstraintPtrs()
{
	eraseInactiveConstraintPtrs(m_ionicBondPtrs);
	eraseInactiveConstraintPtrs(m_metallicBondPtrs);

	eraseInactiveConstraintPtrs(m_anionicConstraintPtrs);
	eraseInactiveConstraintPtrs(m_cationicConstraintPtrs);
	eraseInactiveConstraintPtrs(m_nonIonicBondPtrs);
	eraseInactiveConstraintPtrs(m_nonMetallicBondPtrs);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::clearConstraintPtrs()
{
	m_ionicBondPtrs.clear();
	m_metallicBondPtrs.clear();

	m_anionicConstraintPtrs.clear();
	m_cationicConstraintPtrs.clear();
	m_nonIonicBondPtrs.clear();
	m_nonMetallicBondPtrs.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::clearChemicalBondPtrs()
{
	m_ionicBondPtrs.clear();
	m_metallicBondPtrs.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::clearCationicConstraintPtrs()
{
	m_cationicConstraintPtrs.clear();
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename T>
inline std::vector<std::shared_ptr<T>> MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::getConstraintPtrs(const std::list<std::weak_ptr<T>>& constraintWeakPtrs) const
{
	std::vector<std::shared_ptr<T>> constraintPtrs;
	{
		for (const auto& constraintWeakPtr : constraintWeakPtrs)
		{
			if (std::shared_ptr<T> constraintPtr = constraintWeakPtr.lock())
				constraintPtrs.push_back(std::move(constraintPtr));
		}
	}

	return constraintPtrs;
}

template <typename T>
inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::LigandTracker::eraseInactiveConstraintPtrs(std::list<std::weak_ptr<T>>& constraintPtrs)
{
	for (auto iter = constraintPtrs.begin(); iter != constraintPtrs.end();)
	{
		if (iter->expired())
			iter = constraintPtrs.erase(iter);
		else
			++iter;
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_LIGANDTRACKER_H
