#ifndef OPENMX_FILEIO_MDFILEWRITER_H
#define OPENMX_FILEIO_MDFILEWRITER_H

#include "FileWriter.h"

#include "LinearAlgebra.h"
#include "LengthCasting.h"
#include "TimeCasting.h"

#include "CrystalStructure.h"


namespace OpenMX
{
	namespace FileIO
	{
		class MdFileWriter :public OperatingSystem::IO::FileWriter
		{
			using size_type = std::size_t;
			using NumericalVector = MathToolkit::Generic::NumericalVector<double>;
			using NumericalMatrix = MathToolkit::Generic::NumericalMatrix<double>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			MdFileWriter() = default;
			explicit MdFileWriter(const std::filesystem::path& filePath);
			explicit MdFileWriter(std::filesystem::path&& filePath);

			MdFileWriter(const MdFileWriter&) = default;
			MdFileWriter(MdFileWriter&&) noexcept = default;
			MdFileWriter& operator=(const MdFileWriter&) = default;
			MdFileWriter& operator=(MdFileWriter&&) noexcept = default;

			virtual ~MdFileWriter() = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Static methods

			static std::string correctExtension();
			static bool isCorrectExtension(const std::filesystem::path& filePath);

		// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			void releaseFile() override;

			template <typename A>
			void drawFrame(const ChemToolkit::Crystallography::CrystalStructure<A>&, const size_type elapsedTimeCount);

			template <typename A>
			void drawFrame(const ChemToolkit::Crystallography::CrystalStructure<A>&, const double elapsedTime);

			template <typename A>
			void drawFrame(const ChemToolkit::Crystallography::CrystalStructure<A>&, const size_type elapsedTimeCount, const double totalEnergy);

			template <typename A>
			void drawFrame(const ChemToolkit::Crystallography::CrystalStructure<A>&, const double elapsedTime, const double totalEnergy);

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			std::string writeHeaderMessage(const std::size_t totalAtomCount, const size_type elapsedTimeCount) const;
			std::string writeHeaderMessage(const std::size_t totalAtomCount, const double elapsedTime) const;
			std::string writeHeaderMessage(const std::size_t totalAtomCount, const size_type elapsedTimeCount, const double totalEnergy) const;
			std::string writeHeaderMessage(const std::size_t totalAtomCount, const double elapsedTime, const double totalEnergy) const;

			template <typename A>
			std::string writePrimitiveLatticeVectors(const ChemToolkit::Crystallography::CrystalStructure<A>&) const;

			template <typename A>
			std::string writeAtomicArrangement(const ChemToolkit::Crystallography::CrystalStructure<A>&) const;

		// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			mutable std::size_t m_frameNumber{ 0 };
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline std::string OpenMX::FileIO::MdFileWriter::correctExtension()
{
	return std::string{ ".md" };
}

inline bool OpenMX::FileIO::MdFileWriter::isCorrectExtension(const std::filesystem::path& filePath)
{
	if (filePath.extension().string() == correctExtension())
		return true;
	else
		return false;
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void OpenMX::FileIO::MdFileWriter::releaseFile()
{
	FileBase::releaseFile();
	m_frameNumber = 0;
}

template <typename A>
inline void OpenMX::FileIO::MdFileWriter::drawFrame(const ChemToolkit::Crystallography::CrystalStructure<A>& crystalStructure, const size_type elapsedTimeCount)
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.atomicArrangement().size(), elapsedTimeCount);
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeAtomicArrangement(crystalStructure);
	write(outputString);

	++m_frameNumber;
}

template <typename A>
inline void OpenMX::FileIO::MdFileWriter::drawFrame(const ChemToolkit::Crystallography::CrystalStructure<A>& crystalStructure, const double elapsedTime)
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.atomicArrangement().size(), elapsedTime);
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeAtomicArrangement(crystalStructure);
	write(outputString);

	++m_frameNumber;
}

template <typename A>
inline void OpenMX::FileIO::MdFileWriter::drawFrame(const ChemToolkit::Crystallography::CrystalStructure<A>& crystalStructure, const size_type elapsedTimeCount, const double totalEnergy)
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.atomicArrangement().size(), elapsedTimeCount, totalEnergy);
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeAtomicArrangement(crystalStructure);
	write(outputString);

	++m_frameNumber;
}

template <typename A>
inline void OpenMX::FileIO::MdFileWriter::drawFrame(const ChemToolkit::Crystallography::CrystalStructure<A>& crystalStructure, const double elapsedTime, const double totalEnergy)
{
	std::string outputString;
	outputString += writeHeaderMessage(crystalStructure.atomicArrangement().size(), elapsedTime, totalEnergy);
	outputString += writePrimitiveLatticeVectors(crystalStructure);
	outputString += writeAtomicArrangement(crystalStructure);
	write(outputString);

	++m_frameNumber;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::string OpenMX::FileIO::MdFileWriter::writeHeaderMessage(const std::size_t totalAtomCount, const size_type elapsedTimeCount) const
{
	std::string outputString;
	{
		outputString += std::to_string(totalAtomCount);
		outputString += "\n   time_count=    ";
		outputString += std::to_string(elapsedTimeCount);
		outputString += " ()    ";
	}

	return outputString;
}

inline std::string OpenMX::FileIO::MdFileWriter::writeHeaderMessage(const std::size_t totalAtomCount, const double elapsedTime) const
{
	using MathToolkit::UnitConversion::TimeCasting::Unit::AtomicUnit;
	using FemtoSecond = MathToolkit::UnitConversion::TimeCasting::Unit::Second<std::femto>;

	std::string outputString;
	{
		outputString += std::to_string(totalAtomCount);
		outputString += "\n   time=    ";
		outputString += std::to_string(elapsedTime * MathToolkit::UnitConversion::TimeCasting::getScalingMultiplier<AtomicUnit, FemtoSecond>());
		outputString += " (fs)     ";
	}

	return outputString;
}

inline std::string OpenMX::FileIO::MdFileWriter::writeHeaderMessage(const std::size_t totalAtomCount, const size_type elapsedTimeCount, const double totalEnergy) const
{
	std::string outputString;
	{
		outputString += std::to_string(totalAtomCount);
		outputString += "\n   time_count=    ";
		outputString += std::to_string(elapsedTimeCount);
		outputString += " ()     ";
		outputString += "Energy= ";
		outputString += std::to_string(totalEnergy);
		outputString += " ()  ";
	}

	return outputString;
}

inline std::string OpenMX::FileIO::MdFileWriter::writeHeaderMessage(const std::size_t totalAtomCount, const double elapsedTime, const double totalEnergy) const
{
	using MathToolkit::UnitConversion::TimeCasting::Unit::AtomicUnit;
	using FemtoSecond = MathToolkit::UnitConversion::TimeCasting::Unit::Second<std::femto>;

	std::string outputString;
	{
		outputString += std::to_string(totalAtomCount);
		outputString += "\n   time=    ";
		outputString += std::to_string(elapsedTime * MathToolkit::UnitConversion::TimeCasting::getScalingMultiplier<AtomicUnit, FemtoSecond>());
		outputString += " (fs)     ";
		outputString += "Energy= ";
		outputString += std::to_string(totalEnergy);
		outputString += " ()  ";
	}

	return outputString;
}

template <typename A>
inline std::string OpenMX::FileIO::MdFileWriter::writePrimitiveLatticeVectors(const ChemToolkit::Crystallography::CrystalStructure<A>& crystalStructure) const
{
	std::string outputString;
	{
		outputString += "Cell_Vectors=";

		for (std::size_t columnIndex = 0; columnIndex < 3; ++columnIndex)
		{
			const NumericalMatrix& basisVectors = crystalStructure.unitCell().basisVectors();

			using MathToolkit::UnitConversion::LengthCasting::Unit::Angstrom;
			using MathToolkit::UnitConversion::LengthCasting::Unit::AtomicUnit;

			outputString += "  ";
			outputString += std::to_string(basisVectors(0, columnIndex) * MathToolkit::UnitConversion::LengthCasting::getScalingMultiplier<AtomicUnit, Angstrom>());
			outputString += "  ";
			outputString += std::to_string(basisVectors(1, columnIndex) * MathToolkit::UnitConversion::LengthCasting::getScalingMultiplier<AtomicUnit, Angstrom>());
			outputString += "  ";
			outputString += std::to_string(basisVectors(2, columnIndex) * MathToolkit::UnitConversion::LengthCasting::getScalingMultiplier<AtomicUnit, Angstrom>());
		}

		outputString += "\n";
	}

	return outputString;
}

template <typename A>
inline std::string OpenMX::FileIO::MdFileWriter::writeAtomicArrangement(const ChemToolkit::Crystallography::CrystalStructure<A>& crystalStructure) const
{
	std::string outputString;
	{
		using MathToolkit::UnitConversion::LengthCasting::Unit::Angstrom;
		using MathToolkit::UnitConversion::LengthCasting::Unit::AtomicUnit;

		for (const auto& atom : crystalStructure.atomicArrangement())
		{
			outputString += "   ";
			outputString += atom.elementSymbol();
			outputString += "  ";
			outputString += std::to_string(atom.coordinate()[0] * MathToolkit::UnitConversion::LengthCasting::getScalingMultiplier<AtomicUnit, Angstrom>());
			outputString += "  ";
			outputString += std::to_string(atom.coordinate()[1] * MathToolkit::UnitConversion::LengthCasting::getScalingMultiplier<AtomicUnit, Angstrom>());
			outputString += "  ";
			outputString += std::to_string(atom.coordinate()[2] * MathToolkit::UnitConversion::LengthCasting::getScalingMultiplier<AtomicUnit, Angstrom>());
			outputString += "  ";
			outputString += std::to_string(atom.appliedForce()[0]);
			outputString += "  ";
			outputString += std::to_string(atom.appliedForce()[1]);
			outputString += "  ";
			outputString += std::to_string(atom.appliedForce()[2]);
			outputString += "  ";
			outputString += "0.0";
			outputString += "  ";
			outputString += "0.0";
			outputString += "  ";
			outputString += "0.0";
			outputString += "  0.0  0.0  0.0  0.0\n";
		}
	}

	return outputString;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !OPENMX_FILEIO_MDFILEWRITER_H
