#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_METALLICBOND_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_METALLICBOND_H

#include "ReportException.h"

#include "ConstraintOwner.h"
#include "MinMaxDistanceConstraint.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			class MetallicBond :public Internal::MinMaxDistanceConstraint
			{
				using MetallicBondOwner = ConstraintOwner<MetallicBond>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private constructors and public destructor

			private:
				MetallicBond(ModelAtom&, const Ligand&);
				MetallicBond(ModelAtom&, Ligand&&);
				MetallicBond(const Ligand&, const Ligand&);
				explicit MetallicBond(const std::shared_ptr<MetallicBond>& twinPtr);

			public:
				virtual ~MetallicBond() = default;

			// Private constructors and public destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static MetallicBondOwner create(const ConstraintKey&);
				static bool shouldCreate(const ModelAtom&, const Ligand&, const double cutoffRatio);
				static bool shouldCreate(const ConstraintKey&, const double cutoffRatio);

				static double defaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio(const double cutoffRatio);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				static bool canCreate(const ModelAtom&, const Ligand&);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				static double m_defaultCutoffRatio;


			private:
				MetallicBond() = delete;
				MetallicBond(const MetallicBond&) = delete;
				MetallicBond(MetallicBond&&) noexcept = delete;
				MetallicBond& operator=(const MetallicBond&) = delete;
				MetallicBond& operator=(MetallicBond&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond::MetallicBondOwner MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond::create(const ConstraintKey& constraintKey)
{
	ModelAtom& modelAtom = constraintKey.modelAtomRef();
	const Ligand& ligand = constraintKey.ligand();

	if (canCreate(modelAtom, ligand))
	{
		std::shared_ptr<MetallicBond> metallicBondPtr{ new MetallicBond(modelAtom, ligand) };
		std::shared_ptr<MetallicBond> reverseMetallicBondPtr{ new MetallicBond(metallicBondPtr) };
		metallicBondPtr->setReversePtr(std::static_pointer_cast<InteratomicConstraintBase>(reverseMetallicBondPtr));

		modelAtom.track(metallicBondPtr);
		ligand.get().track(reverseMetallicBondPtr);

		return MetallicBondOwner{ std::move(metallicBondPtr), std::move(reverseMetallicBondPtr) };
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException("create of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond", "Could not create metallic bond."));
	}
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond::shouldCreate(const ModelAtom& modelAtom, const Ligand& ligand, const double cutoffRatio)
{
	if (canCreate(modelAtom, ligand))
	{
		NumericalVector displacement = ligand.coordinate() - modelAtom.coordinate();
		double cutoffLength = cutoffRatio * (modelAtom.coordinationFeature().upperMetallicBondRadius() + ligand.get().coordinationFeature().upperMetallicBondRadius());

		if (displacement.normSquare() < (cutoffLength * cutoffLength))
			return true;
		else
			return false;
	}

	else
		return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond::shouldCreate(const ConstraintKey& constraintKey, const double cutoffRatio)
{
	return shouldCreate(constraintKey.modelAtomRef(), constraintKey.ligand(), cutoffRatio);
}

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond::defaultCutoffRatio() noexcept
{
	return m_defaultCutoffRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond::setDefaultCutoffRatio() noexcept
{
	m_defaultCutoffRatio = 2.0;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond::setDefaultCutoffRatio(const double val)
{
	if (0.0 < val)
		m_defaultCutoffRatio = val;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("setDefaultCutoffRatio of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond", "\"defaultCutoffRatio\" is not more than zero"));
	}
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond::canCreate(const ModelAtom& modelAtom, const Ligand& ligand)
{
	if (modelAtom.coordinationFeature().atomCategory().isMetal() && ligand.get().coordinationFeature().atomCategory().isMetal())
		return true;
	else
		return false;
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_METALLICBOND_H
