#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_MINMAXDISTANCECONSTRAINT_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_MINMAXDISTANCECONSTRAINT_H

#include "ReportException.h"
#include "InteratomicConstraintBase.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			namespace Internal
			{
				class MinMaxDistanceConstraint :public InteratomicConstraintBase
				{
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Protected constructors and public destructor

				protected:
					MinMaxDistanceConstraint(ModelAtom&, const Ligand&);
					MinMaxDistanceConstraint(ModelAtom&, Ligand&&);
					MinMaxDistanceConstraint(const Ligand&, const Ligand&);
					explicit MinMaxDistanceConstraint(const std::shared_ptr<MinMaxDistanceConstraint>& reversePtr);

				public:
					virtual ~MinMaxDistanceConstraint() = default;

				// Protected constructors and public destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					virtual void initialize() noexcept;
					double maximumDistance() const noexcept;
					double minimumDistance() const noexcept;

					bool isFeasible(const double errorRate) const;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Protected methods

				protected:
					void setMinMaxDistances(const double minimumDistance, const double maximumDistance);
					void synchronizeMinMaxDistances();

				// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					double m_maximumDistance;
					double m_minimumDistance;


				private:
					MinMaxDistanceConstraint() = delete;
					MinMaxDistanceConstraint(const MinMaxDistanceConstraint&) = delete;
					MinMaxDistanceConstraint(MinMaxDistanceConstraint&&) noexcept = delete;
					MinMaxDistanceConstraint& operator=(const MinMaxDistanceConstraint&) = delete;
					MinMaxDistanceConstraint& operator=(MinMaxDistanceConstraint&&) noexcept = delete;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::Internal::MinMaxDistanceConstraint::maximumDistance() const noexcept
{
	return m_maximumDistance;
}

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::Internal::MinMaxDistanceConstraint::minimumDistance() const noexcept
{
	return m_minimumDistance;
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::Internal::MinMaxDistanceConstraint::isFeasible(const double errorRate) const
{
	if (errorRate < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "isFeasible", "\"errorRate\" is less than zero."));
	}

	else
	{
		NumericalVector displacement = ligand().coordinate() - modelAtomRef().coordinate();
		double distanceSquare = displacement.normSquare();
		double maxBondLength = (1.0 + errorRate) * m_maximumDistance;


		if ((maxBondLength * maxBondLength) < distanceSquare)
			return false;

		else
		{
			double minBondLength = (1.0 - errorRate) * m_minimumDistance;

			if (distanceSquare < (minBondLength * minBondLength))
				return false;
			else
				return true;
		}
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::Internal::MinMaxDistanceConstraint::setMinMaxDistances(const double minimumDistance, const double maximumDistance)
{
	if (minimumDistance < 0.0 || maximumDistance < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setMinMaxDistance", "\"minimumDistance\" and/or \"maximumDistance\" is less than zero."));
	}

	else
	{
		if (maximumDistance < minimumDistance)
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "setMinMaxDistance", "\"maximumDistance\" is less than \"minimumDistance\"."));
		}

		else
		{
			m_maximumDistance = maximumDistance;
			m_minimumDistance = minimumDistance;
		}
	}
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::Internal::MinMaxDistanceConstraint::synchronizeMinMaxDistances()
{
	std::shared_ptr<MinMaxDistanceConstraint> twinMinMaxDistanceConstraintPtr = std::dynamic_pointer_cast<MinMaxDistanceConstraint>(reversePtr());
	twinMinMaxDistanceConstraintPtr->m_maximumDistance = m_maximumDistance;
	twinMinMaxDistanceConstraintPtr->m_minimumDistance = m_minimumDistance;
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_MINMAXDISTANCECONSTRAINT_H
