#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELATOM_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELATOM_H

#include <limits>
#include <utility>

#include "UnitCell.h"
#include "PeriodicReferenceWrapper.h"

#include "CoordinationCenter.h"
#include "LigandTracker.h"
#include "OxidationManager.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			class AtomicEnvironment;


			class ModelAtom :public Internal::CoordinationCenter, public Internal::LigandTracker, public Internal::OxidationManager
			{
				using size_type = std::size_t;

			public:
				using Ligand = ChemToolkit::Crystallography::PeriodicReferenceWrapper<ModelAtom, ChemToolkit::Crystallography::UnitCell>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				explicit ModelAtom(const std::string& elementSymbol);
				ModelAtom(const std::string& elementSymbol, const CoordinationFeatureSymbol&);
				ModelAtom(const std::string& elementSymbol, const NumericalVector& cartesianCoordinate);
				ModelAtom(const std::string& elementSymbol, const CoordinationFeatureSymbol&, const NumericalVector& cartesianCoordinate);
				virtual ~ModelAtom() = default;

				ModelAtom(const ModelAtom&) = default;
				ModelAtom(ModelAtom&&) noexcept = default;
				ModelAtom& operator=(const ModelAtom&) = default;
				ModelAtom& operator=(ModelAtom&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				const std::string& siteSymmetrySymbol() const;
				const std::string& wyckoffSymbol() const;
				size_type siteIndex() const;
				void setSiteSymmetrySymbol(const std::string& symbol = std::string{});
				void setWyckoffSymbol(const std::string& symbol = std::string{});
				void setSiteIndex(const size_type index = std::numeric_limits<size_type>::max());

				std::vector<Ligand> getIonicLigands() const;
				std::vector<Ligand> getMetallicLigands() const;

				std::vector<Ligand> getAnionicConstrainedLigands() const;
				std::vector<Ligand> getCationicConstrainedLigands() const;
				std::vector<Ligand> getNonIonicBondedLigands() const;
				std::vector<Ligand> getNonMetallicBondedLigands() const;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				AtomicEnvironment getAtomicEnvironment() const;

				void updateConstraintReferences();
				void clearConstraintReferences();
				void clearChemicalBondReferences();

				bool isFeasibleChemicalBondComposition() const;
				bool isFeasibleIonicBondComposition() const;
				bool isFeasibleMetallicBondComposition() const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Customize methods

				void fixChemicalBondCompositions();
				void setCoordinationFeatureFromElementSymbol();

				std::vector<Ligand> getIonicLigandCandidates() const;
				std::vector<Ligand> getIonicLigandCandidates(const double ionicBondCutoffRatio) const;
				std::vector<Ligand> getMetallicLigandCandidates() const;
				std::vector<Ligand> getMetallicLigandCandidates(const double metallicBondCutoffRatio) const;
				std::vector<std::pair<Ligand, Ligand>> getCationicConstraintCandidates() const;

				std::vector<Ligand> getSubstituteIonicLigandCandidates() const;

			// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void recognizeCoordinationPolyhedra();

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::string m_siteSymmetrySymbol;
				std::string m_wyckoffSymbol;
				size_type m_siteIndex;
			};


			inline bool operator<(const ModelAtom& maa, const ModelAtom& mab)
			{
				if (maa.elementSymbol() < mab.elementSymbol())
					return true;
				else if (mab.elementSymbol() < maa.elementSymbol())
					return false;
				else
				{
					if (maa.siteSymmetrySymbol() < mab.siteSymmetrySymbol())
						return true;
					else if (mab.siteSymmetrySymbol() < maa.siteSymmetrySymbol())
						return false;
					else
					{
						if (maa.getCoordinationNumber() < mab.getCoordinationNumber())
							return true;
						else if (mab.getCoordinationNumber() < maa.getCoordinationNumber())
							return false;
						else
						{
							if (maa.coordinationFeature() < mab.coordinationFeature())
								return true;
							else
								return false;
						}
					}
				}
			}

			inline bool operator>(const ModelAtom& maa, const ModelAtom& mab)
			{
				return (mab < maa);
			}

			inline bool operator==(const ModelAtom& maa, const ModelAtom& mab)
			{
				if (maa.coordinationFeature().atomCategory() == mab.coordinationFeature().atomCategory())
				{
					if (maa.getCoordinationNumber() == mab.getCoordinationNumber())
					{
						if (maa.coordinationFeature() == mab.coordinationFeature())
						{
							if (maa.elementSymbol() == mab.elementSymbol())
								return true;
						}
					}
				}

				return false;
			}

			inline bool operator!=(const ModelAtom& maa, const ModelAtom& mab)
			{
				return !(maa == mab);
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const std::string& MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::siteSymmetrySymbol() const
{
	return m_siteSymmetrySymbol;
}

inline const std::string& MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::wyckoffSymbol() const
{
	return m_wyckoffSymbol;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::siteIndex() const
{
	if (m_siteIndex == std::numeric_limits<size_type>::max())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "siteIndex", "The site index is invalid."));
	}

	else
		return m_siteIndex;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::setSiteSymmetrySymbol(const std::string& symbol)
{
	m_siteSymmetrySymbol = symbol;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::setWyckoffSymbol(const std::string& symbol)
{
	m_wyckoffSymbol = symbol;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::setSiteIndex(const size_type index)
{
	m_siteIndex = index;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::updateConstraintReferences()
{
	clearRecognizedLigands();
	eraseInactiveConstraintPtrs();

	recognizeCoordinationPolyhedra();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::clearConstraintReferences()
{
	clearConstraintPtrs();
	clearRecognizedLigands();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::clearChemicalBondReferences()
{
	clearChemicalBondPtrs();
	clearCationicConstraintPtrs();
	clearRecognizedLigands();
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::isFeasibleChemicalBondComposition() const
{
	if (isFeasibleIonicBondComposition() && isFeasibleMetallicBondComposition())
		return true;
	else
		return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::isFeasibleIonicBondComposition() const
{
	if (!(coordinationFeature().feasibleIonicBondComposition().isUnconstrained()))
	{
		if (LigandTracker::getIonicBondComposition() == coordinationFeature().feasibleIonicBondComposition())
			return true;
		else
			return false;
	}

	else
		return true;
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::isFeasibleMetallicBondComposition() const
{
	if (!(coordinationFeature().feasibleMetallicBondComposition().isUnconstrained()))
	{
		if (LigandTracker::getMetallicBondComposition() == coordinationFeature().feasibleMetallicBondComposition())
			return true;
		else
			return false;
	}

	else
		return true;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::fixChemicalBondCompositions()
{
	coordinationFeature().setFeasibleIonicBondComposition(LigandTracker::getIonicBondComposition());
	coordinationFeature().setFeasibleMetallicBondComposition(LigandTracker::getMetallicBondComposition());
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtom::setCoordinationFeatureFromElementSymbol()
{
	CoordinationCenterBase::assignCoordinationFeatureFromElementSymbol(elementSymbol(), oxidationState());
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELATOM_H
