#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELATOMCOLLECTION_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELATOMCOLLECTION_H

#include <vector>

#include "ChemicalComposition.h"
#include "ConstraintComposition.h"
#include "ModelAtom.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			class ModelAtomCollection
			{
				using size_type = std::size_t;
				using ChemicalComposition = ChemToolkit::Generic::ChemicalComposition;

				using ConstraintComposition = MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition;
				using CoordinationFeatureSymbol = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeatureSymbol;

				using iterator = std::vector<ModelAtom>::iterator;
				using const_iterator = std::vector<ModelAtom>::const_iterator;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				ModelAtomCollection() = default;
				virtual ~ModelAtomCollection() = default;

				ModelAtomCollection(const ModelAtomCollection&) = default;
				ModelAtomCollection(ModelAtomCollection&&) noexcept = default;
				ModelAtomCollection& operator=(const ModelAtomCollection&) = default;
				ModelAtomCollection& operator=(ModelAtomCollection&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				iterator begin();
				const_iterator begin() const;
				iterator end();
				const_iterator end() const;

				bool isEmpty() const;
				bool hasChemicalElement(const std::string& elementSymbol);
				bool hasCoordinationFeature(const CoordinationFeatureSymbol& coordinationFeatureSymbol);

				ChemicalComposition toChemicalComposition() const;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Customize methods

				void clear();

				void addAtoms(const ModelAtom& modelAtom, const size_type numAtoms = 1);
				void addAtoms(const std::string& elementSymbol, const size_type numAtoms = 1);
				void eraseAtoms(const std::string& elementSymbol);

				void registerCoordinationFeature(const std::string& elementSymbol, const CoordinationFeatureSymbol&);

			// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::vector<ModelAtom> m_modelAtoms;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::iterator MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::begin()
{
	return m_modelAtoms.begin();
}

inline MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::const_iterator MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::begin() const
{
	return m_modelAtoms.begin();
}

inline MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::iterator MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::end()
{
	return m_modelAtoms.end();
}

inline MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::const_iterator MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::end() const
{
	return m_modelAtoms.end();
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::isEmpty() const
{
	return m_modelAtoms.empty();
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::hasChemicalElement(const std::string& symbol)
{
	for (auto& modelAtom : m_modelAtoms)
	{
		if (modelAtom.elementSymbol() == symbol)
			return true;
	}

	return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::hasCoordinationFeature(const CoordinationFeatureSymbol& symbol)
{
	for (auto& modelAtom : m_modelAtoms)
	{
		if (modelAtom.coordinationFeature().symbol() == symbol)
			return true;
	}

	return false;
}

inline ChemToolkit::Generic::ChemicalComposition MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::toChemicalComposition() const
{
	ChemicalComposition chemicalComposition;
	{
		for (const auto& modelAtom : m_modelAtoms)
			chemicalComposition.add(modelAtom.elementSymbol());
	}

	return chemicalComposition;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::clear()
{
	m_modelAtoms.clear();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::addAtoms(const ModelAtom& modelAtom, const size_type numAtoms)
{
	for (size_type index = 0; index < numAtoms; ++index)
		m_modelAtoms.push_back(ModelAtom{ modelAtom });
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::addAtoms(const std::string& elementSymbol, const size_type numAtoms)
{
	for (size_type index = 0; index < numAtoms; ++index)
		m_modelAtoms.push_back(ModelAtom{ elementSymbol });
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::eraseAtoms(const std::string& symbol)
{
	for (auto iter = m_modelAtoms.begin(); iter != m_modelAtoms.end();)
	{
		if (iter->elementSymbol() == symbol)
			iter = m_modelAtoms.erase(iter);
		else
			++iter;
	}
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection::registerCoordinationFeature(const std::string& elementSymbol, const CoordinationFeatureSymbol& coordinationFeatureSymbol)
{
	for (auto& modelAtom : m_modelAtoms)
	{
		if (modelAtom.elementSymbol() == elementSymbol)
			modelAtom.coordinationFeature() = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature{ coordinationFeatureSymbol };
	}
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELATOMCOLLECTION_H
