#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_STRUCTURECUSTOMIZER_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_STRUCTURECUSTOMIZER_H

#include "CrystalStructure.h"
#include "PeriodicReferenceWrapper.h"

#include "ModelAtom.h"
#include "ConstraintKey.h"
#include "ConstraintManager.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			namespace Internal
			{
				class ModelAtomTracker :public ConstraintManager, public ChemToolkit::Crystallography::CrystalStructure<MathematicalCrystalChemistry::StructureModel::Components::ModelAtom>
				{
				protected:
					using UnitCell = ChemToolkit::Crystallography::UnitCell;
					using ModelAtom = MathematicalCrystalChemistry::StructureModel::Components::ModelAtom;
					using Ligand = ChemToolkit::Crystallography::PeriodicReferenceWrapper<ModelAtom, UnitCell>;
					using ConstraintKey = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Protected constructors, public destructor and operators

				protected:
					ModelAtomTracker();
					ModelAtomTracker(const UnitCell& unitCell, const AtomicArrangement& atomicArrangement = AtomicArrangement{});
					ModelAtomTracker(const UnitCell& unitCell, AtomicArrangement&& atomicArrangement);

				public:
					virtual ~ModelAtomTracker() = default;

					ModelAtomTracker(const ModelAtomTracker&);
					ModelAtomTracker(ModelAtomTracker&&) noexcept;
					ModelAtomTracker& operator=(const ModelAtomTracker&);
					ModelAtomTracker& operator=(ModelAtomTracker&&) noexcept;

				// Protected constructors, public destructor and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void createTrackingAtomPairs();
					void clearTrackingAtomPairs();
					void setAtomPairTrackingCutoffRatio() noexcept;
					void setAtomPairTrackingCutoffRatio(const double);

					static double defaultAtomPairTrackingCutoffRatio();
					static void setCommonAtomPairTrackingCutoffRatio() noexcept;
					static void setCommonAtomPairTrackingCutoffRatio(const double);

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Protected methods

				protected:
					const std::vector<ConstraintKey>& trackingAtomPairs() const;

				// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					bool shouldTrack(const ModelAtom&, const ModelAtom&, const Ligand::LatticePoint&) const;
					std::vector<Ligand::LatticePoint> getSearchingLatticePoints(const ModelAtom&, const ModelAtom&) const;

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					std::vector<ConstraintKey> m_trackingAtomPairs;
					double m_atomPairTrackingCutoffRatio;

					static double m_commonAtomPairTrackingCutoffRatio;
					static double m_defaultAtomPairTrackingCutoffRatio;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::clearTrackingAtomPairs()
{
	ConstraintManager::clearGeometricalConstraintOwners();
	m_trackingAtomPairs.clear();

	for (auto& atom : atomicArrangement())
		atom.clearConstraintReferences();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::setAtomPairTrackingCutoffRatio() noexcept
{
	m_atomPairTrackingCutoffRatio = m_commonAtomPairTrackingCutoffRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::setAtomPairTrackingCutoffRatio(const double val)
{
	if (val < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException(typeid(*this), "setMaxAtomicDisplacement", "\"atomPairTrackingCutoffRatio\" is less than zero."));
	}

	else
		m_atomPairTrackingCutoffRatio = val;
}

inline double MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::defaultAtomPairTrackingCutoffRatio()
{
	return m_defaultAtomPairTrackingCutoffRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::setCommonAtomPairTrackingCutoffRatio() noexcept
{
	m_commonAtomPairTrackingCutoffRatio = m_defaultAtomPairTrackingCutoffRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::setCommonAtomPairTrackingCutoffRatio(const double val)
{
	if (val < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("setCommonAtomPairTrackingCutoffRatio of MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker", "\"commonAtomPairTrackingCutoffRatio\" is less than zero."));
	}

	else
		m_commonAtomPairTrackingCutoffRatio = val;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline const std::vector<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::ConstraintKey>& MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::trackingAtomPairs() const
{
	return m_trackingAtomPairs;
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::shouldTrack(const ModelAtom& modelAtom, const ModelAtom& ligandAtom, const Ligand::LatticePoint& latticePoint) const
{
	double trackingCutoffLength = 0.0;
	{
		if (modelAtom.coordinationFeature().atomCategory().isMetal() && ligandAtom.coordinationFeature().atomCategory().isMetal())
			trackingCutoffLength = m_atomPairTrackingCutoffRatio * (modelAtom.coordinationFeature().upperMetallicBondRadius() + ligandAtom.coordinationFeature().upperMetallicBondRadius());
		else
			trackingCutoffLength = m_atomPairTrackingCutoffRatio * (modelAtom.coordinationFeature().repulsiveIonicRadius() + ligandAtom.coordinationFeature().repulsiveIonicRadius());
	}

	NumericalVector dispacement = ligandAtom.coordinate() - modelAtom.coordinate();
	{
		if (latticePoint[0] != 0 || latticePoint[1] != 0 || latticePoint[2] != 0)
		{
			NumericalVector translationVector(3);
			{
				translationVector[0] = static_cast<double>(latticePoint[0]);
				translationVector[1] = static_cast<double>(latticePoint[1]);
				translationVector[2] = static_cast<double>(latticePoint[2]);
			}

			translationVector.transform(unitCell().basisVectors());
			dispacement += translationVector;
		}
	}
	double distanceSquare = dispacement.normSquare();


	if (distanceSquare < (trackingCutoffLength * trackingCutoffLength))
		return true;
	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_STRUCTURECUSTOMIZER_H
