#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELCRYSTALSTRUCTURE_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELCRYSTALSTRUCTURE_H

#include <filesystem>

#include "AtomCategoryComposition.h"
#include "CrystalCategory.h"

#include "ModelAtomTracker.h"
#include "StructureFingerprint.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			class ModelCrystalStructure :public Internal::ModelAtomTracker
			{
				using size_type = std::size_t;
				using CrystalCategory = MathematicalCrystalChemistry::StructureModel::Features::CrystalCategory;
				using AtomCategoryComposition = MathematicalCrystalChemistry::StructureModel::Features::AtomCategoryComposition;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				ModelCrystalStructure() = default;
				ModelCrystalStructure(const UnitCell& unitCell, const AtomicArrangement& atomicArrangement = AtomicArrangement{});
				ModelCrystalStructure(const UnitCell& unitCell, AtomicArrangement&& atomicArrangement);

				virtual ~ModelCrystalStructure() = default;

				ModelCrystalStructure(const ModelCrystalStructure&) = default;
				ModelCrystalStructure(ModelCrystalStructure&&) noexcept = default;
				ModelCrystalStructure& operator=(const ModelCrystalStructure&) = default;
				ModelCrystalStructure& operator=(ModelCrystalStructure&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				double getLowerIonicPackingFraction() const;
				void expandStructure(const double packingFraction);

				void createGeometricalConstraints();
				void createGeometricalConstraints(const double minimumConstraintCutoffRatio, const double nonBondingCutoffRatio, const double ionicBondCutoffRatio, const double metallicBondCutoffRatio);
				void createChemicalBonds();
				void createChemicalBonds(const double ionicBondCutoffRatio, const double metallicBondCutoffRatio);
				void substituteChemicalBonds(const double feasibleConstraintErrorRate);

				void clearGeometricalConstraints();
				void clearChemicalBonds();

				void optimizeOxidationStates();
				void assignCoordinationFeaturesFromElementSymbols();
				void removeSiteDuplications();

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Utility

				bool isOptimalSolution(const double feasibleConstraintErrorRate) const;
				
				StructureFingerprint getStructureFingerprint() const;
				CrystalCategory getCrystalCategory() const;

				void createStructureInformationFile(const std::filesystem::path& filePath) const;
				AtomCategoryComposition getAtomCategoryComposition() const;

			// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Override methods

				void clearSpaceGroupInfo() override;

				void conventionalizeStructure() override;
				void refineStructure() override;
				void refineStructure(const double spaceGroupPrecision) override;
				void normalizeFractionalCoordinates() override;

				void sort() override;
				void eraseAtoms(const std::string& elementSymbol) override;

			// Override methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::createGeometricalConstraints()
{
	clearGeometricalConstraints();
	ConstraintManager::createGeometricalConstraintOwners(ModelAtomTracker::trackingAtomPairs(), atomicArrangement());

	for (auto& atom : atomicArrangement())
		atom.updateConstraintReferences();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::createGeometricalConstraints(const double minimumConstraintCutoffRatio, const double nonBondingCutoffRatio, const double ionicBondCutoffRatio, const double metallicBondCutoffRatio)
{
	clearGeometricalConstraints();
	ConstraintManager::createGeometricalConstraintOwners(ModelAtomTracker::trackingAtomPairs(), atomicArrangement(), minimumConstraintCutoffRatio, nonBondingCutoffRatio, ionicBondCutoffRatio, metallicBondCutoffRatio);

	for (auto& atom : atomicArrangement())
		atom.updateConstraintReferences();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::createChemicalBonds()
{
	clearChemicalBonds();
	ConstraintManager::createChemicalBondOwners(atomicArrangement());

	for (auto& atom : atomicArrangement())
		atom.updateConstraintReferences();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::createChemicalBonds(const double ionicBondCutoffRatio, const double metallicBondCutoffRatio)
{
	clearChemicalBonds();
	ConstraintManager::createChemicalBondOwners(atomicArrangement(), ionicBondCutoffRatio, metallicBondCutoffRatio);

	for (auto& atom : atomicArrangement())
		atom.updateConstraintReferences();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::substituteChemicalBonds(const double feasibleConstraintErrorRate)
{
	ConstraintManager::substituteIonicBondOwners(atomicArrangement(), feasibleConstraintErrorRate);
	ConstraintManager::clearCationicConstraintOwners();
	ConstraintManager::createCationicConstraintOwners(atomicArrangement());

	for (auto& atom : atomicArrangement())
		atom.updateConstraintReferences();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::clearGeometricalConstraints()
{
	ConstraintManager::clearGeometricalConstraintOwners();

	for (auto& atom : atomicArrangement())
		atom.clearConstraintReferences();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::clearChemicalBonds()
{
	ConstraintManager::clearIonicBondOwners();
	ConstraintManager::clearMetallicBondOwners();
	ConstraintManager::clearCationicConstraintOwners();

	for (auto& atom : atomicArrangement())
		atom.clearChemicalBondReferences();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline bool MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::isOptimalSolution(const double feasibleConstraintErrorRate) const
{
	bool isFeasibleLigandComposition = true;
	{
		for (const auto& atom : atomicArrangement())
		{
			if (!(atom.isFeasibleChemicalBondComposition()))
				isFeasibleLigandComposition = false;
		}
	}

	if (isFeasibleLigandComposition)
		return !(hasInfeasibleGeometricalConstraints(feasibleConstraintErrorRate));
	else
		return false;
}

inline MathematicalCrystalChemistry::StructureModel::Features::AtomCategoryComposition MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::getAtomCategoryComposition() const
{
	AtomCategoryComposition atomCategoryComposition;
	{
		for (const auto& modelAtom : atomicArrangement())
			atomCategoryComposition.add(modelAtom.coordinationFeature().atomCategory().symbol());
	}

	return atomCategoryComposition;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Override methods

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::clearSpaceGroupInfo()
{
	CrystalStructure::clearSpaceGroupInfo();

	for (auto& modelAtom : atomicArrangement())
		modelAtom.setSiteSymmetrySymbol();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::conventionalizeStructure()
{
	ModelAtomTracker::clearTrackingAtomPairs();
	CrystalStructure::conventionalizeStructure();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::refineStructure()
{
	ModelAtomTracker::clearTrackingAtomPairs();
	CrystalStructure::refineStructure();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::refineStructure(const double spaceGroupPrecision)
{
	ModelAtomTracker::clearTrackingAtomPairs();
	CrystalStructure::refineStructure(spaceGroupPrecision);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::normalizeFractionalCoordinates()
{
	ModelAtomTracker::clearTrackingAtomPairs();
	CrystalStructure::normalizeFractionalCoordinates();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::sort()
{
	ModelAtomTracker::clearTrackingAtomPairs();
	CrystalStructure::sort();
}

inline void MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure::eraseAtoms(const std::string& elementSymbol)
{
	ModelAtomTracker::clearTrackingAtomPairs();
	CrystalStructure::eraseAtoms(elementSymbol);
}

// Override methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_MODELCRYSTALSTRUCTURE_H
