#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_NONIONICBOND_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_NONIONICBOND_H

#include "ReportException.h"

#include "ConstraintOwner.h"
#include "MinimumDistanceConstraint.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			class NonIonicBond :public Internal::MinimumDistanceConstraint
			{
				using NonIonicBondOwner = ConstraintOwner<NonIonicBond>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors and destructor

			private:
				NonIonicBond(ModelAtom&, const Ligand&);
				NonIonicBond(ModelAtom&, const Ligand&, const double nonBondingRangeRatio);
				NonIonicBond(ModelAtom&, Ligand&&);
				NonIonicBond(ModelAtom&, Ligand&&, const double nonBondingRangeRatio);
				NonIonicBond(const Ligand&, const Ligand&);
				NonIonicBond(const Ligand&, const Ligand&, const double nonBondingRangeRatio);
				explicit NonIonicBond(const std::shared_ptr<NonIonicBond>& reversePtr);

			public:
				virtual ~NonIonicBond() = default;

			// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				double nonBondingRangeRatio() const noexcept;

				void setNonBondingRangeRatio() noexcept;
				void setNonBondingRangeRatio(const double);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static NonIonicBondOwner create(const ConstraintKey&);
				static bool shouldCreate(const ModelAtom&, const Ligand&, const double cutoffRatio, const double nonBondingRangeRatio);
				static bool shouldCreate(const ConstraintKey&, const double cutoffRatio, const double nonBondingRangeRatio);

				static double defaultNonBondingRangeRatio() noexcept;
				static void setDefaultNonBondingRangeRatio() noexcept;
				static void setDefaultNonBondingRangeRatio(const double);

				static double defaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio(const double);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods
				
			private:
				static bool canCreate(const ModelAtom&, const Ligand&);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				double m_nonBondingRangeRatio;

				static double m_defaultNonBondingRangeRatio;
				static double m_defaultCutoffRatio;


			private:
				NonIonicBond() = delete;
				NonIonicBond(const NonIonicBond&) = delete;
				NonIonicBond(NonIonicBond&&) noexcept = delete;
				NonIonicBond& operator=(const NonIonicBond&) = delete;
				NonIonicBond& operator=(NonIonicBond&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::nonBondingRangeRatio() const noexcept
{
	return m_nonBondingRangeRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::setNonBondingRangeRatio() noexcept
{
	m_nonBondingRangeRatio = m_defaultNonBondingRangeRatio;
	MinimumDistanceConstraint::setMinimumDistance(m_nonBondingRangeRatio * (modelAtomRef().coordinationFeature().upperIonicBondRadius() + ligand().get().coordinationFeature().upperIonicBondRadius()));
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::setNonBondingRangeRatio(const double val)
{
	if (0.0 < val)
	{
		m_defaultNonBondingRangeRatio = val;
		MinimumDistanceConstraint::setMinimumDistance(m_nonBondingRangeRatio * (modelAtomRef().coordinationFeature().upperIonicBondRadius() + ligand().get().coordinationFeature().upperIonicBondRadius()));
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "setNonBondingRangeRatio", "\"nonBondingRangeRatio\" is not more than zero."));
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::NonIonicBondOwner MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::create(const ConstraintKey& constraintKey)
{
	ModelAtom& modelAtom = constraintKey.modelAtomRef();
	const Ligand& ligand = constraintKey.ligand();

	if (canCreate(modelAtom, ligand))
	{
		std::shared_ptr<NonIonicBond> nonIonicBondPtr{ new NonIonicBond(modelAtom, ligand) };
		std::shared_ptr<NonIonicBond> reverseNonIonicBondPtr{ new NonIonicBond(nonIonicBondPtr) };
		nonIonicBondPtr->setReversePtr(std::static_pointer_cast<InteratomicConstraintBase>(reverseNonIonicBondPtr));

		modelAtom.track(nonIonicBondPtr);
		ligand.get().track(reverseNonIonicBondPtr);

		return NonIonicBondOwner{ std::move(nonIonicBondPtr), std::move(reverseNonIonicBondPtr) };
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException("create of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond", "Could not create non-ionic bond."));
	}
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::shouldCreate(const ModelAtom& modelAtom, const Ligand& ligand, const double cutoffRatio, const double nonBondingRangeRatio)
{
	if (canCreate(modelAtom, ligand))
	{
		NumericalVector displacement = ligand.coordinate() - modelAtom.coordinate();
		double cutoffLength = cutoffRatio * nonBondingRangeRatio * (modelAtom.coordinationFeature().upperIonicBondRadius() + ligand.get().coordinationFeature().upperIonicBondRadius());

		if (displacement.normSquare() < (cutoffLength * cutoffLength))
			return true;
		else
			return false;
	}

	else
		return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::shouldCreate(const ConstraintKey& constraintKey, const double cutoffRatio, const double nonBondingRangeRatio)
{
	return shouldCreate(constraintKey.modelAtomRef(), constraintKey.ligand(), cutoffRatio, nonBondingRangeRatio);
}

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::defaultNonBondingRangeRatio() noexcept
{
	return m_defaultNonBondingRangeRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::setDefaultNonBondingRangeRatio() noexcept
{
	m_defaultNonBondingRangeRatio = 1.2;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::setDefaultNonBondingRangeRatio(const double val)
{
	if (0.0 < val)
		m_defaultNonBondingRangeRatio = val;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("setDefaultNonBondingRangeRatio of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond", "\"nonBondingRangeRatio\" is not more than zero."));
	}
}

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::defaultCutoffRatio() noexcept
{
	return m_defaultCutoffRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::setDefaultCutoffRatio() noexcept
{
	m_defaultCutoffRatio = 2.0;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::setDefaultCutoffRatio(const double val)
{
	if (0.0 < val)
		m_defaultCutoffRatio = val;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("setDefaultCutoffRatio of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond", "\"cutoffRatio\" is not more than zero."));
	}
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond::canCreate(const ModelAtom& modelAtom, const Ligand& ligand)
{
	if ((modelAtom.coordinationFeature().atomCategory().isAnion() && ligand.get().coordinationFeature().atomCategory().isCation()) || (modelAtom.coordinationFeature().atomCategory().isCation() && ligand.get().coordinationFeature().atomCategory().isAnion()))
		return true;
	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_NONIONICBOND_H
