#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_NONMETALLICBOND_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_NONMETALLICBOND_H

#include "ReportException.h"

#include "ConstraintOwner.h"
#include "MinimumDistanceConstraint.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace GeometricalConstraints
		{
			class NonMetallicBond :public Internal::MinimumDistanceConstraint
			{
				using NonMetallicBondOwner = ConstraintOwner<NonMetallicBond>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors and destructor

			private:
				NonMetallicBond(ModelAtom&, const Ligand&);
				NonMetallicBond(ModelAtom&, const Ligand&, const double nonBondingRangeRatio);
				NonMetallicBond(ModelAtom&, Ligand&&);
				NonMetallicBond(ModelAtom&, Ligand&&, const double nonBondingRangeRatio);
				NonMetallicBond(const Ligand&, const Ligand&);
				NonMetallicBond(const Ligand&, const Ligand&, const double nonBondingRangeRatio);
				explicit NonMetallicBond(const std::shared_ptr<NonMetallicBond>& reversePtr);

			public:
				virtual ~NonMetallicBond() = default;

			// Constructors and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				double nonBondingRangeRatio() const noexcept;

				void setNonBondingRangeRatio();
				void setNonBondingRangeRatio(const double);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Static methods

				static NonMetallicBondOwner create(const ConstraintKey&);
				static bool shouldCreate(const ModelAtom&, const Ligand&, const double cutoffRatio, const double nonBondingRangeRatio);
				static bool shouldCreate(const ConstraintKey&, const double cutoffRatio, const double nonBondingRangeRatio);

				static double defaultNonBondingRangeRatio() noexcept;
				static void setDefaultNonBondingRangeRatio() noexcept;
				static void setDefaultNonBondingRangeRatio(const double nonBondingRangeRatio);

				static double defaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio() noexcept;
				static void setDefaultCutoffRatio(const double cutoffRatio);

			// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				static bool canCreate(const ModelAtom&, const Ligand&);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				double m_nonBondingRangeRatio;
				
				static double m_defaultNonBondingRangeRatio;
				static double m_defaultCutoffRatio;


			private:
				NonMetallicBond() = delete;
				NonMetallicBond(const NonMetallicBond&) = delete;
				NonMetallicBond(NonMetallicBond&&) noexcept = delete;
				NonMetallicBond& operator=(const NonMetallicBond&) = delete;
				NonMetallicBond& operator=(NonMetallicBond&&) noexcept = delete;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::nonBondingRangeRatio() const noexcept
{
	return m_nonBondingRangeRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::setNonBondingRangeRatio()
{
	m_nonBondingRangeRatio = m_defaultNonBondingRangeRatio;
	MinimumDistanceConstraint::setMinimumDistance(m_nonBondingRangeRatio * (modelAtomRef().coordinationFeature().upperMetallicBondRadius() + ligand().get().coordinationFeature().upperMetallicBondRadius()));
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::setNonBondingRangeRatio(const double val)
{
	if (0.0 < val)
	{
		m_defaultNonBondingRangeRatio = val;
		MinimumDistanceConstraint::setMinimumDistance(m_nonBondingRangeRatio * (modelAtomRef().coordinationFeature().upperIonicBondRadius() + ligand().get().coordinationFeature().upperIonicBondRadius()));
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "setNonBondingRangeRatio", "\"nonBondingRangeRatio\" is not more than zero."));
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

inline MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::NonMetallicBondOwner MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::create(const ConstraintKey& constraintKey)
{
	ModelAtom& modelAtom = constraintKey.modelAtomRef();
	const Ligand& ligand = constraintKey.ligand();

	if (canCreate(modelAtom, ligand))
	{
		std::shared_ptr<NonMetallicBond> nonMetallicBondPtr{ new NonMetallicBond(modelAtom, ligand) };
		std::shared_ptr<NonMetallicBond> reverseNonMetallicBondPtr{ new NonMetallicBond(nonMetallicBondPtr) };
		nonMetallicBondPtr->setReversePtr(std::static_pointer_cast<InteratomicConstraintBase>(reverseNonMetallicBondPtr));

		modelAtom.track(nonMetallicBondPtr);
		ligand.get().track(reverseNonMetallicBondPtr);

		return NonMetallicBondOwner{ std::move(nonMetallicBondPtr), std::move(reverseNonMetallicBondPtr) };
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException("create of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond", "Could not create non-metallic bond."));
	}
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::shouldCreate(const ModelAtom& modelAtom, const Ligand& ligand, const double cutoffRatio, const double nonBondingRangeRatio)
{
	if (canCreate(modelAtom, ligand))
	{
		NumericalVector displacement = ligand.coordinate() - modelAtom.coordinate();
		double cutoffLength = cutoffRatio * nonBondingRangeRatio * (modelAtom.coordinationFeature().upperMetallicBondRadius() + ligand.get().coordinationFeature().upperMetallicBondRadius());

		if (displacement.normSquare() < (cutoffLength * cutoffLength))
			return true;
		else
			return false;
	}

	else
		return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::shouldCreate(const ConstraintKey& constraintKey, const double cutoffRatio, const double nonBondingRangeRatio)
{
	return shouldCreate(constraintKey.modelAtomRef(), constraintKey.ligand(), cutoffRatio, nonBondingRangeRatio);
}

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::defaultNonBondingRangeRatio() noexcept
{
	return m_defaultNonBondingRangeRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::setDefaultNonBondingRangeRatio() noexcept
{
	m_defaultNonBondingRangeRatio = 1.2;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::setDefaultNonBondingRangeRatio(const double val)
{
	if (0.0 < val)
		m_defaultNonBondingRangeRatio = val;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("setDefaultNonBondingRangeRatio of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond", "\"defaultNonBondingRangeRatio\" is not more than zero."));
	}
}

inline double MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::defaultCutoffRatio() noexcept
{
	return m_defaultCutoffRatio;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::setDefaultCutoffRatio() noexcept
{
	m_defaultCutoffRatio = 2.0;
}

inline void MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::setDefaultCutoffRatio(const double val)
{
	if (0.0 < val)
		m_defaultCutoffRatio = val;

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("setDefaultCutoffRatio of MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond", "\"defaultCutoffRatio\" is not more than zero."));
	}
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond::canCreate(const ModelAtom& modelAtom, const Ligand& ligand)
{
	if (modelAtom.coordinationFeature().atomCategory().isMetal() && ligand.get().coordinationFeature().atomCategory().isMetal())
		return true;
	else
		return false;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_GEOMETRICALCONSTRAINTS_NONMETALLICBOND_H
