#ifndef MATHTOOLKIT_GENERIC_NUMERICALMATRIX_H
#define MATHTOOLKIT_GENERIC_NUMERICALMATRIX_H

#include <initializer_list>
#include <utility>
#include <vector>

#include "NumericalMatrixBase.h"
#include "NumericalVector.h"


namespace MathToolkit
{
	namespace Generic
	{
		/*
		Column Major Layout for m_columnSize-by-m_rowSize matrix
		*/

		template <typename T>
		class NumericalMatrix :public MathToolkit::Generic::Internal::NumericalMatrixBase<T>
		{
		public:
			using size_type = std::size_t;
			using value_type = T;
			using pointer = value_type*;
			using const_pointer = const value_type*;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, operators, and destructor

		public:
			NumericalMatrix() = default;
			NumericalMatrix(const std::vector<NumericalVector<T>>&);
			NumericalMatrix(const size_type columnSize, const size_type rowSize, const value_type value = value_type{});
			NumericalMatrix(const size_type columnSize, const size_type rowSize, NumericalVector<value_type>&&);
			NumericalMatrix(const double scalingMultiplier, const NumericalVector<value_type>&, const NumericalVector<value_type>&);

			template <typename U>
			NumericalMatrix(const NumericalMatrix<U>);

			NumericalMatrix(const NumericalMatrix&) = default;
			NumericalMatrix(NumericalMatrix&&) noexcept = default;
			NumericalMatrix& operator=(const NumericalMatrix&) = default;
			NumericalMatrix& operator=(NumericalMatrix&&) noexcept = default;
			NumericalMatrix& operator=(const value_type scalar);

			template <typename U>
			NumericalMatrix& operator=(const NumericalMatrix<U>);

			virtual ~NumericalMatrix() = default;

		// Constructors, operators, and destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Unary operators

			NumericalMatrix& operator+=(const NumericalMatrix&);
			NumericalMatrix& operator-=(const NumericalMatrix&);
			NumericalMatrix& operator*=(const NumericalMatrix&);
			NumericalMatrix& operator*=(const value_type scalingMultiplier);
			NumericalMatrix& operator/=(const value_type scalingDivider);

		// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Mathematical methods

			NumericalMatrix& add(const value_type scalingMultiplier, const NumericalMatrix&);
			NumericalMatrix& add(const value_type scalingMultiplier, const NumericalMatrix&, const value_type selfScalingMultiplier);
			NumericalMatrix& transform(const NumericalMatrix&);
			NumericalMatrix& transform(const value_type scalingMultiplier, const NumericalMatrix&);
			NumericalMatrix& multiply(const value_type scalingMultiplier, const NumericalMatrix&);

			NumericalMatrix& invert();
			NumericalMatrix getInverseMatrix() const;
			NumericalMatrix getTransposedMatrix() const;

			void getSingularValueDecomposition() const;

		// Mathematical methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			NumericalVector<T> columnVector(const size_type columnIndex) const;
			NumericalVector<T> rowVector(const size_type rowIndex) const;
			NumericalVector<T> toNumericalVector();

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

		};


		template <typename T>
		inline NumericalMatrix<T> getIdentityMatrix(const std::size_t rank)
		{
			NumericalMatrix<T> identityMatrix(rank, rank);
			{
				for (std::size_t index = 0; index < rank; ++index)
					identityMatrix(index, index) = 1.0;
			}

			return identityMatrix;
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors and operators

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>::NumericalMatrix(const std::vector<NumericalVector<T>>& numericalVectors)
	: MathToolkit::Generic::Internal::NumericalMatrixBase<T>(numericalVectors.at(0).size(), numericalVectors.size(), value_type{})
{
	size_type columnSize = numericalVectors[0].size();
	auto thisIter = std::begin(this->m_valueArray);

	for (const auto& numericalVector : numericalVectors)
	{
		if (columnSize == numericalVector.size())
		{
			for (const auto value : numericalVector)
			{
				(*thisIter) = value;
				++thisIter;
			}
		}

		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "constructor", "The column size of \"numericalVectors\" is not consistent."));
		}
	}
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>::NumericalMatrix(const size_type columnSize, const size_type rowSize, const value_type value)
	: MathToolkit::Generic::Internal::NumericalMatrixBase<T>{ columnSize, rowSize, value }
{
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>::NumericalMatrix(const size_type columnSize, const size_type rowSize, NumericalVector<T>&& numericalVector)
	: MathToolkit::Generic::Internal::NumericalMatrixBase<T>{ columnSize, rowSize, T{} }
{
	if (this->m_valueArray.size() == numericalVector.size())
		this->m_valueArray = std::move(numericalVector.m_valueArray);

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", "The size of \"numericalVector\" is not equal to (\"columnSize\" * \"rowSize\")."));
	}
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>::NumericalMatrix(const double scalingMultiplier, const NumericalVector<value_type>& nva, const NumericalVector<value_type>& nvb)
	: MathToolkit::Generic::Internal::NumericalMatrixBase<T>{ scalingMultiplier, nva, nvb }
{
}

template <typename T>
template <typename U>
inline MathToolkit::Generic::NumericalMatrix<T>::NumericalMatrix(const NumericalMatrix<U> nm)
	: MathToolkit::Generic::Internal::NumericalMatrixBase<T>(nm)
{
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::operator=(const value_type scalar)
{
	this->m_valueArray = scalar;
	return *this;
}

template <typename T>
template <typename U>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::operator=(const NumericalMatrix<U> nm)
{
	MathToolkit::Generic::Internal::NumericalMatrixBase<T>::operator=(nm);
	return *this;
}

// Constructors and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Unary operators

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::operator+=(const NumericalMatrix& numericalMatrix)
{
	this->m_valueArray += numericalMatrix.m_valueArray;
	return *this;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::operator-=(const NumericalMatrix& numericalMatrix)
{
	this->m_valueArray -= numericalMatrix.m_valueArray;
	return *this;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::operator*=(const NumericalMatrix& nm)
{
	NumericalMatrix resultMatrix(this->columnSize(), nm.rowSize(), 0.0);
	{
		for (size_type rowIndex = 0; rowIndex < resultMatrix.columnSize(); ++rowIndex)
		{
			auto nmIter = nm.begin();
			auto resultIter = resultMatrix.begin();
			resultIter += rowIndex;


			for (size_type columnIndex = 0; columnIndex < resultMatrix.rowSize(); ++columnIndex)
			{
				auto thisIter = this->begin();
				thisIter += rowIndex;

				for (size_type index = 0; index < this->rowSize(); ++index)
				{
					(*resultIter) += (*thisIter) * (*nmIter);
					thisIter += this->columnSize();
					++nmIter;
				}

				resultIter += resultMatrix.columnSize();
			}
		}
	}

	this->swap(resultMatrix);
	return *this;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::operator*=(const value_type scalingMultiplier)
{
	this->m_valueArray *= scalingMultiplier;
	return *this;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::operator/=(const value_type scalingDivider)
{
	const value_type scalingMultiplier = (1.0 / scalingDivider);
	this->m_valueArray *= scalingMultiplier;
	return *this;
}

// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Mathematical methods

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::add(const value_type scalingMultiplier, const NumericalMatrix& numericalMatrix)
{
	auto iter = std::begin(numericalMatrix.m_valueArray);

	for (auto& value : this->m_valueArray)
	{
		value += (*iter) * scalingMultiplier;
		++iter;
	}

	return *this;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::add(const value_type scalingMultiplier, const NumericalMatrix& numericalMatrix, const value_type selfScalingMultiplier)
{
	auto iter = std::begin(numericalMatrix.m_valueArray);

	for (auto& value : this->m_valueArray)
	{
		value *= selfScalingMultiplier;
		value += (*iter) * scalingMultiplier;
		++iter;
	}

	return *this;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::transform(const NumericalMatrix& nm)
{
	NumericalMatrix resultMatrix(nm.columnSize(), this->rowSize());
	{
		for (std::size_t rowIndex = 0; rowIndex < resultMatrix.columnSize(); ++rowIndex)
		{
			auto thisIter = this->begin();
			auto resultIter = resultMatrix.begin();
			resultIter += rowIndex;

			for (std::size_t columnIndex = 0; columnIndex < resultMatrix.rowSize(); ++columnIndex)
			{
				auto nmIter = nm.begin();
				nmIter += rowIndex;

				for (std::size_t index = 0; index < nm.rowSize(); ++index)
				{
					(*resultIter) += (*nmIter) * (*thisIter);
					nmIter += nm.columnSize();
					++thisIter;
				}

				resultIter += resultMatrix.columnSize();
			}
		}
	}

	this->swap(resultMatrix);
	return *this;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::transform(const value_type scalingMultiplier, const NumericalMatrix& nm)
{
	NumericalMatrix resultMatrix(nm.columnSize(), this->rowSize());
	{
		for (std::size_t rowIndex = 0; rowIndex < resultMatrix.columnSize(); ++rowIndex)
		{
			auto thisIter = this->begin();
			auto resultIter = resultMatrix.begin();
			resultIter += rowIndex;

			for (std::size_t columnIndex = 0; columnIndex < resultMatrix.rowSize(); ++columnIndex)
			{
				auto nmIter = nm.begin();
				nmIter += rowIndex;

				for (std::size_t index = 0; index < nm.rowSize(); ++index)
				{
					(*resultIter) += (*nmIter) * (*thisIter);
					nmIter += nm.columnSize();
					++thisIter;
				}

				(*resultIter) *= scalingMultiplier;
				resultIter += resultMatrix.columnSize();
			}
		}
	}

	this->swap(resultMatrix);
	return *this;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T>& MathToolkit::Generic::NumericalMatrix<T>::multiply(const value_type scalingMultiplier, const NumericalMatrix& nm)
{
	NumericalMatrix resultMatrix(this->columnSize(), nm.rowSize(), 0.0);
	{
		for (size_type rowIndex = 0; rowIndex < resultMatrix.columnSize(); ++rowIndex)
		{
			auto nmIter = nm.begin();
			auto resultIter = resultMatrix.begin();
			resultIter += rowIndex;


			for (size_type columnIndex = 0; columnIndex < resultMatrix.rowSize(); ++columnIndex)
			{
				auto thisIter = this->begin();
				thisIter += rowIndex;

				for (size_type index = 0; index < this->rowSize(); ++index)
				{
					(*resultIter) += (*thisIter) * (*nmIter);
					thisIter += this->columnSize();
					++nmIter;
				}

				(*resultIter) *= scalingMultiplier;
				resultIter += resultMatrix.columnSize();
			}
		}
	}

	this->swap(resultMatrix);
	return *this;
}

template <typename T>
MathToolkit::Generic::NumericalMatrix<T> MathToolkit::Generic::NumericalMatrix<T>::getInverseMatrix() const
{
	NumericalMatrix inverseMatrix{ *this };
	inverseMatrix.invert();
	return inverseMatrix;
}

template <typename T>
inline MathToolkit::Generic::NumericalMatrix<T> MathToolkit::Generic::NumericalMatrix<T>::getTransposedMatrix() const
{
	NumericalMatrix<T> transposedMatrix(this->rowSize(), this->columnSize());
	{
		auto thisIter = std::begin(this->m_valueArray);

		for (std::size_t columnIndex = 0; columnIndex < this->rowSize(); ++columnIndex)
		{
			auto transposedIter = transposedMatrix.begin();
			transposedIter += columnIndex;

			for (std::size_t rowIndex = 0; rowIndex < this->columnSize(); ++rowIndex)
			{
				(*transposedIter) = (*thisIter);
				transposedIter += this->rowSize();
				++thisIter;
			}
		}
	}

	return transposedMatrix;
}

// Mathematical methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

template <typename T>
inline MathToolkit::Generic::NumericalVector<T> MathToolkit::Generic::NumericalMatrix<T>::columnVector(const size_type columnIndex) const
{
	if (columnIndex < this->rowSize())
	{
		NumericalVector<T> columnVector;
		columnVector.m_valueArray = this->m_valueArray[std::slice(columnIndex * this->columnSize(), this->columnSize(), 1)];
		return columnVector;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "operator[]", "\"columnIndex\" is out of range."));
	}
}

template <typename T>
inline MathToolkit::Generic::NumericalVector<T> MathToolkit::Generic::NumericalMatrix<T>::rowVector(const size_type rowIndex) const
{
	if (rowIndex < this->columnSize())
	{
		NumericalVector<T> rowVector;
		rowVector.m_valueArray = this->m_valueArray[std::slice(rowIndex, this->rowSize(), this->columnSize())];
		return rowVector;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "rowVectorSlice", "\"rowIndex\" is out of range."));
	}
}

template <typename T>
inline MathToolkit::Generic::NumericalVector<T> MathToolkit::Generic::NumericalMatrix<T>::toNumericalVector()
{
	NumericalVector<T> numericalVector;
	numericalVector.m_valueArray = std::move(this->m_valueArray);
	this->clear();

	return numericalVector;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !MATHTOOLKIT_GENERIC_NUMERICALMATRIX_H
