#ifndef MATHTOOLKIT_GENERIC_INTERNAL_NUMERICALMATRIXBASE_H
#define MATHTOOLKIT_GENERIC_INTERNAL_NUMERICALMATRIXBASE_H

#include <string>
#include <stdexcept>
#include <type_traits>
#include <initializer_list>
#include <algorithm>
#include <limits>
#include <complex>
#include <valarray>
#include <cmath>

#include "ReportException.h"
#include "NumericalVectorBase.h"
#include "NumericalVectorSlice.h"


namespace MathToolkit
{
	namespace Generic
	{
		namespace Internal
		{
			template <typename T>
			class NumericalMatrixBase
			{
				using size_type = std::size_t;
				using value_type = T;
				using pointer = value_type*;
				using const_pointer = const value_type*;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			protected:
				NumericalMatrixBase();
				NumericalMatrixBase(const size_type columnSize, const size_type rowSize, const value_type value);
				NumericalMatrixBase(const double scalingMultiplier, const NumericalVectorBase<value_type>&, const NumericalVectorBase<value_type>&);

			public:
				NumericalMatrixBase(const NumericalMatrixBase&) = default;
				NumericalMatrixBase(NumericalMatrixBase&&) noexcept = default;
				NumericalMatrixBase& operator=(const NumericalMatrixBase&) = default;
				NumericalMatrixBase& operator=(NumericalMatrixBase&&) noexcept = default;

				virtual ~NumericalMatrixBase() = default;

				// Constructors, destructor, and operators
	// **********************************************************************************************************************************************************************************************************************************************************************************************
	// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

				value_type operator()(const size_type rowIndex, const size_type columnIndex) const;
				value_type& operator()(const size_type rowIndex, const size_type columnIndex);
				value_type at(const size_type rowIndex, const size_type columnIndex) const;
				value_type& at(const size_type rowIndex, const size_type columnIndex);

				pointer begin();
				const_pointer begin() const;
				pointer end();
				const_pointer end() const;

				MathToolkit::Generic::NumericalVectorSlice<value_type> columnVectorSlice(const size_type columnIndex);
				MathToolkit::Generic::NumericalVectorSlice<value_type> rowVectorSlice(const size_type rowIndex);

				size_type columnSize() const;
				size_type rowSize() const;

				// Access methods
	// **********************************************************************************************************************************************************************************************************************************************************************************************
	// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Customize methods

				void swap(NumericalMatrixBase&);
				void resize(const size_type columnSize, const size_type rowSize, const value_type value = value_type{});
				void clear();

				void pushbackColumn(const NumericalVectorBase<T>&);
				void pushbackRow(const NumericalVectorBase<T>&);
				void insertColumn(const NumericalVectorBase<T>&, const size_type columnIndex);
				void insertRow(const NumericalVectorBase<T>&, const size_type rowIndex);
				void eraseColumn(const size_type columnIndex);
				void eraseRow(const size_type rowIndex);

				// Customize methods
	// **********************************************************************************************************************************************************************************************************************************************************************************************
	// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Utility

				bool isEmpty() const;
				bool isSquareMatrix() const;
				bool isSame(const NumericalMatrixBase&, const double valuePrecision = std::numeric_limits<double>::epsilon()) const;

				value_type getTrace() const;
				double getHilbertSchmidtNorm() const;
				double getHilbertSchmidtNormSquare() const;

				// Utility
	// **********************************************************************************************************************************************************************************************************************************************************************************************

			protected:
				std::valarray<value_type> m_valueArray;

			private:
				size_type m_columnSize;
				size_type m_rowSize;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

template <typename T>
inline T MathToolkit::Generic::Internal::NumericalMatrixBase<T>::operator()(const size_type rowIndex, const size_type columnIndex) const
{
	return m_valueArray[(columnIndex * m_columnSize) + rowIndex];
}

template <typename T>
inline T& MathToolkit::Generic::Internal::NumericalMatrixBase<T>::operator()(const size_type rowIndex, const size_type columnIndex)
{
	return m_valueArray[(columnIndex * m_columnSize) + rowIndex];
}

template <typename T>
inline T MathToolkit::Generic::Internal::NumericalMatrixBase<T>::at(const size_type rowIndex, const size_type columnIndex) const
{
	if (rowIndex < m_columnSize && columnIndex < m_rowSize)
		return m_valueArray[(columnIndex * m_columnSize) + rowIndex];

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "at", "The index is out of range."));
	}
}

template <typename T>
inline T& MathToolkit::Generic::Internal::NumericalMatrixBase<T>::at(const size_type rowIndex, const size_type columnIndex)
{
	if (rowIndex < m_columnSize && columnIndex < m_rowSize)
		return m_valueArray[(columnIndex * m_columnSize) + rowIndex];

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "at", "The index is out of range."));
	}
}

template <typename T>
inline T* MathToolkit::Generic::Internal::NumericalMatrixBase<T>::begin()
{
	return std::begin(m_valueArray);
}

template <typename T>
inline const T* MathToolkit::Generic::Internal::NumericalMatrixBase<T>::begin() const
{
	return std::begin(m_valueArray);
}

template <typename T>
inline T* MathToolkit::Generic::Internal::NumericalMatrixBase<T>::end()
{
	return std::end(m_valueArray);
}

template <typename T>
inline const T* MathToolkit::Generic::Internal::NumericalMatrixBase<T>::end() const
{
	return std::end(m_valueArray);
}

template <typename T>
inline MathToolkit::Generic::NumericalVectorSlice<T> MathToolkit::Generic::Internal::NumericalMatrixBase<T>::columnVectorSlice(const size_type columnIndex)
{
	if (columnIndex < this->rowSize())
		return NumericalVectorSlice(this->m_valueArray, std::slice(columnIndex * m_columnSize, m_columnSize, 1));

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "columnVectorSlice", "\"columnIndex\" is out of range."));
	}
}

template <typename T>
inline MathToolkit::Generic::NumericalVectorSlice<T> MathToolkit::Generic::Internal::NumericalMatrixBase<T>::rowVectorSlice(const size_type rowIndex)
{
	if (rowIndex < this->columnSize())
		return NumericalVectorSlice(this->m_valueArray, std::slice(rowIndex, m_rowSize, m_columnSize));

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "rowVectorSlice", "\"rowIndex\" is out of range."));
	}
}


template <typename T>
inline std::size_t MathToolkit::Generic::Internal::NumericalMatrixBase<T>::columnSize() const
{
	return m_columnSize;
}

template <typename T>
inline std::size_t MathToolkit::Generic::Internal::NumericalMatrixBase<T>::rowSize() const
{
	return m_rowSize;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::swap(NumericalMatrixBase& numericalMatrixBase)
{
	std::swap(m_columnSize, numericalMatrixBase.m_columnSize);
	std::swap(m_rowSize, numericalMatrixBase.m_rowSize);
	m_valueArray.swap(numericalMatrixBase.m_valueArray);
}

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::resize(const size_type columnSize, const size_type rowSize, const value_type value)
{
	m_columnSize = columnSize;
	m_rowSize = rowSize;
	m_valueArray.resize((m_columnSize * m_rowSize), value);
}

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::clear()
{
	std::valarray<value_type> va;
	m_valueArray.swap(va);
	m_columnSize = 0;
	m_rowSize = 0;
}

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::pushbackColumn(const NumericalVectorBase<T>& columnVectorBase)
{
	insertColumn(columnVectorBase, rowSize());
}

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::pushbackRow(const NumericalVectorBase<T>& rowVectorBase)
{
	insertRow(rowVectorBase, columnSize());
}

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::insertColumn(const NumericalVectorBase<T>& columnVectorBase, const size_type columnIndex)
{
	if (columnIndex <= m_rowSize)
	{
		if (isEmpty())
		{
			m_valueArray = columnVectorBase.m_valueArray;
			m_columnSize = columnVectorBase.size();
			m_rowSize = 1;
		}

		else
		{
			std::valarray<value_type> newValueArray(m_columnSize * (m_rowSize + 1));
			{
				auto insertIter = std::begin(m_valueArray);
				insertIter += (m_columnSize * columnIndex);

				auto newInsertIter = std::begin(newValueArray);
				newInsertIter += (m_columnSize * columnIndex);

				std::copy(std::begin(m_valueArray), insertIter, std::begin(newValueArray));
				std::copy(columnVectorBase.begin(), columnVectorBase.end(), newInsertIter);
				std::copy(insertIter, std::end(m_valueArray), (newInsertIter += m_columnSize));
			}

			m_valueArray.swap(newValueArray);
			m_rowSize += 1;
		}
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "insertColumn", "\"columnIndex\" is out of range."));
	}
}

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::insertRow(const NumericalVectorBase<T>& rowVectorBase, const size_type rowIndex)
{
	if (rowIndex <= m_columnSize)
	{
		if (isEmpty())
		{
			m_valueArray = rowVectorBase.m_valueArray;
			m_columnSize = 1;
			m_rowSize = rowVectorBase.size();
		}

		else
		{
			std::valarray<value_type> newValueArray((m_columnSize + 1) * m_rowSize);
			{
				auto iter = std::begin(m_valueArray);
				auto newIter = std::begin(newValueArray);
				auto rowVectorIter = rowVectorBase.begin();
				size_type remainedSize = m_columnSize - rowIndex;

				for (size_type columnIndex = 0; columnIndex < m_rowSize; ++columnIndex)
				{
					std::copy(iter, (iter + rowIndex), newIter);
					iter += rowIndex;
					newIter += rowIndex;

					(*newIter) = (*rowVectorIter);
					++newIter;
					++rowVectorIter;

					std::copy(iter, (iter + remainedSize), newIter);
					iter += remainedSize;
					newIter += remainedSize;
				}
			}

			m_valueArray.swap(newValueArray);
			m_columnSize += 1;
		}
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "insertRow", "\"rowIndex\" is out of range."));
	}
}

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::eraseColumn(const size_type columnIndex)
{
	if (isEmpty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "eraseColumn", "This is empty."));
	}

	else
	{
		if (columnIndex < m_rowSize)
		{
			std::valarray<value_type> newValueArray(m_columnSize * (m_rowSize - 1));
			{
				auto erasedIter = std::begin(m_valueArray);
				auto newErasedIter = std::begin(newValueArray);
				{
					erasedIter += (m_columnSize * columnIndex);
					newErasedIter += (m_columnSize * columnIndex);
				}

				std::copy(std::begin(m_valueArray), erasedIter, std::begin(newValueArray));
				std::copy((erasedIter + m_columnSize), std::end(m_valueArray), newErasedIter);
			}

			m_valueArray.swap(newValueArray);
			--m_rowSize;
		}

		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::out_of_range(reportException(typeid(*this), "eraseColumn", "\"columnIndex\" is out of range."));
		}
	}
}

template <typename T>
inline void MathToolkit::Generic::Internal::NumericalMatrixBase<T>::eraseRow(const size_type rowIndex)
{
	if (isEmpty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "eraseRow", "This is empty."));
	}

	else
	{
		if (rowIndex < m_columnSize)
		{
			std::valarray<value_type> newValueArray((m_columnSize - 1) * m_rowSize);
			{
				auto iter = std::begin(m_valueArray);
				auto newIter = std::begin(newValueArray);
				size_type remainedSize = m_columnSize - (rowIndex + 1);

				for (size_type columnIndex = 0; columnIndex < m_rowSize; ++columnIndex)
				{
					std::copy(iter, (iter + rowIndex), newIter);
					iter += (rowIndex + 1);
					newIter += rowIndex;

					std::copy(iter, (iter + remainedSize), newIter);
					iter += remainedSize;
					newIter += remainedSize;
				}
			}

			m_valueArray.swap(newValueArray);
			--m_columnSize;
		}

		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::out_of_range(reportException(typeid(*this), "eraseRow", "\"rowIndex\" is out of range."));
		}
	}
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

template <typename T>
inline bool MathToolkit::Generic::Internal::NumericalMatrixBase<T>::isEmpty() const
{
	return !(static_cast<bool>(m_valueArray.size()));
}

template <typename T>
inline bool MathToolkit::Generic::Internal::NumericalMatrixBase<T>::isSquareMatrix() const
{
	if (isEmpty())
		return false;
	else
	{
		if (m_columnSize == m_rowSize)
			return true;
		else
			return false;
	}
}

template <>
inline bool MathToolkit::Generic::Internal::NumericalMatrixBase<double>::isSame(const NumericalMatrixBase& numericalMatrixBase, const double valuePrecision) const
{
	if ((m_columnSize == numericalMatrixBase.m_columnSize) && (m_rowSize == numericalMatrixBase.m_rowSize))
	{
		auto otherIter = numericalMatrixBase.begin();

		for (auto thisIter = begin(); thisIter != end(); ++thisIter)
		{
			double differenceSquare = 0.0;
			{
				value_type difference = ((*otherIter) - (*thisIter));
				differenceSquare += (difference * difference);
			}

			if (differenceSquare < (valuePrecision * valuePrecision))
				++otherIter;
			else
				return false;
		}

		return true;
	}

	else
		return false;
}

template <>
inline bool MathToolkit::Generic::Internal::NumericalMatrixBase<std::complex<double>>::isSame(const NumericalMatrixBase& numericalMatrixBase, const double valuePrecision) const
{
	if ((m_columnSize == numericalMatrixBase.m_columnSize) && (m_rowSize == numericalMatrixBase.m_rowSize))
	{
		auto otherIter = numericalMatrixBase.begin();

		for (auto thisIter = begin(); thisIter != end(); ++thisIter)
		{
			double differenceSquare = 0.0;
			{
				value_type difference = ((*otherIter) - (*thisIter));
				differenceSquare += (difference.real() * difference.real());
				differenceSquare += (difference.imag() * difference.imag());
			}

			if (differenceSquare < (valuePrecision * valuePrecision))
				++otherIter;
			else
				return false;
		}

		return true;
	}

	else
		return false;
}


template <typename T>
inline T MathToolkit::Generic::Internal::NumericalMatrixBase<T>::getTrace() const
{
	if (isSquareMatrix())
	{
		value_type value = 0.0;
		{
			for (size_type rowIndex = 0; rowIndex < m_columnSize; ++rowIndex)
				value += NumericalMatrixBase::operator()(rowIndex, rowIndex);
		}

		return value;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::logic_error(reportException(typeid(*this), "getTrace", "This not a square matrix."));
	}
}

template <typename T>
inline double MathToolkit::Generic::Internal::NumericalMatrixBase<T>::getHilbertSchmidtNorm() const
{
	return std::sqrt(getHilbertSchmidtNormSquare());
}

template <>
inline double MathToolkit::Generic::Internal::NumericalMatrixBase<double>::getHilbertSchmidtNormSquare() const
{
	double normSquare{ 0.0 };
	{
		for (const auto value : this->m_valueArray)
			normSquare += (value * value);
	}

	return normSquare;
}

template <>
inline double MathToolkit::Generic::Internal::NumericalMatrixBase<std::complex<double>>::getHilbertSchmidtNormSquare() const
{
	double normSquare{ 0.0 };
	{
		for (const auto value : this->m_valueArray)
		{
			normSquare += (value.real() * value.real());
			normSquare += (value.imag() * value.imag());
		}
	}

	return normSquare;
}


// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !MATHTOOLKIT_GENERIC_INTERNAL_NUMERICALMATRIXBASE_H
