#ifndef MATHTOOLKIT_GENERIC_INTERNAL_NUMERICALSYMMETRICMATRIXBASE_H
#define MATHTOOLKIT_GENERIC_INTERNAL_NUMERICALSYMMETRICMATRIXBASE_H

#include <string>
#include <stdexcept>
#include <type_traits>
#include <initializer_list>
#include <algorithm>
#include <limits>
#include <complex>
#include <valarray>
#include <cmath>

#include "ReportException.h"
#include "NumericalVectorBase.h"
#include "NumericalVectorSlice.h"


namespace MathToolkit
{
	namespace Generic
	{
		namespace Internal
		{
			class NumericalSymmetricMatrixBase
			{
				using size_type = std::size_t;
				using value_type = double;
				using pointer = value_type*;
				using const_pointer = const value_type*;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			protected:
				NumericalSymmetricMatrixBase();
				NumericalSymmetricMatrixBase(const size_type dimension, const value_type value);

			public:
				NumericalSymmetricMatrixBase(const NumericalSymmetricMatrixBase&) = default;
				NumericalSymmetricMatrixBase(NumericalSymmetricMatrixBase&&) noexcept = default;
				NumericalSymmetricMatrixBase& operator=(const NumericalSymmetricMatrixBase&) = default;
				NumericalSymmetricMatrixBase& operator=(NumericalSymmetricMatrixBase&&) noexcept = default;

				virtual ~NumericalSymmetricMatrixBase() = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				value_type operator()(const size_type rowIndex, const size_type columnIndex) const;
				value_type& operator()(const size_type rowIndex, const size_type columnIndex);
				value_type at(const size_type rowIndex, const size_type columnIndex) const;
				value_type& at(const size_type rowIndex, const size_type columnIndex);

				size_type dimension() const;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Customize methods

				void swap(NumericalSymmetricMatrixBase&);
				void resize(const size_type dimension, const value_type value = value_type{});
				void clear();

			// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Utility

				bool isEmpty() const;
				bool isSame(const NumericalSymmetricMatrixBase&, const double valuePrecision = std::numeric_limits<double>::epsilon()) const;

				value_type getTrace() const;

			// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Protected and private methods

			protected:
				pointer begin();
				pointer end();

			private:
				size_type getIndex(const size_type rowIndex, const size_type columnIndex) const;

			// Protected and private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			protected:
				std::valarray<value_type> m_valueArray;

			private:
				size_type m_dimension;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::NumericalSymmetricMatrixBase()
	: m_valueArray{}
	, m_dimension{ 0 }
{
}

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::NumericalSymmetricMatrixBase(const size_type dimension, const value_type value)
	: m_valueArray(value, ((dimension* (dimension + 1)) / 2))
	, m_dimension{ dimension }
{
	if (m_dimension == 0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", "\"dimension\" is zero."));
	}
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::value_type MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::operator()(const size_type rowIndex, const size_type columnIndex) const
{
	return m_valueArray[getIndex(rowIndex, columnIndex)];
}

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::value_type& MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::operator()(const size_type rowIndex, const size_type columnIndex)
{
	return m_valueArray[getIndex(rowIndex, columnIndex)];
}

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::value_type MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::at(const size_type rowIndex, const size_type columnIndex) const
{
	if (rowIndex < m_dimension && columnIndex < m_dimension)
		return m_valueArray[getIndex(rowIndex, columnIndex)];

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "at", "The index is out of range."));
	}
}

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::value_type& MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::at(const size_type rowIndex, const size_type columnIndex)
{
	if (rowIndex < m_dimension && columnIndex < m_dimension)
		return m_valueArray[getIndex(rowIndex, columnIndex)];

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "at", "The index is out of range."));
	}
}

inline std::size_t MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::dimension() const
{
	return m_dimension;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

inline void MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::swap(NumericalSymmetricMatrixBase& numericalMatrixBase)
{
	m_valueArray.swap(numericalMatrixBase.m_valueArray);
	std::swap(m_dimension, numericalMatrixBase.m_dimension);
}

inline void MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::resize(const size_type dimension, const value_type value)
{
	m_dimension = dimension;
	m_valueArray.resize(((dimension * (dimension + 1)) / 2), value);
}

inline void MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::clear()
{
	std::valarray<value_type> va;
	m_valueArray.swap(va);
	m_dimension = 0;
}


// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline bool MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::isEmpty() const
{
	return !(static_cast<bool>(m_valueArray.size()));
}

inline bool MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::isSame(const NumericalSymmetricMatrixBase& numericalMatrixBase, const double valuePrecision) const
{
	if (m_dimension == numericalMatrixBase.m_dimension)
	{
		auto otherIter = std::begin(numericalMatrixBase.m_valueArray);

		for (auto thisIter = std::begin(m_valueArray); thisIter != std::end(m_valueArray); ++thisIter)
		{
			double differenceSquare = 0.0;
			{
				value_type difference = ((*otherIter) - (*thisIter));
				differenceSquare += (difference * difference);
			}

			if (differenceSquare < (valuePrecision * valuePrecision))
				++otherIter;
			else
				return false;
		}

		return true;
	}

	else
		return false;
}

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::value_type MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::getTrace() const
{
	value_type value = 0.0;
	{
		for (size_type rowIndex = 0; rowIndex < m_dimension; ++rowIndex)
			value += NumericalSymmetricMatrixBase::operator()(rowIndex, rowIndex);
	}

	return value;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected and private methods

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::pointer MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::begin()
{
	return std::begin(m_valueArray);
}

inline MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::pointer MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::end()
{
	return std::end(m_valueArray);
}

inline std::size_t MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase::getIndex(const size_type rowIndex, const size_type columnIndex) const
{
	if (columnIndex < rowIndex)
	{
		size_type index = columnIndex;
		index += ((rowIndex * (rowIndex + 1)) / 2);
		return index;
	}

	else
	{
		size_type index = rowIndex;
		index += ((columnIndex * (columnIndex + 1)) / 2);
		return index;
	}
}

// Protected and private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !MATHTOOLKIT_GENERIC_INTERNAL_NUMERICALSYMMETRICMATRIXBASE_H
