#ifndef MATHTOOLKIT_GENERIC_NUMERICALVECTORSLICE_H
#define MATHTOOLKIT_GENERIC_NUMERICALVECTORSLICE_H

#include "ReportException.h"
#include "NumericalVectorBase.h"


namespace MathToolkit
{
	namespace Generic
	{
		namespace Internal
		{
			template <typename T>
			class NumericalMatrixBase;

			class NumericalSymmetricMatrixBase;
		}

		template <typename T>
		class NumericalVectorSlice final
		{
			friend MathToolkit::Generic::Internal::NumericalMatrixBase<T>;
			friend MathToolkit::Generic::Internal::NumericalSymmetricMatrixBase;

			using NumericalVectorBase = MathToolkit::Generic::Internal::NumericalVectorBase<T>;
			using size_type = std::size_t;
			using value_type = T;
			using pointer = value_type*;
			using const_pointer = const value_type*;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		private:
			NumericalVectorSlice(std::valarray<value_type>&, const std::slice&);

		public:
			NumericalVectorSlice(const NumericalVectorSlice&) = default;
			NumericalVectorSlice(NumericalVectorSlice&&) noexcept = default;
			~NumericalVectorSlice() = default;

			const NumericalVectorSlice& operator=(const NumericalVectorSlice&);
			void operator=(const NumericalVectorBase&);
			void operator=(const value_type scalar);

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Arithmetic methods

			void add(const value_type scalingMultiplier, const NumericalVectorBase&);
			void add(const value_type scalingMultiplier, const NumericalVectorSlice&);
			void add(const value_type scalingMultiplier, const NumericalVectorBase&, const value_type selfScalingMultiplier);
			void add(const value_type scalingMultiplier, const NumericalVectorSlice&, const value_type selfScalingMultiplier);

		// Arithmetic methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Unary operators

			void operator+=(const NumericalVectorBase&);
			void operator+=(const NumericalVectorSlice&);
			void operator-=(const NumericalVectorBase&);
			void operator-=(const NumericalVectorSlice&);
			void operator*=(const value_type scalingMultiplier);
			void operator/=(const value_type scalingDivider);

		// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			value_type operator[](const size_type index) const;
			value_type& operator[](const size_type index);
			value_type at(const size_type index) const;
			value_type& at(const size_type index);

			size_type start() const;
			size_type size() const;
			size_type stride() const;

			double norm() const;
			double normSquare() const;

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Friends and private methods

		private:
			pointer begin();
			const_pointer begin() const;

		// Friends and private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::valarray<value_type>& m_valueArraySlice;
			size_type m_start;
			size_type m_size;
			size_type m_stride;

		private:
			NumericalVectorSlice() = delete;
			NumericalVectorSlice& operator=(NumericalVectorSlice&&) = delete;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Arithmetic methods

template <typename T>
inline const MathToolkit::Generic::NumericalVectorSlice<T>& MathToolkit::Generic::NumericalVectorSlice<T>::operator=(const NumericalVectorSlice& numericalVectorSlice)
{
	if (m_size == numericalVectorSlice.m_size)
	{
		auto thisIter = begin();
		auto iter = numericalVectorSlice.begin();

		for (size_type index = 0; index < m_size; ++index)
		{
			*thisIter = (*iter);

			thisIter += m_stride;
			iter += numericalVectorSlice.m_stride;
		}

		return *this;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "operator=", "The size of \"numericalVectorSlice\" is not equal to the size of this."));
	}
}

template <typename T>
void MathToolkit::Generic::NumericalVectorSlice<T>::operator=(const NumericalVectorBase& numericalVectorBase)
{
	if (m_size == numericalVectorBase.size())
	{
		auto thisIter = begin();
		auto iter = numericalVectorBase.begin();

		for (size_type index = 0; index < m_size; ++index)
		{
			*thisIter = (*iter);

			thisIter += m_stride;
			++iter;
		}
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "operator=", "The size of \"numericalVectorBase\" is not equal to the size of this."));
	}
}

template <typename T>
void MathToolkit::Generic::NumericalVectorSlice<T>::operator=(const value_type scalar)
{
	auto thisIter = begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) = scalar;
		thisIter += m_stride;
	}
}


// Arithmetic methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Arithmetic methods

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::add(const value_type scalingMultiplier, const NumericalVectorBase& numericalVectorBase)
{
	auto thisIter = begin();
	auto iter = numericalVectorBase.begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) += (*iter) * scalingMultiplier;

		thisIter += m_stride;
		++iter;
	}
}

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::add(const value_type scalingMultiplier, const NumericalVectorSlice& numericalVectorSlice)
{
	auto thisIter = begin();
	auto iter = numericalVectorSlice.begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) += (*iter) * scalingMultiplier;

		thisIter += m_stride;
		iter += numericalVectorSlice.m_stride;
	}
}

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::add(const value_type scalingMultiplier, const NumericalVectorBase& numericalVectorBase, const value_type selfScalingMultiplier)
{
	auto thisIter = begin();
	auto iter = numericalVectorBase.begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) *= selfScalingMultiplier;
		(*thisIter) += (*iter) * scalingMultiplier;

		thisIter += m_stride;
		++iter;
	}
}

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::add(const value_type scalingMultiplier, const NumericalVectorSlice& numericalVectorSlice, const value_type selfScalingMultiplier)
{
	auto thisIter = begin();
	auto iter = numericalVectorSlice.begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) *= selfScalingMultiplier;
		(*thisIter) += (*iter) * scalingMultiplier;

		thisIter += m_stride;
		iter += numericalVectorSlice.m_stride;
	}
}

// Arithmetic methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Unary operators

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::operator+=(const NumericalVectorBase& numericalVectorBase)
{
	auto thisIter = begin();
	auto iter = numericalVectorBase.begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) += (*iter);

		thisIter += m_stride;
		++iter;
	}
}

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::operator+=(const NumericalVectorSlice& numericalVectorSlice)
{
	auto thisIter = begin();
	auto iter = numericalVectorSlice.begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) += (*iter);

		thisIter += m_stride;
		iter += numericalVectorSlice.m_stride;
	}
}

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::operator-=(const NumericalVectorBase& numericalVectorBase)
{
	auto thisIter = begin();
	auto iter = numericalVectorBase.begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) -= (*iter);

		thisIter += m_stride;
		++iter;
	}
}

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::operator-=(const NumericalVectorSlice& numericalVectorSlice)
{
	auto thisIter = begin();
	auto iter = numericalVectorSlice.begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) -= (*iter);

		thisIter += m_stride;
		iter += numericalVectorSlice.m_stride;
	}
}

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::operator*=(const value_type scalingMultiplier)
{
	auto thisIter = begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) *= scalingMultiplier;
		thisIter += m_stride;
	}
}

template <typename T>
inline void MathToolkit::Generic::NumericalVectorSlice<T>::operator/=(const value_type scalingDivider)
{
	const value_type scalingMultiplier = (1.0 / scalingDivider);
	auto thisIter = begin();

	for (size_type index = 0; index < m_size; ++index)
	{
		(*thisIter) *= scalingMultiplier;
		thisIter += m_stride;
	}
}

// Unary operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

template <typename T>
inline T MathToolkit::Generic::NumericalVectorSlice<T>::operator[](const size_type index) const
{
	return m_valueArraySlice[m_start + (m_stride * index)];
}

template <typename T>
inline T& MathToolkit::Generic::NumericalVectorSlice<T>::operator[](const size_type index)
{
	return m_valueArraySlice[m_start + (m_stride * index)];
}

template <typename T>
inline T MathToolkit::Generic::NumericalVectorSlice<T>::at(const size_type index) const
{
	if (index < m_size)
		return m_valueArraySlice[m_start + (m_stride * index)];

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "at", "The index is out of range."));
	}
}

template <typename T>
inline T& MathToolkit::Generic::NumericalVectorSlice<T>::at(const size_type index)
{
	if (index < m_size)
		return m_valueArraySlice[m_start + (m_stride * index)];

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::out_of_range(reportException(typeid(*this), "at", "The index is out of range."));
	}
}

template <typename T>
inline std::size_t MathToolkit::Generic::NumericalVectorSlice<T>::start() const
{
	return m_start;
}

template <typename T>
inline std::size_t MathToolkit::Generic::NumericalVectorSlice<T>::size() const
{
	return m_size;
}

template <typename T>
inline std::size_t MathToolkit::Generic::NumericalVectorSlice<T>::stride() const
{
	return m_stride;
}

template <typename T>
inline double MathToolkit::Generic::NumericalVectorSlice<T>::norm() const
{
	return std::sqrt(normSquare());
}

template <>
inline double MathToolkit::Generic::NumericalVectorSlice<double>::normSquare() const
{
	double result{ 0.0 };
	{
		auto thisIter = begin();

		for (size_type index = 0; index < m_size; ++index)
		{
			result += (*thisIter) * (*thisIter);
			thisIter += m_stride;
		}
	}

	return result;
}

template <>
inline double MathToolkit::Generic::NumericalVectorSlice<std::complex<double>>::normSquare() const
{
	double result{ 0.0 };
	{
		auto thisIter = begin();

		for (size_type index = 0; index < m_size; ++index)
		{
			result += (thisIter->real()) * (thisIter->real());
			result += (thisIter->imag()) * (thisIter->imag());
			thisIter += m_stride;
		}
	}

	return result;
}


// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename T>
inline T* MathToolkit::Generic::NumericalVectorSlice<T>::begin()
{
	return (std::begin(m_valueArraySlice) + m_start);
}

template <typename T>
inline const T* MathToolkit::Generic::NumericalVectorSlice<T>::begin() const
{
	return (std::begin(m_valueArraySlice) + m_start);
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !MATHTOOLKIT_GENERIC_INTERNAL_NUMERICALVECTORSLICE_H
