#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_OXIDATIONMANAGER_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_OXIDATIONMANAGER_H

#include <stdexcept>
#include <string>

#include "ReportException.h"

#include "OxidationFeature.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			namespace Internal
			{
				class OxidationManager
				{
					using OxidationFeature = MathematicalCrystalChemistry::StructureModel::Features::OxidationFeature;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructor, destructor, and operators

				protected:
					OxidationManager(const std::string& elementSymbol);

				public:
					virtual ~OxidationManager() = default;

					OxidationManager(const OxidationManager&) = default;
					OxidationManager(OxidationManager&&) noexcept = default;
					OxidationManager& operator=(const OxidationManager&) = default;
					OxidationManager& operator=(OxidationManager&&) noexcept = default;

				// Constructor, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

					int oxidationState() const;
					double ionizationPotential() const;
					const OxidationFeature& oxidationFeature() const;

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void initializeOxidationState();
					void addOxidationState();
					void reduceOxidationState();

					bool canAddOxidationState() const;
					bool canReduceOxidationState() const;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					int m_oxidationState;
					double m_ionizationPotential;
					OxidationFeature m_oxidationFeature;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline int MathematicalCrystalChemistry::StructureModel::Components::Internal::OxidationManager::oxidationState() const
{
	return m_oxidationState;
}

inline double MathematicalCrystalChemistry::StructureModel::Components::Internal::OxidationManager::ionizationPotential() const
{
	return m_ionizationPotential;
}

inline const MathematicalCrystalChemistry::StructureModel::Features::OxidationFeature& MathematicalCrystalChemistry::StructureModel::Components::Internal::OxidationManager::oxidationFeature() const
{
	return m_oxidationFeature;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::OxidationManager::initializeOxidationState()
{
	m_oxidationState = 0;
	m_ionizationPotential = m_oxidationFeature.getIonizationPotential(m_oxidationState);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::OxidationManager::addOxidationState()
{
	if (m_oxidationState < 0)
	{
		++m_oxidationState;

		if (m_oxidationState == 0)
			m_ionizationPotential = m_oxidationFeature.getIonizationPotential(m_oxidationState);
	}

	else
	{
		auto iter = m_oxidationFeature.ionizationPotentials().find(m_oxidationState + 1);

		if (iter == m_oxidationFeature.ionizationPotentials().end())
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "oxidize", "Could not oxidize."));
		}

		else
		{
			++m_oxidationState;
			m_ionizationPotential = m_oxidationFeature.getIonizationPotential(m_oxidationState);
		}
	}
}

inline void MathematicalCrystalChemistry::StructureModel::Components::Internal::OxidationManager::reduceOxidationState()
{
	if (0 < m_oxidationState)
	{
		--m_oxidationState;
		m_ionizationPotential = m_oxidationFeature.getIonizationPotential(m_oxidationState);
	}

	else
	{
		if (canReduceOxidationState())
		{
			--m_oxidationState;
			m_ionizationPotential = 0.0;
		}

		else
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "reduce", "Could not reduce."));
		}
	}
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::OxidationManager::canAddOxidationState() const
{
	if (m_oxidationState < m_oxidationFeature.maxOxidationState())
		return true;
	else
		return false;
}

inline bool MathematicalCrystalChemistry::StructureModel::Components::Internal::OxidationManager::canReduceOxidationState() const
{
	if (m_oxidationFeature.minOxidationState() < m_oxidationState)
		return true;
	else
		return false;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_INTERNAL_OXIDATIONMANAGER_H
