#ifndef CHEMTOOLKIT_CRYSTALLOGRAPHY_PERIODICREFERENCEWRAPPER_H
#define CHEMTOOLKIT_CRYSTALLOGRAPHY_PERIODICREFERENCEWRAPPER_H

#include <vector>
#include <functional>

#include "LinearAlgebra.h"


namespace ChemToolkit
{
	namespace Crystallography
	{
		template <typename T, typename L>
		class PeriodicReferenceWrapper
		{
			using size_type = std::size_t;
			using NumericalMatrix = typename T::NumericalMatrix;
			using NumericalVector = typename T::NumericalVector;

		public:
			using LatticeSystem = L;
			using LatticePoint = std::vector<int>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			PeriodicReferenceWrapper(T& object, LatticeSystem& latticeSystem);
			PeriodicReferenceWrapper(T& object, LatticeSystem& latticeSystem, const LatticePoint& latticePoint);
			PeriodicReferenceWrapper(T& object, LatticeSystem& latticeSystem, LatticePoint&& latticePoint);
			PeriodicReferenceWrapper(T& object, const PeriodicReferenceWrapper&);

			template <typename U>
			PeriodicReferenceWrapper(T& object, const PeriodicReferenceWrapper<U, L>&);

			virtual ~PeriodicReferenceWrapper() = default;

			PeriodicReferenceWrapper(const PeriodicReferenceWrapper&) = default;
			PeriodicReferenceWrapper(PeriodicReferenceWrapper&&) noexcept;
			PeriodicReferenceWrapper& operator=(const PeriodicReferenceWrapper&) = default;
			PeriodicReferenceWrapper& operator=(PeriodicReferenceWrapper&&) noexcept;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			T& get() const;
			operator T& () const;
			LatticeSystem& latticeSystemRef() const;

			LatticePoint getLatticePoint() const;

			void resetLatticeSystemRef(LatticeSystem&);
			void resetLatticePoint(const LatticePoint& latticePoint);
			void resetLatticePoint(LatticePoint&& latticePoint);

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			NumericalVector coordinate() const;
			NumericalVector translationVector() const;

			bool isSameObject(const T&) const;
			bool hasSameLatticePoint(const PeriodicReferenceWrapper&) const;
			bool hasOriginalLatticePoint() const;

			void copyLatticePoint(const PeriodicReferenceWrapper&) const;
			LatticePoint getRelativeLatticePointTo(const PeriodicReferenceWrapper&) const;

			size_type getHashValue() const;

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Customize methods

			void moveToOriginalLatticePoint();
			void reverseLatticePoint();
			void translateLatticePoint(const LatticePoint& latticePoint);

		// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			bool isOriginalLatticePoint(const LatticePoint&) const;
			NumericalVector toLatticePointVector(const LatticePoint&) const;

		// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			std::reference_wrapper<T> m_objectRef;
			std::reference_wrapper<LatticeSystem> m_latticeSystemRef;
			LatticePoint m_latticePoint;
			NumericalVector m_latticePointVector;
		};


		template <typename T, typename L>
		inline bool operator==(const PeriodicReferenceWrapper<T, L>& prwa, const PeriodicReferenceWrapper<T, L>& prwb)
		{
			if (&(prwa.get()) != &(prwb.get()))
				return false;
			else
				return prwa.hasSameLatticePoint(prwb);
		}

		template <typename T, typename L>
		inline bool operator!=(const PeriodicReferenceWrapper<T, L>& prwa, const PeriodicReferenceWrapper<T, L>& prwb)
		{
			return !(prwa == prwb);
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors and operator

template <typename T, typename L>
inline ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::PeriodicReferenceWrapper(T& object, LatticeSystem& latticeSystem)
	: m_objectRef{ std::ref(object) }
	, m_latticeSystemRef{ std::ref(latticeSystem) }
	, m_latticePoint{}
	, m_latticePointVector(object.coordinate().size(), 0.0)
{
}

template <typename T, typename L>
inline ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::PeriodicReferenceWrapper(T& object, LatticeSystem& latticeSystem, const LatticePoint& latticePoint)
	: m_objectRef{ std::ref(object) }
	, m_latticeSystemRef{ std::ref(latticeSystem) }
	, m_latticePoint{ latticePoint }
	, m_latticePointVector(latticePoint.size(), 0.0)
{
	bool isOriginalLattice = true;
	auto latticePointVectorIter = m_latticePointVector.begin();
	{
		for (const auto value : m_latticePoint)
		{
			if (value != 0)
			{
				(*latticePointVectorIter) = static_cast<typename NumericalVector::value_type>(value);
				isOriginalLattice = false;
			}

			++latticePointVectorIter;
		}
	}

	if (isOriginalLattice)
		m_latticePoint.clear();
}

template <typename T, typename L>
inline ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::PeriodicReferenceWrapper(T& object, LatticeSystem& latticeSystem, LatticePoint&& latticePoint)
	: m_objectRef{ std::ref(object) }
	, m_latticeSystemRef{ std::ref(latticeSystem) }
	, m_latticePoint{ std::move(latticePoint) }
	, m_latticePointVector{}
{
	m_latticePointVector.resize(m_latticePoint.size(), 0.0);

	bool isOriginalLattice = true;
	auto latticePointVectorIter = m_latticePointVector.begin();
	{
		for (const auto value : m_latticePoint)
		{
			if (value != 0)
			{
				(*latticePointVectorIter) = static_cast<typename NumericalVector::value_type>(value);
				isOriginalLattice = false;
			}

			++latticePointVectorIter;
		}
	}

	if (isOriginalLattice)
		m_latticePoint.clear();
}

template <typename T, typename L>
inline ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::PeriodicReferenceWrapper(T& object, const PeriodicReferenceWrapper& prw)
	: m_objectRef{ std::ref(object) }
	, m_latticeSystemRef{ prw.m_latticeSystemRef }
	, m_latticePoint{ prw.m_latticePoint }
	, m_latticePointVector{ prw.m_latticePointVector }
{
}

template <typename T, typename L>
template <typename U>
inline ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::PeriodicReferenceWrapper(T& object, const PeriodicReferenceWrapper<U, L>& prw)
	: m_objectRef{ std::ref(object) }
	, m_latticeSystemRef{ prw.latticeSystemRef() }
	, m_latticePoint{ prw.getLatticePoint() }
	, m_latticePointVector{}
{
	m_latticePointVector = NumericalVector(m_latticePoint.size(), 0.0);
	auto latticePointVectorIter = m_latticePointVector.begin();
	bool isOriginalLattice = true;
	{
		for (const auto value : m_latticePoint)
		{
			if (value != 0)
			{
				(*latticePointVectorIter) = static_cast<typename NumericalVector::value_type>(value);
				isOriginalLattice = false;
			}

			++latticePointVectorIter;
		}
	}

	if (isOriginalLattice)
		m_latticePoint.clear();
}

template <typename T, typename L>
inline ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::PeriodicReferenceWrapper(PeriodicReferenceWrapper&& prw) noexcept
	: m_objectRef{ prw.m_objectRef }
	, m_latticeSystemRef{ prw.m_latticeSystemRef }
	, m_latticePoint{ std::move(prw.m_latticePoint) }
	, m_latticePointVector{ std::move(prw.m_latticePointVector) }
{
}

template <typename T, typename L>
inline ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>& ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::operator=(PeriodicReferenceWrapper&& prw) noexcept
{
	m_objectRef = prw.m_objectRef;
	m_latticeSystemRef = prw.m_latticeSystemRef;
	m_latticePoint = std::move(prw.m_latticePoint);
	m_latticePointVector = std::move(prw.m_latticePointVector);

	return *this;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

template <typename T, typename L>
inline T& ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::get() const
{
	return m_objectRef.get();
}

template <typename T, typename L>
inline ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::operator T& () const
{
	return m_objectRef.get();
}

template <typename T, typename L>
inline L& ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::latticeSystemRef() const
{
	return m_latticeSystemRef.get();
}

template <typename T, typename L>
ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::LatticePoint ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::getLatticePoint() const
{
	if (m_latticePoint.empty())
		return LatticePoint(m_latticeSystemRef.get().basisVectors().columnSize(), 0);
	else
		return m_latticePoint;
}

template <typename T, typename L>
inline void ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::resetLatticeSystemRef(LatticeSystem& latticeSystem)
{
	m_latticeSystemRef = std::ref(latticeSystem);
}

template <typename T, typename L>
inline void ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::resetLatticePoint(const LatticePoint& latticePoint)
{
	if (isOriginalLatticePoint(latticePoint))
		m_latticePointVector = 0.0;

	else
	{
		m_latticePoint = latticePoint;
		m_latticePointVector = toLatticePointVector(m_latticePoint);
	}
}

template <typename T, typename L>
inline void ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::resetLatticePoint(LatticePoint&& latticePoint)
{
	if (isOriginalLatticePoint(latticePoint))
		m_latticePointVector = 0.0;

	else
	{
		m_latticePoint = std::move(latticePoint);
		m_latticePointVector = toLatticePointVector(m_latticePoint);
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

template <typename T, typename L>
inline typename ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::NumericalVector ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::coordinate() const
{
	if (m_latticePoint.empty())
		return m_objectRef.get().coordinate();
	else
		return (m_objectRef.get().coordinate() + m_latticeSystemRef.get().basisVectors() * m_latticePointVector);
}

template <typename T, typename L>
inline typename ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::NumericalVector ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::translationVector() const
{
	if (m_latticePoint.empty())
		return m_latticePointVector;
	else
		return (m_latticeSystemRef.get().basisVectors() * m_latticePointVector);
}

template <typename T, typename L>
inline bool ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::isSameObject(const T& object) const
{
	if (&(m_objectRef.get()) == &(object))
		return true;
	else
		return false;
}

template <typename T, typename L>
inline bool ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::hasSameLatticePoint(const PeriodicReferenceWrapper& anotherRef) const
{
	if (hasOriginalLatticePoint())
		return anotherRef.hasOriginalLatticePoint();

	else
	{
		if (anotherRef.hasOriginalLatticePoint())
			return false;

		else
		{
			auto thisLatticePointIter = m_latticePoint.begin();

			for (const auto value : anotherRef.m_latticePoint)
			{
				if (value != (*thisLatticePointIter))
					return false;

				++thisLatticePointIter;
			}

			return true;
		}
	}
}

template <typename T, typename L>
inline bool ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::hasOriginalLatticePoint() const
{
	return m_latticePoint.empty();
}

template <typename T, typename L>
inline void ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::copyLatticePoint(const PeriodicReferenceWrapper& anotherRef) const
{
	m_latticePoint = anotherRef.m_latticePoint;
	m_latticePointVector = anotherRef.m_latticePointVector;
}

template <typename T, typename L>
inline typename ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::LatticePoint ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::getRelativeLatticePointTo(const PeriodicReferenceWrapper& anotherRef) const
{
	if (hasOriginalLatticePoint())
	{
		if (anotherRef.hasOriginalLatticePoint())
			return LatticePoint(m_latticeSystemRef.get().basisVectors().columnSize(), 0);
		else
			return anotherRef.m_latticePoint;
	}

	else
	{
		if (anotherRef.hasOriginalLatticePoint())
		{
			LatticePoint relativeLatticePoint = m_latticePoint;

			for (auto& value : relativeLatticePoint)
				value *= (-1);

			return relativeLatticePoint;
		}

		else
		{
			LatticePoint resultLatticePoint = anotherRef.m_latticePoint;
			{
				auto resultIter = resultLatticePoint.begin();

				for (const auto value : m_latticePoint)
				{
					(*resultIter) -= value;
					++resultIter;
				}
			}

			return resultLatticePoint;
		}
	}
}

template <typename T, typename L>
inline std::size_t ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::getHashValue() const
{
	size_type hashValue = 0;
	{
		if (!(m_latticePoint.empty()))
		{
			std::string latticeString;
			{
				for (const auto num : m_latticePoint)
					latticeString += std::to_string(num);
			}

			hashValue ^= std::hash<std::string>{}(latticeString);
		}
	}

	hashValue ^= std::hash<const T*>{}(&(m_objectRef.get()));
	return hashValue;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

template <typename T, typename L>
inline void ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::moveToOriginalLatticePoint()
{
	m_latticePoint.clear();
	m_latticePointVector = 0.0;
}

template <typename T, typename L>
inline void ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::reverseLatticePoint()
{
	if (!(m_latticePoint.empty()))
	{
		for (auto& latticeValue : m_latticePoint)
			latticeValue *= (-1);

		for (auto& latticeVectorValue : m_latticePointVector)
			latticeVectorValue *= (-1.0);
	}
}

template <typename T, typename L>
inline void ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::translateLatticePoint(const LatticePoint& latticePoint)
{
	if (hasOriginalLatticePoint())
	{
		if (!(isOriginalLatticePoint(latticePoint)))
		{
			m_latticePoint = latticePoint;
			m_latticePointVector = toLatticePointVector(m_latticePoint);
		}
	}

	else
	{
		auto thisLatticeIter = m_latticePoint.begin();

		for (const auto value : latticePoint)
		{
			(*thisLatticeIter) += value;
			++thisLatticeIter;
		}

		m_latticePointVector = toLatticePointVector(m_latticePoint);
	}
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

template <typename T, typename L>
inline bool ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::isOriginalLatticePoint(const LatticePoint& latticePoint) const
{
	for (const auto value : latticePoint)
	{
		if (value != 0)
			return false;
	}

	return true;
}

template <typename T, typename L>
inline typename ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::NumericalVector ChemToolkit::Crystallography::PeriodicReferenceWrapper<T, L>::toLatticePointVector(const LatticePoint& latticePoint) const
{
	NumericalVector latticePointVector(latticePoint.size());
	{
		auto latticePointVectorIter = latticePointVector.begin();

		for (const auto value : latticePoint)
		{
			(*latticePointVectorIter) = static_cast<typename NumericalVector::value_type>(value);
			++latticePointVectorIter;
		}
	}

	return latticePointVector;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !CHEMTOOLKIT_CRYSTALLOGRAPHY_PERIODICREFERENCEWRAPPER_H
