#ifndef CHEMTOOLKIT_GENERIC_PERIODICTABLE_H
#define CHEMTOOLKIT_GENERIC_PERIODICTABLE_H

#include <stdexcept>
#include <string>
#include <map>
#include <set>

#include "ReportException.h"


namespace ChemToolkit
{
	namespace Generic
	{
		class PeriodicTable
		{
			using size_type = std::size_t;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Destructor

		public:
			PeriodicTable() = default;
			virtual ~PeriodicTable() = default;

		// Destructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			static std::string getElementSymbol(const size_type atomicNumber);
			static size_type getAtomicNumber(const std::string& elementSymbol);

			static double getAtomicRadius(const std::string& elementSymbol);
			static double getMetallicRadius(const std::string& elementSymbol);
			static double getIonicRadius(const std::string& elementSymbol, const int oxidationState);
			static std::map<int, double> getIonicRadii(const std::string& elementSymbol);

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			static bool isAtomicNumber(const size_type atomicNumber);
			static bool isElementSymbol(const std::string& elementSymbol);

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			static std::string translateIntoElementSymbol(const size_type atomicNumber);
			static size_type translateIntoAtomicNumber(const std::string& elementSymbol);

		// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
		

		private:
			PeriodicTable(const PeriodicTable&) = delete;
			PeriodicTable(PeriodicTable&&) = delete;
			PeriodicTable& operator=(const PeriodicTable&) = delete;
			PeriodicTable& operator=(PeriodicTable&&) = delete;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline std::string ChemToolkit::Generic::PeriodicTable::getElementSymbol(const size_type atomicNumber)
{
	std::string elementSymbol = translateIntoElementSymbol(atomicNumber);

	if (elementSymbol.empty())
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("ChemToolkit::ChemicalElements::PeriodicTable::getElementSymbol", "\"atomicNumber\" is out of definitions."));
	}

	else
		return elementSymbol;
}

inline std::size_t ChemToolkit::Generic::PeriodicTable::getAtomicNumber(const std::string& elementSymbol)
{
	size_type atomicNumber = translateIntoAtomicNumber(elementSymbol);

	if (atomicNumber == 0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("ChemToolkit::ChemicalElements::PeriodicTable::getAtomicNumber", "\"elementSymbol\" is out of definitions."));
	}

	else
		return atomicNumber;
}

inline double ChemToolkit::Generic::PeriodicTable::getIonicRadius(const std::string& elementSymbol, const int oxidationState)
{
	if (oxidationState == 0)
		return getAtomicRadius(elementSymbol);
	else
	{
		std::map<int, double> ionicRadii = getIonicRadii(elementSymbol);
		auto iter = ionicRadii.find(oxidationState);

		if (iter == ionicRadii.end())
			return 0.0;
		else
			return iter->second;
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline bool ChemToolkit::Generic::PeriodicTable::isAtomicNumber(const size_type atomicNumber)
{
	if (translateIntoElementSymbol(atomicNumber).empty())
		return false;
	else
		return true;
}

inline bool ChemToolkit::Generic::PeriodicTable::isElementSymbol(const std::string& elementSymbol)
{
	if (translateIntoAtomicNumber(elementSymbol) == 0)
		return false;
	else
		return true;
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************

#endif // !CHEMTOOLKIT_GENERIC_PERIODICTABLE_H
