#ifndef MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_SPECIFYCOORDINATIONFEATURES_H
#define MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_SPECIFYCOORDINATIONFEATURES_H

#include <map>
#include <utility>
#include <mutex>

#include "ModelCrystalStructure.h"

#include "CoordinationFeatureSpecifyOptions.h"
#include "CoordinationFeatureAssigner.h"
#include "StructureOptimalityAppraiser.h"


namespace MathematicalCrystalChemistry
{
	namespace Runtime
	{
		namespace Analysis
		{
			namespace Internal
			{
				class SpecifyCoordinationFeatures
				{
					using size_type = std::size_t;
					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;
					using StructureFingerprint = MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				public:
					explicit SpecifyCoordinationFeatures(const size_type threadRank);
					SpecifyCoordinationFeatures(const size_type threadRank, const CoordinationFeatureSpecifyOptions&);
					virtual ~SpecifyCoordinationFeatures() = default;

					SpecifyCoordinationFeatures(const SpecifyCoordinationFeatures&) = default;
					SpecifyCoordinationFeatures(SpecifyCoordinationFeatures&&) noexcept = default;
					SpecifyCoordinationFeatures& operator=(const SpecifyCoordinationFeatures&) = default;
					SpecifyCoordinationFeatures& operator=(SpecifyCoordinationFeatures&&) noexcept = default;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void operator()();

					void setCoordinationFeatureSpecifyOptions(const CoordinationFeatureSpecifyOptions&);
					void setCoordinationFeatureSpecifyOptions(CoordinationFeatureSpecifyOptions&&);

					static void initializeNumExcavatedCifFile() noexcept;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					std::string writeCifFileExcavationLogMessage(const size_type threadRank, const std::filesystem::path& cifFilePath) const;
					void createDirectoriesIfNecessary(const std::filesystem::path&) const;

					void initializeInputModelCrystalStructure(ModelCrystalStructure&) const;

					StructureReoptimizer::ReoptimizationResult surveyStructure(ModelCrystalStructure, const std::filesystem::path& workDirectoryPath) const;
					bool isSurveyedStructure(const ModelCrystalStructure&, const std::filesystem::path& cifFilePath) const;

					static void addNumExcavatedCifFile() noexcept;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private utility

					void reportReoptimizationResult(const StructureReoptimizer::ReoptimizationResult&, const std::filesystem::path& featuredStructureDirectoryPath) const;
					void upgradeTotalReoptimizationResult(StructureReoptimizer::ReoptimizationResult& totalReoptimizationResult, const StructureReoptimizer::ReoptimizationResult& reoptimizationResult) const;

					std::filesystem::path getStructureWorkDirectoryPath(const std::filesystem::path& cifFilePath) const;

					std::filesystem::path getResultDirectoryPath(const ModelCrystalStructure& inputModelCrystalStructure, const std::filesystem::path& cifFilePath, const StructureReoptimizer::ReoptimizationResult&) const;
					std::filesystem::path getExceptionDirectoryPath(const ModelCrystalStructure& inputModelCrystalStructure, const std::filesystem::path& cifFilePath) const;

					std::filesystem::path getFeaturedStructureDirectoryPath(const std::filesystem::path& structureWorkDirectoryPath, const size_type featureIndex) const;
					std::filesystem::path getAppraisedDirectoryPath(const std::filesystem::path& featuredStructureDirectoryPath, const StructureReoptimizer::ReoptimizationResult&) const;

					void moveDirectory(const std::filesystem::path& fromPath, const std::filesystem::path& toPath) const;

				// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					size_type m_threadRank;
					std::filesystem::path m_mpiWorkDirectoryPath;

					CoordinationFeatureSpecifyOptions m_coordinationFeatureSpecifyOptions;
					StructureOptimalityAppraiser m_structureOptimalityAppraiser;

					static size_type m_numExcavatedCifFile;
					static std::mutex m_numMutex;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::setCoordinationFeatureSpecifyOptions(const CoordinationFeatureSpecifyOptions& coordinationFeatureSpecifyOptions)
{
	m_coordinationFeatureSpecifyOptions = coordinationFeatureSpecifyOptions;

	m_structureOptimalityAppraiser.setMaxReoptimizing(m_coordinationFeatureSpecifyOptions.maxReoptimizing());
	m_structureOptimalityAppraiser.setMaxAtomFractuationRatio(m_coordinationFeatureSpecifyOptions.maxAtomFractuationRatio());
	m_structureOptimalityAppraiser.setStructureDesignRecordOptions(m_coordinationFeatureSpecifyOptions.structureDesignRecordOptions());
	m_structureOptimalityAppraiser.setStructuralOptimizationParameters(m_coordinationFeatureSpecifyOptions.structuralOptimizationParameters());
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::setCoordinationFeatureSpecifyOptions(CoordinationFeatureSpecifyOptions&& coordinationFeatureSpecifyOptions)
{
	m_coordinationFeatureSpecifyOptions = std::move(coordinationFeatureSpecifyOptions);

	m_structureOptimalityAppraiser.setMaxReoptimizing(m_coordinationFeatureSpecifyOptions.maxReoptimizing());
	m_structureOptimalityAppraiser.setMaxAtomFractuationRatio(m_coordinationFeatureSpecifyOptions.maxAtomFractuationRatio());
	m_structureOptimalityAppraiser.setStructureDesignRecordOptions(m_coordinationFeatureSpecifyOptions.structureDesignRecordOptions());
	m_structureOptimalityAppraiser.setStructuralOptimizationParameters(m_coordinationFeatureSpecifyOptions.structuralOptimizationParameters());
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::string MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::writeCifFileExcavationLogMessage(const size_type threadRank, const std::filesystem::path& cifFilePath) const
{
	std::string logMessage;
	{
		logMessage += "Thread ";
		logMessage += std::to_string(threadRank);
		logMessage += ": Excavate the cif file of ";
		logMessage += std::to_string(m_numExcavatedCifFile);
		logMessage += "th cif file of ";
		logMessage += cifFilePath.generic_string();
	}

	return logMessage;
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::createDirectoriesIfNecessary(const std::filesystem::path& directoryPath) const
{
	if (!(std::filesystem::is_directory(directoryPath)))
	{
		if (!(std::filesystem::create_directories(directoryPath)))
		{
			std::string errorMessage;
			{
				errorMessage += "Could not create the directory of \"";
				errorMessage += directoryPath.generic_string();
				errorMessage += "\"";
			}

			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "createDirectoriesIfNecessary", errorMessage));
		}
	}
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::initializeInputModelCrystalStructure(ModelCrystalStructure& inputModelCrystalStructure) const
{
	inputModelCrystalStructure.removeSiteDuplications();
	inputModelCrystalStructure.assignCoordinationFeaturesFromElementSymbols();
	inputModelCrystalStructure.optimizeOxidationStates();
	inputModelCrystalStructure.assignCoordinationFeaturesFromElementSymbols();

	inputModelCrystalStructure.refineStructure();
	inputModelCrystalStructure.updateSpaceGroupInfo();

	inputModelCrystalStructure.createTrackingAtomPairs();
	inputModelCrystalStructure.createGeometricalConstraints(m_coordinationFeatureSpecifyOptions.initialMinimumConstraintCutoffRatio(), 1.0, m_coordinationFeatureSpecifyOptions.initialIonicBondCutoffRatio(), m_coordinationFeatureSpecifyOptions.initialMetallicBondCutoffRatio());

	for (auto& modelAtom : inputModelCrystalStructure.atomicArrangement())
	{
		if (modelAtom.coordinationFeature().atomCategory().isCation())
		{
			modelAtom.coordinationFeature().setMetallicBondRadii("O", "o");
			modelAtom.fixChemicalBondCompositions();
		}
	}
}

inline bool MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::isSurveyedStructure(const ModelCrystalStructure& modelCrystalStructure, const std::filesystem::path& cifFilePath) const
{
	if (std::filesystem::is_directory(getResultDirectoryPath(modelCrystalStructure, cifFilePath, StructureReoptimizer::ReoptimizationResult::isotypic)))
		return true;
	if (std::filesystem::is_directory(getResultDirectoryPath(modelCrystalStructure, cifFilePath, StructureReoptimizer::ReoptimizationResult::homeotypic)))
		return true;
	if (std::filesystem::is_directory(getResultDirectoryPath(modelCrystalStructure, cifFilePath, StructureReoptimizer::ReoptimizationResult::same_symmetry)))
		return true;
	if (std::filesystem::is_directory(getResultDirectoryPath(modelCrystalStructure, cifFilePath, StructureReoptimizer::ReoptimizationResult::same_space_group_type)))
		return true;
	if (std::filesystem::is_directory(getResultDirectoryPath(modelCrystalStructure, cifFilePath, StructureReoptimizer::ReoptimizationResult::optimal)))
		return true;
	if (std::filesystem::is_directory(getResultDirectoryPath(modelCrystalStructure, cifFilePath, StructureReoptimizer::ReoptimizationResult::infeasible)))
		return true;
	if (std::filesystem::is_directory(getExceptionDirectoryPath(modelCrystalStructure, cifFilePath)))
		return true;

	return false;
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::initializeNumExcavatedCifFile() noexcept
{
	std::lock_guard<std::mutex> guard{ m_numMutex };
	m_numExcavatedCifFile = 0;
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::addNumExcavatedCifFile() noexcept
{
	std::lock_guard<std::mutex> guard{ m_numMutex };
	++m_numExcavatedCifFile;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

inline std::filesystem::path MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::getStructureWorkDirectoryPath(const std::filesystem::path& cifFilePath) const
{
	std::filesystem::path structureWorkDirectoryPath = m_mpiWorkDirectoryPath;
	structureWorkDirectoryPath /= cifFilePath.stem();

	return structureWorkDirectoryPath;
}

inline std::filesystem::path MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::getResultDirectoryPath(const ModelCrystalStructure& inputModelCrystalStructure, const std::filesystem::path& cifFilePath, const StructureReoptimizer::ReoptimizationResult& reoptimizationResult) const
{
	std::filesystem::path resultDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
	{
		switch (reoptimizationResult)
		{
		case StructureReoptimizer::ReoptimizationResult::isotypic:
			resultDirectoryPath /= "Isotypic";
			break;

		case StructureReoptimizer::ReoptimizationResult::homeotypic:
			resultDirectoryPath /= "Homeotypic";
			break;

		case StructureReoptimizer::ReoptimizationResult::same_symmetry:
			resultDirectoryPath /= "Same_symmetry";
			break;

		case StructureReoptimizer::ReoptimizationResult::same_space_group_type:
			resultDirectoryPath /= "Same_space_group_type";
			break;

		case StructureReoptimizer::ReoptimizationResult::optimal:
			resultDirectoryPath /= "Optimal";
			break;

		case StructureReoptimizer::ReoptimizationResult::infeasible:
			resultDirectoryPath /= "Infeasible";
			break;

		default:
			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "getResultDirectoryPath", "\"reoptimizationResult\" is out of definitions."));
		}
	}

	auto chemicalComposition = inputModelCrystalStructure.getChemicalComposition();
	resultDirectoryPath /= chemicalComposition.getChemicalSystemName();
	resultDirectoryPath /= chemicalComposition.name();
	resultDirectoryPath /= std::to_string(inputModelCrystalStructure.spaceGroupNumber());
	resultDirectoryPath /= cifFilePath.stem();


	return resultDirectoryPath;
}

inline std::filesystem::path MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::getExceptionDirectoryPath(const ModelCrystalStructure& inputModelCrystalStructure, const std::filesystem::path& cifFilePath) const
{
	std::filesystem::path exceptionDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
	exceptionDirectoryPath /= "Exception";

	auto chemicalComposition = inputModelCrystalStructure.getChemicalComposition();
	exceptionDirectoryPath /= chemicalComposition.getChemicalSystemName();
	exceptionDirectoryPath /= chemicalComposition.name();
	exceptionDirectoryPath /= std::to_string(inputModelCrystalStructure.spaceGroupNumber());
	exceptionDirectoryPath /= cifFilePath.stem();

	return exceptionDirectoryPath;
}

inline std::filesystem::path MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::getFeaturedStructureDirectoryPath(const std::filesystem::path& structureWorkDirectoryPath, const size_type featureIndex) const
{
	std::filesystem::path featuredStructureDirectoryPath = structureWorkDirectoryPath;
	{
		featuredStructureDirectoryPath /= "Feature-";
		featuredStructureDirectoryPath += std::to_string(featureIndex);
	}

	if (std::filesystem::is_directory(featuredStructureDirectoryPath))
		std::filesystem::remove_all(featuredStructureDirectoryPath);


	return featuredStructureDirectoryPath;
}

inline std::filesystem::path MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::getAppraisedDirectoryPath(const std::filesystem::path& featuredStructureDirectoryPath, const StructureReoptimizer::ReoptimizationResult& reoptimizationResult) const
{
	std::filesystem::path featuredStructureResultDirectoryPath = featuredStructureDirectoryPath.parent_path();
	{
		switch (reoptimizationResult)
		{
		case StructureReoptimizer::ReoptimizationResult::isotypic:
			featuredStructureResultDirectoryPath /= "Isotypic";
			break;

		case StructureReoptimizer::ReoptimizationResult::homeotypic:
			featuredStructureResultDirectoryPath /= "Homeotypic";
			break;

		case StructureReoptimizer::ReoptimizationResult::same_symmetry:
			featuredStructureResultDirectoryPath /= "Same_symmetry";
			break;

		case StructureReoptimizer::ReoptimizationResult::same_space_group_type:
			featuredStructureResultDirectoryPath /= "Same_space_group_type";
			break;

		case StructureReoptimizer::ReoptimizationResult::optimal:
			featuredStructureResultDirectoryPath /= "Optimal";
			break;

		case StructureReoptimizer::ReoptimizationResult::infeasible:
			featuredStructureResultDirectoryPath /= "Infeasibles";
			break;

		default:
			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "getFeaturedStructureResultDirectoryPath", "\"reoptimizationResult\" is out of definitions."));
		}
	}

	featuredStructureResultDirectoryPath /= std::filesystem::relative(featuredStructureDirectoryPath, featuredStructureDirectoryPath.parent_path());
	return featuredStructureResultDirectoryPath;
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::SpecifyCoordinationFeatures::moveDirectory(const std::filesystem::path& fromPath, const std::filesystem::path& toPath) const
{
	if (std::filesystem::is_directory(toPath))
		std::filesystem::remove_all(toPath);


	std::filesystem::rename(fromPath, toPath);
}

// Private methods for file system
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_SPECIFYCOORDINATIONFEATURES_H
