#ifndef CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_SPGLIBWRAPPER_H
#define CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_SPGLIBWRAPPER_H

#include <stdexcept>
#include <string>
#include <vector>
#include <memory>

#include "LinearAlgebra.h"
#include "ReportException.h"

#include "AtomBase.h"
#include "UnitCell.h"


namespace ChemToolkit
{
	namespace Crystallography
	{
		namespace SpaceGroups
		{
			class SpglibWrapper final
			{
				using size_type = std::size_t;
				using AtomBase = ChemToolkit::Generic::AtomBase;

				struct SpglibVariables
				{
					int numAtoms{ 0 };
					double lattice[3][3]{ {0,0,0}, {0,0,0}, {0,0,0} };
					std::unique_ptr<int[]> types;
					std::unique_ptr<double[][3]> positions;

					size_type spaceGroupNumber{ 0 };
					std::vector<int> mappingToInputIndices;
					std::vector<std::string> siteSymmetrySymbols;
					std::vector<std::string> wyckoffSymbols;
					std::vector<size_type> siteIndices;
				};

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				SpglibWrapper();
				~SpglibWrapper() = default;

				SpglibWrapper(const SpglibWrapper&) = default;
				SpglibWrapper(SpglibWrapper&&) noexcept = default;
				SpglibWrapper& operator=(const SpglibWrapper&) = default;
				SpglibWrapper& operator=(SpglibWrapper&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				template <typename A>
				size_type getSpaceGroupInformation(const UnitCell&, std::vector<A>&);

				template <typename A>
				size_type findConventionalStructure(UnitCell&, std::vector<A>&);

				template <typename A>
				void findPrimitiveStructure(UnitCell&, std::vector<A>&);

				template <typename A>
				void findPrimitiveStructure(UnitCell&, std::vector<A>&, const double spaceGroupPrecision);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods
				
			private:
				SpglibVariables getSpglibVariablesWithSpaceGroupInfo(const UnitCell&, const std::vector<AtomBase>&) const;
				SpglibVariables getConventionalSpglibVariables(const UnitCell&, const std::vector<AtomBase>&) const;
				void findPrimitiveSpglibVariables(SpglibVariables&) const;
				void findPrimitiveSpglibVariables(SpglibVariables&, const double spaceGroupPrecision) const;

				void updateUnitCell(UnitCell&, const SpglibVariables&);

				template <typename A>
				void updateAtomicArrangement(std::vector<A>&, const UnitCell&, const SpglibVariables&) const;

				template <typename A>
				void assignSiteSymmetrySymbols(std::vector<A>&, const SpglibVariables&);
				
			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private utility

				std::string toWyckoffSymbol(const size_type) const;

				std::string toElementSymbol(const size_type atomicNumber) const;
				SpglibVariables getInputSpglibVariables(const UnitCell&, const std::vector<AtomBase>&) const;

				template <typename A>
				std::vector<AtomBase> toAtomBases(const std::vector<A>&) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				double m_spaceGroupPrecision;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

template <typename A>
inline std::size_t ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::getSpaceGroupInformation(const UnitCell& unitCell, std::vector<A>& atomicArrangement)
{
	std::vector<AtomBase> atomBases = toAtomBases(atomicArrangement);
	SpglibVariables primitiveSpglibVariables = getSpglibVariablesWithSpaceGroupInfo(unitCell, atomBases);

	assignSiteSymmetrySymbols(atomicArrangement, primitiveSpglibVariables);
	return primitiveSpglibVariables.spaceGroupNumber;
}

template <typename A>
inline std::size_t ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::findConventionalStructure(UnitCell& unitCell, std::vector<A>& atomicArrangement)
{
	std::vector<AtomBase> atomBases = toAtomBases(atomicArrangement);
	SpglibVariables conventionalSpglibVariables = getConventionalSpglibVariables(unitCell, atomBases);

	updateUnitCell(unitCell, conventionalSpglibVariables);
	updateAtomicArrangement(atomicArrangement, unitCell, conventionalSpglibVariables);
	assignSiteSymmetrySymbols(atomicArrangement, conventionalSpglibVariables);

	return conventionalSpglibVariables.spaceGroupNumber;
}

template <typename A>
inline void ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::findPrimitiveStructure(UnitCell& unitCell, std::vector<A>& atomicArrangement)
{
	std::vector<AtomBase> atomBases = toAtomBases(atomicArrangement);
	SpglibVariables spglibVariables = getInputSpglibVariables(unitCell, atomBases);
	findPrimitiveSpglibVariables(spglibVariables);

	updateUnitCell(unitCell, spglibVariables);
	updateAtomicArrangement(atomicArrangement, unitCell, spglibVariables);
}

template <typename A>
inline void ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::findPrimitiveStructure(UnitCell& unitCell, std::vector<A>& atomicArrangement, const double spaceGroupPrecision)
{
	std::vector<AtomBase> atomBases = toAtomBases(atomicArrangement);
	SpglibVariables spglibVariables = getInputSpglibVariables(unitCell, atomBases);
	findPrimitiveSpglibVariables(spglibVariables, spaceGroupPrecision);

	updateUnitCell(unitCell, spglibVariables);
	updateAtomicArrangement(atomicArrangement, unitCell, spglibVariables);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::findPrimitiveSpglibVariables(SpglibVariables& spglibVariables) const
{
	findPrimitiveSpglibVariables(spglibVariables, m_spaceGroupPrecision);
}

inline void ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::updateUnitCell(UnitCell& unitCell, const SpglibVariables& spglibVariables)
{
	MathToolkit::Generic::NumericalMatrix<double> basisVectorMatrix(3, 3);

	basisVectorMatrix(0, 0) = spglibVariables.lattice[0][0];
	basisVectorMatrix(0, 1) = spglibVariables.lattice[0][1];
	basisVectorMatrix(0, 2) = spglibVariables.lattice[0][2];
	basisVectorMatrix(1, 0) = spglibVariables.lattice[1][0];
	basisVectorMatrix(1, 1) = spglibVariables.lattice[1][1];
	basisVectorMatrix(1, 2) = spglibVariables.lattice[1][2];
	basisVectorMatrix(2, 0) = spglibVariables.lattice[2][0];
	basisVectorMatrix(2, 1) = spglibVariables.lattice[2][1];
	basisVectorMatrix(2, 2) = spglibVariables.lattice[2][2];

	unitCell.setBasisVectors(std::move(basisVectorMatrix));
}

template <typename A>
inline void ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::updateAtomicArrangement(std::vector<A>& atomicArrangement, const UnitCell& unitCell, const SpglibVariables& spglibVariables) const
{
	std::vector<A> previousAtomicArrangement = atomicArrangement;
	atomicArrangement.clear();

	for (size_type index = 0; index < spglibVariables.numAtoms; ++index)
	{
		A atom = previousAtomicArrangement.at(spglibVariables.mappingToInputIndices.at(index));
		{
			MathToolkit::Generic::NumericalVector<double> fractionalCoordinate(3);
			{
				fractionalCoordinate[0] = spglibVariables.positions[index][0];
				fractionalCoordinate[1] = spglibVariables.positions[index][1];
				fractionalCoordinate[2] = spglibVariables.positions[index][2];
			}

			atom.coordinate() = unitCell.basisVectors() * fractionalCoordinate;
		}

		atomicArrangement.push_back(std::move(atom));
	}
}

template <typename A>
inline void ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::assignSiteSymmetrySymbols(std::vector<A>& atomicArrangement, const SpglibVariables& spglibVariables)
{
	for (size_type index = 0; index < atomicArrangement.size(); ++index)
	{
		atomicArrangement[index].setSiteSymmetrySymbol(spglibVariables.siteSymmetrySymbols.at(index));
		atomicArrangement[index].setWyckoffSymbol(spglibVariables.wyckoffSymbols.at(index));
		atomicArrangement[index].setSiteIndex(spglibVariables.siteIndices.at(index));
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

template <typename A>
inline std::vector<ChemToolkit::Generic::AtomBase> ChemToolkit::Crystallography::SpaceGroups::SpglibWrapper::toAtomBases(const std::vector<A>& atomicArrangement) const
{
	std::vector<AtomBase> atomBases;
	{
		for (const auto& atom : atomicArrangement)
		{
			AtomBase atomBase{ atom.elementSymbol(), atom.coordinate() };
			atomBases.push_back(std::move(atomBase));
		}
	}

	return atomBases;
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !CHEMTOOLKIT_CRYSTALLOGRAPHY_SPACEGROUPS_SPGLIBWRAPPER_H
