#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALOPTIMIZATIONPARAMETERS_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALOPTIMIZATIONPARAMETERS_H

#include <stdexcept>
#include <string>

#include "ReportException.h"

#include "StructuralRelaxationParameters.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Optimization
		{
			class StructuralOptimizationParameters
			{
				using size_type = std::size_t;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				StructuralOptimizationParameters();
				virtual ~StructuralOptimizationParameters() = default;

				StructuralOptimizationParameters(const StructuralOptimizationParameters&) = default;
				StructuralOptimizationParameters(StructuralOptimizationParameters&&) noexcept = default;
				StructuralOptimizationParameters& operator=(const StructuralOptimizationParameters&) = default;
				StructuralOptimizationParameters& operator=(StructuralOptimizationParameters&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept;
				void initializeFromInputFileTexts(const std::string& inputTexts);

				bool isValid() const noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				size_type maxTotalStructureOptimizing() const noexcept;
				size_type atomPairTrackersLifetime() const noexcept;
				size_type constraintsLifetime() const noexcept;
				size_type minMaxConstraintsLifetime() const noexcept;
				size_type unitCellLifetime() const noexcept;
				
				double atomPairTrackingCutoffRatio() const;
				double pressure() const;
				double attractiveForceConstant() const;
				double repulsiveForceConstant() const;
				
				const StructuralRelaxationParameters& globalRelaxationParameters() const noexcept;
				const StructuralRelaxationParameters& localRelaxationParameters() const noexcept;
				const StructuralRelaxationParameters& preciseRelaxationParameters() const noexcept;

				void setMaxTotalStructureOptimizing(const size_type) noexcept;
				void setAtomPairTrackersLifetime(const size_type);
				void setConstraintsLifetime(const size_type);
				void setMinMaxConstraintsLifetime(const size_type);
				void setUnitCellLifetime(const size_type);
				void setGlobalRelaxationParameters(const StructuralRelaxationParameters&) noexcept;
				void setLocalRelaxationParameters(const StructuralRelaxationParameters&) noexcept;
				void setPreciseRelaxationParameters(const StructuralRelaxationParameters&) noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void initializeGlobalStructuralRelaxationParameters(const std::string& inputTexts);
				void initializeLocalStructuralRelaxationParameters(const std::string& inputTexts);
				void initializePreciseStructuralRelaxationParameters(const std::string& inputTexts);

				void validateInitializedValues() const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				size_type m_maxTotalStructureOptimizing;
				size_type m_atomPairTrackersLifetime;
				size_type m_constraintsLifetime;
				size_type m_minMaxConstraintsLifetime;
				size_type m_unitCellLifetime;

				double m_atomPairTrackingCutoffRatio;
				double m_pressure;
				double m_attractiveForceConstant;
				double m_repulsiveForceConstant;

				StructuralRelaxationParameters m_globalRelaxationParameters;
				StructuralRelaxationParameters m_localRelaxationParameters;
				StructuralRelaxationParameters m_preciseRelaxationParameters;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::isValid() const noexcept
{
	if (m_maxTotalStructureOptimizing == 0)
		return false;
	else
		return true;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline std::size_t MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::maxTotalStructureOptimizing() const noexcept
{
	return m_maxTotalStructureOptimizing;
}

inline std::size_t MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::atomPairTrackersLifetime() const noexcept
{
	return m_atomPairTrackersLifetime;
}

inline std::size_t MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::constraintsLifetime() const noexcept
{
	return m_constraintsLifetime;
}

inline std::size_t MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::minMaxConstraintsLifetime() const noexcept
{
	return m_minMaxConstraintsLifetime;
}

inline std::size_t MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::unitCellLifetime() const noexcept
{
	return m_unitCellLifetime;
}

inline double MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::atomPairTrackingCutoffRatio() const
{
	return m_atomPairTrackingCutoffRatio;
}

inline double MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::pressure() const
{
	return m_pressure;
}

inline double MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::attractiveForceConstant() const
{
	return m_attractiveForceConstant;
}

inline double MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::repulsiveForceConstant() const
{
	return m_repulsiveForceConstant;
}

inline const MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters& MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::globalRelaxationParameters() const noexcept
{
	return m_globalRelaxationParameters;
}

inline const MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters& MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::localRelaxationParameters() const noexcept
{
	return m_localRelaxationParameters;
}

inline const MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters& MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::preciseRelaxationParameters() const noexcept
{
	return m_preciseRelaxationParameters;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::setMaxTotalStructureOptimizing(const size_type value) noexcept
{
	m_maxTotalStructureOptimizing = value;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::setAtomPairTrackersLifetime(const size_type value)
{
	if (value == 0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setAtomPairTrackersLifetime", "Input value is zero."));
	}

	else
		m_atomPairTrackersLifetime = value;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::setConstraintsLifetime(const size_type value)
{
	if (value == 0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setConstraintsLifetime", "Input value is zero."));
	}

	else
		m_constraintsLifetime = value;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::setMinMaxConstraintsLifetime(const size_type value)
{
	if (value == 0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setMinMaxConstraintsLifetime", "Input value is zero."));
	}

	else
		m_minMaxConstraintsLifetime = value;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::setUnitCellLifetime(const size_type value)
{
	if (value == 0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setUnitCellLifetime", "Input value is zero."));
	}

	else
		m_unitCellLifetime = value;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::setGlobalRelaxationParameters(const StructuralRelaxationParameters& structureRelaxationParameters) noexcept
{
	m_globalRelaxationParameters = structureRelaxationParameters;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::setLocalRelaxationParameters(const StructuralRelaxationParameters& structureRelaxationParameters) noexcept
{
	m_localRelaxationParameters = structureRelaxationParameters;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters::setPreciseRelaxationParameters(const StructuralRelaxationParameters& structureRelaxationParameters) noexcept
{
	m_preciseRelaxationParameters = structureRelaxationParameters;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALOPTIMIZATIONPARAMETERS_H
