#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALOPTIMIZER_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALOPTIMIZER_H

#include <filesystem>

#include "ModelCrystalStructure.h"

#include "StructureDesignRecorder.h"

#include "StructuralOptimizationParameters.h"
#include "StructuralRelaxer.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Optimization
		{
			class StructuralOptimizer
			{
				using size_type = std::size_t;
				using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;
				
				using StructureDesignRecorder = MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				StructuralOptimizer();
				StructuralOptimizer(const StructuralOptimizationParameters&);
				virtual ~StructuralOptimizer() = default;

				StructuralOptimizer(const StructuralOptimizer&) = default;
				StructuralOptimizer(StructuralOptimizer&&) noexcept = default;
				StructuralOptimizer& operator=(const StructuralOptimizer&) = default;
				StructuralOptimizer& operator=(StructuralOptimizer&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void optimize(ModelCrystalStructure&) const;
				void optimize(ModelCrystalStructure&, StructureDesignRecorder&) const;

				const StructuralOptimizationParameters& structuralOptimizationParameters() const noexcept;
				void setStructuralOptimizationParameters(const StructuralOptimizationParameters&);
				void setStructuralOptimizationParameters(StructuralOptimizationParameters&&);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				bool manageGeometricalConstraints(ModelCrystalStructure&) const;
				bool manageUnitCell(ModelCrystalStructure&) const;

				bool applyGlobalStructuralRelaxation(ModelCrystalStructure&) const;
				bool applyGlobalStructuralRelaxation(ModelCrystalStructure&, StructureDesignRecorder&) const;

				bool applyLocalStructuralRelaxation(ModelCrystalStructure&) const;
				bool applyLocalStructuralRelaxation(ModelCrystalStructure&, StructureDesignRecorder&) const;

				bool applyPreciseStructuralRelaxation(ModelCrystalStructure&) const;
				bool applyPreciseStructuralRelaxation(ModelCrystalStructure&, StructureDesignRecorder&) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private utility

				void initializeTimers() const;
				void addTimers(const size_type structureOptimizing) const;

			// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				StructuralOptimizationParameters m_structuralOptimizationParameters;

				mutable size_type m_totalStructureOptimizing{ 0 };
				mutable size_type m_geometricalConstraintsOptimizing{ 0 };

				mutable size_type m_atomPairTrackersTimer{ 0 };
				mutable size_type m_constraintsTimer{ 0 };
				mutable size_type m_minMaxConstraintsTimer{ 0 };
				mutable size_type m_unitCellTimer{ 0 };
				mutable std::filesystem::path m_workDirectoryPath{};
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline const MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters& MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::structuralOptimizationParameters() const noexcept
{
	return m_structuralOptimizationParameters;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::setStructuralOptimizationParameters(const StructuralOptimizationParameters& optimizationParameters)
{
	m_structuralOptimizationParameters = optimizationParameters;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::setStructuralOptimizationParameters(StructuralOptimizationParameters&& optimizationParameters)
{
	m_structuralOptimizationParameters = std::move(optimizationParameters);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::applyGlobalStructuralRelaxation(ModelCrystalStructure& modelCrystalStructure) const
{
	if (0 < m_structuralOptimizationParameters.globalRelaxationParameters().maxStructureOptimizing())
	{
		StructuralRelaxationParameters globalRelaxationParameters = m_structuralOptimizationParameters.globalRelaxationParameters();
		globalRelaxationParameters.setMaxStructureOptimizing(m_structuralOptimizationParameters.minMaxConstraintsLifetime());
	
		MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer globalStructuralRelaxer;
		globalStructuralRelaxer.setStructuralRelaxationParameters(globalRelaxationParameters);


		bool isOptimalSolution = false;
		{
			for (size_type globalOptimizing = 0; globalOptimizing < m_structuralOptimizationParameters.globalRelaxationParameters().maxStructureOptimizing();)
			{
				addTimers(globalRelaxationParameters.maxStructureOptimizing());
				isOptimalSolution = globalStructuralRelaxer.run(modelCrystalStructure);
				globalOptimizing += m_structuralOptimizationParameters.minMaxConstraintsLifetime();

				manageUnitCell(modelCrystalStructure);
				manageGeometricalConstraints(modelCrystalStructure);
			}
		}

		return isOptimalSolution;
	}

	else
		return true;
}

inline bool MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::applyGlobalStructuralRelaxation(ModelCrystalStructure& modelCrystalStructure, StructureDesignRecorder& structureDesignRecorder) const
{
	if (0 < m_structuralOptimizationParameters.globalRelaxationParameters().maxStructureOptimizing())
	{
		StructuralRelaxationParameters globalRelaxationParameters = m_structuralOptimizationParameters.globalRelaxationParameters();
		globalRelaxationParameters.setMaxStructureOptimizing(m_structuralOptimizationParameters.minMaxConstraintsLifetime());

		MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer globalStructuralRelaxer;
		globalStructuralRelaxer.setStructuralRelaxationParameters(globalRelaxationParameters);


		bool isOptimalSolution = false;
		{
			for (size_type globalOptimizing = 0; globalOptimizing < m_structuralOptimizationParameters.globalRelaxationParameters().maxStructureOptimizing();)
			{
				addTimers(globalRelaxationParameters.maxStructureOptimizing());
				isOptimalSolution = globalStructuralRelaxer.run(modelCrystalStructure, structureDesignRecorder);
				globalOptimizing += m_structuralOptimizationParameters.minMaxConstraintsLifetime();

				manageUnitCell(modelCrystalStructure);
				manageGeometricalConstraints(modelCrystalStructure);
			}
		}

		return isOptimalSolution;
	}

	else
		return true;
}

inline bool MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::applyLocalStructuralRelaxation(ModelCrystalStructure& modelCrystalStructure) const
{
	if (0 < m_structuralOptimizationParameters.localRelaxationParameters().maxStructureOptimizing())
	{
		MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer localStructuralRelaxer{ m_structuralOptimizationParameters.localRelaxationParameters() };
		addTimers(m_structuralOptimizationParameters.localRelaxationParameters().maxStructureOptimizing());

		if (localStructuralRelaxer.run(modelCrystalStructure))
			return true;

		else
		{
			addTimers(m_structuralOptimizationParameters.localRelaxationParameters().maxStructureOptimizing());
			return localStructuralRelaxer.rerun(modelCrystalStructure);
		}
	}

	else
		return true;
}

inline bool MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::applyLocalStructuralRelaxation(ModelCrystalStructure& modelCrystalStructure, StructureDesignRecorder& structureDesignRecorder) const
{
	if (0 < m_structuralOptimizationParameters.localRelaxationParameters().maxStructureOptimizing())
	{
		MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer localStructuralRelaxer{ m_structuralOptimizationParameters.localRelaxationParameters() };
		addTimers(m_structuralOptimizationParameters.localRelaxationParameters().maxStructureOptimizing());

		if (localStructuralRelaxer.run(modelCrystalStructure, structureDesignRecorder))
			return true;

		else
		{
			addTimers(m_structuralOptimizationParameters.localRelaxationParameters().maxStructureOptimizing());
			return localStructuralRelaxer.rerun(modelCrystalStructure, structureDesignRecorder);
		}
	}

	else
		return true;
}

inline bool MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::applyPreciseStructuralRelaxation(ModelCrystalStructure& modelCrystalStructure) const
{
	if (0 < m_structuralOptimizationParameters.preciseRelaxationParameters().maxStructureOptimizing())
	{
		MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer preciseStructuralRelaxer{ m_structuralOptimizationParameters.preciseRelaxationParameters() };

		addTimers(m_structuralOptimizationParameters.preciseRelaxationParameters().maxStructureOptimizing());
		return preciseStructuralRelaxer.run(modelCrystalStructure);
	}

	else
		return modelCrystalStructure.isOptimalSolution(m_structuralOptimizationParameters.preciseRelaxationParameters().feasibleConstraintErrorRate());
}

inline bool MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::applyPreciseStructuralRelaxation(ModelCrystalStructure& modelCrystalStructure, StructureDesignRecorder& structureDesignRecorder) const
{
	if (0 < m_structuralOptimizationParameters.preciseRelaxationParameters().maxStructureOptimizing())
	{
		MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer preciseStructuralRelaxer{ m_structuralOptimizationParameters.preciseRelaxationParameters() };

		addTimers(m_structuralOptimizationParameters.preciseRelaxationParameters().maxStructureOptimizing());
		return preciseStructuralRelaxer.run(modelCrystalStructure, structureDesignRecorder);
	}

	else
		return modelCrystalStructure.isOptimalSolution(m_structuralOptimizationParameters.preciseRelaxationParameters().feasibleConstraintErrorRate());
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::initializeTimers() const
{
	m_totalStructureOptimizing = 0;
	m_geometricalConstraintsOptimizing = 0;

	m_atomPairTrackersTimer = 0;
	m_constraintsTimer = 0;
	m_minMaxConstraintsTimer = 0;
	m_unitCellTimer = 0;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer::addTimers(const size_type structureOptimizing) const
{
	m_totalStructureOptimizing += structureOptimizing;
	m_atomPairTrackersTimer += structureOptimizing;
	m_constraintsTimer += structureOptimizing;
	m_minMaxConstraintsTimer += structureOptimizing;
	m_unitCellTimer += structureOptimizing;
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALOPTIMIZER_H
