#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURERELAXATIONPARAMETERS_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURERELAXATIONPARAMETERS_H

#include <stdexcept>
#include <string>

#include "ReportException.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Optimization
		{
			class StructuralRelaxationParameters
			{
				using size_type = std::size_t;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				StructuralRelaxationParameters();
				virtual ~StructuralRelaxationParameters() = default;

				StructuralRelaxationParameters(const StructuralRelaxationParameters&) = default;
				StructuralRelaxationParameters(StructuralRelaxationParameters&&) noexcept = default;
				StructuralRelaxationParameters& operator=(const StructuralRelaxationParameters&) = default;
				StructuralRelaxationParameters& operator=(StructuralRelaxationParameters&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize() noexcept;
				void initializeFromInputFileTexts(const std::string& inputTexts);

				bool isValid() const noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				size_type maxStructureOptimizing() const noexcept;
				double initialMaxAtomicDisplacement() const noexcept;
				double finalMaxAtomicDisplacement() const noexcept;
				double maxUnitCellDisplacementFactor() const noexcept;
				double feasibleConstraintErrorRate() const noexcept;

				void setMaxStructureOptimizing(const size_type);
				void setMaxAtomicDisplacements(const double initialMaxAtomicDisplacement, const double finalMaxAtomicDisplacement);
				void setMaxUnitCellDisplacementFactor(const double);
				void setFeasibleConstraintErrorRate(const double);

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void validateInitializedValues() const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				size_type m_maxStructureOptimizing;
				double m_initialMaxAtomicDisplacement;
				double m_finalMaxAtomicDisplacement;
				double m_maxUnitCellDisplacementFactor;
				double m_feasibleConstraintErrorRate;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::isValid() const noexcept
{
	if (m_maxStructureOptimizing == 0)
		return false;
	else
	{
		if (0.0 < m_initialMaxAtomicDisplacement)
			return true;
		else
			return false;
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline std::size_t MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::maxStructureOptimizing() const noexcept
{
	return m_maxStructureOptimizing;
}

inline double MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::initialMaxAtomicDisplacement() const noexcept
{
	return m_initialMaxAtomicDisplacement;
}

inline double MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::finalMaxAtomicDisplacement() const noexcept
{
	return m_finalMaxAtomicDisplacement;
}

inline double MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::maxUnitCellDisplacementFactor() const noexcept
{
	return m_maxUnitCellDisplacementFactor;
}

inline double MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::feasibleConstraintErrorRate() const noexcept
{
	return m_feasibleConstraintErrorRate;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::setMaxStructureOptimizing(const size_type value)
{
	m_maxStructureOptimizing = value;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::setMaxAtomicDisplacements(const double initialValue, const double finalValue)
{
	if (initialValue < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setInitialMaxAtomicDisplacement", "Initial maximum displacement of atoms is less than zero."));
	}

	else
	{
		if (finalValue < 0.0)
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "setInitialMaxAtomicDisplacement", "Final maximum displacement of atoms is less than zero."));
		}

		else
		{
			if (initialValue < finalValue)
			{
				OperatingSystem::Configuration::ReportException reportException;
				throw std::invalid_argument(reportException(typeid(*this), "setInitialMaxAtomicDisplacement", "Initial maximum displacement of atoms is less than the final maximum displacement."));
			}

			else
			{
				m_initialMaxAtomicDisplacement = initialValue;
				m_finalMaxAtomicDisplacement = finalValue;
			}
		}
	}
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::setMaxUnitCellDisplacementFactor(const double value)
{
	if (value < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setInitialMaxAtomicDisplacement", "Input value is less than zero."));
	}

	else
		m_maxUnitCellDisplacementFactor = value;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters::setFeasibleConstraintErrorRate(const double errorRate)
{
	if (errorRate < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setFeasibleConstraintErrorRate", "Feasible error rate of geometrical constraints is less than zero."));
	}

	else
		m_feasibleConstraintErrorRate = errorRate;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURERELAXATIONPARAMETERS_H
