#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALRELAXER_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALRELAXER_H

#include "ModelCrystalStructure.h"

#include "StructuralRelaxationParameters.h"
#include "StructureDesignRecorder.h"

#include "GeometricalBarrierHandler.h"
#include "StructureCustomizer.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Optimization
		{
			class StructuralRelaxer :public Internal::GeometricalBarrierHandler, public Internal::StructureCustomizer
			{
				using size_type = std::size_t;
				using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

				using StructureDesignRecorder = MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				StructuralRelaxer();
				explicit StructuralRelaxer(const StructuralRelaxationParameters&);
				virtual ~StructuralRelaxer() = default;

				StructuralRelaxer(const StructuralRelaxer&) = default;
				StructuralRelaxer(StructuralRelaxer&&) noexcept = default;
				StructuralRelaxer& operator=(const StructuralRelaxer&) = default;
				StructuralRelaxer& operator=(StructuralRelaxer&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				bool run(ModelCrystalStructure&);
				bool run(ModelCrystalStructure&, StructureDesignRecorder&);

				bool rerun(ModelCrystalStructure&);
				bool rerun(ModelCrystalStructure&, StructureDesignRecorder&);

				void setStructuralRelaxationParameters(const StructuralRelaxationParameters&);

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void preprocess(ModelCrystalStructure&);

				void relaxStructure();
				void recordStructure(ModelCrystalStructure&, StructureDesignRecorder&);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				StructuralRelaxationParameters m_structureRelaxationParameters;
				double m_maxUnitCellDisplacement;
				double m_maxAtomicDisplacement;
				double m_displacementReductionRate;

				mutable size_type m_mdRecordFrequency{ 0 };
				mutable size_type m_chemicalBondRecordFrequency{ 0 };
				mutable size_type m_mdRecordTimer{ 0 };
				mutable size_type m_chemicalBondRecordTimer{ 0 };
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer::setStructuralRelaxationParameters(const StructuralRelaxationParameters& structureRelaxationParameters)
{
	m_structureRelaxationParameters = structureRelaxationParameters;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer::preprocess(ModelCrystalStructure& modelCrystalStructure)
{
	m_maxAtomicDisplacement = m_structureRelaxationParameters.initialMaxAtomicDisplacement();
	m_maxUnitCellDisplacement = static_cast<double>(modelCrystalStructure.atomicArrangement().size()) * m_structureRelaxationParameters.initialMaxAtomicDisplacement() * m_structureRelaxationParameters.maxUnitCellDisplacementFactor();
	m_displacementReductionRate = std::pow((m_structureRelaxationParameters.finalMaxAtomicDisplacement() / m_structureRelaxationParameters.initialMaxAtomicDisplacement()), (1.0 / static_cast<double>(m_structureRelaxationParameters.maxStructureOptimizing())));
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer::relaxStructure()
{
	for (const auto& key : simpleMinimumBarrierKeys())
		applyMinimumBarrierForce(key.atomIndex, key.ligandIndex, key.minimumDistance);
	for (const auto& key : minimumBarrierKeys())
		applyMinimumBarrierForce(key.atomIndex, key.ligandIndex, key.latticePoint, key.minimumDistance);
	for (const auto& key : simpleMinMaxBarrierKeys())
		applyMinMaxBarrierForce(key.atomIndex, key.ligandIndex, key.minimumDistance, key.maximumDistance);
	for (const auto& key : minMaxBarrierKeys())
		applyMinMaxBarrierForce(key.atomIndex, key.ligandIndex, key.latticePoint, key.minimumDistance, key.maximumDistance);

	applyPressure();

	moveAtoms(m_maxAtomicDisplacement);
	transformBasisVectors(m_maxUnitCellDisplacement);

	m_maxUnitCellDisplacement *= m_displacementReductionRate;
	m_maxAtomicDisplacement *= m_displacementReductionRate;
}

inline void MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxer::recordStructure(ModelCrystalStructure& modelCrystalStructure, StructureDesignRecorder& structureDesignRecorder)
{
	++m_mdRecordTimer;
	++m_chemicalBondRecordTimer;
	{
		bool shouldExport = false;
		{
			if (m_mdRecordFrequency <= m_mdRecordTimer)
			{
				shouldExport = true;
				m_mdRecordTimer = 0;
			}

			if (m_chemicalBondRecordFrequency <= m_chemicalBondRecordTimer)
			{
				shouldExport = true;
				m_chemicalBondRecordTimer = 0;
			}
		}

		if (shouldExport)
			StructureCustomizer::exportStructure(modelCrystalStructure);
	}
	

	structureDesignRecorder.chemicalBondRecorder().record(modelCrystalStructure, m_structureRelaxationParameters.feasibleConstraintErrorRate());
	structureDesignRecorder.dynamicsRecorder().record(modelCrystalStructure);
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_STRUCTURALRELAXER_H
