#ifndef MATHEMATICALCRYSTALCHEMISTRY_IO_STRUCTURECLASSIFIER_H
#define MATHEMATICALCRYSTALCHEMISTRY_IO_STRUCTURECLASSIFIER_H

#include <stdexcept>
#include <string>
#include <filesystem>
#include <utility>
#include <mutex>

#include "ReportException.h"

#include "ModelCrystalStructure.h"
#include "StructureFingerprint.h"

#include "StructureReporter.h"


namespace MathematicalCrystalChemistry
{
	namespace IO
	{
		class StructureClassifier
		{
			using size_type = std::size_t;

			using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;
			using StructureFingerprint = MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor and operators

		public:
			StructureClassifier();
			explicit StructureClassifier(const std::filesystem::path& outputDirectoryPath);
			StructureClassifier(const std::filesystem::path& outputDirectoryPath, const bool needConventionalStructure);
			virtual ~StructureClassifier() = default;

			StructureClassifier(const StructureClassifier&) = default;
			StructureClassifier(StructureClassifier&&) noexcept = default;
			StructureClassifier& operator=(const StructureClassifier&) = default;
			StructureClassifier& operator=(StructureClassifier&&) noexcept = default;

		// Constructors, destructor and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Public methods

			std::filesystem::path classify(const ModelCrystalStructure&, const std::string& outputDirectoryName = std::string{}) const;
			std::filesystem::path classify(const std::filesystem::path& cifFilePath) const;

			bool needConventionalStructure() const noexcept;
			const std::filesystem::path& outputDirectoryPath() const noexcept;
			void setOutputDirectoryPath(const std::filesystem::path& outputDirectoryPath);

			static std::string fingerprintFilename() noexcept;

		// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			std::filesystem::path getStructureDirectoryPath(const ModelCrystalStructure& primitiveStructure, const StructureFingerprint&) const;
			void outputModelCrystalStructure(const ModelCrystalStructure&, const std::filesystem::path& directoryPath, const std::string filename) const;

		// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private utility

			void createFingerprintFile(const StructureFingerprint&, const std::filesystem::path& directoryPath) const;
			std::string readFingerprintFile(const std::filesystem::path& directoryPath) const;

			void createDirectoryIfNecessary(const std::filesystem::path& directoryPath) const;
			
		// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

		private:
			bool m_needConventionalStructure;
			std::filesystem::path m_outputDirectoryPath;
			StructureReporter m_structureReporter;

			static std::string m_fingerprintFilename;
			static std::mutex m_fingerprintMutex;
			static std::mutex m_directoryMutex;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool MathematicalCrystalChemistry::IO::StructureClassifier::needConventionalStructure() const noexcept
{
	return m_needConventionalStructure;
}

inline const std::filesystem::path& MathematicalCrystalChemistry::IO::StructureClassifier::outputDirectoryPath() const noexcept
{
	return m_outputDirectoryPath;
}

inline void MathematicalCrystalChemistry::IO::StructureClassifier::setOutputDirectoryPath(const std::filesystem::path& outputDirectoryPath)
{
	if (std::filesystem::is_directory(outputDirectoryPath))
		m_outputDirectoryPath = outputDirectoryPath;

	else
	{
		std::string errorMessage;
		{
			errorMessage += "\"";
			errorMessage += outputDirectoryPath.generic_string();
			errorMessage += "\" is not a path to a directory.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setOutputDirectoryPath", errorMessage));
	}
}

inline std::string MathematicalCrystalChemistry::IO::StructureClassifier::fingerprintFilename() noexcept
{
	return m_fingerprintFilename;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::filesystem::path MathematicalCrystalChemistry::IO::StructureClassifier::getStructureDirectoryPath(const ModelCrystalStructure& primitiveStructure, const StructureFingerprint& structureFingerprint) const
{
	std::string structureFingerprintSymbol = structureFingerprint.getSymbol();
	MathematicalCrystalChemistry::StructureModel::Features::AtomCategoryComposition atomCategoryComposition = primitiveStructure.getAtomCategoryComposition();

	std::filesystem::path structureDirectoryPath = m_outputDirectoryPath;
	{
		structureDirectoryPath /= atomCategoryComposition.name();
		createDirectoryIfNecessary(structureDirectoryPath);
	}
	{
		structureDirectoryPath /= std::to_string(primitiveStructure.spaceGroupNumber());
		createDirectoryIfNecessary(structureDirectoryPath);
	}

	std::lock_guard<std::mutex> guard{ m_fingerprintMutex };
	{
		size_type numDirectory = 0;
		bool isAlreadyExist = false;

		for (const auto& directoryEntry : std::filesystem::directory_iterator(structureDirectoryPath))
		{
			if (directoryEntry.is_directory())
			{
				std::string directoryFingerprintSymbol = readFingerprintFile(directoryEntry.path());

				if (!(directoryFingerprintSymbol.empty()))
				{
					++numDirectory;

					if (structureFingerprintSymbol == directoryFingerprintSymbol)
					{
						structureDirectoryPath = directoryEntry.path();
						isAlreadyExist = true;
						break;
					}
				}
			}
		}

		if (!isAlreadyExist)
		{
			++numDirectory;

			std::string directoryName;
			{
				directoryName += "Structure_";
				directoryName += std::to_string(numDirectory);
			}

			structureDirectoryPath /= directoryName;
			createDirectoryIfNecessary(structureDirectoryPath);

			createFingerprintFile(structureFingerprint, structureDirectoryPath);
		}
	}

	return structureDirectoryPath;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

inline void MathematicalCrystalChemistry::IO::StructureClassifier::createDirectoryIfNecessary(const std::filesystem::path& directoryPath) const
{
	std::lock_guard<std::mutex> guard{ m_directoryMutex };


	if (!(std::filesystem::is_directory(directoryPath)))
	{
		if (!(std::filesystem::create_directory(directoryPath)))
		{
			std::string errorMessage;
			{
				errorMessage += "Could not create the directory of \"";
				errorMessage += directoryPath.generic_string();
				errorMessage += "\"";
			}

			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "createDirectoryIfNecessary", errorMessage));
		}
	}
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_IO_STRUCTURECLASSIFIER_H
