#ifndef MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_INTERNAL_STRUCTURECREATIONREPORTER_H
#define MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_INTERNAL_STRUCTURECREATIONREPORTER_H

#include <stdexcept>
#include <string>
#include <filesystem>
#include <mutex>

#include "ModelCrystalStructure.h"
#include "StructureClassifier.h"

#include "StructureCreationReportOptions.h"
#include "StructureDesignRecorder.h"


namespace MathematicalCrystalChemistry
{
	namespace Runtime
	{
		namespace Creation
		{
			namespace Internal
			{
				class StructureCreationReporter
				{
					using size_type = std::size_t;
					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;
					using StructureDesignRecorder = MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor and operators

				public:
					StructureCreationReporter();
					explicit StructureCreationReporter(const std::filesystem::path& designedStructuresDirectoryPath);
					virtual ~StructureCreationReporter() = default;

					StructureCreationReporter(const StructureCreationReporter&) = default;
					StructureCreationReporter(StructureCreationReporter&&) noexcept = default;
					StructureCreationReporter& operator=(const StructureCreationReporter&) = default;
					StructureCreationReporter& operator=(StructureCreationReporter&&) noexcept = default;

				// Constructors, destructor and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

					void setDesignedStructuresDirectoryPath(const std::filesystem::path& designedStructuresDirectoryPath);

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					void reportOptimalStructures() const;
					void reportInfeasibleStructures() const;
					void reportExceptionalStructures() const;

					void eraseMpiDirectories() const;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					bool isMpiDirectory(const std::filesystem::path&) const;
					bool isFingerprintFilePath(const std::filesystem::path&) const;

					std::filesystem::path findMergedStructureDirectoryPath(const std::string& fingerprintSymbol, const std::filesystem::path& fingerprintFilePath, const std::filesystem::path& mpiOptimaDirectoryPath, const std::filesystem::path& mergedOptimaDirectoryPath) const;

					void moveToMergedDirectory(const std::filesystem::path& structureDirectoryPath, const std::filesystem::path& mergedStructureDirectoryPath) const;

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private utility

					std::string getDirectoryName(const std::filesystem::path&) const;
					std::string readFingerprintFile(const std::filesystem::path& filePath) const;

					void createDirectoryIfNecessary(const std::filesystem::path& directoryPath) const;
					void mergeResultFile(const std::filesystem::path& resultFilePath, const std::filesystem::path& mergedResultFilePath) const;

				// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					std::filesystem::path m_designedStructuresDirectoryPath;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::StructureCreationReporter::setDesignedStructuresDirectoryPath(const std::filesystem::path& designedStructuresDirectoryPath)
{
	if (std::filesystem::is_directory(designedStructuresDirectoryPath))
		m_designedStructuresDirectoryPath = designedStructuresDirectoryPath;
	
	else
	{
		std::string errorMessage;
		{
			errorMessage += "\"";
			errorMessage += designedStructuresDirectoryPath.generic_string();
			errorMessage += "\" is not a directory path.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "constructor", errorMessage));
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::StructureCreationReporter::eraseMpiDirectories() const
{
	for (const auto& directoryEntry : std::filesystem::directory_iterator(m_designedStructuresDirectoryPath))
	{
		std::filesystem::path mpiDirectoryPath = directoryEntry.path();

		if (isMpiDirectory(mpiDirectoryPath))
			std::filesystem::remove_all(mpiDirectoryPath);
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

inline std::string MathematicalCrystalChemistry::Runtime::Creation::Internal::StructureCreationReporter::getDirectoryName(const std::filesystem::path& directoryPath) const
{
	std::string directoryPathName = directoryPath.generic_string();
	{
		if (directoryPathName.back() == '/')
			directoryPathName.pop_back();
	}

	return std::filesystem::path{ directoryPathName }.filename().generic_string();
}

inline void MathematicalCrystalChemistry::Runtime::Creation::Internal::StructureCreationReporter::createDirectoryIfNecessary(const std::filesystem::path& directoryPath) const
{
	if (!(std::filesystem::is_directory(directoryPath)))
	{
		if (!(std::filesystem::create_directory(directoryPath)))
		{
			std::string errorMessage;
			{
				errorMessage += "Could not create the directory of \"";
				errorMessage += directoryPath.generic_string();
				errorMessage += "\"";
			}

			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "createDirectoryIfNecessary", errorMessage));
		}
	}
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_INTERNAL_STRUCTURECREATIONREPORTER_H
