#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_INTERNAL_STRUCTURECUSTOMIZER_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_INTERNAL_STRUCTURECUSTOMIZER_H

#include <array>

#include "LinearAlgebra.h"
#include "ModelCrystalStructure.h"

#include "StructureCustomizerBase.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Optimization
		{
			namespace Internal
			{
				class StructureCustomizer :public StructureCustomizerBase
				{
					using size_type = std::size_t;
					using NumericalVector = MathToolkit::Generic::NumericalVector<double>;
					using NumericalMatrix = MathToolkit::Generic::NumericalMatrix<double>;
					using LatticePoint = std::array<int, 3>;

					using ModelCrystalStruture = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

// **********************************************************************************************************************************************************************************************************************************************************************************************
						// Constructors, destructor, and operators

				protected:
					StructureCustomizer();

				public:
					virtual ~StructureCustomizer() = default;

					StructureCustomizer(const StructureCustomizer&) = default;
					StructureCustomizer(StructureCustomizer&&) noexcept = default;
					StructureCustomizer& operator=(const StructureCustomizer&) = default;
					StructureCustomizer& operator=(StructureCustomizer&&) noexcept = default;

					// Constructors, destructor, and operators
	// **********************************************************************************************************************************************************************************************************************************************************************************************
	// **********************************************************************************************************************************************************************************************************************************************************************************************
					// Public methods

					void exportStructure(ModelCrystalStruture&) const;
					void importStructure(const ModelCrystalStruture&);
					virtual void clear() noexcept;

					void setPressure() noexcept;
					void setPressure(const double);
					void setAttractiveForceConstant() noexcept;
					void setAttractiveForceConstant(const double);
					void setRepulsiveForceConstant() noexcept;
					void setRepulsiveForceConstant(const double);

					// Public methods
	// **********************************************************************************************************************************************************************************************************************************************************************************************
	// **********************************************************************************************************************************************************************************************************************************************************************************************
					// Protected methods

				protected:
					void transformBasisVectors(const double maxBasisVectorDisplacement);
					void moveAtoms(const double maxAtomDisplacement);

					void applyMinimumBarrierForce(const size_type atomIndex, const size_type ligandIndex, const double minimumDistance);
					void applyMinimumBarrierForce(const size_type atomIndex, const size_type ligandIndex, const LatticePoint& latticePoint, const double minimumDistance);
					void applyMinMaxBarrierForce(const size_type atomIndex, const size_type ligandIndex, const double minimumDistance, const double maximumDistance);
					void applyMinMaxBarrierForce(const size_type atomIndex, const size_type ligandIndex, const LatticePoint& latticePoint, const double minimumDistance, const double maximumDistance);

					void applyPressure();

					// Protected methods
	// **********************************************************************************************************************************************************************************************************************************************************************************************
	// **********************************************************************************************************************************************************************************************************************************************************************************************
					// Private methods

				private:
					NumericalVector getDisplacement(const size_type atomIndex, const size_type ligandIndex) const;
					NumericalVector getDisplacement(const size_type atomIndex, const size_type ligandIndex, const LatticePoint& latticePoint) const;

					void applyAttractiveForce(const size_type atomIndex, const size_type ligandIndex, NumericalVector displacement, const double distanceSquare, const double maximumDistance);
					void applyRepulsiveForce(const size_type atomIndex, const size_type ligandIndex, NumericalVector displacement, const double distanceSquare, const double minimumDistance);

					// Private methods
	// **********************************************************************************************************************************************************************************************************************************************************************************************
	// **********************************************************************************************************************************************************************************************************************************************************************************************
					// Private utility

					NumericalMatrix::const_pointer getAtomCoordinateIterator(const size_type atomIndex) const;
					NumericalMatrix::pointer getAtomCoordinateIterator(const size_type atomIndex);

					void applyForce(const size_type atomIndex, const NumericalVector&);
					void addBasisVectorsTransformation(const size_type columnIndex, const NumericalVector&);

					// Public utility
	// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					double m_pressure;
					double m_attractiveForceConstant;
					double m_repulsiveForceConstant;

					NumericalMatrix m_basisVectors;
					NumericalMatrix m_inverseBasisVectors;
					NumericalMatrix m_basisVectorsTransformation;
					NumericalMatrix m_atomCoordinates;
					NumericalMatrix m_atomAppliedForces;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::exportStructure(ModelCrystalStruture& modelCrystalStruture) const
{
	modelCrystalStruture.unitCell().setBasisVectors(m_basisVectors);
	size_type numAtoms = modelCrystalStruture.atomicArrangement().size();

	for (size_type atomIndex = 0; atomIndex < numAtoms; ++atomIndex)
		modelCrystalStruture.atomicArrangement()[atomIndex].coordinate() = m_atomCoordinates.columnVector(atomIndex);
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::importStructure(const ModelCrystalStruture& modelCrystalStruture)
{
	m_basisVectors = modelCrystalStruture.unitCell().basisVectors();
	m_inverseBasisVectors = m_basisVectors.getInverseMatrix();
	m_basisVectorsTransformation = NumericalMatrix(3, 3);

	size_type numAtoms = modelCrystalStruture.atomicArrangement().size();
	m_atomAppliedForces = NumericalMatrix(3, numAtoms);
	m_atomCoordinates = NumericalMatrix(3, numAtoms);
	{
		auto atomCoordinateIter = m_atomCoordinates.begin();

		for (const auto& modelAtom : modelCrystalStruture.atomicArrangement())
		{
			(*atomCoordinateIter) = modelAtom.coordinate()[0];
			++atomCoordinateIter;
			(*atomCoordinateIter) = modelAtom.coordinate()[1];
			++atomCoordinateIter;
			(*atomCoordinateIter) = modelAtom.coordinate()[2];
			++atomCoordinateIter;
		}
	}
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::clear() noexcept
{
	m_pressure = StructureCustomizerBase::commonPressure();
	m_attractiveForceConstant = StructureCustomizerBase::commonAttractiveForceConstant();
	m_repulsiveForceConstant = StructureCustomizerBase::commonRepulsiveForceConstant();

	m_basisVectors = NumericalMatrix(3, 3);
	m_inverseBasisVectors = NumericalMatrix(3, 3);
	m_basisVectorsTransformation = NumericalMatrix(3, 3);
	m_atomCoordinates.clear();
	m_atomAppliedForces.clear();
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::setPressure() noexcept
{
	m_pressure = StructureCustomizerBase::commonPressure();
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::setPressure(const double val)
{
	if (val < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setPressure", "\"pressure\" is less than zero."));
	}

	else
		m_pressure = val;
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::setAttractiveForceConstant() noexcept
{
	m_attractiveForceConstant = StructureCustomizerBase::commonAttractiveForceConstant();
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::setAttractiveForceConstant(const double val)
{
	if (val < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setAttractiveForceConstant", "\"attractiveForceConstant\" is less than zero."));
	}

	else
		m_attractiveForceConstant = val;
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::setRepulsiveForceConstant() noexcept
{
	m_repulsiveForceConstant = StructureCustomizerBase::commonRepulsiveForceConstant();
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::setRepulsiveForceConstant(const double val)
{
	if (0.0 < val)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setRepulsiveForceConstant", "\"repulsiveForceConstant\" is more than zero."));
	}

	else
		m_repulsiveForceConstant = val;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::applyMinimumBarrierForce(const size_type atomIndex, const size_type ligandIndex, const double minimumDistance)
{
	NumericalVector displacement = getDisplacement(atomIndex, ligandIndex);
	const double distanceSquare = displacement.normSquare();
	applyRepulsiveForce(atomIndex, ligandIndex, displacement, distanceSquare, minimumDistance);
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::applyMinimumBarrierForce(const size_type atomIndex, const size_type ligandIndex, const LatticePoint& latticePoint, const double minimumDistance)
{
	NumericalVector displacement = getDisplacement(atomIndex, ligandIndex, latticePoint);
	const double distanceSquare = displacement.normSquare();
	applyRepulsiveForce(atomIndex, ligandIndex, displacement, distanceSquare, minimumDistance);
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::applyMinMaxBarrierForce(const size_type atomIndex, const size_type ligandIndex, const double minimumDistance, const double maximumDistance)
{
	NumericalVector displacement = getDisplacement(atomIndex, ligandIndex);
	const double distanceSquare = displacement.normSquare();

	applyAttractiveForce(atomIndex, ligandIndex, displacement, distanceSquare, maximumDistance);
	applyRepulsiveForce(atomIndex, ligandIndex, displacement, distanceSquare, minimumDistance);
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::applyMinMaxBarrierForce(const size_type atomIndex, const size_type ligandIndex, const LatticePoint& latticePoint, const double minimumDistance, const double maximumDistance)
{
	NumericalVector displacement = getDisplacement(atomIndex, ligandIndex, latticePoint);
	const double distanceSquare = displacement.normSquare();

	applyAttractiveForce(atomIndex, ligandIndex, displacement, distanceSquare, maximumDistance);
	applyRepulsiveForce(atomIndex, ligandIndex, displacement, distanceSquare, minimumDistance);
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::applyPressure()
{
	m_basisVectorsTransformation(0, 0) += m_pressure * (m_basisVectors(2, 1) * m_basisVectors(1, 2) - m_basisVectors(1, 1) * m_basisVectors(2, 2));
	m_basisVectorsTransformation(1, 0) += m_pressure * (m_basisVectors(0, 1) * m_basisVectors(2, 2) - m_basisVectors(2, 1) * m_basisVectors(0, 2));
	m_basisVectorsTransformation(2, 0) += m_pressure * (m_basisVectors(1, 1) * m_basisVectors(0, 2) - m_basisVectors(0, 1) * m_basisVectors(1, 2));

	m_basisVectorsTransformation(0, 1) += m_pressure * (m_basisVectors(2, 2) * m_basisVectors(1, 0) - m_basisVectors(1, 2) * m_basisVectors(2, 0));
	m_basisVectorsTransformation(1, 1) += m_pressure * (m_basisVectors(0, 2) * m_basisVectors(2, 0) - m_basisVectors(2, 2) * m_basisVectors(0, 0));
	m_basisVectorsTransformation(2, 1) += m_pressure * (m_basisVectors(1, 2) * m_basisVectors(0, 0) - m_basisVectors(0, 2) * m_basisVectors(1, 0));

	m_basisVectorsTransformation(0, 2) += m_pressure * (m_basisVectors(2, 0) * m_basisVectors(1, 1) - m_basisVectors(1, 0) * m_basisVectors(2, 1));
	m_basisVectorsTransformation(1, 2) += m_pressure * (m_basisVectors(0, 0) * m_basisVectors(2, 1) - m_basisVectors(2, 0) * m_basisVectors(0, 1));
	m_basisVectorsTransformation(2, 2) += m_pressure * (m_basisVectors(1, 0) * m_basisVectors(0, 1) - m_basisVectors(0, 0) * m_basisVectors(1, 1));
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline MathToolkit::Generic::NumericalVector<double> MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::getDisplacement(const size_type atomIndex, const size_type ligandIndex) const
{
	NumericalVector displacement(3);
	{
		auto displacementIter = displacement.begin();
		auto atomIter = getAtomCoordinateIterator(atomIndex);
		auto ligandIter = getAtomCoordinateIterator(ligandIndex);

		(*displacementIter) = ((*ligandIter) - (*atomIter));
		++displacementIter;
		++atomIter;
		++ligandIter;
		(*displacementIter) = ((*ligandIter) - (*atomIter));
		++displacementIter;
		++atomIter;
		++ligandIter;
		(*displacementIter) = ((*ligandIter) - (*atomIter));
	}

	return displacement;
}

inline MathToolkit::Generic::NumericalVector<double> MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::getDisplacement(const size_type atomIndex, const size_type ligandIndex, const LatticePoint& latticePoint) const
{
	NumericalVector displacement = getDisplacement(atomIndex, ligandIndex);
	auto displacementIter = displacement.begin();
	{
		auto basisVectorsIter = m_basisVectors.begin();
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[0]);
		basisVectorsIter += 3;
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[1]);
		basisVectorsIter += 3;
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[2]);
		++displacementIter;

		basisVectorsIter = m_basisVectors.begin();
		basisVectorsIter += 1;
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[0]);
		basisVectorsIter += 3;
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[1]);
		basisVectorsIter += 3;
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[2]);
		++displacementIter;

		basisVectorsIter = m_basisVectors.begin();
		basisVectorsIter += 2;
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[0]);
		basisVectorsIter += 3;
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[1]);
		basisVectorsIter += 3;
		(*displacementIter) += (*basisVectorsIter) * static_cast<double>(latticePoint[2]);
	}

	return displacement;
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::applyAttractiveForce(const size_type atomIndex, const size_type ligandIndex, NumericalVector displacement, const double distanceSquare, const double maximumDistance)
{
	if ((maximumDistance * maximumDistance) < distanceSquare)
	{
		NumericalVector fractionalDisplacement = m_inverseBasisVectors * displacement;

		displacement *= (m_attractiveForceConstant / std::sqrt(distanceSquare));
		applyForce(atomIndex, displacement);

		displacement *= (-1.0);
		applyForce(ligandIndex, displacement);

		addBasisVectorsTransformation(0, (fractionalDisplacement[0] * displacement));
		addBasisVectorsTransformation(1, (fractionalDisplacement[1] * displacement));
		addBasisVectorsTransformation(2, (fractionalDisplacement[2] * displacement));
	}
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::applyRepulsiveForce(const size_type atomIndex, const size_type ligandIndex, NumericalVector displacement, const double distanceSquare, const double minimumDistance)
{
	if (distanceSquare < (minimumDistance * minimumDistance))
	{
		NumericalVector fractionalDisplacement = m_inverseBasisVectors * displacement;

		displacement *= (m_repulsiveForceConstant / std::sqrt(distanceSquare));
		applyForce(atomIndex, displacement);

		displacement *= (-1.0);
		applyForce(ligandIndex, displacement);

		addBasisVectorsTransformation(0, (fractionalDisplacement[0] * displacement));
		addBasisVectorsTransformation(1, (fractionalDisplacement[1] * displacement));
		addBasisVectorsTransformation(2, (fractionalDisplacement[2] * displacement));
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

inline MathToolkit::Generic::NumericalMatrix<double>::const_pointer MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::getAtomCoordinateIterator(const size_type atomIndex) const
{
	auto iter = m_atomCoordinates.begin();
	iter += (3 * atomIndex);
	return iter;
}

inline MathToolkit::Generic::NumericalMatrix<double>::pointer MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::getAtomCoordinateIterator(const size_type atomIndex)
{
	auto iter = m_atomCoordinates.begin();
	iter += (3 * atomIndex);
	return iter;
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::applyForce(const size_type atomIndex, const NumericalVector& appliedForce)
{
	auto atomAppliedForcesIter = m_atomAppliedForces.begin();
	atomAppliedForcesIter += (3 * atomIndex);
	{
		auto appliedForceIter = appliedForce.begin();

		(*atomAppliedForcesIter) += (*appliedForceIter);
		++atomAppliedForcesIter;
		++appliedForceIter;
		(*atomAppliedForcesIter) += (*appliedForceIter);
		++atomAppliedForcesIter;
		++appliedForceIter;
		(*atomAppliedForcesIter) += (*appliedForceIter);
	}
}

inline void MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizer::addBasisVectorsTransformation(const size_type columnIndex, const NumericalVector& displacement)
{
	auto basisVectorsTransformationIter = m_basisVectorsTransformation.begin();
	basisVectorsTransformationIter += (3 * columnIndex);
	{
		auto displacementIter = displacement.begin();

		(*basisVectorsTransformationIter) += (*displacementIter);
		++basisVectorsTransformationIter;
		++displacementIter;
		(*basisVectorsTransformationIter) += (*displacementIter);
		++basisVectorsTransformationIter;
		++displacementIter;
		(*basisVectorsTransformationIter) += (*displacementIter);
	}
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_OPTIMIZATION_INTERNAL_STRUCTURECUSTOMIZER_H
