#ifndef MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_STRUCTUREDESIGNRECORDER_H
#define MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_STRUCTUREDESIGNRECORDER_H

#include <stdexcept>
#include <string>
#include <filesystem>
#include <mutex>
#include <chrono>

#include "ChemicalBondRecorder.h"
#include "DynamicsRecorder.h"

#include "ModelCrystalStructure.h"


namespace MathematicalCrystalChemistry
{
	namespace Design
	{
		namespace Recording
		{
			class StructureDesignRecorder
			{
				using size_type = std::size_t;
				using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

				using ChemicalBondRecorder = Internal::ChemicalBondRecorder;
				using DynamicsRecorder = Internal::DynamicsRecorder;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				StructureDesignRecorder();
				virtual ~StructureDesignRecorder() = default;

				StructureDesignRecorder(const StructureDesignRecorder&) = default;
				StructureDesignRecorder(StructureDesignRecorder&&) noexcept = default;
				StructureDesignRecorder& operator=(const StructureDesignRecorder&) = default;
				StructureDesignRecorder& operator=(StructureDesignRecorder&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize();

				void reportOptimalStructure(ModelCrystalStructure, const std::filesystem::path& designedStructuresDirectoryPath);
				void reportOptimalStructure(ModelCrystalStructure, const std::filesystem::path& designedStructuresDirectoryPath, const std::filesystem::path workDirectoryPath);

				void reportInfeasibleStructure(ModelCrystalStructure, const std::filesystem::path& designedStructuresDirectoryPath, const std::filesystem::path workDirectoryPath);
				void reportExceptionalStructure(ModelCrystalStructure, const std::filesystem::path& designedStructuresDirectoryPath, const std::filesystem::path workDirectoryPath);

				bool needRecords() const noexcept;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				size_type structureOptimizing() const;
				size_type geometricalConstraintsOptimizing() const;

				const ChemicalBondRecorder& chemicalBondRecorder() const;
				ChemicalBondRecorder& chemicalBondRecorder();
				const DynamicsRecorder& dynamicsRecorder() const;
				DynamicsRecorder& dynamicsRecorder();

				void addStructureOptimizing(const size_type additionalStructureOptimizing = 1);
				void addGeometricalConstraintsOptimizing(const size_type additionalConstraintOptimizing = 1);

				static const std::string& basicRecordFilename() noexcept;
				static const std::string& resultFilename() noexcept;
				static const std::string& optimalStructuresDirectoryName() noexcept;
				static const std::string& infeasibleStructuresDirectoryName() noexcept;
				static const std::string& exceptionalStructuresDirectoryName() noexcept;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				std::string getDirectoryName(const std::filesystem::path& directoryPath) const;

				void createDirectoryIfNecessary(const std::filesystem::path& directoryPath) const;
				void createBasicRecordFile(const std::filesystem::path& outputDirectoryPath) const;
				void createResultFile(const ModelCrystalStructure&, const std::filesystem::path& outputDirectoryPath) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::chrono::system_clock::time_point m_initializedTimePoint;

				size_type m_structureOptimizing;
				size_type m_geometricalConstraintsOptimizing;
				ChemicalBondRecorder m_chemicalBondRecorder;
				DynamicsRecorder m_dynamicsRecorder;

				static std::mutex m_directoryMutex;
				static std::mutex m_reportMutex;
				static std::string m_basicRecordFilename;
				static std::string m_resultFilename;
				static std::string m_optimalStructuresDirectoryName;
				static std::string m_infeasibleStructuresDirectoryName;
				static std::string m_exceptionalStructuresDirectoryName;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline bool MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::needRecords() const noexcept
{
	if (m_chemicalBondRecorder.needRecord() || m_dynamicsRecorder.needRecord())
		return true;
	else
		return false;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline void MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::initialize()
{
	m_initializedTimePoint = std::chrono::system_clock::now();

	m_structureOptimizing = 0;
	m_geometricalConstraintsOptimizing = 0;

	m_chemicalBondRecorder.initialize();
	m_dynamicsRecorder.initialize();
}

inline std::size_t MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::structureOptimizing() const
{
	return m_structureOptimizing;
}

inline std::size_t MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::geometricalConstraintsOptimizing() const
{
	return m_geometricalConstraintsOptimizing;
}

inline const MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::chemicalBondRecorder() const
{
	return m_chemicalBondRecorder;
}

inline MathematicalCrystalChemistry::Design::Recording::Internal::ChemicalBondRecorder& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::chemicalBondRecorder()
{
	return m_chemicalBondRecorder;
}

inline const MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::dynamicsRecorder() const
{
	return m_dynamicsRecorder;
}

inline MathematicalCrystalChemistry::Design::Recording::Internal::DynamicsRecorder& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::dynamicsRecorder()
{
	return m_dynamicsRecorder;
}

inline void MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::addStructureOptimizing(const size_type additionalStructureOptimizing)
{
	m_structureOptimizing += additionalStructureOptimizing;
}

inline void MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::addGeometricalConstraintsOptimizing(const size_type additionalConstraintOptimizing)
{
	m_geometricalConstraintsOptimizing += additionalConstraintOptimizing;
}

inline const std::string& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::basicRecordFilename() noexcept
{
	return m_basicRecordFilename;
}

inline const std::string& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::resultFilename() noexcept
{
	return m_resultFilename;
}

inline const std::string& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::optimalStructuresDirectoryName() noexcept
{
	return m_optimalStructuresDirectoryName;
}

inline const std::string& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::infeasibleStructuresDirectoryName() noexcept
{
	return m_infeasibleStructuresDirectoryName;
}

inline const std::string& MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::exceptionalStructuresDirectoryName() noexcept
{
	return m_exceptionalStructuresDirectoryName;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline std::string MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::getDirectoryName(const std::filesystem::path& directoryPath) const
{
	std::string directoryPathName = directoryPath.generic_string();
	{
		if (directoryPathName.back() == '/')
			directoryPathName.pop_back();
	}

	return std::filesystem::path{ directoryPathName }.filename().generic_string();
}

inline void MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder::createDirectoryIfNecessary(const std::filesystem::path& directoryPath) const
{
	std::lock_guard<std::mutex> guard{ m_directoryMutex };


	if (!(std::filesystem::is_directory(directoryPath)))
	{
		if (!(std::filesystem::create_directory(directoryPath)))
		{
			std::string errorMessage;
			{
				errorMessage += "Could not create the directory of \"";
				errorMessage += directoryPath.generic_string();
				errorMessage += "\"";
			}

			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "createDirectoryIfNecessary", errorMessage));
		}
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_DESIGN_RECORDING_STRUCTUREDESIGNRECORDER_H
