#ifndef MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_STRUCTUREFINGERPRINT_H
#define MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_STRUCTUREFINGERPRINT_H

#include <string>
#include <filesystem>
#include <vector>

#include "AtomCategoryComposition.h"
#include "AtomicEnvironment.h"


namespace MathematicalCrystalChemistry
{
	namespace StructureModel
	{
		namespace Components
		{
			class ModelCrystalStructure;


			class StructureFingerprint
			{
				friend ModelCrystalStructure;

				using size_type = std::size_t;
				using AtomCategoryComposition = MathematicalCrystalChemistry::StructureModel::Features::AtomCategoryComposition;

			public:
				struct Hasher
				{
					std::size_t operator()(const StructureFingerprint& structureFingerprint) const { return structureFingerprint.getHashValue(); }
				};

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor and operators

			private:
				StructureFingerprint();

			public:
				virtual ~StructureFingerprint() = default;

				StructureFingerprint(const StructureFingerprint&) = default;
				StructureFingerprint(StructureFingerprint&&) noexcept = default;
				StructureFingerprint& operator=(const StructureFingerprint&) = default;
				StructureFingerprint& operator=(StructureFingerprint&&) noexcept = default;

			// Constructors, destructor and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				const AtomCategoryComposition& atomCategoryComposition() const;
				size_type spaceGroupNumber() const;
				const std::vector<AtomicEnvironment>& atomicEnvironments() const;

				std::string getSymbol() const;
				std::string getPlainSymbol() const;
				size_type getHashValue() const;

				void report(const std::filesystem::path& filePath) const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void setAtomCategoryComposition(const AtomCategoryComposition&);
				void setAtomCategoryComposition(AtomCategoryComposition&&);
				void setSpaceGroupNumber(const size_type);
				void setAtomicEnvironments(const std::vector<AtomicEnvironment>&);
				void setAtomicEnvironments(std::vector<AtomicEnvironment>&&);

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				AtomCategoryComposition m_atomCategoryComposition;
				size_type m_spaceGroupNumber;
				std::vector<AtomicEnvironment> m_atomicEnvironments;

				mutable bool m_isHashValueValid{ false };
				mutable size_type m_hashValue{ 0 };
			};


			inline bool isIsotypicStructureFingerprint(const StructureFingerprint& sfa, const StructureFingerprint& sfb)
			{
				if (sfa.atomCategoryComposition() == sfb.atomCategoryComposition())
				{
					if (sfa.spaceGroupNumber() == sfb.spaceGroupNumber())
					{
						if (sfa.atomicEnvironments() == sfb.atomicEnvironments())
							return true;
						else
							return false;
					}
					else
						return false;
				}
				else
					return false;
			}

			inline bool isHomeotypicStructureFingerprint(const StructureFingerprint& sfa, const StructureFingerprint& sfb)
			{
				return (sfa.getPlainSymbol() == sfb.getPlainSymbol());
			}

			inline bool operator==(const StructureFingerprint& sfa, const StructureFingerprint& sfb)
			{
				return isIsotypicStructureFingerprint(sfa, sfb);
			}

			inline bool operator!=(const StructureFingerprint& sfa, const StructureFingerprint& sfb)
			{
				return !(sfa == sfb);
			}

			inline bool operator<(const StructureFingerprint& sfa, const StructureFingerprint& sfb)
			{
				if (sfa.atomCategoryComposition() < sfb.atomCategoryComposition())
					return true;
				else if (sfb.atomCategoryComposition() < sfa.atomCategoryComposition())
					return false;
				else
				{
					if (sfa.spaceGroupNumber() < sfb.spaceGroupNumber())
						return true;
					else if (sfb.spaceGroupNumber() < sfa.spaceGroupNumber())
						return false;
					else
					{
						if (sfa.atomicEnvironments() < sfb.atomicEnvironments())
							return true;
						else
							return false;
					}
				}
			}

			inline bool operator>(const StructureFingerprint& sfa, const StructureFingerprint& sfb)
			{
				return (sfb < sfa);
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const MathematicalCrystalChemistry::StructureModel::Features::AtomCategoryComposition& MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint::atomCategoryComposition() const
{
	return m_atomCategoryComposition;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint::spaceGroupNumber() const
{
	return m_spaceGroupNumber;
}

inline const std::vector<MathematicalCrystalChemistry::StructureModel::Components::AtomicEnvironment>& MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint::atomicEnvironments() const
{
	return m_atomicEnvironments;
}

inline std::size_t MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint::getHashValue() const
{
	if (!m_isHashValueValid)
	{
		m_isHashValueValid = true;
		m_hashValue = 0;
		{
			m_hashValue ^= std::hash<size_type>{}(m_spaceGroupNumber);
			m_hashValue ^= m_atomCategoryComposition.getHashValue();

			for (const auto& atomicEnvironment : m_atomicEnvironments)
				m_hashValue ^= atomicEnvironment.getHashValue();
		}
	}

	return m_hashValue;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint::setAtomCategoryComposition(const AtomCategoryComposition& atomCategoryComposition)
{
	m_isHashValueValid = false;
	m_atomCategoryComposition = atomCategoryComposition;
}

inline void MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint::setAtomCategoryComposition(AtomCategoryComposition&& atomCategoryComposition)
{
	m_isHashValueValid = false;
	m_atomCategoryComposition = std::move(atomCategoryComposition);
}

inline void MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint::setSpaceGroupNumber(const size_type num)
{
	m_isHashValueValid = false;
	m_spaceGroupNumber = num;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_STRUCTUREMODEL_COMPONENTS_STRUCTUREFINGERPRINT_H
