#ifndef MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_STRUCTUREOPTIMALITYAPPRAISER_H
#define MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_STRUCTUREOPTIMALITYAPPRAISER_H

#include "ModelCrystalStructure.h"

#include "StructureReoptimizer.h"


namespace MathematicalCrystalChemistry
{
	namespace Runtime
	{
		namespace Analysis
		{
			namespace Internal
			{
				class StructureOptimalityAppraiser
				{
					using size_type = std::size_t;
					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;
					using StructureFingerprint = MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint;

					using StructureDesignRecordOptions = MathematicalCrystalChemistry::Design::Recording::StructureDesignRecordOptions;
					using StructuralOptimizationParameters = MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters;

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				public:
					StructureOptimalityAppraiser();
					virtual ~StructureOptimalityAppraiser() = default;

					StructureOptimalityAppraiser(const StructureOptimalityAppraiser&) = default;
					StructureOptimalityAppraiser(StructureOptimalityAppraiser&&) noexcept = default;
					StructureOptimalityAppraiser& operator=(const StructureOptimalityAppraiser&) = default;
					StructureOptimalityAppraiser& operator=(StructureOptimalityAppraiser&&) noexcept = default;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Access methods

					void setMaxReoptimizing(const size_type) noexcept;

					void setMaxAtomFractuationRatio(const double);
					void setStructureDesignRecordOptions(const StructureDesignRecordOptions&);
					void setStructuralOptimizationParameters(const StructuralOptimizationParameters&);

				// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					StructureReoptimizer::ReoptimizationResult appraise(const ModelCrystalStructure&, const StructureFingerprint&, const std::filesystem::path& workDirectoryPath) const;

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					void downgradeTotalReoptimizationResult(StructureReoptimizer::ReoptimizationResult& totalReoptimizationResult, const StructureReoptimizer::ReoptimizationResult& reoptimizationResult) const;
					StructureReoptimizer::ReoptimizationResult analyzeReoptimizationResults(const std::filesystem::path& workDirectoryPath) const;
					StructureReoptimizer::ReoptimizationResult toReoptimizationResult(const std::string&) const;

					std::filesystem::path getSoftReoptimizationDirectoryPath(const std::filesystem::path& workDirectoryPath) const;
					std::filesystem::path getHardReoptimizationDirectoryPath(const std::filesystem::path& workDirectoryPath, const size_type numReoptimizing) const;

					bool isDiscoveredReoptimizationResult(const std::filesystem::path& structureWorkDirectoryPath, const StructureReoptimizer::ReoptimizationResult&) const;

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					size_type m_maxReoptimizing;

					StructureReoptimizer m_softStructureReoptimizer;
					StructureReoptimizer m_hardStructureReoptimizer;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureOptimalityAppraiser::setMaxReoptimizing(const size_type val) noexcept
{
	m_maxReoptimizing = val;
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureOptimalityAppraiser::setMaxAtomFractuationRatio(const double val)
{
	m_hardStructureReoptimizer.setMaxAtomFractuationRatio(val);
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureOptimalityAppraiser::setStructureDesignRecordOptions(const StructureDesignRecordOptions& options)
{
	m_softStructureReoptimizer.setStructureDesignRecordOptions(options);
	m_hardStructureReoptimizer.setStructureDesignRecordOptions(options);
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureOptimalityAppraiser::setStructuralOptimizationParameters(const StructuralOptimizationParameters& parameters)
{
	StructuralOptimizationParameters softStructuralOptimizationParameters = parameters;
	{
		using MathematicalCrystalChemistry::Design::Optimization::StructuralRelaxationParameters;

		StructuralRelaxationParameters preciseStructuralRelaxationParameters = parameters.preciseRelaxationParameters();
		{
			preciseStructuralRelaxationParameters.setMaxStructureOptimizing(10000);
			preciseStructuralRelaxationParameters.setMaxAtomicDisplacements(0.1, 0.001);
		}

		softStructuralOptimizationParameters.setGlobalRelaxationParameters(StructuralRelaxationParameters{});
		softStructuralOptimizationParameters.setLocalRelaxationParameters(StructuralRelaxationParameters{});
		softStructuralOptimizationParameters.setPreciseRelaxationParameters(preciseStructuralRelaxationParameters);
	}

	m_softStructureReoptimizer.setStructuralOptimizationParameters(softStructuralOptimizationParameters);
	m_hardStructureReoptimizer.setStructuralOptimizationParameters(parameters);
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureOptimalityAppraiser::downgradeTotalReoptimizationResult(StructureReoptimizer::ReoptimizationResult& totalReoptimizationResult, const StructureReoptimizer::ReoptimizationResult& reoptimizationResult) const
{
	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::isotypic)
		totalReoptimizationResult = reoptimizationResult;

	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::homeotypic)
	{
		if (reoptimizationResult != StructureReoptimizer::ReoptimizationResult::isotypic)
			totalReoptimizationResult = reoptimizationResult;
	}

	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::same_symmetry)
	{
		if ((reoptimizationResult != StructureReoptimizer::ReoptimizationResult::isotypic) && (reoptimizationResult != StructureReoptimizer::ReoptimizationResult::homeotypic))
			totalReoptimizationResult = reoptimizationResult;
	}

	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::same_space_group_type)
	{
		if ((reoptimizationResult == StructureReoptimizer::ReoptimizationResult::optimal) || (reoptimizationResult != StructureReoptimizer::ReoptimizationResult::infeasible))
			totalReoptimizationResult = reoptimizationResult;
	}

	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::optimal)
	{
		if (reoptimizationResult == StructureReoptimizer::ReoptimizationResult::infeasible)
			totalReoptimizationResult = reoptimizationResult;
	}
}

inline MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureReoptimizer::ReoptimizationResult MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureOptimalityAppraiser::toReoptimizationResult(const std::string& text) const
{
	if (text == "isotypic")
		return StructureReoptimizer::ReoptimizationResult::isotypic;

	if (text == "homeotypic")
		return StructureReoptimizer::ReoptimizationResult::homeotypic;

	if (text == "same_symmetry")
		return StructureReoptimizer::ReoptimizationResult::same_symmetry;

	if (text == "same_space_group_type")
		return StructureReoptimizer::ReoptimizationResult::same_space_group_type;

	if (text == "optimal")
		return StructureReoptimizer::ReoptimizationResult::optimal;

	if (text == "infeasible")
		return StructureReoptimizer::ReoptimizationResult::infeasible;


	std::string errorMessage;
	{
		errorMessage += text;
		errorMessage += " is out of definitions.";
	}

	OperatingSystem::Configuration::ReportException reportException;
	throw std::runtime_error(reportException(typeid(*this), "getReoptimizationResult", errorMessage));
}

inline std::filesystem::path MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureOptimalityAppraiser::getSoftReoptimizationDirectoryPath(const std::filesystem::path& workDirectoryPath) const
{
	std::filesystem::path reoptimizationDirectoryPath = workDirectoryPath;
	reoptimizationDirectoryPath /= "Soft-reoptimization";

	if (std::filesystem::is_directory(reoptimizationDirectoryPath))
		std::filesystem::remove_all(reoptimizationDirectoryPath);

	if (!(std::filesystem::create_directories(reoptimizationDirectoryPath)))
	{
		std::string errorMessage;
		{
			errorMessage += "Could not create the directory of \"";
			errorMessage += reoptimizationDirectoryPath.generic_string();
			errorMessage += "\"";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "getSoftReoptimizationDirectoryPath", errorMessage));
	}

	return reoptimizationDirectoryPath;
}

inline std::filesystem::path MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureOptimalityAppraiser::getHardReoptimizationDirectoryPath(const std::filesystem::path& workDirectoryPath, const size_type numReoptimizing) const
{
	std::filesystem::path reoptimizationDirectoryPath = workDirectoryPath;
	reoptimizationDirectoryPath /= "Reoptimization-";
	reoptimizationDirectoryPath += std::to_string(numReoptimizing);

	if (std::filesystem::is_directory(reoptimizationDirectoryPath))
		std::filesystem::remove_all(reoptimizationDirectoryPath);

	if (!(std::filesystem::create_directories(reoptimizationDirectoryPath)))
	{
		std::string errorMessage;
		{
			errorMessage += "Could not create the directory of \"";
			errorMessage += reoptimizationDirectoryPath.generic_string();
			errorMessage += "\"";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "getHardReoptimizationDirectoryPath", errorMessage));
	}

	return reoptimizationDirectoryPath;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_STRUCTUREOPTIMALITYAPPRAISER_H
