#ifndef MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_STRUCTUREREOPTIMIZER_H
#define MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_STRUCTUREREOPTIMIZER_H

#include <random>

#include "ModelCrystalStructure.h"

#include "StructuralOptimizer.h"
#include "StructureDesignRecordOptions.h"


namespace MathematicalCrystalChemistry
{
	namespace Runtime
	{
		namespace Analysis
		{
			namespace Internal
			{
				class StructureReoptimizer
				{
					using size_type = std::size_t;
					using ModelCrystalStructure = MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;
					using StructureFingerprint = MathematicalCrystalChemistry::StructureModel::Components::StructureFingerprint;

					using StructureDesignRecordOptions = MathematicalCrystalChemistry::Design::Recording::StructureDesignRecordOptions;
					using StructuralOptimizationParameters = MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizationParameters;
					using StructuralOptimizer = MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer;

				public:
					enum class ReoptimizationResult
					{
						isotypic,
						homeotypic,
						same_symmetry,
						same_space_group_type,
						optimal,
						infeasible
					};

// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Constructors, destructor, and operators

				public:
					StructureReoptimizer();
					virtual ~StructureReoptimizer() = default;

					StructureReoptimizer(const StructureReoptimizer&) = default;
					StructureReoptimizer(StructureReoptimizer&&) noexcept = default;
					StructureReoptimizer& operator=(const StructureReoptimizer&) = default;
					StructureReoptimizer& operator=(StructureReoptimizer&&) noexcept = default;

				// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Public methods

					ReoptimizationResult reoptimize(const ModelCrystalStructure&, const StructureFingerprint&, const std::filesystem::path& workDirectoryPath) const;

					void setMaxAtomFractuationRatio(const double);
					void setStructureDesignRecordOptions(const StructureDesignRecordOptions&);
					void setStructuralOptimizationParameters(const StructuralOptimizationParameters&);

				// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
				// Private methods

				private:
					void applyFractuation(ModelCrystalStructure&) const;

					bool hasSameSymmetry(const ModelCrystalStructure&, const ModelCrystalStructure&) const;
					bool hasSameSpaceGroupNumber(const ModelCrystalStructure&, const ModelCrystalStructure&) const;

					void createPrimitiveCifFile(const ModelCrystalStructure&, const std::filesystem::path& directoryPath) const;
					void createConventionalCifFile(const ModelCrystalStructure&, const std::filesystem::path& directoryPath) const;
					void createReoptimizationResultFile(const ModelCrystalStructure&, const ReoptimizationResult&, const std::filesystem::path& directoryPath) const;

				// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

				private:
					mutable size_type m_randomValueGenerating;
					mutable std::mt19937 m_engine;

					double m_maxAtomFractuationRatio;
					StructureDesignRecordOptions m_structureDesignRecordOptions;

					StructuralOptimizer m_structuralOptimizer;
				};
			}
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureReoptimizer::setMaxAtomFractuationRatio(const double val)
{
	if (val < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setMaxAtomFractuationRatio", "Maximum fractuation of atoms is less than zero."));
	}

	else
		m_maxAtomFractuationRatio = val;
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureReoptimizer::setStructureDesignRecordOptions(const StructureDesignRecordOptions& options)
{
	m_structureDesignRecordOptions = options;
}

inline void MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureReoptimizer::setStructuralOptimizationParameters(const StructuralOptimizationParameters& parameters)
{
	m_structuralOptimizer.setStructuralOptimizationParameters(parameters);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline bool MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureReoptimizer::hasSameSymmetry(const ModelCrystalStructure& mcsa, const ModelCrystalStructure& mcsb) const
{
	if (hasSameSpaceGroupNumber(mcsa, mcsb))
	{
		auto iterA = mcsa.atomicArrangement().begin();
		auto iterB = mcsb.atomicArrangement().begin();

		for (; (iterA != mcsa.atomicArrangement().end() && iterB != mcsb.atomicArrangement().end()); ++iterA, ++iterB)
		{
			if (iterA->elementSymbol() != iterB->elementSymbol())
				return false;
			if (iterA->siteSymmetrySymbol() != iterB->siteSymmetrySymbol())
				return false;
		}

		return true;
	}

	else
		return false;
}

inline bool MathematicalCrystalChemistry::Runtime::Analysis::Internal::StructureReoptimizer::hasSameSpaceGroupNumber(const ModelCrystalStructure& mcsa, const ModelCrystalStructure& mcsb) const
{
	if (mcsa.spaceGroupNumber() == mcsb.spaceGroupNumber())
		return true;
	else
		return false;
}

// Private methods for file system
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_ANALYSIS_INTERNAL_STRUCTUREREOPTIMIZER_H
