#ifndef MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_TARGETMODELATOMCOMPOSITIONS_H
#define MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_TARGETMODELATOMCOMPOSITIONS_H

#include <stdexcept>
#include <string>
#include <filesystem>
#include <map>
#include <vector>
#include <iterator>

#include "ReportException.h"
#include "ChemicalComposition.h"

#include "ModelAtomCollection.h"


namespace MathematicalCrystalChemistry
{
	namespace Runtime
	{
		namespace Creation
		{
			class TargetModelAtomCompositions
			{
				using size_type = std::size_t;
				using ChemicalComposition = ChemToolkit::Generic::ChemicalComposition;
				using ModelAtomCollection = MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection;

// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Constructors, destructor, and operators

			public:
				TargetModelAtomCompositions();
				virtual ~TargetModelAtomCompositions() = default;

				TargetModelAtomCompositions(const TargetModelAtomCompositions&) = default;
				TargetModelAtomCompositions(TargetModelAtomCompositions&&) noexcept = default;
				TargetModelAtomCompositions& operator=(const TargetModelAtomCompositions&) = default;
				TargetModelAtomCompositions& operator=(TargetModelAtomCompositions&&) noexcept = default;

			// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Public methods

				void initialize();
				void initializeFromInputFileTexts(const std::string& inputTexts);

				std::vector<ModelAtomCollection> getModelAtomCollections() const;

			// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Access methods

				const std::vector<ChemicalComposition>& targetChemicalCompositions() const;
				std::vector<std::string> coordinationFeatureSymbolNames(const std::string& elementSymbol) const;

			// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
			// Private methods

			private:
				void initializeTargetChemicalCompositions(const std::string& inputTexts);
				void initializeCoordinationFeatureSymbols(const std::string& inputTexts);

				std::vector<ModelAtomCollection> getModelAtomCollections(const ChemicalComposition&) const;

				ModelAtomCollection createModelAtomCollection(const ChemicalComposition&, const std::map<std::string, std::vector<std::string>::const_iterator>&) const;
				bool shouldCreateModelAtomCollection(const std::map<std::string, std::vector<std::string>::const_iterator>&) const;
				void advanceCoordinationFeatureSymbolIters(std::map<std::string, std::vector<std::string>::const_iterator>&) const;

			// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************

			private:
				std::vector<ChemicalComposition> m_targetChemicalCompositions;
				std::map<std::string, std::vector<std::string>> m_coordinationFeatureSymbolNames;
			};
		}
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

inline std::vector<MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection> MathematicalCrystalChemistry::Runtime::Creation::TargetModelAtomCompositions::getModelAtomCollections() const
{
	std::vector<ModelAtomCollection> modelAtomCollections;
	{
		for (const auto& chemicalComposition : m_targetChemicalCompositions)
		{
			for (auto& modelAtomCollection : getModelAtomCollections(chemicalComposition))
				modelAtomCollections.push_back(std::move(modelAtomCollection));
		}
	}

	return modelAtomCollections;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const std::vector<ChemToolkit::Generic::ChemicalComposition>& MathematicalCrystalChemistry::Runtime::Creation::TargetModelAtomCompositions::targetChemicalCompositions() const
{
	return m_targetChemicalCompositions;
}

inline std::vector<std::string> MathematicalCrystalChemistry::Runtime::Creation::TargetModelAtomCompositions::coordinationFeatureSymbolNames(const std::string& elementSymbol) const
{
	auto iter = m_coordinationFeatureSymbolNames.find(elementSymbol);

	if (iter != m_coordinationFeatureSymbolNames.end())
		return iter->second;
	else
		return std::vector<std::string>{};
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline std::vector<MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection> MathematicalCrystalChemistry::Runtime::Creation::TargetModelAtomCompositions::getModelAtomCollections(const ChemicalComposition& chemicalComposition) const
{
	std::vector<ModelAtomCollection> modelAtomCollections;
	{
		std::map<std::string, std::vector<std::string>::const_iterator> coordinationFeatureSymbolIters;
		{
			for (const auto& elementAndFeatures : m_coordinationFeatureSymbolNames)
				coordinationFeatureSymbolIters.emplace(elementAndFeatures.first, elementAndFeatures.second.begin());
		}


		while (shouldCreateModelAtomCollection(coordinationFeatureSymbolIters))
		{
			modelAtomCollections.push_back(createModelAtomCollection(chemicalComposition, coordinationFeatureSymbolIters));
			advanceCoordinationFeatureSymbolIters(coordinationFeatureSymbolIters);
		}
	}

	return modelAtomCollections;
}

inline MathematicalCrystalChemistry::StructureModel::Components::ModelAtomCollection MathematicalCrystalChemistry::Runtime::Creation::TargetModelAtomCompositions::createModelAtomCollection(const ChemicalComposition& chemicalComposition, const std::map<std::string, std::vector<std::string>::const_iterator>& coordinationFeatureSymbolIters) const
{
	ModelAtomCollection modelAtomCollection;
	{
		for (const auto& symbolAndNum : chemicalComposition)
		{
			const std::string elementSymbol = symbolAndNum.first;
			modelAtomCollection.addAtoms(elementSymbol, symbolAndNum.second);


			if (m_coordinationFeatureSymbolNames.find(elementSymbol) != m_coordinationFeatureSymbolNames.end())
			{
				using CoordinationFeatureSymbol = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeatureSymbol;
				modelAtomCollection.registerCoordinationFeature(elementSymbol, CoordinationFeatureSymbol{ *(coordinationFeatureSymbolIters.at(elementSymbol)) });
			}
		}
	}

	return modelAtomCollection;
}

inline bool MathematicalCrystalChemistry::Runtime::Creation::TargetModelAtomCompositions::shouldCreateModelAtomCollection(const std::map<std::string, std::vector<std::string>::const_iterator>& coordinationFeatureSymbolIters) const
{
	for (const auto& elementAndIter : coordinationFeatureSymbolIters)
	{
		if (elementAndIter.second != m_coordinationFeatureSymbolNames.at(elementAndIter.first).end())
			return true;
	}

	return false;
}

inline void MathematicalCrystalChemistry::Runtime::Creation::TargetModelAtomCompositions::advanceCoordinationFeatureSymbolIters(std::map<std::string, std::vector<std::string>::const_iterator>& coordinationFeatureSymbolIters) const
{
	bool hasNoBeginingIter = false;
	{
		for (auto& elementAndIter : coordinationFeatureSymbolIters)
		{
			if (elementAndIter.second == std::prev(m_coordinationFeatureSymbolNames.at(elementAndIter.first).end()))
				elementAndIter.second = m_coordinationFeatureSymbolNames.at(elementAndIter.first).begin();

			else
			{
				++elementAndIter.second;
				hasNoBeginingIter = true;
				break;
			}
		}
	}

	if (!hasNoBeginingIter)
	{
		for (auto& elementAndIter : coordinationFeatureSymbolIters)
			elementAndIter.second = m_coordinationFeatureSymbolNames.at(elementAndIter.first).end();
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !MATHEMATICALCRYSTALCHEMISTRY_RUNTIME_CREATION_TARGETMODELATOMCOMPOSITIONS_H
