#ifndef CHEMTOOLKIT_CRYSTALLOGRAPHY_UNITCELL_H
#define CHEMTOOLKIT_CRYSTALLOGRAPHY_UNITCELL_H

#include "LinearAlgebra.h"


namespace ChemToolkit
{
	namespace Crystallography
	{
		struct LatticeParameters final
		{
			struct Lengths
			{
				double a;
				double b;
				double c;
			};

			struct Angles
			{
				double alpha;
				double beta;
				double gamma;
			};

			Lengths lengths;
			Angles angles;
		};


		class UnitCell final
		{
			using size_type = std::size_t;
			using NumericalVector = MathToolkit::Generic::NumericalVector<double>;
			using NumericalMatrix = MathToolkit::Generic::NumericalMatrix<double>;
			using NumericalVectorSlice = MathToolkit::Generic::NumericalVectorSlice<double>;

// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Constructors, destructor, and operators

		public:
			UnitCell();
			explicit UnitCell(const NumericalMatrix& basisVectors);
			explicit UnitCell(NumericalMatrix&& basisVectors);
			explicit UnitCell(const LatticeParameters&);

			UnitCell(const UnitCell&) = default;
			UnitCell(UnitCell&&) noexcept = default;
			UnitCell& operator=(const UnitCell&) = default;
			UnitCell& operator=(UnitCell&&) noexcept = default;

			~UnitCell() = default;

		// Constructors, destructor, and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Access methods

			const NumericalMatrix& basisVectors() const;
			const NumericalMatrix& inverseBasisVectors() const;
			NumericalMatrix reciprocalVectors() const;

			const NumericalMatrix& appliedTransformation() const;
			NumericalMatrix& appliedTransformation();

			void setBasisVectors(const NumericalMatrix& basisVectors);
			void setBasisVectors(NumericalMatrix&& basisVectors);

		// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Utility

			bool isValid() const;
			void clear();

			void transform();
			void transform(const double maxDisplacement);

			bool hasPositiveVolume() const;
			double getCellVolume() const;
			LatticeParameters getLatticeParameters() const;

		// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
		// Private methods

		private:
			double getAngle(const NumericalVector&, const NumericalVector&) const;

		// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************


		private:
			NumericalMatrix m_basisVectors;
			NumericalMatrix m_inverseBasisVectors;
			NumericalMatrix m_appliedTransformation;
		};
	}
}

// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

inline const MathToolkit::Generic::NumericalMatrix<double>& ChemToolkit::Crystallography::UnitCell::basisVectors() const
{
	return m_basisVectors;
}

inline const MathToolkit::Generic::NumericalMatrix<double>& ChemToolkit::Crystallography::UnitCell::inverseBasisVectors() const
{
	return m_inverseBasisVectors;
}

inline const MathToolkit::Generic::NumericalMatrix<double>& ChemToolkit::Crystallography::UnitCell::appliedTransformation() const
{
	return m_appliedTransformation;
}

inline MathToolkit::Generic::NumericalMatrix<double>& ChemToolkit::Crystallography::UnitCell::appliedTransformation()
{
	return m_appliedTransformation;
}

inline void ChemToolkit::Crystallography::UnitCell::setBasisVectors(const NumericalMatrix& basisVectors)
{
	if (basisVectors.columnSize() == 3 && basisVectors.rowSize() == 3)
	{
		m_basisVectors = basisVectors;
		m_inverseBasisVectors = m_basisVectors.getInverseMatrix();
		m_appliedTransformation = 0.0;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setBasisVectors", "The sizes of the column and/or row of \"basisVectors\" are not three."));
	}
}

inline void ChemToolkit::Crystallography::UnitCell::setBasisVectors(NumericalMatrix&& basisVectors)
{
	if (basisVectors.columnSize() == 3 && basisVectors.rowSize() == 3)
	{
		m_basisVectors = std::move(basisVectors);
		m_inverseBasisVectors = m_basisVectors.getInverseMatrix();
		m_appliedTransformation = 0.0;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "setBasisVectors", "The sizes of the column and/or row of \"basisVectors\" are not three."));
	}
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

inline void ChemToolkit::Crystallography::UnitCell::transform()
{
	m_basisVectors += m_appliedTransformation;
	m_inverseBasisVectors = m_basisVectors.getInverseMatrix();
	m_appliedTransformation = 0.0;
}

inline void ChemToolkit::Crystallography::UnitCell::transform(const double maxDisplacement)
{
	double transformationNormSquare = m_appliedTransformation.getHilbertSchmidtNormSquare();
	{
		if ((maxDisplacement * maxDisplacement) < transformationNormSquare)
			m_appliedTransformation *= (maxDisplacement / std::sqrt(transformationNormSquare));
	}

	m_basisVectors += m_appliedTransformation;
	m_inverseBasisVectors = m_basisVectors.getInverseMatrix();
	m_appliedTransformation = 0.0;
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

inline bool ChemToolkit::Crystallography::UnitCell::isValid() const
{
	return true;
}

inline void ChemToolkit::Crystallography::UnitCell::clear()
{
	m_basisVectors = 0.0;
	m_basisVectors(0, 0) = 1.0;
	m_basisVectors(1, 1) = 1.0;
	m_basisVectors(2, 2) = 1.0;

	m_inverseBasisVectors = 0.0;
	m_inverseBasisVectors(0, 0) = 1.0;
	m_inverseBasisVectors(1, 1) = 1.0;
	m_inverseBasisVectors(2, 2) = 1.0;

	m_appliedTransformation = 0.0;
}

inline bool ChemToolkit::Crystallography::UnitCell::hasPositiveVolume() const
{
	if (0.0 < getCellVolume())
		return true;
	else
		return false;
}

inline double ChemToolkit::Crystallography::UnitCell::getCellVolume() const
{
	double primitiveCellVolume{ 0.0 };
	{
		primitiveCellVolume += m_basisVectors(0, 0) * ((m_basisVectors(1, 1) * m_basisVectors(2, 2)) - (m_basisVectors(2, 1) * m_basisVectors(1, 2)));
		primitiveCellVolume += m_basisVectors(1, 0) * ((m_basisVectors(2, 1) * m_basisVectors(0, 2)) - (m_basisVectors(0, 1) * m_basisVectors(2, 2)));
		primitiveCellVolume += m_basisVectors(2, 0) * ((m_basisVectors(0, 1) * m_basisVectors(1, 2)) - (m_basisVectors(1, 1) * m_basisVectors(0, 2)));
	}

	return primitiveCellVolume;
}

inline ChemToolkit::Crystallography::LatticeParameters ChemToolkit::Crystallography::UnitCell::getLatticeParameters() const
{
	NumericalVector latticeVectorA = m_basisVectors.columnVector(0);
	NumericalVector latticeVectorB = m_basisVectors.columnVector(1);
	NumericalVector latticeVectorC = m_basisVectors.columnVector(2);

	double a = latticeVectorA.norm();
	double b = latticeVectorB.norm();
	double c = latticeVectorC.norm();
	double alpha = getAngle(latticeVectorB, latticeVectorC);
	double beta = getAngle(latticeVectorC, latticeVectorA);
	double gamma = getAngle(latticeVectorA, latticeVectorB);

	LatticeParameters::Lengths lengths{ a,b,c };
	LatticeParameters::Angles angles{ alpha, beta, gamma };
	return LatticeParameters{ lengths, angles };
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

inline double ChemToolkit::Crystallography::UnitCell::getAngle(const NumericalVector& nva, const NumericalVector& nvb) const
{
	double angle = 0.0;
	{
		double cosine = MathToolkit::Generic::getInnerProduct(nva, nvb);
		cosine /= std::sqrt(nva.normSquare() * nvb.normSquare());
		angle = std::acos(cosine);
	}

	return angle;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************


#endif // !CHEMTOOLKIT_CRYSTALLOGRAPHY_UNITCELL_H
