#include "ChemicalBondRecorder.h"

using namespace MathematicalCrystalChemistry::Design::Recording::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

std::string ChemicalBondRecorder::m_feasibleAtomRatesFileStem{ "feasible_atom_rates" };
std::string ChemicalBondRecorder::m_feasibleChemicalBondRatesFileStem{ "feasible_chemical_bond_rates" };
std::string ChemicalBondRecorder::m_averageChemicalBondErrorsFileStem{ "average_chemical_bond_errors" };


ChemicalBondRecorder::ChemicalBondRecorder()
	: m_recordFrequency{ 0 }
	, m_feasibleAtomRatesWriter{}
	, m_feasibleChemicalBondRatesWriter{}
	, m_averageChemicalBondErrorsWriter{}
	, m_feasibleAtomRatesGplFileWriter{}
	, m_feasibleChemicalBondRatesGplFileWriter{}
	, m_averageChemicalBondErrorsGplFileWriter{}
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void ChemicalBondRecorder::record(const ModelCrystalStructure& modelCrystalStructure, const double feasibleConstraintErrorRate) const
{
	++m_recordTimer;
	++m_structureOptimizing;

	if (m_recordTimer == m_recordFrequency)
	{
		size_type numFeasibleAtom = 0;
		size_type numFeasibleChemicalBonds = 0;
		size_type numChemcalBonds = 0;
		double chemicalBondErrorSum = 0.0;

		for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
		{
			std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond>> ionicBondPtrs = modelAtom.getIonicBondPtrs();
			std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond>> metallicBondPtrs = modelAtom.getMetallicBondPtrs();
			numChemcalBonds += ionicBondPtrs.size();
			numChemcalBonds += metallicBondPtrs.size();

			bool hasInfeasibleChemicalBonds = false;
			{
				for (const auto& ionicBondPtr : ionicBondPtrs)
				{
					if (ionicBondPtr->isFeasible(feasibleConstraintErrorRate))
						++numFeasibleChemicalBonds;

					else
					{
						hasInfeasibleChemicalBonds = true;
						double interatomicDistanceSquare = (ionicBondPtr->ligand().coordinate() - ionicBondPtr->modelAtomRef().coordinate()).normSquare();
						double minInteratomicDistance = ionicBondPtr->minimumDistance();

						if (interatomicDistanceSquare < (minInteratomicDistance * minInteratomicDistance))
							chemicalBondErrorSum += (minInteratomicDistance - std::sqrt(interatomicDistanceSquare));

						else
						{
							double maxInteratomicDistance = ionicBondPtr->maximumDistance();
							chemicalBondErrorSum += (std::sqrt(interatomicDistanceSquare) - maxInteratomicDistance);
						}
					}
				}
			}

			if (!hasInfeasibleChemicalBonds)
				++numFeasibleAtom;
		}


		recordFeasibleAtomRate(numFeasibleAtom / static_cast<double>(modelCrystalStructure.atomicArrangement().size()));
		recordFeasibleChemicalBondRate(numFeasibleChemicalBonds / static_cast<double>(numChemcalBonds));
		recordAverageChemicalBondError(chemicalBondErrorSum / static_cast<double>(numChemcalBonds));
		m_recordTimer = 0;
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void ChemicalBondRecorder::createFeasibleAtomRatesGplFile(const std::filesystem::path& workDirectoryPath)
{
	std::string outputText = writeCommonGplHeaderText(m_feasibleAtomRatesFileStem);
	{
		outputText += "set xlabel \"Number of structural optimization\" font \"Times-Roman,22\"\n";
		outputText += "set ylabel \"Rate of feasible atoms\" font \"Times-Roman,22\"\n";

		outputText += "plot \\\n";
		{
			outputText += "     \"";
			outputText += m_feasibleAtomRatesWriter.filePath().filename().generic_string();
			outputText += "\" u 1:2 with line lc rgb \"red\" t \"Feasible Atom :\", \\\n";
		}
	}


	std::filesystem::path feasibleAtomRatesGplFilePath = workDirectoryPath;
	{
		std::string filename = m_feasibleAtomRatesFileStem;
		filename += ".gpl";
		feasibleAtomRatesGplFilePath /= filename;
	}
	m_feasibleAtomRatesGplFileWriter = FileWriter{ feasibleAtomRatesGplFilePath };
	m_feasibleAtomRatesGplFileWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing, outputText);
}

void ChemicalBondRecorder::createFeasibleChemicalBondRatesGplFile(const std::filesystem::path& workDirectoryPath)
{
	std::string outputText = writeCommonGplHeaderText(m_feasibleChemicalBondRatesFileStem);
	{
		outputText += "set xlabel \"Number of structural optimization\" font \"Times-Roman,22\"\n";
		outputText += "set ylabel \"Rate of feasible chemical bonds\" font \"Times-Roman,22\"\n";

		outputText += "plot \\\n";
		{
			outputText += "     \"";
			outputText += m_feasibleChemicalBondRatesWriter.filePath().filename().generic_string();
			outputText += "\" u 1:2 with line lc rgb \"red\" t \"Chemical Bonding :\", \\\n";
		}
	}


	std::filesystem::path feasibleChemicalBondRatesGplFilePath = workDirectoryPath;
	{
		std::string filename = m_feasibleChemicalBondRatesFileStem;
		filename += ".gpl";
		feasibleChemicalBondRatesGplFilePath /= filename;
	}
	m_feasibleChemicalBondRatesGplFileWriter = FileWriter{ feasibleChemicalBondRatesGplFilePath };
	m_feasibleChemicalBondRatesGplFileWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing, outputText);
}

void ChemicalBondRecorder::createAverageChemicalBondErrorsGplFile(const std::filesystem::path& workDirectoryPath)
{
	std::string outputText = writeCommonGplHeaderText(m_averageChemicalBondErrorsFileStem);
	{
		outputText += "set xlabel \"Number of structural optimization\" font \"Times-Roman,22\"\n";
		outputText += "set ylabel \"Average error\" font \"Times-Roman,22\"\n";

		outputText += "plot \\\n";
		{
			outputText += "     \"";
			outputText += m_averageChemicalBondErrorsWriter.filePath().filename().generic_string();
			outputText += "\" u 1:2 with line lc rgb \"red\" t \"Chemical Bonding :\", \\\n";
		}
	}


	std::filesystem::path averageChemicalBondErrorsGplFilePath = workDirectoryPath;
	{
		std::string filename = m_averageChemicalBondErrorsFileStem;
		filename += ".gpl";
		averageChemicalBondErrorsGplFilePath /= filename;
	}
	m_averageChemicalBondErrorsGplFileWriter = FileWriter{ averageChemicalBondErrorsGplFilePath };
	m_averageChemicalBondErrorsGplFileWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing, outputText);
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
