#include "CifFileReader.h"

#include <sstream>
#include <regex>

#include "LogWriter.h"
#include "ReportException.h"
#include "LengthCasting.h"
#include "AngleCasting.h"

using namespace ChemToolkit::Crystallography::FileIO;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

CifFileReader::CifFileReader(const std::filesystem::path& cifFilePath)
	: FileReader{ cifFilePath }
{
	if (!(isCorrectExtension(filePath())))
	{
		std::string errorMessage;
		{
			errorMessage += "The extension of cifFilePath, which is \"";
			errorMessage += filePath().generic_string();
			errorMessage += "\", is not \"";
			errorMessage += correctExtension();
			errorMessage += "\".";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", errorMessage));
	}
}

CifFileReader::CifFileReader(std::filesystem::path&& cifFilePath)
	: FileReader{ std::move(cifFilePath) }
{
	if (!(isCorrectExtension(filePath())))
	{
		std::string errorMessage;
		{
			errorMessage += "The extension of cifFilePath, which is \"";
			errorMessage += filePath().generic_string();
			errorMessage += "\", is not \"";
			errorMessage += correctExtension();
			errorMessage += "\".";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", errorMessage));
	}
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

ChemToolkit::Crystallography::UnitCell CifFileReader::readUnitCell(const std::string& inputText) const
{
	ChemToolkit::Crystallography::LatticeParameters latticeParameters;
	{
		using  namespace MathToolkit::UnitConversion;
		using Angstrom = LengthCasting::Unit::Angstrom;
		using AtomicUnit = LengthCasting::Unit::AtomicUnit;
		using Degree = AngleCasting::Unit::Degree;
		using Radian = AngleCasting::Unit::Radian;
		{
			double lengthA = 0.0;
			readData(inputText, "_cell_length_a", lengthA);
			latticeParameters.lengths.a = LengthCasting::cast<Angstrom, AtomicUnit>(lengthA);
		}
		{
			double lengthB = 0.0;
			readData(inputText, "_cell_length_b", lengthB);
			latticeParameters.lengths.b = LengthCasting::cast<Angstrom, AtomicUnit>(lengthB);
		}
		{
			double lengthC = 0.0;
			readData(inputText, "_cell_length_c", lengthC);
			latticeParameters.lengths.c = LengthCasting::cast<Angstrom, AtomicUnit>(lengthC);
		}
		{
			double angleA = 0.0;
			readData(inputText, "_cell_angle_alpha", angleA);
			latticeParameters.angles.alpha = AngleCasting::cast<Degree, Radian>(angleA);
		}
		{
			double angleB = 0.0;
			readData(inputText, "_cell_angle_beta", angleB);
			latticeParameters.angles.beta = AngleCasting::cast<Degree, Radian>(angleB);
		}
		{
			double angleC = 0.0;
			readData(inputText, "_cell_angle_gamma", angleC);
			latticeParameters.angles.gamma = AngleCasting::cast<Degree, Radian>(angleC);
		}
	}

	return ChemToolkit::Crystallography::UnitCell{ latticeParameters };
}

std::vector<ChemToolkit::Crystallography::SpaceGroups::AtomSite> CifFileReader::readAtomSites(const std::string& inputText) const
{
	std::string atomSiteText = getLoopText(inputText, "_atom_site_fract_");
	std::vector<std::string> loopDataNames;
	{
		if (atomSiteText.empty())
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "readAtomSites", "Could not read atom sites."));
		}

		else
			loopDataNames = readLoopDataNames(atomSiteText, "_atom_site_");
	}

	std::vector<AtomSite> atomSites;
	{
		std::stringstream atomSiteStream{ atomSiteText };

		std::regex pat("([[:alnum:]\\+\\-\\.\\(\\)]+)[\\s]*");
		std::regex totalPat;
		{
			std::string patText{ "[\\s]*" };;
			{
				for (const auto& loopDataName : loopDataNames)
					patText += std::string{ "([[:alnum:]\\+\\-\\.\\(\\)]+)[\\s]*" };
			}

			totalPat = std::regex{ patText };
		}


		for (std::string stringLine; std::getline(atomSiteStream, stringLine);)
		{
			std::smatch totalSm;

			if (std::regex_match(stringLine, totalSm, totalPat))
			{
				AtomSite atomSite;
				{
					MathToolkit::Generic::NumericalVector<double> fractionalCoordinate(3);
					auto loopDataNameIter = loopDataNames.begin();

					for (std::sregex_iterator iter(stringLine.begin(), stringLine.end(), pat); (iter != std::sregex_iterator{} && loopDataNameIter != loopDataNames.end()); ++iter, ++loopDataNameIter)
					{
						if ((*loopDataNameIter) == "_atom_site_label")
						{
							std::regex siteLabelPat("([[:alpha:]]+)([\\d]*)");
							std::smatch siteLabelSm;
							std::string siteLabel = iter->str(1);

							if (std::regex_match(siteLabel, siteLabelSm, siteLabelPat))
							{
								atomSite.setElementSymbol(siteLabelSm.str(1));
								atomSite.setSiteLabel(siteLabel);
							}

							else
							{
								OperatingSystem::Configuration::ReportException reportException;
								throw std::invalid_argument(reportException(typeid(*this), "readAtomSites", "Could not read _atom_site_label."));
							}
						}

						if ((*loopDataNameIter) == "_atom_site_Wyckoff_symbol")
							atomSite.setWyckoffSymbol(iter->str(1));
						if ((*loopDataNameIter) == "_atom_site_fract_x")
							translateString(iter->str(1), fractionalCoordinate[0]);
						if ((*loopDataNameIter) == "_atom_site_fract_y")
							translateString(iter->str(1), fractionalCoordinate[1]);
						if ((*loopDataNameIter) == "_atom_site_fract_z")
							translateString(iter->str(1), fractionalCoordinate[2]);

						if ((*loopDataNameIter) == "_atom_site_occupancy")
						{
							double occupancy = 1.0;
							translateString(iter->str(1), occupancy);
							atomSite.setOccupancy(occupancy);
						}
					}

					atomSite.setFractionalCoordinate(fractionalCoordinate);
				}

				if (atomSite.isValid())
					atomSites.push_back(std::move(atomSite));
				else
				{
					OperatingSystem::Configuration::ReportException reportException;
					throw std::runtime_error(reportException(typeid(*this), "readAtomSites", "Could not read atom sites."));
				}
			}
		}
	}

	return atomSites;
}

void CifFileReader::readSpaceGroupInformation(const std::string& inputText, CrystallographicInformation& crystallographicInformation) const
{
	std::string symmetryOperationsText;
	{
		symmetryOperationsText = getLoopText(inputText, "_space_group_symop_operation_xyz");
		{
			if (symmetryOperationsText.empty())
			{
				symmetryOperationsText = getLoopText(inputText, "_symmetry_equiv_pos_as_xyz");

				if (symmetryOperationsText.empty())
				{
					OperatingSystem::Configuration::ReportException reportException;
					throw std::runtime_error(reportException(typeid(*this), "readSpaceGroupInformation", "Could not read symmetry operations."));
				}
			}
		}
	}

	std::vector<AffineMapping> symmetryOperations;
	{
		std::stringstream symmetryOperationStream{ symmetryOperationsText };

		for (std::string stringLine; std::getline(symmetryOperationStream, stringLine);)
		{
			std::regex totalPat;
			std::smatch totalSm;
			{
				std::string totalPatString{ "[\\']{1}[\\s]*" };
				{
					std::string operationPatString{ "([xyz\\d\\+\\-\\/]+)" };

					totalPatString += operationPatString;
					totalPatString += "[\\,]{1}[\\s]*";
					totalPatString += operationPatString;
					totalPatString += "[\\,]{1}[\\s]*";
					totalPatString += operationPatString;
					totalPatString += "[\\s]*[\\']{1}";
				}

				totalPat = std::regex(totalPatString);
			}

			if (std::regex_search(stringLine, totalSm, totalPat))
				symmetryOperations.push_back(AffineMapping{ readSymmetryOperation(totalSm.str(1), totalSm.str(2), totalSm.str(3)) });
		}
	}


	crystallographicInformation.setSpaceGroupNumber(readSpaceGroupNumber(inputText));
	crystallographicInformation.setSymmetryOperations(std::move(symmetryOperations));
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

std::size_t CifFileReader::readSpaceGroupNumber(const std::string& inputText) const
{
	size_type spaceGroupNumber = 0;

	if (tryReadData(inputText, "_space_group_IT_number", spaceGroupNumber))
		return spaceGroupNumber;
	else
	{
		if (tryReadData(inputText, "_symmetry_Int_Tables_number", spaceGroupNumber))
			return spaceGroupNumber;
		else
		{
			std::string message;
			{
				message += "Could not read the space group number of cif file of \"";
				message += filePath().generic_string();
				message += "\".";
			}

			return 1;
		}
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods for loop data

std::string CifFileReader::getLoopText(const std::string& inputText, const std::string& dataSign) const
{
	auto fractPos = inputText.find(dataSign);
	{
		if (fractPos == std::string::npos)
			return std::string{};
	}

	auto initialLoopPos = inputText.rfind("loop_", fractPos);
	{
		if (initialLoopPos == std::string::npos)
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "getloopText", "Could not find initial \"loop_\"."));
		}

		else
			initialLoopPos += 6;
	}

	auto finalLoopPos = inputText.find("loop_", fractPos);
	{
		if (finalLoopPos == std::string::npos)
		{
			finalLoopPos = inputText.find("#End", fractPos);

			if (finalLoopPos == std::string::npos)
				finalLoopPos = inputText.size();
		}
	}


	if (initialLoopPos < finalLoopPos)
		return std::string{ inputText, initialLoopPos, (finalLoopPos - initialLoopPos) };
	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "getloopText", "Could not find valid initial and/or final \"loop_\"."));
	}
}

std::vector<std::string> CifFileReader::readLoopDataNames(const std::string& inputText, const std::string& dataSign) const
{
	std::stringstream dataStream{ inputText };

	std::vector<std::string> dataNames;
	{
		for (std::string stringLine; std::getline(dataStream, stringLine);)
		{
			auto dataNamePos = stringLine.find(dataSign);

			if (dataNamePos == std::string::npos)
				break;

			else
			{
				std::regex pat("[\\s]*[\\_]{1}([\\_[:alpha:]]+)[\\s]*");
				std::smatch sm;

				if (std::regex_match(stringLine, sm, pat))
				{
					std::string dataName;
					{
						dataName += "_";
						dataName += sm.str(1);
					}

					dataNames.push_back(dataName);
				}

				else
				{
					OperatingSystem::Configuration::ReportException reportException;
					throw std::runtime_error(reportException(typeid(*this), "readAtomSites", "Invalid data name."));
				}
			}
		}
	}

	return dataNames;
}

// Private methods for loop data
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

MathToolkit::Generic::AffineMapping<double> CifFileReader::readSymmetryOperation(const std::string& xOperation, const std::string& yOperation, const std::string& zOperation) const
{
	MathToolkit::Generic::NumericalMatrix<double> matrixPart(3, 3);
	MathToolkit::Generic::NumericalVector<double> columnPart(3);

	std::vector<std::string> operationStrings{ xOperation, yOperation, zOperation };

	for (std::size_t index = 0; index < 3; ++index)
	{
		std::string operationString{ operationStrings[index] };
		std::regex termPat("([xyz\\/\\d]+)");

		for (std::sregex_iterator iter(operationString.begin(), operationString.end(), termPat); iter != std::sregex_iterator{}; ++iter)
		{
			std::string termString = iter->str(1);
			double pmCoefficient{ 1.0 };
			{
				std::string prevText = iter->prefix().str();

				if (!(prevText.empty()))
				{
					if (prevText.back() == '-')
						pmCoefficient = (-1.0);
				}
			}

			if (termString.back() == 'x')
			{
				termString.pop_back();
				matrixPart(index, 0) = pmCoefficient * translateIntoTermValue(termString);
			}

			else if (termString.back() == 'y')
			{
				termString.pop_back();
				matrixPart(index, 1) = pmCoefficient * translateIntoTermValue(termString);
			}

			else if (termString.back() == 'z')
			{
				termString.pop_back();
				matrixPart(index, 2) = pmCoefficient * translateIntoTermValue(termString);
			}

			else
				columnPart[index] = pmCoefficient * translateIntoTermValue(termString);
		}
	}

	return AffineMapping{ matrixPart, columnPart };
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
