#include "CifFileWriter.h"

#include "PeriodicTable.h"

#include "AngleCasting.h"
#include "LengthCasting.h"

using namespace ChemToolkit::Crystallography::FileIO;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

CifFileWriter::CifFileWriter(const std::filesystem::path& cifFilePath)
	: FileWriter{ cifFilePath }
{
	if (!(isCorrectExtension(filePath())))
	{
		std::string errorMessage;
		{
			errorMessage += "The extension of cifFilePath, which is \"";
			errorMessage += filePath().generic_string();
			errorMessage += "\", is not \"";
			errorMessage += correctExtension();
			errorMessage += "\".";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", errorMessage));
	}
}

CifFileWriter::CifFileWriter(std::filesystem::path&& cifFilePath)
	: FileWriter{ std::move(cifFilePath) }
{
	if (!(isCorrectExtension(filePath())))
	{
		std::string errorMessage;
		{
			errorMessage += "The extension of cifFilePath, which is \"";
			errorMessage += filePath().generic_string();
			errorMessage += "\", is not \"";
			errorMessage += correctExtension();
			errorMessage += "\".";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", errorMessage));
	}
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

std::string CifFileWriter::writeHeaderMessage(const std::string& materialName) const
{
	std::string headerString;
	headerString += "data_";
	{
		headerString += materialName;
		headerString += "\n";
	}
	headerString += "_symmetry_space_group_name_H-M    'P1'\n";

	return headerString;
}

std::string CifFileWriter::writeCrystallographicInfo(const ChemToolkit::Generic::ChemicalComposition& chemicalComposition, const double cellVolume) const
{
	ChemToolkit::Generic::ChemicalComposition reducedChemicalComposition = chemicalComposition.getReducedChemicalComposition();

	std::string outputText;
	outputText += "_symmetry_Int_Tables_number   1\n";
	outputText += "_chemical_formula_structural   ";
	std::size_t firstElementCountStructural = 0;
	{
		for (const auto& symbolAndCounting : reducedChemicalComposition)
		{
			outputText += symbolAndCounting.first;
			outputText += std::to_string(symbolAndCounting.second);
		}

		outputText += "\n";
		firstElementCountStructural = reducedChemicalComposition.begin()->second;
	}
	outputText += "_chemical_formula_sum   \'";

	std::size_t firstElementCountSum = 0;
	{
		for (const auto& symbolAndCounting : chemicalComposition)
		{
			outputText += symbolAndCounting.first;
			outputText += std::to_string(symbolAndCounting.second);
			outputText += " ";
		}

		outputText.pop_back();
		outputText += "\'\n";
		firstElementCountSum = chemicalComposition.begin()->second;
	}
	outputText += "_cell_volume   ";
	{
		using MathToolkit::UnitConversion::LengthCasting::Unit::AtomicUnit;
		using MathToolkit::UnitConversion::LengthCasting::Unit::Angstrom;
		const double lengthScalingMultiplier = MathToolkit::UnitConversion::LengthCasting::getScalingMultiplier<AtomicUnit, Angstrom>();
		const double volumeMultiplier = (lengthScalingMultiplier * lengthScalingMultiplier * lengthScalingMultiplier);

		outputText += std::to_string(volumeMultiplier * cellVolume);
		outputText += "\n";
	}
	outputText += "_cell_formula_units_Z   ";
	{
		outputText += std::to_string(firstElementCountSum / firstElementCountStructural);
		outputText += "\n";
	}
	outputText += "loop_\n";
	outputText += " _symmetry_equiv_pos_site_id\n";
	outputText += " _symmetry_equiv_pos_as_xyz\n";
	outputText += "  1  \'x, y, z\'\n";
	outputText += "loop_\n";

	return outputText;
}

std::string ChemToolkit::Crystallography::FileIO::CifFileWriter::writeUnitCellParameters(const ChemToolkit::Crystallography::UnitCell& unitCell) const
{
	ChemToolkit::Crystallography::LatticeParameters latticeParameters = unitCell.getLatticeParameters();

	std::stringstream outputStream;
	{
		using MathToolkit::UnitConversion::LengthCasting::Unit::AtomicUnit;
		using MathToolkit::UnitConversion::LengthCasting::Unit::Angstrom;
		using MathToolkit::UnitConversion::AngleCasting::Unit::Degree;
		using MathToolkit::UnitConversion::AngleCasting::Unit::Radian;

		const double angleScalingMultiplier = MathToolkit::UnitConversion::AngleCasting::getScalingMultiplier<Radian, Degree>();
		const double lengthScalingMultiplier = MathToolkit::UnitConversion::LengthCasting::getScalingMultiplier<AtomicUnit, Angstrom>();

		outputStream.precision(12);
		outputStream.flags(std::ios_base::fixed);
		outputStream << "_cell_length_a        " << (lengthScalingMultiplier * latticeParameters.lengths.a) << "\n";
		outputStream << "_cell_length_b        " << (lengthScalingMultiplier * latticeParameters.lengths.b) << "\n";
		outputStream << "_cell_length_c        " << (lengthScalingMultiplier * latticeParameters.lengths.c) << "\n";
		outputStream << "_cell_angle_alpha     " << (angleScalingMultiplier * latticeParameters.angles.alpha) << "\n";
		outputStream << "_cell_angle_beta      " << (angleScalingMultiplier * latticeParameters.angles.beta) << "\n";
		outputStream << "_cell_angle_gamma     " << (angleScalingMultiplier * latticeParameters.angles.gamma) << "\n";
	}

	return outputStream.str();
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
