#include "ConstraintManager.h"

#include <algorithm>
#include <iterator>
#include <limits>

#include "MathematicalConstants.h"

using namespace MathematicalCrystalChemistry::StructureModel::Components::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

ConstraintManager::ConstraintManager()
	: m_randomValueGenerating{ 0 }
	, m_engine{}
	, m_ionicBondOwners{}
	, m_metallicBondOwners{}
	, m_anionicConstraintOwners{}
	, m_cationicConstraintOwners{}
	, m_nonIonicBondOwners{}
	, m_nonMetallicBondOwners{}
{
	std::random_device seed_gen;
	m_engine = std::mt19937(seed_gen());
}

ConstraintManager::ConstraintManager(const ConstraintManager& constraintManager)
	: m_randomValueGenerating{ 0 }
	, m_engine{}
	, m_ionicBondOwners{}
	, m_metallicBondOwners{}
	, m_anionicConstraintOwners{}
	, m_cationicConstraintOwners{}
	, m_nonIonicBondOwners{}
	, m_nonMetallicBondOwners{}
{
}

ConstraintManager::ConstraintManager(ConstraintManager&& constraintManager) noexcept
	: m_randomValueGenerating{ 0 }
	, m_engine{}
	, m_ionicBondOwners{}
	, m_metallicBondOwners{}
	, m_anionicConstraintOwners{}
	, m_cationicConstraintOwners{}
	, m_nonIonicBondOwners{}
	, m_nonMetallicBondOwners{}
{
	constraintManager.clearGeometricalConstraintOwners();
}

ConstraintManager& ConstraintManager::operator=(const ConstraintManager& constraintManager)
{
	clearGeometricalConstraintOwners();
	return *this;
}

ConstraintManager& ConstraintManager::operator=(ConstraintManager&& constraintManager) noexcept
{
	clearGeometricalConstraintOwners();
	constraintManager.clearGeometricalConstraintOwners();

	return *this;
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

void ConstraintManager::createCationicConstraintOwners(std::vector<ModelAtom>& atomicArrangement)
{
	using Ligand = ChemToolkit::Crystallography::PeriodicReferenceWrapper<ModelAtom, ChemToolkit::Crystallography::UnitCell>;
	using CationicConstraint = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint;

	for (auto& modelAtom : atomicArrangement)
	{
		for (auto& ligands : modelAtom.getCationicConstraintCandidates())
		{
			Ligand commonBridgingAnion{ modelAtom, ConstraintKey::selectModelAtomKey(ligands.first, ligands.second) };
			commonBridgingAnion.reverseLatticePoint();

			ConstraintKey constraintKey = ConstraintKey::createConstraintKey(ligands.first, ligands.second);
			auto iter = m_cationicConstraintOwners.find(constraintKey);


			if (iter == m_cationicConstraintOwners.end())
			{
				CationicConstraintOwner cationicConstraintOwner = CationicConstraint::create(constraintKey);
				cationicConstraintOwner.constraintPtr()->addCommonBridgingAnions(commonBridgingAnion);
				m_cationicConstraintOwners.emplace(constraintKey, std::move(cationicConstraintOwner));

				m_nonMetallicBondOwners.erase(constraintKey);
				m_metallicBondOwners.erase(constraintKey);
			}

			else
				iter->second.constraintPtr()->addCommonBridgingAnions(commonBridgingAnion);
		}
	}
}

void ConstraintManager::substituteIonicBondOwners(std::vector<ModelAtom>& atomicArrangement, const double feasibleConstraintErrorRate)
{
	using IonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond;
	using NonIonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond;

	ModelAtom* infeasibleModelAtomPtr;
	{
		std::vector<ModelAtom*> infeasibleModelAtomPtrs = getInfeasibleAtomPtrs(atomicArrangement, feasibleConstraintErrorRate);
		{
			if (infeasibleModelAtomPtrs.empty())
				return;
			else
			{
				std::shuffle(infeasibleModelAtomPtrs.begin(), infeasibleModelAtomPtrs.end(), m_engine);
				++m_randomValueGenerating;
				{
					if (1000 < m_randomValueGenerating)
					{
						std::random_device seed_gen;
						m_engine = std::mt19937(seed_gen());
					}
				}

				infeasibleModelAtomPtr = infeasibleModelAtomPtrs.front();
			}
		}
	}

	std::vector<ModelAtom::Ligand> substituteIonicLigands;
	{
		auto substituteIonicLigandCandidates = infeasibleModelAtomPtr->getSubstituteIonicLigandCandidates();;
		{
			std::shuffle(substituteIonicLigandCandidates.begin(), substituteIonicLigandCandidates.end(), m_engine);
			++m_randomValueGenerating;
			{
				if (1000 < m_randomValueGenerating)
				{
					std::random_device seed_gen;
					m_engine = std::mt19937(seed_gen());
				}
			}
		}

		for (auto iter = substituteIonicLigandCandidates.begin(); iter != substituteIonicLigandCandidates.end(); ++iter)
		{
			if (substituteIonicLigands.size() < 1)
				substituteIonicLigands.push_back(*iter);
			else
				break;
		}
	}


	for (const auto& trashIonicBondPtr : infeasibleModelAtomPtr->getIonicBondPtrs())
	{
		ConstraintKey trashConstraintKey = ConstraintKey::createConstraintKey(trashIonicBondPtr->modelAtomRef(), trashIonicBondPtr->ligand());
		{
			m_ionicBondOwners.erase(trashConstraintKey);

			if (m_nonIonicBondOwners.find(trashConstraintKey) == m_nonIonicBondOwners.end())
				m_nonIonicBondOwners.emplace(trashConstraintKey, NonIonicBond::create(trashConstraintKey));
		}
	}

	for (const auto& substituteIonicLigandCandidate : substituteIonicLigands)
	{
		ConstraintKey newConstraintKey = ConstraintKey::createConstraintKey((*infeasibleModelAtomPtr), substituteIonicLigandCandidate);
		{
			if (m_ionicBondOwners.find(newConstraintKey) == m_ionicBondOwners.end())
			{
				m_ionicBondOwners.emplace(newConstraintKey, IonicBond::create(newConstraintKey));
				m_nonIonicBondOwners.erase(newConstraintKey);
			}
		}
	}
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

std::vector<MathematicalCrystalChemistry::StructureModel::Components::ModelAtom*> ConstraintManager::getInfeasibleAtomPtrs(std::vector<ModelAtom>& atomicArrangement, const double feasibleConstraintErrorRate) const
{
	std::vector<MathematicalCrystalChemistry::StructureModel::Components::ModelAtom*> infeasibleAtomPtrs;
	{
		for (auto& modelAtom : atomicArrangement)
		{
			if (modelAtom.coordinationFeature().atomCategory().isCation())
			{
				for (const auto& ionicBondPtr : modelAtom.getIonicBondPtrs())
				{
					double distanceSquare = (ionicBondPtr->ligand().coordinate() - ionicBondPtr->modelAtomRef().coordinate()).normSquare();
					double upperFeasibleDistance = (1.0 + feasibleConstraintErrorRate) * ionicBondPtr->maximumDistance();

					if ((upperFeasibleDistance * upperFeasibleDistance) < distanceSquare)
						infeasibleAtomPtrs.push_back(&modelAtom);
				}
			}
		}
	}

	return infeasibleAtomPtrs;
}

void ConstraintManager::createBasicMinimumConstraintOwners(const std::vector<ConstraintKey>& trackingAtomPairs)
{
	using AnionicConstraint = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint;
	using NonIonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond;
	using NonMetallicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond;

	double anionicConstraintCutoffRatio = AnionicConstraint::defaultCutoffRatio();
	double nonIonicBondCutoffRatio = NonIonicBond::defaultCutoffRatio();
	double nonMetallicBondCutoffRatio = NonMetallicBond::defaultCutoffRatio();
	double nonIonicBondRangeRatio = NonIonicBond::defaultNonBondingRangeRatio();
	double nonMetallicBondRangeRatio = NonMetallicBond::defaultNonBondingRangeRatio();


	for (const auto& trackingAtomPair : trackingAtomPairs)
	{
		if (AnionicConstraint::shouldCreate(trackingAtomPair, anionicConstraintCutoffRatio))
		{
			if (m_anionicConstraintOwners.find(trackingAtomPair) == m_anionicConstraintOwners.end())
				m_anionicConstraintOwners.emplace(trackingAtomPair, AnionicConstraint::create(trackingAtomPair));
		}

		else if (NonIonicBond::shouldCreate(trackingAtomPair, nonIonicBondCutoffRatio, nonIonicBondRangeRatio))
		{
			if (m_nonIonicBondOwners.find(trackingAtomPair) == m_nonIonicBondOwners.end())
				m_nonIonicBondOwners.emplace(trackingAtomPair, NonIonicBond::create(trackingAtomPair));
		}

		else if (NonMetallicBond::shouldCreate(trackingAtomPair, nonMetallicBondCutoffRatio, nonMetallicBondRangeRatio))
		{
			if (m_nonMetallicBondOwners.find(trackingAtomPair) == m_nonMetallicBondOwners.end())
				m_nonMetallicBondOwners.emplace(trackingAtomPair, NonMetallicBond::create(trackingAtomPair));
		}

		else
			continue;
	}
}

void ConstraintManager::createBasicMinimumConstraintOwners(const std::vector<ConstraintKey>& trackingAtomPairs, const double minimumConstraintCutoffRatio, const double nonBondingRangeRatio)
{
	using AnionicConstraint = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint;
	using NonIonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond;
	using NonMetallicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond;
	
	double anionicConstraintCutoffRatio = minimumConstraintCutoffRatio;
	double nonIonicBondCutoffRatio = minimumConstraintCutoffRatio;
	double nonMetallicBondCutoffRatio = minimumConstraintCutoffRatio;
	double nonIonicBondRangeRatio = nonBondingRangeRatio;
	double nonMetallicBondRangeRatio = nonBondingRangeRatio;


	for (const auto& trackingAtomPair : trackingAtomPairs)
	{
		if (AnionicConstraint::shouldCreate(trackingAtomPair, anionicConstraintCutoffRatio))
		{
			if (m_anionicConstraintOwners.find(trackingAtomPair) == m_anionicConstraintOwners.end())
				m_anionicConstraintOwners.emplace(trackingAtomPair, AnionicConstraint::create(trackingAtomPair));
		}

		else if (NonIonicBond::shouldCreate(trackingAtomPair, nonIonicBondCutoffRatio, nonIonicBondRangeRatio))
		{
			if (m_nonIonicBondOwners.find(trackingAtomPair) == m_nonIonicBondOwners.end())
				m_nonIonicBondOwners.emplace(trackingAtomPair, NonIonicBond::create(trackingAtomPair));
		}

		else if (NonMetallicBond::shouldCreate(trackingAtomPair, nonMetallicBondCutoffRatio, nonMetallicBondRangeRatio))
		{
			if (m_nonMetallicBondOwners.find(trackingAtomPair) == m_nonMetallicBondOwners.end())
				m_nonMetallicBondOwners.emplace(trackingAtomPair, NonMetallicBond::create(trackingAtomPair));
		}

		else
			continue;
	}
}

void ConstraintManager::createIonicBondOwners(std::vector<ModelAtom>& atomicArrangement)
{
	using IonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond;

	for (auto& modelAtom : atomicArrangement)
	{
		for (const auto& ligand : modelAtom.getIonicLigandCandidates())
		{
			ConstraintKey constraintKey = ConstraintKey::createConstraintKey(modelAtom, ligand);

			if (m_ionicBondOwners.find(constraintKey) == m_ionicBondOwners.end())
			{
				m_ionicBondOwners.emplace(constraintKey, IonicBond::create(constraintKey));
				m_nonIonicBondOwners.erase(constraintKey);
			}
		}
	}
}

void ConstraintManager::createIonicBondOwners(std::vector<ModelAtom>& atomicArrangement, const double ionicBondCutoffRatio)
{
	using IonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond;

	for (auto& modelAtom : atomicArrangement)
	{
		for (const auto& ligand : modelAtom.getIonicLigandCandidates(ionicBondCutoffRatio))
		{
			ConstraintKey constraintKey = ConstraintKey::createConstraintKey(modelAtom, ligand);

			if (m_ionicBondOwners.find(constraintKey) == m_ionicBondOwners.end())
			{
				m_ionicBondOwners.emplace(constraintKey, IonicBond::create(constraintKey));
				m_nonIonicBondOwners.erase(constraintKey);
			}
		}
	}
}

void ConstraintManager::createMetallicBondOwners(std::vector<ModelAtom>& atomicArrangement)
{
	using MetallicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond;

	for (auto& modelAtom : atomicArrangement)
	{
		for (const auto& ligand : modelAtom.getMetallicLigandCandidates())
		{
			ConstraintKey constraintKey = ConstraintKey::createConstraintKey(modelAtom, ligand);

			if (m_metallicBondOwners.find(constraintKey) == m_metallicBondOwners.end())
			{
				m_metallicBondOwners.emplace(constraintKey, MetallicBond::create(constraintKey));
				m_nonMetallicBondOwners.erase(constraintKey);
			}
		}
	}
}

void ConstraintManager::createMetallicBondOwners(std::vector<ModelAtom>& atomicArrangement, const double metallicBondCutoffRatio)
{
	using MetallicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond;

	for (auto& modelAtom : atomicArrangement)
	{
		for (const auto& ligand : modelAtom.getMetallicLigandCandidates(metallicBondCutoffRatio))
		{
			ConstraintKey constraintKey = ConstraintKey::createConstraintKey(modelAtom, ligand);

			if (m_metallicBondOwners.find(constraintKey) == m_metallicBondOwners.end())
			{
				m_metallicBondOwners.emplace(constraintKey, MetallicBond::create(constraintKey));
				m_nonMetallicBondOwners.erase(constraintKey);
			}
		}
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
