#include "CoordinationFeature.h"

#include <locale>

#include "LengthCasting.h"
#include "PeriodicTable.h"

using namespace MathematicalCrystalChemistry::StructureModel::Features;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

CoordinationFeature::CoordinationFeature()
	: m_coordinationFeatureSymbol{}
	, m_atomCategory{}
	, m_upperIonicBondRadius{ 0.0 }
	, m_lowerIonicBondRadius{ 0.0 }
	, m_repulsiveIonicRadius{ 0.0 }
	, m_upperMetallicBondRadius{ 0.0 }
	, m_lowerMetallicBondRadius{ 0.0 }
	, m_feasibleIonicBondComposition{}
	, m_feasibleMetallicBondComposition{}
{
}

CoordinationFeature::CoordinationFeature(const CoordinationFeatureSymbol& coordinationFeatureSymbol)
	: m_coordinationFeatureSymbol{ coordinationFeatureSymbol }
	, m_atomCategory{ coordinationFeatureSymbol.atomCategorySymbol() }
	, m_upperIonicBondRadius{ getDefinedRadius(coordinationFeatureSymbol.upperIonicBondRadiusSymbol()) }
	, m_lowerIonicBondRadius{ getDefinedRadius(coordinationFeatureSymbol.lowerIonicBondRadiusSymbol()) }
	, m_repulsiveIonicRadius{ getDefinedRadius(coordinationFeatureSymbol.repulsiveIonicRadiusSymbol()) }
	, m_upperMetallicBondRadius{ getDefinedRadius(coordinationFeatureSymbol.upperMetallicBondRadiusSymbol()) }
	, m_lowerMetallicBondRadius{ getDefinedRadius(coordinationFeatureSymbol.lowerMetallicBondRadiusSymbol()) }
	, m_feasibleIonicBondComposition{ coordinationFeatureSymbol.ionicBondCompositionSymbol() }
	, m_feasibleMetallicBondComposition{ coordinationFeatureSymbol.metallicBondCompositionSymbol() }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

void CoordinationFeature::setCoordinationFeatureSymbol(const CoordinationFeatureSymbol& coordinationFeatureSymbol)
{
	m_coordinationFeatureSymbol = coordinationFeatureSymbol;
	m_atomCategory = AtomCategory{ coordinationFeatureSymbol.atomCategorySymbol() };
	m_upperIonicBondRadius = getDefinedRadius(coordinationFeatureSymbol.upperIonicBondRadiusSymbol());
	m_lowerIonicBondRadius = getDefinedRadius(coordinationFeatureSymbol.lowerIonicBondRadiusSymbol());
	m_repulsiveIonicRadius = getDefinedRadius(coordinationFeatureSymbol.repulsiveIonicRadiusSymbol());
	m_upperMetallicBondRadius = getDefinedRadius(coordinationFeatureSymbol.upperMetallicBondRadiusSymbol());
	m_lowerMetallicBondRadius = getDefinedRadius(coordinationFeatureSymbol.lowerMetallicBondRadiusSymbol());
	m_feasibleIonicBondComposition = ConstraintComposition{ coordinationFeatureSymbol.ionicBondCompositionSymbol() };
	m_feasibleMetallicBondComposition = ConstraintComposition{ coordinationFeatureSymbol.metallicBondCompositionSymbol() };
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void CoordinationFeature::initialize() noexcept
{
	m_coordinationFeatureSymbol.clear();

	m_atomCategory = AtomCategory{};	
	m_upperIonicBondRadius = 0.0;
	m_lowerIonicBondRadius = 0.0;
	m_repulsiveIonicRadius = 0.0;
	m_upperMetallicBondRadius = 0.0;
	m_lowerMetallicBondRadius = 0.0;

	m_feasibleIonicBondComposition.initialize();
	m_feasibleMetallicBondComposition.initialize();
}

void CoordinationFeature::initializeFromElementSymbol(const std::string& elementSymbol, const int oxidationState)
{
	m_coordinationFeatureSymbol.initializeFromElementSymbol(elementSymbol);
	m_coordinationFeatureSymbol.setIonicBondCompositionSymbol(std::string{ "*" });
	m_coordinationFeatureSymbol.setMetallicBondCompositionSymbol(std::string{ "*" });

	m_atomCategory = AtomCategory::toAtomCategory(elementSymbol);
	m_upperIonicBondRadius = ChemToolkit::Generic::PeriodicTable::getIonicRadius(elementSymbol, oxidationState);
	m_lowerIonicBondRadius = ChemToolkit::Generic::PeriodicTable::getIonicRadius(elementSymbol, oxidationState);
	m_repulsiveIonicRadius = ChemToolkit::Generic::PeriodicTable::getIonicRadius(elementSymbol, oxidationState);
	m_upperMetallicBondRadius = ChemToolkit::Generic::PeriodicTable::getMetallicRadius(elementSymbol);
	m_lowerMetallicBondRadius = ChemToolkit::Generic::PeriodicTable::getMetallicRadius(elementSymbol);

	m_feasibleIonicBondComposition.disableConstraints();
	m_feasibleMetallicBondComposition.disableConstraints();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static utility

double CoordinationFeature::getDefinedRadius(const std::string& radiusSymbol)
{
	using MathToolkit::UnitConversion::LengthCasting::Unit::AtomicUnit;
	using Picometre = MathToolkit::UnitConversion::LengthCasting::Unit::Metre<std::pico>;


	if (radiusSymbol == "Z" || radiusSymbol == "z")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(0.0);


	if (radiusSymbol == "O" || radiusSymbol == "o")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(140.0);


	if (radiusSymbol == "T" || radiusSymbol == "t")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(21.65807537);
	if (radiusSymbol == "Q" || radiusSymbol == "q")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(31.46428199);
	if (radiusSymbol == "P" || radiusSymbol == "p")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(57.98989873);
	if (radiusSymbol == "H" || radiusSymbol == "h")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(57.98989873);
	if (radiusSymbol == "S" || radiusSymbol == "s")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(82.77554212);
	if (radiusSymbol == "E" || radiusSymbol == "e")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(90.34602864);
	if (radiusSymbol == "N" || radiusSymbol == "n")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(102.48711306);
	if (radiusSymbol == "D" || radiusSymbol == "d")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(116.5450385);
	if (radiusSymbol == "UD" || radiusSymbol == "ud")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(126.29582456);
	if (radiusSymbol == "WD" || radiusSymbol == "wd")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(126.29582456);
	if (radiusSymbol == "TD" || radiusSymbol == "td")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(152.7603622844);
	if (radiusSymbol == "QD" || radiusSymbol == "qd")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(159.8299674928);
	if (radiusSymbol == "PD" || radiusSymbol == "pd")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(170.1956244814);
	if (radiusSymbol == "HD" || radiusSymbol == "hd")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(177.9743716381);
	if (radiusSymbol == "SD" || radiusSymbol == "sd")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(184.6584294428);
	if (radiusSymbol == "ED" || radiusSymbol == "ed")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(194.0422362585);
	if (radiusSymbol == "ND" || radiusSymbol == "nd")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(206.2954547144);
	if (radiusSymbol == "V" || radiusSymbol == "v")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(208.0891974575);
	if (radiusSymbol == "UV" || radiusSymbol == "uv")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(221.1766467615);
	if (radiusSymbol == "WV" || radiusSymbol == "wv")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(227.8523011150);
	if (radiusSymbol == "TV" || radiusSymbol == "tv")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(236.0845691980);
	if (radiusSymbol == "QV" || radiusSymbol == "qv")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(236.2397446484);
	if (radiusSymbol == "PV" || radiusSymbol == "pv")
		return MathToolkit::UnitConversion::LengthCasting::cast<Picometre, AtomicUnit>(253.9355561376);


	OperatingSystem::Configuration::ReportException reportException;
	throw std::invalid_argument(reportException("getRadius of MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature", "\"radiusSymbol\" is out of definitions."));
}

std::string CoordinationFeature::getUppercaseRadiusSymbol(const size_type radiusNumber)
{
	switch (radiusNumber)
	{
	case 0:
		return std::string{ "Z" };
	case 1:
		return std::string{ "T" };
	case 2:
		return std::string{ "Q" };
	case 3:
		return std::string{ "H" };
	case 4:
		return std::string{ "S" };
	case 5:
		return std::string{ "E" };
	case 6:
		return std::string{ "N" };
	case 7:
		return std::string{ "D" };
	case 8:
		return std::string{ "WD" };
	case 9:
		return std::string{ "O" };
	case 10:
		return std::string{ "TD" };
	case 11:
		return std::string{ "QD" };
	case 12:
		return std::string{ "PD" };
	case 13:
		return std::string{ "HD" };
	case 14:
		return std::string{ "SD" };
	case 15:
		return std::string{ "ED" };
	case 16:
		return std::string{ "ND" };
	case 17:
		return std::string{ "V" };
	case 18:
		return std::string{ "UV" };
	case 19:
		return std::string{ "WV" };
	case 20:
		return std::string{ "TV" };
	case 21:
		return std::string{ "QV" };
	case 22:
		return std::string{ "PV" };


	default:
		return std::string{};
	}
}

std::string CoordinationFeature::getLowercaseRadiusSymbol(const size_type radiusNumber)
{
	std::string radiusSymbol;
	{
		for (const auto& charText : getUppercaseRadiusSymbol(radiusNumber))
			radiusSymbol.push_back(std::tolower(charText));
	}

	return radiusSymbol;
}

CoordinationFeature::size_type CoordinationFeature::getFloorRadiusNumber(const double radius)
{
	if (0.0 < radius)
	{
		size_type radiusNumber = 1;

		while (true)
		{
			std::string radiusSymbol = getUppercaseRadiusSymbol(radiusNumber);

			if (radiusSymbol.empty())
			{
				--radiusNumber;
				break;
			}

			else
			{
				double ceilRadiusCandidate = getDefinedRadius(radiusSymbol);

				if (radius < ceilRadiusCandidate)
				{
					--radiusNumber;
					break;
				}

				else
					++radiusNumber;
			}
		}

		return radiusNumber;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException("getFloorRadiusNumber of MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature", "\"radius\" is less than zero."));
	}
}

std::size_t CoordinationFeature::getMinimumFeasibleRadiusNumber(const size_type coordinationNumber)
{
	switch (coordinationNumber)
	{
	case 0:
		return 0;
	case 3:
		return 1;
	case 4:
		return 2;
	case 5:
		return 3;
	case 6:
		return 3;
	case 7:
		return 4;
	case 8:
		return 5;
	case 9:
		return 6;
	case 10:
		return 7;
	case 11:
		return 8;
	case 12:
		return 8;
	case 13:
		return 10;
	case 14:
		return 11;
	case 15:
		return 12;
	case 16:
		return 13;
	case 17:
		return 14;
	case 18:
		return 15;
	case 19:
		return 16;
	case 20:
		return 17;
	case 21:
		return 18;
	case 22:
		return 19;
	case 23:
		return 20;
	case 24:
		return 21;
	case 25:
		return 22;

	default:
		return 0;
	}
}

// Static utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
