#include "CoordinationFeatureAssigner.h"

#include <set>

#include "FileWriter.h"

using namespace MathematicalCrystalChemistry::Runtime::Analysis::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void CoordinationFeatureAssigner::setDefaultMinimumIonicBondRadiusNumbersRange(const ModelCrystalStructure& modelCrystalStructure)
{
	m_lowerIonicBondRadiusNumbersRange.clear();
	std::set<size_type> surveyedSiteIndices;


	for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
	{
		if (surveyedSiteIndices.find(modelAtom.siteIndex()) == surveyedSiteIndices.end())
		{
			surveyedSiteIndices.emplace(modelAtom.siteIndex());

			double defaultIonicBondRadius = (modelAtom.coordinationFeature().upperIonicBondRadius() + 0.0001);
			size_type defaultIonicRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getFloorRadiusNumber(defaultIonicBondRadius);


			if (modelAtom.coordinationFeature().atomCategory().isAnion())
			{
				std::pair<size_type, size_type> lowerAndUpper{ defaultIonicRadiusNumber, defaultIonicRadiusNumber };
				m_lowerIonicBondRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::move(lowerAndUpper));
			}

			else
			{
				const size_type coordinationNumber = modelAtom.getCoordinationNumber();
				const size_type minimumFeasibleRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMinimumFeasibleRadiusNumber(coordinationNumber);
				{
					if (defaultIonicRadiusNumber < minimumFeasibleRadiusNumber)
						defaultIonicRadiusNumber = minimumFeasibleRadiusNumber;

					if (MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber() < defaultIonicRadiusNumber)
						defaultIonicRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber();
				}

				m_lowerIonicBondRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::pair<size_type, size_type>{ defaultIonicRadiusNumber, defaultIonicRadiusNumber });
			}
		}
	}
}

void CoordinationFeatureAssigner::setDefaultMaximumIonicBondRadiusNumbersRange(const ModelCrystalStructure& modelCrystalStructure)
{
	m_upperIonicBondRadiusNumbersRange.clear();
	std::set<size_type> surveyedSiteIndices;


	for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
	{
		if (surveyedSiteIndices.find(modelAtom.siteIndex()) == surveyedSiteIndices.end())
		{
			surveyedSiteIndices.emplace(modelAtom.siteIndex());

			double defaultIonicBondRadius = (modelAtom.coordinationFeature().upperIonicBondRadius() + 0.0001);
			size_type defaultIonicRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getFloorRadiusNumber(defaultIonicBondRadius);


			if (modelAtom.coordinationFeature().atomCategory().isAnion())
				m_upperIonicBondRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::pair<size_type, size_type>{ defaultIonicRadiusNumber, defaultIonicRadiusNumber });

			else
			{
				const size_type coordinationNumber = modelAtom.getCoordinationNumber();
				const size_type minimumFeasibleRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMinimumFeasibleRadiusNumber(coordinationNumber);
				{
					if (defaultIonicRadiusNumber < minimumFeasibleRadiusNumber)
						defaultIonicRadiusNumber = minimumFeasibleRadiusNumber;

					if (MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber() < defaultIonicRadiusNumber)
						defaultIonicRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber();
				}

				m_upperIonicBondRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::pair<size_type, size_type>{ defaultIonicRadiusNumber, defaultIonicRadiusNumber });
			}
		}
	}
}

void CoordinationFeatureAssigner::setIonicBondRadiusNumbersRange(const ModelCrystalStructure& modelCrystalStructure, const size_type upperIonicBondRadiusNumberExtensionWidth)
{
	m_upperIonicBondRadiusNumbersRange.clear();
	m_lowerIonicBondRadiusNumbersRange.clear();
	std::set<size_type> surveyedSiteIndices;


	for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
	{
		if (surveyedSiteIndices.find(modelAtom.siteIndex()) == surveyedSiteIndices.end())
		{
			surveyedSiteIndices.emplace(modelAtom.siteIndex());

			double defaultIonicBondRadius = (modelAtom.coordinationFeature().upperIonicBondRadius() + 0.0001);
			size_type defaultIonicRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getFloorRadiusNumber(defaultIonicBondRadius);


			if (modelAtom.coordinationFeature().atomCategory().isAnion())
			{
				std::pair<size_type, size_type> lowerAndUpper{ defaultIonicRadiusNumber, defaultIonicRadiusNumber };
				m_upperIonicBondRadiusNumbersRange.emplace(modelAtom.siteIndex(), lowerAndUpper);
				m_lowerIonicBondRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::move(lowerAndUpper));
			}

			else
			{
				const size_type coordinationNumber = modelAtom.getCoordinationNumber();
				const size_type minimumFeasibleRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMinimumFeasibleRadiusNumber(coordinationNumber);
				{
					if (defaultIonicRadiusNumber < minimumFeasibleRadiusNumber)
						defaultIonicRadiusNumber = minimumFeasibleRadiusNumber;

					if (MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber() < defaultIonicRadiusNumber)
						defaultIonicRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber();
				}

				size_type maxIonicRadiusNumber = 0;
				size_type minIonicRadiusNumber = 0;
				{
					double maxIonicRadius = 0.0;
					double minIonicRadius = std::numeric_limits<double>::max();
					{
						if (modelAtom.getIonicBondPtrs().empty())
						{
							maxIonicRadius = defaultIonicBondRadius;
							minIonicRadius = defaultIonicBondRadius;
						}

						else
						{
							for (const auto& ionicBondPtr : modelAtom.getIonicBondPtrs())
							{
								double largeIonicRadius = (ionicBondPtr->ligand().coordinate() - modelAtom.coordinate()).norm();
								double smallIonicRadius = largeIonicRadius;
								largeIonicRadius -= ionicBondPtr->ligand().get().coordinationFeature().lowerIonicBondRadius();
								smallIonicRadius -= ionicBondPtr->ligand().get().coordinationFeature().upperIonicBondRadius();

								if (maxIonicRadius < largeIonicRadius)
									maxIonicRadius = largeIonicRadius;

								if (smallIonicRadius < minIonicRadius)
									minIonicRadius = smallIonicRadius;
							}
						}
					}


					maxIonicRadiusNumber = 1 + upperIonicBondRadiusNumberExtensionWidth + MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getFloorRadiusNumber(maxIonicRadius);
					minIonicRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getFloorRadiusNumber(minIonicRadius);

					if (maxIonicRadiusNumber < defaultIonicRadiusNumber)
						maxIonicRadiusNumber = defaultIonicRadiusNumber;

					if (MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber() < defaultIonicRadiusNumber)
						maxIonicRadiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getMaxRadiusNumber();

					if (defaultIonicRadiusNumber < minIonicRadiusNumber)
						minIonicRadiusNumber = defaultIonicRadiusNumber;

					if (minIonicRadiusNumber == 0)
						minIonicRadiusNumber = 1;
				}


				m_upperIonicBondRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::pair<size_type, size_type>{ defaultIonicRadiusNumber, maxIonicRadiusNumber });
				m_lowerIonicBondRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::pair<size_type, size_type>{ minIonicRadiusNumber, defaultIonicRadiusNumber });
			}
		}
	}
}

void CoordinationFeatureAssigner::setRepulsiveIonicRadiusNumbersRange(const ModelCrystalStructure& modelCrystalStructure, const size_type repulsiveIonicRadiusNumberExtensionWidth)
{
	m_repulsiveIonicRadiusNumbersRange.clear();


	for (const auto& modelAtom : modelCrystalStructure.atomicArrangement())
	{
		if (m_repulsiveIonicRadiusNumbersRange.find(modelAtom.siteIndex()) == m_repulsiveIonicRadiusNumbersRange.end())
		{
			if (modelAtom.coordinationFeature().atomCategory().isAnion())
			{
				double ionicBondRadius = (0.0001 + modelAtom.coordinationFeature().upperIonicBondRadius());
				size_type radiusNumber = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature::getFloorRadiusNumber(ionicBondRadius);

				std::pair<size_type, size_type> lowerAndUpper{ radiusNumber, radiusNumber };
				m_repulsiveIonicRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::move(lowerAndUpper));
			}

			else
			{
				double minUpperRepulsiveIonicRadius = std::numeric_limits<double>::max();
				{
					if (modelAtom.getCationicConstraintPtrs().empty())
						minUpperRepulsiveIonicRadius = (modelAtom.coordinationFeature().upperIonicBondRadius() + 0.0001);

					else
					{
						for (const auto& cationicConstraintPtr : modelAtom.getCationicConstraintPtrs())
						{
							double repulsiveIonicRadius = (cationicConstraintPtr->ligand().coordinate() - modelAtom.coordinate()).norm();
							repulsiveIonicRadius *= 0.5;

							if (repulsiveIonicRadius < minUpperRepulsiveIonicRadius)
								minUpperRepulsiveIonicRadius = repulsiveIonicRadius;
						}
					}
				}

				std::pair<size_type, size_type> lowerAndUpper{ 0, 0 };
				{
					using MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeature;

					size_type floorRadiusNumber = CoordinationFeature::getFloorRadiusNumber(minUpperRepulsiveIonicRadius);
					lowerAndUpper.second = (floorRadiusNumber + repulsiveIonicRadiusNumberExtensionWidth);
					{
						if (CoordinationFeature::getMaxRadiusNumber() < lowerAndUpper.second)
							lowerAndUpper.second = CoordinationFeature::getMaxRadiusNumber();
					}

					
					lowerAndUpper.first = floorRadiusNumber;

					if (CoordinationFeature::getMaxRadiusNumber() < lowerAndUpper.first)
						lowerAndUpper.first = CoordinationFeature::getMaxRadiusNumber();
				}

				m_repulsiveIonicRadiusNumbersRange.emplace(modelAtom.siteIndex(), std::move(lowerAndUpper));
			}
		}
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
