#include "CoordinationFeatureSpecifier.h"

#include <thread>
#include <mpi.h>

#include "ThreadController.h"
#include "MpiController.h"
#include "FileWriter.h"

#include "InternationalTables.h"

#include "IonicBond.h"
#include "MetallicBond.h"
#include "AnionicConstraint.h"
#include "CationicConstraint.h"
#include "NonIonicBond.h"
#include "NonMetallicBond.h"
#include "StructureCustomizerBase.h"

#include "SpecifyCoordinationFeatures.h"
#include "StructureReoptimizer.h"

using namespace MathematicalCrystalChemistry::Runtime::Analysis;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

CoordinationFeatureSpecifier::CoordinationFeatureSpecifier(const CoordinationFeatureSpecifyOptions& coordinationFeatureSpecifyOptions)
	: m_coordinationFeatureSpecifyOptions{ coordinationFeatureSpecifyOptions }
{
}

CoordinationFeatureSpecifier::CoordinationFeatureSpecifier(CoordinationFeatureSpecifyOptions&& coordinationFeatureSpecifyOptions)
	: m_coordinationFeatureSpecifyOptions{ std::move(coordinationFeatureSpecifyOptions) }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void CoordinationFeatureSpecifier::specify() const
{
	initializeDirectories();
	initializeChemicalParameters();
	initializeGeometricalConstraints();

	Internal::SpecifyCoordinationFeatures::initializeNumExcavatedCifFile();


	std::vector<Internal::SpecifyCoordinationFeatures> specifyCoordinationFeaturesCollection;
	{
		for (std::size_t threadRank = 0; threadRank < OperatingSystem::Configuration::ThreadController::maxThreading(); ++threadRank)
			specifyCoordinationFeaturesCollection.push_back(Internal::SpecifyCoordinationFeatures{ threadRank, m_coordinationFeatureSpecifyOptions });
	}

	std::vector<std::thread> operatingSystems;
	{
		for (std::size_t threadRank = 0; threadRank < OperatingSystem::Configuration::ThreadController::maxThreading(); ++threadRank)
			operatingSystems.push_back(std::thread{ specifyCoordinationFeaturesCollection[threadRank] });
	}

	for (auto& operatingSystem : operatingSystems)
		operatingSystem.join();


	MPI_Barrier(MPI_COMM_WORLD);

	if (OperatingSystem::Configuration::MpiController::mpiRank() == 0)
		createResultFile();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void CoordinationFeatureSpecifier::initializeDirectories() const
{
	std::filesystem::path workDirectoryPath = m_coordinationFeatureSpecifyOptions.workDirectoryPath();
	workDirectoryPath /= OperatingSystem::Configuration::MpiController::getMpiDirectoryName();
	{
		if (std::filesystem::is_directory(workDirectoryPath))
			std::filesystem::remove_all(workDirectoryPath);

		if (!(std::filesystem::create_directories(workDirectoryPath)))
		{
			std::string errorMessage;
			{
				errorMessage += "Could not create directories of the path \"";
				errorMessage += workDirectoryPath.generic_string();
				errorMessage += "\"";
			}

			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "initializeDirectories", errorMessage));
		}
	}

	std::filesystem::path outputDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
	{
		if (!(std::filesystem::is_directory(outputDirectoryPath)))
		{
			if (!(std::filesystem::create_directories(outputDirectoryPath)))
			{
				std::string errorMessage;
				{
					errorMessage += "Could not create directories of the path \"";
					errorMessage += outputDirectoryPath.generic_string();
					errorMessage += "\"";
				}

				OperatingSystem::Configuration::ReportException reportException;
				throw std::invalid_argument(reportException(typeid(*this), "initializeDirectories", errorMessage));
			}
		}
	}
}

void CoordinationFeatureSpecifier::initializeChemicalParameters() const
{
	ChemToolkit::Crystallography::SpaceGroups::InternationalTables::setSpaceGroupPrecision(m_coordinationFeatureSpecifyOptions.chemicalParameters().spaceGroupPrecision());
}

void CoordinationFeatureSpecifier::initializeGeometricalConstraints() const
{
	using MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

	using MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond;


	ModelCrystalStructure::setCommonAtomPairTrackingCutoffRatio(m_coordinationFeatureSpecifyOptions.structuralOptimizationParameters().atomPairTrackingCutoffRatio());
	StructureCustomizerBase::setCommonPressure(m_coordinationFeatureSpecifyOptions.structuralOptimizationParameters().pressure());
	StructureCustomizerBase::setCommonAttractiveForceConstant(m_coordinationFeatureSpecifyOptions.structuralOptimizationParameters().attractiveForceConstant());
	StructureCustomizerBase::setCommonRepulsiveForceConstant(m_coordinationFeatureSpecifyOptions.structuralOptimizationParameters().repulsiveForceConstant());

	IonicBond::setDefaultCutoffRatio(m_coordinationFeatureSpecifyOptions.geometricalConstraintParameters().ionicBondCutoffLengthRatio());
	MetallicBond::setDefaultCutoffRatio(m_coordinationFeatureSpecifyOptions.geometricalConstraintParameters().metallicBondCutoffLengthRatio());
	AnionicConstraint::setDefaultCutoffRatio(m_coordinationFeatureSpecifyOptions.geometricalConstraintParameters().anionicConstraintCutoffLengthRatio());
	NonIonicBond::setDefaultCutoffRatio(m_coordinationFeatureSpecifyOptions.geometricalConstraintParameters().nonIonicBondCutoffLengthRatio());
	NonMetallicBond::setDefaultCutoffRatio(m_coordinationFeatureSpecifyOptions.geometricalConstraintParameters().nonMetallicBondCutoffLengthRatio());

	NonIonicBond::setDefaultNonBondingRangeRatio(m_coordinationFeatureSpecifyOptions.geometricalConstraintParameters().nonIonicBondRangeRatio());
	NonMetallicBond::setDefaultNonBondingRangeRatio(m_coordinationFeatureSpecifyOptions.geometricalConstraintParameters().nonMetallicBondRangeRatio());
}

void CoordinationFeatureSpecifier::createResultFile() const
{
	size_type numIsotypicStructures = 0;
	{
		std::filesystem::path isotypicDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
		isotypicDirectoryPath /= "Isotypic";
		numIsotypicStructures = countInputCrystalStructureFingerprintFiles(isotypicDirectoryPath);
	}
	size_type numHomeotypicStructures = 0;
	{
		std::filesystem::path homeotypicDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
		homeotypicDirectoryPath /= "Homeotypic";
		numHomeotypicStructures = countInputCrystalStructureFingerprintFiles(homeotypicDirectoryPath);
	}
	size_type numSameSymmetryStructures = 0;
	{
		std::filesystem::path sameSymmetryDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
		sameSymmetryDirectoryPath /= "Same_symmetry";
		numSameSymmetryStructures = countInputCrystalStructureFingerprintFiles(sameSymmetryDirectoryPath);
	}
	size_type numSameSpaceGroupTypeStructures = 0;
	{
		std::filesystem::path sameSpaceGroupTypeDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
		sameSpaceGroupTypeDirectoryPath /= "Same_space_group_type";
		numSameSpaceGroupTypeStructures = countInputCrystalStructureFingerprintFiles(sameSpaceGroupTypeDirectoryPath);
	}
	size_type numOptimalStructures = 0;
	{
		std::filesystem::path optimalDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
		optimalDirectoryPath /= "Optimal";
		numOptimalStructures = countInputCrystalStructureFingerprintFiles(optimalDirectoryPath);
	}
	size_type numInfeasibleStructures = 0;
	{
		std::filesystem::path infeasibleDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
		infeasibleDirectoryPath /= "Infeasible";
		numInfeasibleStructures = countInputCrystalStructureFingerprintFiles(infeasibleDirectoryPath);
	}
	size_type numExceptionStructures = 0;
	{
		std::filesystem::path exceptionDirectoryPath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
		exceptionDirectoryPath /= "Exception";
		numInfeasibleStructures = countInputCrystalStructureFingerprintFiles(exceptionDirectoryPath);
	}

	std::string outputText;
	{
		outputText += "Number.of.Isotypic.Structure                ";
		outputText += std::to_string(numIsotypicStructures);
		outputText += "\n";

		outputText += "Number.of.Homeotypic.Structure              ";
		outputText += std::to_string(numHomeotypicStructures);
		outputText += "\n";

		outputText += "Number.of.Same.Symmetry.Structure           ";
		outputText += std::to_string(numSameSymmetryStructures);
		outputText += "\n";

		outputText += "Number.of.Same.Space.Group.Type.Structure   ";
		outputText += std::to_string(numSameSpaceGroupTypeStructures);
		outputText += "\n";

		outputText += "Number.of.Optimal.Structure                 ";
		outputText += std::to_string(numOptimalStructures);
		outputText += "\n";

		outputText += "Number.of.Infeasible.Structure              ";
		outputText += std::to_string(numInfeasibleStructures);
		outputText += "\n";

		outputText += "Number.of.Exception.Structure               ";
		outputText += std::to_string(numExceptionStructures);
		outputText += "\n";
	}

	std::filesystem::path filePath = m_coordinationFeatureSpecifyOptions.outputDirectoryPath();
	filePath /= "analysis_result.txt";
	OperatingSystem::IO::FileWriter fileWriter{ filePath };
	fileWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing, outputText);
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
