#include "CoordinationPolyhedraConstraints.h"

#include <limits>
#include <regex>
#include <sstream>

#include "LengthCasting.h"
#include "FileReader.h"
#include "PeriodicTable.h"

#include "SymmetryOperationTable.h"
#include "CoordinationCenterBase.h"

using namespace MathematicalCrystalChemistry::Analysis;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

CoordinationPolyhedraConstraints::CoordinationPolyhedraConstraints()
	: m_centralElementSymbol{}
	, m_centralCoordinationFeatureSymbol{}
	, m_feasibleCentralAtomCategories{}
	, m_coordinationNumber{ std::numeric_limits<size_type>::max() }
	, m_ionicCoordinationNumber{ std::numeric_limits<size_type>::max() }
	, m_metallicCoordinationNumber{ std::numeric_limits<size_type>::max() }
	, m_chemicalBondComposition{}
	, m_ionicBondComposition{}
	, m_metallicBondComposition{}
	, m_pointGroupPrecision{ MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenterBase::defaultPointGroupPrecision() }
	, m_symmetryConstraints{}
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void CoordinationPolyhedraConstraints::initialize()
{
	m_centralElementSymbol.clear();
	m_centralCoordinationFeatureSymbol.clear();
	m_feasibleCentralAtomCategories.clear();

	m_coordinationNumber = std::numeric_limits<size_type>::max();
	m_ionicCoordinationNumber = std::numeric_limits<size_type>::max();
	m_metallicCoordinationNumber = std::numeric_limits<size_type>::max();
	m_chemicalBondComposition.initialize();
	m_ionicBondComposition.initialize();
	m_metallicBondComposition.initialize();

	m_pointGroupPrecision = MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenterBase::defaultPointGroupPrecision();
	m_symmetryConstraints.clear();
}

void CoordinationPolyhedraConstraints::initializeFromInputFileTexts(const std::string& inputTexts)
{
	initialize();
	using FileReader = OperatingSystem::IO::FileReader;

	FileReader::tryReadData(inputTexts, "Seeking.Central.Element.Symbol", m_centralElementSymbol);
	{
		std::string centralCoordinantionFeatureSymbolString;

		if (FileReader::tryReadData(inputTexts, "Seeking.Central.Coordination.Feature.Symbol", centralCoordinantionFeatureSymbolString))
			m_centralCoordinationFeatureSymbol = MathematicalCrystalChemistry::StructureModel::Features::CoordinationFeatureSymbol{ centralCoordinantionFeatureSymbolString };
	}
	readCentralAtomCategories(inputTexts);

	FileReader::tryReadData(inputTexts, "Seeking.Coordination.Number", m_coordinationNumber);
	FileReader::tryReadData(inputTexts, "Seeking.Ionic.Coordination.Number", m_ionicCoordinationNumber);
	FileReader::tryReadData(inputTexts, "Seeking.Metallic.Coordination.Number", m_metallicCoordinationNumber);

	readConstraintComposition(inputTexts, "Seeking.Chemical.Bond.Composition", m_chemicalBondComposition);
	readConstraintComposition(inputTexts, "Seeking.Ionic.Bond.Composition", m_ionicBondComposition);
	readConstraintComposition(inputTexts, "Seeking.Metallic.Bond.Composition", m_metallicBondComposition);

	{
		double angstromPrecision = 0.0;
		FileReader::tryReadData(inputTexts, "Point.Group.Precision", angstromPrecision);
		m_pointGroupPrecision = MathToolkit::UnitConversion::LengthCasting::cast<MathToolkit::UnitConversion::LengthCasting::Unit::Angstrom, MathToolkit::UnitConversion::LengthCasting::Unit::AtomicUnit>(angstromPrecision);
	}
	readSymmetryConstraints(inputTexts);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void CoordinationPolyhedraConstraints::readCentralAtomCategories(const std::string& inputTexts)
{
	std::vector<std::string> categoryNames;
	OperatingSystem::IO::FileReader::tryReadData(inputTexts, "Seeking.Central.Atom.Categories", categoryNames);

	for (const auto& categoryName : categoryNames)
		m_feasibleCentralAtomCategories.push_back(AtomCategory{ categoryName });
}

void CoordinationPolyhedraConstraints::readConstraintComposition(const std::string& inputTexts, const std::string& flag, ConstraintComposition& constraintComposition)
{
	for (const auto& lineData : OperatingSystem::IO::FileReader::readDataEnumeration(inputTexts, flag))
	{
		std::regex pat("([[:alpha:]]+)([\\d]+)");

		for (const auto& symbolAndNum : lineData)
		{
			std::smatch sm;

			if (std::regex_match(symbolAndNum, sm, pat))
				constraintComposition.add(sm.str(1), static_cast<size_type>(std::stoi(sm.str(2))));
		}
	}

	if (constraintComposition.isInfeasible())
		constraintComposition.disableConstraints();
}

void CoordinationPolyhedraConstraints::readSymmetryConstraints(const std::string& inputTexts)
{
	std::vector<std::string> symmetryOperationSymbols;
	OperatingSystem::IO::FileReader::tryReadData(inputTexts, "Symmetry.Operation.Symbols", symmetryOperationSymbols);

	for (const auto& symmetryOperationSymbol : symmetryOperationSymbols)
	{
		for (const auto& symmetryOperation : ChemToolkit::Crystallography::SpaceGroups::SymmetryOperationTable::getSymmetryOperations(symmetryOperationSymbol))
			m_symmetryConstraints.push_back(symmetryOperation);
	}		
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
