#include "CreateStructures.h"

#include <regex>

#include "LogWriter.h"
#include "MpiController.h"

#include "StructureDesignRecorder.h"


using namespace MathematicalCrystalChemistry::Runtime::Creation::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

std::size_t CreateStructures::m_structureDesigning;
std::size_t CreateStructures::m_maxStructureDesigning;
std::mutex CreateStructures::m_designMutex;


CreateStructures::CreateStructures()
	: m_structureCreationReportOptions{}
	, m_initialStructureGenerator{}
	, m_structuralOptimizer{}
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void CreateStructures::createWithRecord(const ModelAtomCollection& modelAtomCollection) const
{
	OperatingSystem::IO::LogWriter logWriter;
	{
		std::filesystem::path directoryPath = m_structureCreationReportOptions.workDirectoryPath();
		directoryPath /= OperatingSystem::Configuration::MpiController::getMpiDirectoryName();

		logWriter.setDirectoryPath(directoryPath);
	}

	std::filesystem::path designedStructuresDirectoryPath = m_structureCreationReportOptions.designedStructuresDirectoryPath();
	designedStructuresDirectoryPath /= OperatingSystem::Configuration::MpiController::getMpiDirectoryName();


	while (shouldDesign())
	{
		ModelCrystalStructure modelCrystalStructure;

		try
		{
			initializeDesignDirectoryPath();

			MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder structureDesignRecorder;
			{
				structureDesignRecorder.chemicalBondRecorder().setRecordFrequency(m_structureCreationReportOptions.structureDesignRecordOptions().chemicalBondsRecordFrequency());
				structureDesignRecorder.chemicalBondRecorder().createDataFiles(m_designDirectoryPath);

				structureDesignRecorder.dynamicsRecorder().setRecordFrequency(m_structureCreationReportOptions.structureDesignRecordOptions().mdRecordFrequency());
				structureDesignRecorder.dynamicsRecorder().createFile(m_designDirectoryPath, modelAtomCollection.toChemicalComposition().name());
			}


			modelCrystalStructure = m_initialStructureGenerator.generate(modelAtomCollection);
			m_structuralOptimizer.optimize(modelCrystalStructure, structureDesignRecorder);


			modelCrystalStructure.createTrackingAtomPairs();
			modelCrystalStructure.createGeometricalConstraints();
			{
				structureDesignRecorder.chemicalBondRecorder().createGplFiles(m_designDirectoryPath);

				if (modelCrystalStructure.isOptimalSolution(m_structuralOptimizer.structuralOptimizationParameters().preciseRelaxationParameters().feasibleConstraintErrorRate()))
					structureDesignRecorder.reportOptimalStructure(modelCrystalStructure, designedStructuresDirectoryPath, m_designDirectoryPath);
				else
				{
					if (m_structureCreationReportOptions.needInfeasibleStructureData())
						structureDesignRecorder.reportInfeasibleStructure(modelCrystalStructure, designedStructuresDirectoryPath, m_designDirectoryPath);
				}
			}


			std::filesystem::remove_all(m_designDirectoryPath);
			m_designDirectoryPath.clear();
		}

		catch (const std::exception& e)
		{
			if (m_structureCreationReportOptions.needInfeasibleStructureData())
			{
				MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder structureDesignRecorder;
				structureDesignRecorder.reportExceptionalStructure(modelCrystalStructure, designedStructuresDirectoryPath, m_designDirectoryPath);
			}

			std::filesystem::remove_all(m_designDirectoryPath);
			logWriter.write(e);
		}
	}
}

void CreateStructures::createWithoutRecord(const ModelAtomCollection& modelAtomCollection) const
{
	OperatingSystem::IO::LogWriter logWriter;
	{
		std::filesystem::path directoryPath = m_structureCreationReportOptions.workDirectoryPath();
		directoryPath /= OperatingSystem::Configuration::MpiController::getMpiDirectoryName();

		logWriter.setDirectoryPath(directoryPath);
	}

	std::filesystem::path designedStructuresDirectoryPath = m_structureCreationReportOptions.designedStructuresDirectoryPath();
	designedStructuresDirectoryPath /= OperatingSystem::Configuration::MpiController::getMpiDirectoryName();


	while (shouldDesign())
	{
		ModelCrystalStructure modelCrystalStructure;

		try
		{
			initializeDesignDirectoryPath();

			modelCrystalStructure = m_initialStructureGenerator.generate(modelAtomCollection);
			m_structuralOptimizer.optimize(modelCrystalStructure);


			modelCrystalStructure.createTrackingAtomPairs();
			modelCrystalStructure.createGeometricalConstraints();
			{
				MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder structureDesignRecorder;

				if (modelCrystalStructure.isOptimalSolution(m_structuralOptimizer.structuralOptimizationParameters().preciseRelaxationParameters().feasibleConstraintErrorRate()))
					structureDesignRecorder.reportOptimalStructure(modelCrystalStructure, designedStructuresDirectoryPath);
				else
				{
					if (m_structureCreationReportOptions.needInfeasibleStructureData())
						structureDesignRecorder.reportInfeasibleStructure(modelCrystalStructure, designedStructuresDirectoryPath, m_designDirectoryPath);
				}
			}


			std::filesystem::remove_all(m_designDirectoryPath);
			m_designDirectoryPath.clear();
		}

		catch (const std::exception& e)
		{
			if (m_structureCreationReportOptions.needInfeasibleStructureData())
			{
				MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder structureDesignRecorder;
				structureDesignRecorder.reportExceptionalStructure(modelCrystalStructure, designedStructuresDirectoryPath, m_designDirectoryPath);
			}

			std::filesystem::remove_all(m_designDirectoryPath);
			logWriter.write(e);
		}
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

void CreateStructures::initializeDesignDirectoryPath() const
{
	m_designDirectoryPath.clear();
	{
		m_designDirectoryPath = m_structureCreationReportOptions.workDirectoryPath();
		m_designDirectoryPath /= OperatingSystem::Configuration::MpiController::getMpiDirectoryName();

		std::string directoryName;
		{
			directoryName += OperatingSystem::Configuration::MpiController::getMpiDirectoryName();
			directoryName += "_Sample_";
			directoryName += std::to_string(m_sampleIndex);
		}
		m_designDirectoryPath /= directoryName;
	}
	std::filesystem::remove_all(m_designDirectoryPath);


	if (!(std::filesystem::create_directory(m_designDirectoryPath)))
	{
		std::string errorMessage;
		{
			errorMessage += "Could not create directory of the path \"";
			errorMessage += m_designDirectoryPath.generic_string();
			errorMessage += "\"";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "initializeDesignDirectoryPath", errorMessage));
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
