#include "CrystalSelector.h"

#include <thread>

#include "ThreadController.h"

#include "InternationalTables.h"

#include "IonicBond.h"
#include "MetallicBond.h"
#include "AnionicConstraint.h"
#include "CationicConstraint.h"
#include "NonIonicBond.h"
#include "NonMetallicBond.h"
#include "StructureCustomizerBase.h"

#include "SelectCrystals.h"

using namespace MathematicalCrystalChemistry::Runtime::Selection;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

CrystalSelector::CrystalSelector(const CrystalSelectPlan& crystalSelectPlan)
	: m_crystalSelectPlan{ crystalSelectPlan }
{
}

CrystalSelector::CrystalSelector(CrystalSelectPlan&& crystalSelectPlan)
	: m_crystalSelectPlan{ std::move(crystalSelectPlan) }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void CrystalSelector::execute() const
{
	initializeDirectories();
	initializeChemicalParameters();
	initializeGeometricalConstraints();


	std::vector<Internal::SelectCrystals> selectCrystalsCollection;
	{
		for (std::size_t threadIndex = 0; threadIndex < OperatingSystem::Configuration::ThreadController::maxThreading(); ++threadIndex)
			selectCrystalsCollection.push_back(Internal::SelectCrystals{ threadIndex, m_crystalSelectPlan });
	}

	std::vector<std::thread> operatingSystems;
	{
		for (std::size_t threadIndex = 0; threadIndex < OperatingSystem::Configuration::ThreadController::maxThreading(); ++threadIndex)
			operatingSystems.push_back(std::thread{ selectCrystalsCollection[threadIndex] });
	}

	for (auto& operatingSystem : operatingSystems)
		operatingSystem.join();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void CrystalSelector::initializeDirectories() const
{
	std::filesystem::path cifDataDirectoryPath = m_crystalSelectPlan.cifDataDirectoryPath();
	{
		if (!(std::filesystem::is_directory(cifDataDirectoryPath)))
		{
			std::string errorMessage;
			{
				errorMessage += "Could not find directory of \"";
				errorMessage += cifDataDirectoryPath.generic_string();
				errorMessage += "\".";
			}

			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "initializeDirectories", errorMessage));
		}
	}

	std::filesystem::path outputDirectoryPath = m_crystalSelectPlan.outputDirectoryPath();
	{
		if (!(std::filesystem::is_directory(outputDirectoryPath)))
		{
			if (!(std::filesystem::create_directories(outputDirectoryPath)))
			{
				std::string errorMessage;
				{
					errorMessage += "Could not create directories of the path \"";
					errorMessage += outputDirectoryPath.generic_string();
					errorMessage += "\"";
				}

				OperatingSystem::Configuration::ReportException reportException;
				throw std::invalid_argument(reportException(typeid(*this), "initializeDirectories", errorMessage));
			}
		}
	}
}

void CrystalSelector::initializeChemicalParameters() const
{
	ChemToolkit::Crystallography::SpaceGroups::InternationalTables::setSpaceGroupPrecision(m_crystalSelectPlan.chemicalParameters().spaceGroupPrecision());
	MathematicalCrystalChemistry::StructureModel::Components::Internal::CoordinationCenterBase::setPointGroupPrecision(m_crystalSelectPlan.crystalAnalysisOptions().coordinationPolyhedraConstraints().pointGroupPrecision());
}

void CrystalSelector::initializeGeometricalConstraints() const
{
	using MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

	using MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond;


	ModelCrystalStructure::setCommonAtomPairTrackingCutoffRatio(m_crystalSelectPlan.structuralOptimizationParameters().atomPairTrackingCutoffRatio());
	StructureCustomizerBase::setCommonPressure(m_crystalSelectPlan.structuralOptimizationParameters().pressure());
	StructureCustomizerBase::setCommonAttractiveForceConstant(m_crystalSelectPlan.structuralOptimizationParameters().attractiveForceConstant());
	StructureCustomizerBase::setCommonRepulsiveForceConstant(m_crystalSelectPlan.structuralOptimizationParameters().repulsiveForceConstant());

	IonicBond::setDefaultCutoffRatio(m_crystalSelectPlan.geometricalConstraintParameters().ionicBondCutoffLengthRatio());
	MetallicBond::setDefaultCutoffRatio(m_crystalSelectPlan.geometricalConstraintParameters().metallicBondCutoffLengthRatio());
	AnionicConstraint::setDefaultCutoffRatio(m_crystalSelectPlan.geometricalConstraintParameters().anionicConstraintCutoffLengthRatio());
	NonIonicBond::setDefaultCutoffRatio(m_crystalSelectPlan.geometricalConstraintParameters().nonIonicBondCutoffLengthRatio());
	NonMetallicBond::setDefaultCutoffRatio(m_crystalSelectPlan.geometricalConstraintParameters().nonMetallicBondCutoffLengthRatio());

	NonIonicBond::setDefaultNonBondingRangeRatio(m_crystalSelectPlan.geometricalConstraintParameters().nonIonicBondRangeRatio());
	NonMetallicBond::setDefaultNonBondingRangeRatio(m_crystalSelectPlan.geometricalConstraintParameters().nonMetallicBondRangeRatio());
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
