#include "LigandTracker.h"

#include "IonicBond.h"
#include "MetallicBond.h"
#include "AnionicConstraint.h"
#include "CationicConstraint.h"
#include "NonIonicBond.h"
#include "NonMetallicBond.h"

using namespace MathematicalCrystalChemistry::StructureModel::Components::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors and operators

LigandTracker::LigandTracker(const LigandTracker& ligandTracker)
	: m_ionicBondPtrs{}
	, m_metallicBondPtrs{}
	, m_anionicConstraintPtrs{}
	, m_cationicConstraintPtrs{}
	, m_nonIonicBondPtrs{}
	, m_nonMetallicBondPtrs{}
{
}

LigandTracker::LigandTracker(LigandTracker&& ligandTracker) noexcept
	: m_ionicBondPtrs{}
	, m_metallicBondPtrs{}
	, m_anionicConstraintPtrs{}
	, m_cationicConstraintPtrs{}
	, m_nonIonicBondPtrs{}
	, m_nonMetallicBondPtrs{}
{
	ligandTracker.clearConstraintPtrs();
}

LigandTracker& LigandTracker::operator=(const LigandTracker& ligandTracker)
{
	clearConstraintPtrs();
	return *this;
}

LigandTracker& LigandTracker::operator=(LigandTracker&& ligandTracker) noexcept
{
	clearConstraintPtrs();
	ligandTracker.clearConstraintPtrs();
	return *this;
}

// Constructors and operators
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition LigandTracker::getIonicBondComposition() const
{
	ConstraintComposition bondComposition;
	{
		for (const auto& ionicBondWeakPtr : m_ionicBondPtrs)
		{
			if (std::shared_ptr<IonicBond> ionicBondPtr = ionicBondWeakPtr.lock())
				bondComposition.add(ionicBondPtr->ligand().get().coordinationFeature().atomCategory().symbol());
		}
	}

	return bondComposition;
}

MathematicalCrystalChemistry::StructureModel::Features::ConstraintComposition LigandTracker::getMetallicBondComposition() const
{
	ConstraintComposition bondComposition;
	{
		for (const auto& metallicBondWeakPtr : m_metallicBondPtrs)
		{
			if (std::shared_ptr<MetallicBond> metallicBondPtr = metallicBondWeakPtr.lock())
				bondComposition.add(metallicBondPtr->ligand().get().coordinationFeature().atomCategory().symbol());
		}
	}

	return bondComposition;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Protected methods

std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond>> LigandTracker::selectIonicLigandCandidates() const
{
	return selectIonicLigandCandidates(IonicBond::defaultCutoffRatio());
}

std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond>> LigandTracker::selectIonicLigandCandidates(const double ionicBondCutoffRatio) const
{
	std::vector<std::shared_ptr<NonIonicBond>> ligandCandidates;
	{
		for (const auto& nonIonicBondWeakPtr : m_nonIonicBondPtrs)
		{
			if (std::shared_ptr<NonIonicBond> nonIonicBondPtr = nonIonicBondWeakPtr.lock())
			{
				if (IonicBond::shouldCreate(nonIonicBondPtr->modelAtomRef(), nonIonicBondPtr->ligand(), ionicBondCutoffRatio))
					ligandCandidates.push_back(std::move(nonIonicBondPtr));
			}
		}
	}

	return ligandCandidates;
}

std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond>> LigandTracker::selectIonicLigandCandidates(const ConstraintComposition& fixedIonicBondComposition) const
{
	return selectIonicLigandCandidates(fixedIonicBondComposition, IonicBond::defaultCutoffRatio());
}

std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond>> LigandTracker::selectIonicLigandCandidates(const ConstraintComposition& fixedIonicBondComposition, const double ionicBondCutoffRatio) const
{
	std::map<std::string, std::multimap<double, std::shared_ptr<NonIonicBond>>> possibleLigands;
	{
		for (const auto& nonIonicBondWeakPtr : m_nonIonicBondPtrs)
		{
			if (std::shared_ptr<NonIonicBond> nonIonicBondPtr = nonIonicBondWeakPtr.lock())
			{
				const std::string atomCategorySymbol = nonIonicBondPtr->ligand().get().coordinationFeature().atomCategory().symbol();

				if (fixedIonicBondComposition.contains(atomCategorySymbol))
				{
					if (IonicBond::shouldCreate(nonIonicBondPtr->modelAtomRef(), nonIonicBondPtr->ligand(), ionicBondCutoffRatio))
					{
						const double distanceSquare = (nonIonicBondPtr->ligand().coordinate() - nonIonicBondPtr->modelAtomRef().coordinate()).normSquare();
						possibleLigands[atomCategorySymbol].emplace(distanceSquare, std::move(nonIonicBondPtr));
					}
				}
			}
		}
	}

	std::vector<std::shared_ptr<NonIonicBond>> ligandCandidates;
	{
		for (const auto& symbolAndNum : fixedIonicBondComposition)
		{
			size_type ligandIndex = 0;

			for (const auto& distanceAndPtr : possibleLigands[symbolAndNum.first])
			{
				if (ligandIndex < symbolAndNum.second)
				{
					ligandCandidates.push_back(distanceAndPtr.second);
					++ligandIndex;
				}

				else
					break;
			}
		}
	}

	return ligandCandidates;
}

std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond>> LigandTracker::selectMetallicLigandCandidates() const
{
	return selectMetallicLigandCandidates(MetallicBond::defaultCutoffRatio());
}

std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond>> LigandTracker::selectMetallicLigandCandidates(const double metallicBondCutoffRatio) const
{
	std::vector<std::shared_ptr<NonMetallicBond>> ligandCandidates;
	{
		for (const auto& nonMetallicBondWeakPtr : m_nonMetallicBondPtrs)
		{
			if (std::shared_ptr<NonMetallicBond> nonMetallicBondPtr = nonMetallicBondWeakPtr.lock())
			{
				if (MetallicBond::shouldCreate(nonMetallicBondPtr->modelAtomRef(), nonMetallicBondPtr->ligand(), metallicBondCutoffRatio))
					ligandCandidates.push_back(std::move(nonMetallicBondPtr));
			}
		}
	}

	return ligandCandidates;
}

std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond>> LigandTracker::selectMetallicLigandCandidates(const ConstraintComposition& fixedMetallicBondComposition) const
{
	return selectMetallicLigandCandidates(fixedMetallicBondComposition, MetallicBond::defaultCutoffRatio());
}

std::vector<std::shared_ptr<MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond>> LigandTracker::selectMetallicLigandCandidates(const ConstraintComposition& fixedMetallicBondComposition, const double metallicBondCutoffRatio) const
{
	std::map<std::string, std::multimap<double, std::shared_ptr<NonMetallicBond>>> possibleLigands;
	{
		for (const auto& nonMetallicBondWeakPtr : m_nonMetallicBondPtrs)
		{
			if (std::shared_ptr<NonMetallicBond> nonMetallicBondPtr = nonMetallicBondWeakPtr.lock())
			{
				const std::string atomCategorySymbol = nonMetallicBondPtr->ligand().get().coordinationFeature().atomCategory().symbol();

				if (fixedMetallicBondComposition.contains(atomCategorySymbol))
				{
					if (MetallicBond::shouldCreate(nonMetallicBondPtr->modelAtomRef(), nonMetallicBondPtr->ligand(), metallicBondCutoffRatio))
					{
						const double distanceSquare = (nonMetallicBondPtr->ligand().coordinate() - nonMetallicBondPtr->modelAtomRef().coordinate()).normSquare();
						possibleLigands[atomCategorySymbol].emplace(distanceSquare, std::move(nonMetallicBondPtr));
					}
				}
			}
		}
	}

	std::vector<std::shared_ptr<NonMetallicBond>> ligandCandidates;
	{
		for (const auto& symbolAndNum : fixedMetallicBondComposition)
		{
			size_type ligandIndex = 0;

			for (const auto& distanceAndPtr : possibleLigands[symbolAndNum.first])
			{
				if (ligandIndex < symbolAndNum.second)
				{
					ligandCandidates.push_back(distanceAndPtr.second);
					++ligandIndex;
				}

				else
					break;
			}
		}
	}

	return ligandCandidates;
}

// Protected methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
