#include "ModelAtom.h"

#include <utility>
#include <functional>

#include "IonicBond.h"
#include "MetallicBond.h"
#include "AnionicConstraint.h"
#include "CationicConstraint.h"
#include "NonIonicBond.h"
#include "NonMetallicBond.h"

#include "AtomicEnvironment.h"

using namespace MathematicalCrystalChemistry::StructureModel::Components;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

ModelAtom::ModelAtom(const std::string& elementSymbol)
	: CoordinationCenter{ elementSymbol }
	, LigandTracker{}
	, OxidationManager{ elementSymbol }
	, m_siteSymmetrySymbol{}
	, m_wyckoffSymbol{ 'a' }
	, m_siteIndex{ std::numeric_limits<size_type>::max() }
{
}

ModelAtom::ModelAtom(const std::string& elementSymbol, const CoordinationFeatureSymbol& coordinationFeatureSymbol)
	: CoordinationCenter{ elementSymbol, coordinationFeatureSymbol }
	, LigandTracker{}
	, OxidationManager{ elementSymbol }
	, m_siteSymmetrySymbol{}
	, m_wyckoffSymbol{ 'a' }
	, m_siteIndex{ std::numeric_limits<size_type>::max() }
{
}

ModelAtom::ModelAtom(const std::string& elementSymbol, const NumericalVector& cartesianCoordinate)
	: CoordinationCenter{ elementSymbol, cartesianCoordinate }
	, LigandTracker{}
	, OxidationManager{ elementSymbol }
	, m_siteSymmetrySymbol{}
	, m_wyckoffSymbol{ 'a' }
	, m_siteIndex{ std::numeric_limits<size_type>::max() }
{
}

ModelAtom::ModelAtom(const std::string& elementSymbol, const CoordinationFeatureSymbol& coordinationFeatureSymbol, const NumericalVector& cartesianCoordinate)
	: CoordinationCenter{ elementSymbol, coordinationFeatureSymbol, cartesianCoordinate }
	, LigandTracker{}
	, OxidationManager{ elementSymbol }
	, m_siteSymmetrySymbol{}
	, m_wyckoffSymbol{ 'a' }
	, m_siteIndex{ std::numeric_limits<size_type>::max() }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Access methods

std::vector<ModelAtom::Ligand> ModelAtom::getIonicLigands() const
{
	std::vector<Ligand> ligands;
	{
		for (const auto& ionicBondPtr : getIonicBondPtrs())
			ligands.push_back(ionicBondPtr->ligand());
	}

	return ligands;
}

std::vector<ModelAtom::Ligand> ModelAtom::getMetallicLigands() const
{
	std::vector<Ligand> ligands;
	{
		for (const auto& metallicBondPtr : getMetallicBondPtrs())
			ligands.push_back(metallicBondPtr->ligand());
	}

	return ligands;
}

std::vector<ModelAtom::Ligand> ModelAtom::getAnionicConstrainedLigands() const
{
	std::vector<Ligand> ligands;
	{
		for (const auto& anionicConstraintPtr : getAnionicConstraintPtrs())
			ligands.push_back(anionicConstraintPtr->ligand());
	}

	return ligands;
}

std::vector<ModelAtom::Ligand> ModelAtom::getCationicConstrainedLigands() const
{
	std::vector<Ligand> ligands;
	{
		for (const auto& cationicConstraintPtr : getCationicConstraintPtrs())
			ligands.push_back(cationicConstraintPtr->ligand());
	}

	return ligands;
}

std::vector<ModelAtom::Ligand> ModelAtom::getNonIonicBondedLigands() const
{
	std::vector<Ligand> ligands;
	{
		for (const auto& nonIonicBondPtr : getNonIonicBondPtrs())
			ligands.push_back(nonIonicBondPtr->ligand());
	}

	return ligands;
}

std::vector<ModelAtom::Ligand> ModelAtom::getNonMetallicBondedLigands() const
{
	std::vector<Ligand> ligands;
	{
		for (const auto& nonMetallicBondPtr : getNonMetallicBondPtrs())
			ligands.push_back(nonMetallicBondPtr->ligand());
	}

	return ligands;
}

// Access methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

AtomicEnvironment ModelAtom::getAtomicEnvironment() const
{
	AtomicEnvironment atomicEnvironment;
	{
		atomicEnvironment.setElementSymbol(elementSymbol());
		atomicEnvironment.setAtomCategorySymbol(coordinationFeature().atomCategory().symbol());
		atomicEnvironment.setSiteSymmetrySymbol(m_siteSymmetrySymbol);
		atomicEnvironment.setSiteIndex(m_siteIndex);
		atomicEnvironment.setWyckoffSymbol(m_wyckoffSymbol);
		atomicEnvironment.setIonicBondFingerprint(getIonicLigands());
		atomicEnvironment.setMetallicBondFingerprint(getMetallicLigands());


		std::vector<std::pair<ModelAtom::Ligand, std::vector<ModelAtom::Ligand>>> ligandsAndBridges;
		{
			for (const auto& constraintPtr : getCationicConstraintPtrs())
			{
				std::pair<ModelAtom::Ligand, std::vector<ModelAtom::Ligand>> ligandAndBridges = std::make_pair(constraintPtr->ligand(), constraintPtr->commonBridgingAnions());
				ligandsAndBridges.push_back(std::move(ligandAndBridges));
			}
		}
		atomicEnvironment.setCationicConstraintFingerprint(ligandsAndBridges);
	}

	return atomicEnvironment;
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Customize methods

std::vector<ModelAtom::Ligand> ModelAtom::getIonicLigandCandidates() const
{
	std::vector<ModelAtom::Ligand> ionicLigandCandidates;
	{
		if (coordinationFeature().atomCategory().isCation())
		{
			if (coordinationFeature().feasibleIonicBondComposition().isUnconstrained())
			{
				for (const auto& nonBondIter : selectIonicLigandCandidates())
					ionicLigandCandidates.push_back(nonBondIter->ligand());
			}

			else
			{
				for (const auto& nonBondIter : selectIonicLigandCandidates(coordinationFeature().feasibleIonicBondComposition()))
					ionicLigandCandidates.push_back(nonBondIter->ligand());
			}
		}
	}

	return ionicLigandCandidates;
}

std::vector<ModelAtom::Ligand> ModelAtom::getIonicLigandCandidates(const double ionicBondCutoffRatio) const
{
	std::vector<ModelAtom::Ligand> ionicLigandCandidates;
	{
		if (coordinationFeature().atomCategory().isCation())
		{
			if (coordinationFeature().feasibleIonicBondComposition().isUnconstrained())
			{
				for (const auto& nonBondIter : selectIonicLigandCandidates(ionicBondCutoffRatio))
					ionicLigandCandidates.push_back(nonBondIter->ligand());
			}

			else
			{
				for (const auto& nonBondIter : selectIonicLigandCandidates(coordinationFeature().feasibleIonicBondComposition(), ionicBondCutoffRatio))
					ionicLigandCandidates.push_back(nonBondIter->ligand());
			}
		}
	}

	return ionicLigandCandidates;
}

std::vector<ModelAtom::Ligand> ModelAtom::getMetallicLigandCandidates() const
{
	std::vector<ModelAtom::Ligand> metallicLigandCandidates;
	{
		if (coordinationFeature().atomCategory().isMetal())
		{
			if (coordinationFeature().feasibleMetallicBondComposition().isUnconstrained())
			{
				for (const auto& nonBondIter : selectMetallicLigandCandidates())
					metallicLigandCandidates.push_back(nonBondIter->ligand());
			}

			else
			{
				for (const auto& nonBondIter : selectMetallicLigandCandidates(coordinationFeature().feasibleMetallicBondComposition()))
					metallicLigandCandidates.push_back(nonBondIter->ligand());
			}
		}
	}

	return metallicLigandCandidates;
}

std::vector<ModelAtom::Ligand> ModelAtom::getMetallicLigandCandidates(const double metallicBondCutoffRatio) const
{
	std::vector<ModelAtom::Ligand> metallicLigandCandidates;
	{
		if (coordinationFeature().atomCategory().isMetal())
		{
			if (coordinationFeature().feasibleMetallicBondComposition().isUnconstrained())
			{
				for (const auto& nonBondIter : selectMetallicLigandCandidates(metallicBondCutoffRatio))
					metallicLigandCandidates.push_back(nonBondIter->ligand());
			}

			else
			{
				for (const auto& nonBondIter : selectMetallicLigandCandidates(coordinationFeature().feasibleMetallicBondComposition(), metallicBondCutoffRatio))
					metallicLigandCandidates.push_back(nonBondIter->ligand());
			}
		}
	}

	return metallicLigandCandidates;
}

std::vector<std::pair<ModelAtom::Ligand, ModelAtom::Ligand>> ModelAtom::getCationicConstraintCandidates() const
{
	std::vector<std::pair<Ligand, Ligand>> cationicConstraintCandidates;
	{
		if (coordinationFeature().atomCategory().isAnion())
		{
			using IonicBond = MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond;
			const std::vector<std::shared_ptr<IonicBond>> ionicBondPtrs = LigandTracker::getIonicBondPtrs();

			for (auto iter = ionicBondPtrs.begin(); iter != ionicBondPtrs.end(); ++iter)
			{
				for (auto anotherIter = std::next(iter); anotherIter != ionicBondPtrs.end(); ++anotherIter)
				{
					std::pair<Ligand, Ligand> cationicConstraintCandidate = std::make_pair((*iter)->ligand(), (*anotherIter)->ligand());
					cationicConstraintCandidates.push_back(std::move(cationicConstraintCandidate));
				}
			}
		}
	}

	return cationicConstraintCandidates;
}

std::vector<ModelAtom::Ligand> ModelAtom::getSubstituteIonicLigandCandidates() const
{
	std::vector<ModelAtom::Ligand> substituteIonicLigandCandidates;
	{
		if (coordinationFeature().atomCategory().isMetal())
		{
			if (!(coordinationFeature().feasibleIonicBondComposition().isUnconstrained()))
			{
				for (const auto& nonIonicBondIter : selectIonicLigandCandidates())
					substituteIonicLigandCandidates.push_back(nonIonicBondIter->ligand());
			}
		}
	}

	return substituteIonicLigandCandidates;
}

// Customize methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void ModelAtom::recognizeCoordinationPolyhedra()
{
	using CoordinatedLigand = ChemToolkit::Crystallography::PeriodicReferenceWrapper<CoordinationCenter, ChemToolkit::Crystallography::UnitCell>;

	std::vector<CoordinatedLigand> ionicLigands;
	{
		for (const auto& ionicBondPtr : getIonicBondPtrs())
		{
			CoordinationCenter& coordinationCenter = dynamic_cast<CoordinationCenter&>(ionicBondPtr->ligand().get());
			CoordinatedLigand ionicLigand{ coordinationCenter, ionicBondPtr->ligand() };

			ionicLigands.push_back(std::move(ionicLigand));
		}
	}

	std::vector<CoordinatedLigand> metallicLigands;
	{
		for (const auto& metallicBondPtr : getMetallicBondPtrs())
		{
			CoordinationCenter& coordinationCenter = dynamic_cast<CoordinationCenter&>(metallicBondPtr->ligand().get());
			CoordinatedLigand metallicLigand{ coordinationCenter, metallicBondPtr->ligand() };

			metallicLigands.push_back(std::move(metallicLigand));
		}
	}

	recognizeIonicLigands(std::move(ionicLigands));
	recognizeMetallicLigands(std::move(metallicLigands));
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
