#include "ModelAtomTracker.h"

#include <cmath>

using namespace MathematicalCrystalChemistry::StructureModel::Components::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

double ModelAtomTracker::m_commonAtomPairTrackingCutoffRatio{ 4.0 };
double ModelAtomTracker::m_defaultAtomPairTrackingCutoffRatio{ 4.0 };


ModelAtomTracker::ModelAtomTracker()
	: ConstraintManager{}
	, CrystalStructure{}
	, m_trackingAtomPairs{}
	, m_atomPairTrackingCutoffRatio{ m_commonAtomPairTrackingCutoffRatio }
{
}

ModelAtomTracker::ModelAtomTracker(const UnitCell& unitCell, const AtomicArrangement& atomicArrangement)
	: ConstraintManager{}
	, CrystalStructure{ unitCell, atomicArrangement }
	, m_trackingAtomPairs{}
	, m_atomPairTrackingCutoffRatio{ m_commonAtomPairTrackingCutoffRatio }
{
}

ModelAtomTracker::ModelAtomTracker(const UnitCell& unitCell, AtomicArrangement&& atomicArrangement)
	: ConstraintManager{}
	, CrystalStructure{ unitCell, std::move(atomicArrangement) }
	, m_trackingAtomPairs{}
	, m_atomPairTrackingCutoffRatio{ m_commonAtomPairTrackingCutoffRatio }
{
}

ModelAtomTracker::ModelAtomTracker(const ModelAtomTracker& modelAtomTracker)
	: ConstraintManager{}
	, CrystalStructure{ dynamic_cast<const CrystalStructure&>(modelAtomTracker) }
	, m_trackingAtomPairs{}
	, m_atomPairTrackingCutoffRatio{ modelAtomTracker.m_atomPairTrackingCutoffRatio }
{
}

ModelAtomTracker::ModelAtomTracker(ModelAtomTracker&& modelAtomTracker) noexcept
	: ConstraintManager{}
	, CrystalStructure{ modelAtomTracker.unitCell(), std::move(modelAtomTracker.atomicArrangement())}
	, m_trackingAtomPairs{}
	, m_atomPairTrackingCutoffRatio{ modelAtomTracker.m_atomPairTrackingCutoffRatio }
{
	modelAtomTracker.clearTrackingAtomPairs();
}

ModelAtomTracker& ModelAtomTracker::operator=(const ModelAtomTracker& modelAtomTracker)
{
	CrystalStructure::operator=(modelAtomTracker);
	clearTrackingAtomPairs();
	m_atomPairTrackingCutoffRatio = modelAtomTracker.m_atomPairTrackingCutoffRatio;

	return *this;
}

ModelAtomTracker& ModelAtomTracker::operator=(ModelAtomTracker&& modelAtomTracker) noexcept
{
	CrystalStructure::operator=(modelAtomTracker);
	clearTrackingAtomPairs();
	m_atomPairTrackingCutoffRatio = modelAtomTracker.m_atomPairTrackingCutoffRatio;

	modelAtomTracker.clearTrackingAtomPairs();

	return *this;
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void ModelAtomTracker::createTrackingAtomPairs()
{
	clearTrackingAtomPairs();

	for (auto iter = atomicArrangement().begin(); iter != atomicArrangement().end(); ++iter)
	{
		for (auto anotherIter = std::next(iter); anotherIter != atomicArrangement().end(); ++anotherIter)
		{
			for (const auto& latticePoint : getSearchingLatticePoints((*iter), (*anotherIter)))
			{
				if (shouldTrack((*iter), (*anotherIter), latticePoint))
				{
					Ligand ligand{ (*anotherIter), unitCell(), latticePoint };
					m_trackingAtomPairs.push_back(ConstraintKey::createConstraintKey((*iter), ligand));
				}
			}
		}

		for (const auto& latticePoint : getSearchingLatticePoints((*iter), (*iter)))
		{
			if (latticePoint[0] != 0 || latticePoint[1] != 0 || latticePoint[2] != 0)
			{

				if (shouldTrack((*iter), (*iter), latticePoint))
				{
					Ligand ligand{ (*iter), unitCell(), latticePoint };
					m_trackingAtomPairs.push_back(ConstraintKey::createConstraintKey((*iter), ligand));
				}
			}
		}
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

std::vector<ChemToolkit::Crystallography::PeriodicReferenceWrapper<MathematicalCrystalChemistry::StructureModel::Components::ModelAtom, ChemToolkit::Crystallography::UnitCell>::LatticePoint> ModelAtomTracker::getSearchingLatticePoints(const ModelAtom& modelAtom, const ModelAtom& ligandAtom) const
{
	std::vector<Ligand::LatticePoint> latticePoints;
	{
		const NumericalVector fractionalCoordinate = unitCell().inverseBasisVectors() * modelAtom.coordinate();
		double neighborsDetectionZoneRadius = 0.0;
		{
			if (modelAtom.coordinationFeature().atomCategory().isMetal() && ligandAtom.coordinationFeature().atomCategory().isMetal())
				neighborsDetectionZoneRadius = m_atomPairTrackingCutoffRatio * (modelAtom.coordinationFeature().upperMetallicBondRadius() + ligandAtom.coordinationFeature().upperMetallicBondRadius());
			else
				neighborsDetectionZoneRadius = m_atomPairTrackingCutoffRatio * (modelAtom.coordinationFeature().repulsiveIonicRadius() + ligandAtom.coordinationFeature().repulsiveIonicRadius());

		}

		std::vector<int> indicesA;
		{
			double lengthA = (neighborsDetectionZoneRadius * unitCell().inverseBasisVectors().rowVector(0).norm());
			int min = static_cast<int>(std::floor(fractionalCoordinate[0] - lengthA));
			int max = static_cast<int>(std::floor(fractionalCoordinate[0] + lengthA));

			for (int index = min; index <= max; ++index)
				indicesA.push_back(index);
		}

		std::vector<int> indicesB;
		{
			double lengthB = (neighborsDetectionZoneRadius * unitCell().inverseBasisVectors().rowVector(1).norm());
			int min = static_cast<int>(std::floor(fractionalCoordinate[1] - lengthB));
			int max = static_cast<int>(std::floor(fractionalCoordinate[1] + lengthB));

			for (int index = min; index <= max; ++index)
				indicesB.push_back(index);
		}

		std::vector<int> indicesC;
		{
			double lengthC = (neighborsDetectionZoneRadius * unitCell().inverseBasisVectors().rowVector(2).norm());
			int min = static_cast<int>(std::floor(fractionalCoordinate[2] - lengthC));
			int max = static_cast<int>(std::floor(fractionalCoordinate[2] + lengthC));

			for (int index = min; index <= max; ++index)
				indicesC.push_back(index);
		}

		for (const auto numA : indicesA)
		{
			for (const auto numB : indicesB)
			{
				for (const auto numC : indicesC)
					latticePoints.push_back(Ligand::LatticePoint{ numA, numB, numC });
			}
		}
	}

	return latticePoints;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
