#include "ModelCrystalStructure.h"

#include <map>
#include <set>
#include <algorithm>
#include <iterator>

#include "FileWriter.h"
#include "InternationalTables.h"
#include "MathematicalConstants.h"

using namespace MathematicalCrystalChemistry::StructureModel::Components;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

ModelCrystalStructure::ModelCrystalStructure(const UnitCell& unitCell, const AtomicArrangement& atomicArrangement)
	: ModelAtomTracker{ unitCell, atomicArrangement }
{
}

ModelCrystalStructure::ModelCrystalStructure(const UnitCell& unitCell, AtomicArrangement&& atomicArrangement)
	: ModelAtomTracker{ unitCell, std::move(atomicArrangement) }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

double ModelCrystalStructure::getLowerIonicPackingFraction() const
{
	double unitCellVolume = unitCell().getCellVolume();

	if (0.0 < unitCellVolume)
	{
		double packingFraction = 0.0;
		{
			for (const auto& modelAtom : atomicArrangement())
			{
				double thisVolume = 1.33333333333333333333 * MathToolkit::Numerics::MathematicalConstants::getPi();
				{
					double radius = modelAtom.coordinationFeature().lowerIonicBondRadius();
					thisVolume *= (radius * radius * radius);
				}

				packingFraction += thisVolume;
			}

			packingFraction /= unitCellVolume;
		}

		return packingFraction;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "getIonicPackingFraction", "The volume of unit cell is not more than zero."));
	}
}

void ModelCrystalStructure::expandStructure(const double packingFraction)
{
	if (0.0 < packingFraction)
	{
		double sphereVolume = 0.0;
		{
			for (auto& modelAtom : atomicArrangement())
			{
				modelAtom.coordinate().transform(unitCell().inverseBasisVectors());

				double thisVolume = 1.33333333333333333333 * MathToolkit::Numerics::MathematicalConstants::getPi();
				{
					double radius = modelAtom.coordinationFeature().upperIonicBondRadius();
					thisVolume *= (radius * radius * radius);
				}

				sphereVolume += thisVolume;
			}
		}


		double scalingRatio = std::pow((sphereVolume / (packingFraction * unitCell().getCellVolume())), (1.0 / 3.0));

		NumericalMatrix newBasisVectors = unitCell().basisVectors();
		newBasisVectors *= scalingRatio;
		unitCell().setBasisVectors(std::move(newBasisVectors));

		for (auto& modelAtom : atomicArrangement())
			modelAtom.coordinate().transform(unitCell().basisVectors());
	}
}

void ModelCrystalStructure::optimizeOxidationStates()
{
	for (auto& atom : atomicArrangement())
		atom.initializeOxidationState();

	std::vector<ModelAtom*> anionPtrs;
	std::vector<ModelAtom*> cationPtrs;
	{
		for (auto& atom : atomicArrangement())
		{
			if (atom.coordinationFeature().atomCategory().isAnion())
				anionPtrs.push_back(&atom);
		}

		for (auto& atom : atomicArrangement())
		{
			if (atom.coordinationFeature().atomCategory().isCation())
				cationPtrs.push_back(&atom);
		}
	}

	while (true)
	{
		for (auto iter = anionPtrs.begin(); iter != anionPtrs.end();)
		{
			if ((*iter)->canReduceOxidationState())
				++iter;
			else
				iter = anionPtrs.erase(iter);
		}

		for (auto iter = cationPtrs.begin(); iter != cationPtrs.end();)
		{
			if ((*iter)->canAddOxidationState())
				++iter;
			else
				iter = cationPtrs.erase(iter);
		}

		if (anionPtrs.empty() || cationPtrs.empty())
			break;
		else
		{
			std::sort(anionPtrs.begin(), anionPtrs.end(), [](const ModelAtom* maaPtr, const ModelAtom* mabPtr) { return (maaPtr->oxidationState() > mabPtr->oxidationState()); });
			std::sort(cationPtrs.begin(), cationPtrs.end(), [](const ModelAtom* maaPtr, const ModelAtom* mabPtr) { return (maaPtr->ionizationPotential() < mabPtr->ionizationPotential()); });
	
			(*anionPtrs.begin())->reduceOxidationState();
			(*cationPtrs.begin())->addOxidationState();
		}
	}
}

void ModelCrystalStructure::assignCoordinationFeaturesFromElementSymbols()
{
	for (auto& atom : atomicArrangement())
		atom.setCoordinationFeatureFromElementSymbol();
}

void ModelCrystalStructure::removeSiteDuplications()
{
	sort();
	updateSpaceGroupInfo();
	const double spaceGroupPrecision = ChemToolkit::Crystallography::SpaceGroups::InternationalTables::spaceGroupPrecision();

	std::set<size_type> surveyedSiteIndices;
	std::set<size_type> trashSiteIndices;


	for (auto iter = atomicArrangement().begin(); iter != atomicArrangement().end(); ++iter)
	{
		if (surveyedSiteIndices.find(iter->siteIndex()) == surveyedSiteIndices.end())
		{
			surveyedSiteIndices.emplace(iter->siteIndex());

			for (auto anotherIter = std::next(iter); anotherIter != atomicArrangement().end(); ++anotherIter)
			{
				double differenceSquare = (anotherIter->coordinate() - iter->coordinate()).normSquare();

				if (differenceSquare < (spaceGroupPrecision * spaceGroupPrecision))
				{
					if (iter->oxidationFeature().maxOxidationState() < anotherIter->oxidationFeature().maxOxidationState())
						trashSiteIndices.emplace(iter->siteIndex());

					else
					{
						surveyedSiteIndices.emplace(anotherIter->siteIndex());
						trashSiteIndices.emplace(anotherIter->siteIndex());
					}
				}
			}
		}
	}


	for (auto iter = atomicArrangement().begin(); iter != atomicArrangement().end();)
	{
		if (trashSiteIndices.find(iter->siteIndex()) == trashSiteIndices.end())
			++iter;
		else
			iter = atomicArrangement().erase(iter);
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Utility

StructureFingerprint ModelCrystalStructure::getStructureFingerprint() const
{
	StructureFingerprint structureFingerprint;
	{
		structureFingerprint.setSpaceGroupNumber(spaceGroupNumber());
		structureFingerprint.setAtomCategoryComposition(getAtomCategoryComposition());

		std::vector<AtomicEnvironment> atomicEnvironments;
		{
			for (const auto& atom : atomicArrangement())
				atomicEnvironments.push_back(atom.getAtomicEnvironment());
		}

		structureFingerprint.setAtomicEnvironments(std::move(atomicEnvironments));
	}

	return structureFingerprint;
}

MathematicalCrystalChemistry::StructureModel::Features::CrystalCategory ModelCrystalStructure::getCrystalCategory() const
{
	size_type numAnion = 0;
	size_type numMetal = 0;
	{
		for (const auto& modelAtom : atomicArrangement())
		{
			if (modelAtom.coordinationFeature().atomCategory().isAnion())
				++numAnion;
			if (modelAtom.coordinationFeature().atomCategory().isMetal())
				++numMetal;
		}
	}

	if (0 < numAnion)
	{
		if (numAnion < numMetal)
			return CrystalCategory{ "metallic_ionic" };
		else
			return CrystalCategory{ "ionic" };
	}

	else
	{
		if (0 < numMetal)
			return CrystalCategory{ "metallic" };
		else
			return CrystalCategory{ "covalent" };
	}
}

void ModelCrystalStructure::createStructureInformationFile(const std::filesystem::path& filePath) const
{
	std::string outputText;
	{
		outputText += "Space.Group.Number     ";
		outputText += std::to_string(spaceGroupNumber());
		outputText += "\n\n";

		outputText += "<Atom.Sites\n";
		outputText += "_element_symbol\n";
		outputText += "_site_index\n";
		outputText += "_site_symmetry\n";
		outputText += "_wyckoff_symbol\n";
		outputText += "_coordination_feature_symbol\n\n";
		{
			for (const auto& modelAtom : atomicArrangement())
			{
				outputText += modelAtom.elementSymbol();
				outputText += "   ";
				outputText += std::to_string(modelAtom.siteIndex());
				outputText += "   ";
				outputText += modelAtom.siteSymmetrySymbol();
				outputText += "   ";
				outputText += modelAtom.wyckoffSymbol();
				outputText += "   ";
				outputText += modelAtom.coordinationFeature().symbol().name();
				outputText += "\n";
			}
		}
		outputText += "\nAtom.Sites>\n";
	}


	OperatingSystem::IO::FileWriter fileWriter{ filePath };
	fileWriter.createFile(OperatingSystem::IO::FileCreateOptions::skip_existing, outputText);
}

// Utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
