#include "OxidationFeature.h"

#include <stdexcept>

#include "ReportException.h"

using namespace MathematicalCrystalChemistry::StructureModel::Features;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

OxidationFeature::OxidationFeature(const std::string& elementSymbol)
	: m_feasibleOxidationStates{ getFeasibleOxidationStates(elementSymbol) }
	, m_ionizationPotentials{ getIonizationPotentials(elementSymbol) }
	, m_electronegativity{ getElectronegativity(elementSymbol) }
{
	if (m_feasibleOxidationStates.size() == 1)
	{
		std::string errorMessage;
		{
			errorMessage += elementSymbol;
			errorMessage += " is out of definitions.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "constructor", errorMessage));
	}
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Static methods

std::set<int> OxidationFeature::getFeasibleOxidationStates(const std::string& elementSymbol)
{
	std::set<int> feasibleOxidationStates;
	{
		for (const auto& stateAndEnergy : getIonizationPotentials(elementSymbol))
		{
			feasibleOxidationStates.emplace(0);
			feasibleOxidationStates.emplace(stateAndEnergy.first);
		}
	}

	if (feasibleOxidationStates.empty())
	{
		feasibleOxidationStates.emplace(0);

		if (elementSymbol == "N" || elementSymbol == "P" || elementSymbol == "As" || elementSymbol == "Sb")
		{
			feasibleOxidationStates.emplace(-1);
			feasibleOxidationStates.emplace(-2);
			feasibleOxidationStates.emplace(-3);
		}
		if (elementSymbol == "O" || elementSymbol == "S" || elementSymbol == "Se" || elementSymbol == "Te")
		{
			feasibleOxidationStates.emplace(-1);
			feasibleOxidationStates.emplace(-2);
		}
		if (elementSymbol == "F" || elementSymbol == "Cl" || elementSymbol == "Br" || elementSymbol == "I")
			feasibleOxidationStates.emplace(-1);
	}

	return feasibleOxidationStates;
}

std::map<int, double> OxidationFeature::getIonizationPotentials(const std::string& elementSymbol)
{
	std::map<int, double> potentials;

	if (elementSymbol == "H")
		potentials.emplace(1, 13.59);
	if (elementSymbol == "Li")
		potentials.emplace(1, 5.39);
	if (elementSymbol == "Be")
	{
		potentials.emplace(1, 9.32);
		potentials.emplace(2, 18.21);
	}
	if (elementSymbol == "B")
	{
		potentials.emplace(1, 8.3);
		potentials.emplace(2, 25.15);
		potentials.emplace(3, 37.93);
	}
	if (elementSymbol == "C")
	{
		potentials.emplace(1, 11.26);
		potentials.emplace(2, 24.38);
		potentials.emplace(3, 47.89);
		potentials.emplace(4, 64.49);
	}
	if (elementSymbol == "Na")
		potentials.emplace(1, 5.14);
	if (elementSymbol == "Mg")
	{
		potentials.emplace(1, 7.65);
		potentials.emplace(2, 15.04);
	}
	if (elementSymbol == "Al")
	{
		potentials.emplace(1, 5.99);
		potentials.emplace(2, 18.83);
		potentials.emplace(3, 28.45);
	}
	if (elementSymbol == "Si")
	{
		potentials.emplace(1, 8.15);
		potentials.emplace(2, 16.35);
		potentials.emplace(3, 33.49);
		potentials.emplace(4, 45.14);
	}
	if (elementSymbol == "K")
		potentials.emplace(1, 4.34);
	if (elementSymbol == "Ca")
	{
		potentials.emplace(1, 6.11);
		potentials.emplace(2, 11.87);
	}
	if (elementSymbol == "Sc")
	{
		potentials.emplace(1, 6.56);
		potentials.emplace(2, 12.80);
		potentials.emplace(3, 24.76);
	}
	if (elementSymbol == "Ti")
	{
		potentials.emplace(1, 6.83);
		potentials.emplace(2, 13.58);
		potentials.emplace(3, 27.49);
		potentials.emplace(4, 43.27);
	}
	if (elementSymbol == "V")
	{
		potentials.emplace(1, 6.75);
		potentials.emplace(2, 14.62);
		potentials.emplace(3, 29.31);
		potentials.emplace(4, 46.71);
		potentials.emplace(5, 65.28);
	}
	if (elementSymbol == "Cr")
	{
		potentials.emplace(1, 6.77);
		potentials.emplace(2, 16.48);
		potentials.emplace(3, 30.96);
		potentials.emplace(4, 49.16);
		potentials.emplace(5, 69.46);
		potentials.emplace(6, 90.63);
	}
	if (elementSymbol == "Mn")
	{
		potentials.emplace(1, 7.43);
		potentials.emplace(2, 15.64);
		potentials.emplace(3, 33.67);
		potentials.emplace(4, 51.2);
		potentials.emplace(5, 72.4);
		potentials.emplace(6, 95.6);
	}
	if (elementSymbol == "Fe")
	{
		potentials.emplace(1, 7.9);
		potentials.emplace(2, 16.19);
		potentials.emplace(3, 30.65);
		potentials.emplace(4, 54.8);
		potentials.emplace(5, 75.0);
		potentials.emplace(6, 99.1);
	}
	if (elementSymbol == "Co")
	{
		potentials.emplace(1, 7.88);
		potentials.emplace(2, 17.08);
		potentials.emplace(3, 33.50);
		potentials.emplace(4, 51.3);
	}
	if (elementSymbol == "Ni")
	{
		potentials.emplace(1, 7.64);
		potentials.emplace(2, 18.17);
		potentials.emplace(3, 35.19);
		potentials.emplace(4, 54.9);
	}
	if (elementSymbol == "Cu")
	{
		potentials.emplace(1, 7.73);
		potentials.emplace(2, 20.29);
	}
	if (elementSymbol == "Zn")
	{
		potentials.emplace(1, 7.73);
		potentials.emplace(2, 17.96);
	}
	if (elementSymbol == "Ga")
	{
		potentials.emplace(1, 6.00);
		potentials.emplace(2, 20.51);
		potentials.emplace(3, 30.71);
	}
	if (elementSymbol == "Ge")
	{
		potentials.emplace(1, 7.9);
		potentials.emplace(2, 15.93);
		potentials.emplace(3, 34.22);
		potentials.emplace(4, 45.71);
	}
	if (elementSymbol == "Rb")
		potentials.emplace(1, 4.18);
	if (elementSymbol == "Sr")
	{
		potentials.emplace(1, 5.69);
		potentials.emplace(2, 11.03);
	}
	if (elementSymbol == "Y")
	{
		potentials.emplace(1, 6.22);
		potentials.emplace(2, 12.22);
		potentials.emplace(3, 20.52);
	}
	if (elementSymbol == "Zr")
	{
		potentials.emplace(1, 6.63);
		potentials.emplace(2, 13.1);
		potentials.emplace(3, 22.99);
		potentials.emplace(4, 34.34);
	}
	if (elementSymbol == "Nb")
	{
		potentials.emplace(1, 6.76);
		potentials.emplace(2, 14.0);
		potentials.emplace(3, 25.04);
		potentials.emplace(4, 38.3);
		potentials.emplace(5, 50.55);
	}
	if (elementSymbol == "Mo")
	{
		potentials.emplace(1, 7.09);
		potentials.emplace(2, 16.16);
		potentials.emplace(3, 27.13);
		potentials.emplace(4, 46.4);
		potentials.emplace(5, 54.49);
		potentials.emplace(6, 68.83);
	}
	if (elementSymbol == "Ru")
	{
		potentials.emplace(1, 7.36);
		potentials.emplace(2, 16.76);
		potentials.emplace(3, 28.47);
		potentials.emplace(4, 50.0);
		potentials.emplace(5, 60.0);
		potentials.emplace(6, 92.0);
	}
	if (elementSymbol == "Rh")
	{
		potentials.emplace(1, 7.46);
		potentials.emplace(2, 18.08);
		potentials.emplace(3, 31.06);
		potentials.emplace(4, 48.0);
		potentials.emplace(5, 65.0);
	}
	if (elementSymbol == "Pd")
	{
		potentials.emplace(1, 8.34);
		potentials.emplace(2, 19.43);
		potentials.emplace(3, 32.93);
		potentials.emplace(4, 53.0);
	}
	if (elementSymbol == "Ag")
		potentials.emplace(1, 7.58);
	if (elementSymbol == "Cd")
	{
		potentials.emplace(1, 8.99);
		potentials.emplace(2, 16.91);
	}
	if (elementSymbol == "In")
	{
		potentials.emplace(1, 5.79);
		potentials.emplace(2, 18.87);
		potentials.emplace(3, 28.03);
	}
	if (elementSymbol == "Sn")
	{
		potentials.emplace(1, 7.34);
		potentials.emplace(2, 14.63);
		potentials.emplace(3, 30.5);
		potentials.emplace(4, 40.73);
	}
	if (elementSymbol == "Cs")
		potentials.emplace(1, 3.89);
	if (elementSymbol == "Ba")
	{
		potentials.emplace(1, 5.21);
		potentials.emplace(2, 10.0);
	}
	if (elementSymbol == "La")
	{
		potentials.emplace(1, 5.58);
		potentials.emplace(2, 11.06);
		potentials.emplace(3, 19.18);
	}
	if (elementSymbol == "Hf")
	{
		potentials.emplace(1, 6.82);
		potentials.emplace(2, 14.9);
		potentials.emplace(3, 23.3);
		potentials.emplace(4, 33.33);
	}
	if (elementSymbol == "Ta")
	{
		potentials.emplace(1, 7.55);
		potentials.emplace(2, 16.2);
		potentials.emplace(3, 22.0);
		potentials.emplace(4, 33.0);
		potentials.emplace(5, 45.0);
	}
	if (elementSymbol == "W")
	{
		potentials.emplace(1, 7.86);
		potentials.emplace(2, 16.1);
		potentials.emplace(3, 26.0);
		potentials.emplace(4, 38.2);
		potentials.emplace(5, 51.6);
		potentials.emplace(6, 64.8);
	}
	if (elementSymbol == "Re")
	{
		potentials.emplace(1, 7.83);
		potentials.emplace(2, 16.6);
		potentials.emplace(3, 26.0);
		potentials.emplace(4, 38.0);
		potentials.emplace(5, 51.0);
		potentials.emplace(6, 64.0);
	}
	if (elementSymbol == "Os")
	{
		potentials.emplace(1, 8.44);
		potentials.emplace(2, 17.0);
		potentials.emplace(3, 25.0);
		potentials.emplace(4, 40.0);
		potentials.emplace(5, 54.0);
		potentials.emplace(6, 68.0);
	}
	if (elementSymbol == "Ir")
	{
		potentials.emplace(1, 8.97);
		potentials.emplace(2, 17.0);
		potentials.emplace(3, 27.0);
		potentials.emplace(4, 39.0);
		potentials.emplace(5, 57.0);
	}
	if (elementSymbol == "Pt")
	{
		potentials.emplace(1, 8.96);
		potentials.emplace(2, 18.6);
		potentials.emplace(3, 28.0);
		potentials.emplace(4, 41.0);
		potentials.emplace(5, 55.0);
	}
	if (elementSymbol == "Au")
	{
		potentials.emplace(1, 9.22);
		potentials.emplace(2, 20.5);
	}
	if (elementSymbol == "Hg")
	{
		potentials.emplace(1, 10.44);
		potentials.emplace(2, 18.76);
	}
	if (elementSymbol == "Tl")
	{
		potentials.emplace(1, 6.11);
		potentials.emplace(2, 20.43);
		potentials.emplace(3, 29.83);
	}
	if (elementSymbol == "Pb")
	{
		potentials.emplace(1, 7.42);
		potentials.emplace(2, 15.03);
		potentials.emplace(3, 31.94);
		potentials.emplace(4, 42.32);
	}
	if (elementSymbol == "Bi")
	{
		potentials.emplace(1, 7.28);
		potentials.emplace(2, 16.7);
		potentials.emplace(3, 25.56);
		potentials.emplace(4, 45.3);
		potentials.emplace(5, 56.0);
	}
	if (elementSymbol == "Ac")
	{
		potentials.emplace(1, 5.17);
		potentials.emplace(2, 12.1);
		potentials.emplace(3, 27.1);
	}
	if (elementSymbol == "Th")
	{
		potentials.emplace(1, 6.31);
		potentials.emplace(2, 11.9);
		potentials.emplace(3, 20.0);
		potentials.emplace(4, 28.8);
	}
	if (elementSymbol == "Pa")
		potentials.emplace(1, 5.89);
	if (elementSymbol == "U")
	{
		potentials.emplace(1, 6.05);
		potentials.emplace(2, 10.6);
		potentials.emplace(3, 17.9);
		potentials.emplace(4, 31.4);
	}


	if (elementSymbol == "Ce" || elementSymbol == "Pr" || elementSymbol == "Nd" || elementSymbol == "Pm" || elementSymbol == "Sm" || elementSymbol == "Eu" || elementSymbol == "Gd" || elementSymbol == "Tb")
	{
		potentials.emplace(1, 5.58);
		potentials.emplace(2, 11.06);
		potentials.emplace(3, 39.18);
	}

	if (elementSymbol == "Dy" || elementSymbol == "Ho" || elementSymbol == "Er" || elementSymbol == "Tm" || elementSymbol == "Yb" || elementSymbol == "Lu")
	{
		potentials.emplace(1, 5.58);
		potentials.emplace(2, 11.06);
		potentials.emplace(3, 39.18);
	}

	return potentials;
}

double OxidationFeature::getElectronegativity(const std::string& elementSymbol)
{
	if (elementSymbol == "N")
		return 2.9;
	if (elementSymbol == "O")
		return 3.2;
	if (elementSymbol == "F")
		return 3.5;
	if (elementSymbol == "P")
		return 2.1;
	if (elementSymbol == "S")
		return 2.5;
	if (elementSymbol == "Cl")
		return 2.7;
	if (elementSymbol == "As")
		return 2.0;
	if (elementSymbol == "Se")
		return 2.4;
	if (elementSymbol == "Br")
		return 2.6;
	if (elementSymbol == "Sb")
		return 1.65;
	if (elementSymbol == "Te")
		return 1.9;
	if (elementSymbol == "I")
		return 2.3;

	return 0.0;
}

// Static methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
