#include "SpecifyCoordinationFeatures.h"

#include <cmath>

#include "MpiController.h"

#include "FileReader.h"
#include "LogWriter.h"
#include "CifFileExcavator.h"

#include "CrystallographicInformation.h"
#include "StructureFingerprint.h"

using namespace MathematicalCrystalChemistry::Runtime::Analysis::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

std::size_t SpecifyCoordinationFeatures::m_numExcavatedCifFile;
std::mutex SpecifyCoordinationFeatures::m_numMutex;


SpecifyCoordinationFeatures::SpecifyCoordinationFeatures(const size_type threadRank)
	: m_threadRank{ threadRank }
	, m_mpiWorkDirectoryPath{}
	, m_coordinationFeatureSpecifyOptions{}
	, m_structureOptimalityAppraiser{}
{
}

SpecifyCoordinationFeatures::SpecifyCoordinationFeatures(const size_type threadRank, const CoordinationFeatureSpecifyOptions& coordinationFeatureSpecifyOptions)
	: m_threadRank{ threadRank }
	, m_mpiWorkDirectoryPath{ coordinationFeatureSpecifyOptions.workDirectoryPath() }
	, m_coordinationFeatureSpecifyOptions{ coordinationFeatureSpecifyOptions }
	, m_structureOptimalityAppraiser{}
{
	m_mpiWorkDirectoryPath /= OperatingSystem::Configuration::MpiController::getMpiDirectoryName();

	m_structureOptimalityAppraiser.setMaxReoptimizing(m_coordinationFeatureSpecifyOptions.maxReoptimizing());
	m_structureOptimalityAppraiser.setMaxAtomFractuationRatio(m_coordinationFeatureSpecifyOptions.maxAtomFractuationRatio());
	m_structureOptimalityAppraiser.setStructureDesignRecordOptions(m_coordinationFeatureSpecifyOptions.structureDesignRecordOptions());
	m_structureOptimalityAppraiser.setStructuralOptimizationParameters(m_coordinationFeatureSpecifyOptions.structuralOptimizationParameters());
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void SpecifyCoordinationFeatures::operator()()
{
	OperatingSystem::IO::LogWriter logWriter;
	logWriter.setDirectoryPath(m_mpiWorkDirectoryPath);

	ChemToolkit::Crystallography::FileIO::CifFileExcavator cifFileExcavator{ m_threadRank };
	cifFileExcavator.initialize(m_coordinationFeatureSpecifyOptions.cifFilesDirectoryPath());


	while (cifFileExcavator.cifFileReader())
	{
		addNumExcavatedCifFile();
		logWriter.write(writeCifFileExcavationLogMessage(m_threadRank, cifFileExcavator.cifFileReader().filePath()));

		ModelCrystalStructure inputModelCrystalStructure;
		std::filesystem::path structureWorkDirectoryPath;


		try
		{
			ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation crystallographicInformation = cifFileExcavator.cifFileReader().readStructure();
			crystallographicInformation.toCrystalStructure(inputModelCrystalStructure);
			initializeInputModelCrystalStructure(inputModelCrystalStructure);


			if (!(isSurveyedStructure(inputModelCrystalStructure, cifFileExcavator.cifFileReader().filePath())))
			{
				structureWorkDirectoryPath = getStructureWorkDirectoryPath(cifFileExcavator.cifFileReader().filePath());
				createDirectoriesIfNecessary(structureWorkDirectoryPath);

				std::filesystem::path copyCifFilePath = structureWorkDirectoryPath;
				copyCifFilePath /= cifFileExcavator.cifFileReader().filename();
				cifFileExcavator.cifFileReader().copyFile(copyCifFilePath, OperatingSystem::IO::FileCopyOptions::skip_existing);

				StructureReoptimizer::ReoptimizationResult surveyReoptimizationResult = surveyStructure(inputModelCrystalStructure, structureWorkDirectoryPath);
				std::filesystem::path resultDirectoryPath = getResultDirectoryPath(inputModelCrystalStructure, cifFileExcavator.cifFileReader().filePath(), surveyReoptimizationResult);
				createDirectoriesIfNecessary(resultDirectoryPath);

				moveDirectory(structureWorkDirectoryPath, resultDirectoryPath);
				std::filesystem::remove_all(structureWorkDirectoryPath);
			}
		}

		catch (const std::exception& e)
		{
			std::filesystem::path exceptionDirectoryPath = getExceptionDirectoryPath(inputModelCrystalStructure, cifFileExcavator.cifFileReader().filePath());
			createDirectoriesIfNecessary(exceptionDirectoryPath);
			moveDirectory(structureWorkDirectoryPath, exceptionDirectoryPath);

			logWriter.write(e);
		}

		cifFileExcavator.excavateNextCifFile();
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

StructureReoptimizer::ReoptimizationResult SpecifyCoordinationFeatures::surveyStructure(ModelCrystalStructure inputModelCrystalStructure, const std::filesystem::path& workDirectoryPath) const
{
	OperatingSystem::IO::LogWriter logWriter;
	logWriter.setDirectoryPath(workDirectoryPath);

	try
	{
		inputModelCrystalStructure.createTrackingAtomPairs();
		inputModelCrystalStructure.createGeometricalConstraints();
		StructureFingerprint inputStructureFingerprint = inputModelCrystalStructure.getStructureFingerprint();
		{
			std::filesystem::path inputFingerprintFilePath = workDirectoryPath;
			inputFingerprintFilePath /= "input_crystal_structure_fingerprint.txt";
			inputStructureFingerprint.report(inputFingerprintFilePath);

			std::filesystem::path structureInfoFilePath = workDirectoryPath;
			structureInfoFilePath /= "input_crystal_structure_info.txt";
			inputModelCrystalStructure.createStructureInformationFile(structureInfoFilePath);
		}

		StructureReoptimizer::ReoptimizationResult totalReoptimizationResult = StructureReoptimizer::ReoptimizationResult::infeasible;
		{
			size_type coordinationFeatureIndex = 0;

			CoordinationFeatureAssigner coordinationFeatureAssigner;
			coordinationFeatureAssigner.setNumbersRanges(inputModelCrystalStructure, m_coordinationFeatureSpecifyOptions.upperIonicBondRadiusNumberExtensionWidth(), m_coordinationFeatureSpecifyOptions.repulsiveIonicRadiusNumberExtensionWidth());
			coordinationFeatureAssigner.initializeRadiusNumbers();


			while (true)
			{
				++coordinationFeatureIndex;
				const std::filesystem::path featuredStructureDirectoryPath = getFeaturedStructureDirectoryPath(workDirectoryPath, coordinationFeatureIndex);
				createDirectoriesIfNecessary(featuredStructureDirectoryPath);


				try
				{
					ModelCrystalStructure modelCrystalStructure = coordinationFeatureAssigner.getModelCrystalStructure(inputModelCrystalStructure);
					StructureReoptimizer::ReoptimizationResult reoptimizationResult = m_structureOptimalityAppraiser.appraise(modelCrystalStructure, inputStructureFingerprint, featuredStructureDirectoryPath);

					upgradeTotalReoptimizationResult(totalReoptimizationResult, reoptimizationResult);
					reportReoptimizationResult(reoptimizationResult, featuredStructureDirectoryPath);
				}

				catch (const std::exception& e)
				{
					logWriter.write(e);

					std::filesystem::path exceptionDirectoryPath = workDirectoryPath;
					exceptionDirectoryPath /= "Exceptions";
					createDirectoriesIfNecessary(exceptionDirectoryPath);
					exceptionDirectoryPath /= std::filesystem::relative(featuredStructureDirectoryPath, featuredStructureDirectoryPath.parent_path());

					if (std::filesystem::exists(exceptionDirectoryPath))
						std::filesystem::remove_all(exceptionDirectoryPath);

					std::filesystem::rename(featuredStructureDirectoryPath, exceptionDirectoryPath);
					std::filesystem::remove_all(featuredStructureDirectoryPath);
				}


				if ((totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::isotypic) || (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::same_symmetry))
					break;
				else
				{
					if (!(coordinationFeatureAssigner.advanceRadiusNumbers()))
						break;
				}
			}
		}

		return totalReoptimizationResult;
	}


	catch (const std::exception& e)
	{
		logWriter.write(e);
		throw;
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

void SpecifyCoordinationFeatures::reportReoptimizationResult(const StructureReoptimizer::ReoptimizationResult& reoptimizationResult, const std::filesystem::path& featuredStructureDirectoryPath) const
{
	if (!(m_coordinationFeatureSpecifyOptions.needInfeasibleResult()) && (reoptimizationResult == StructureReoptimizer::ReoptimizationResult::infeasible))
		std::filesystem::remove_all(featuredStructureDirectoryPath);

	else
	{
		std::filesystem::path appraisedDirectoryPath = getAppraisedDirectoryPath(featuredStructureDirectoryPath, reoptimizationResult);
		createDirectoriesIfNecessary(appraisedDirectoryPath);

		moveDirectory(featuredStructureDirectoryPath, appraisedDirectoryPath);
		std::filesystem::remove_all(featuredStructureDirectoryPath);
	}
}

void SpecifyCoordinationFeatures::upgradeTotalReoptimizationResult(StructureReoptimizer::ReoptimizationResult& totalReoptimizationResult, const StructureReoptimizer::ReoptimizationResult& reoptimizationResult) const
{
	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::homeotypic)
	{
		if (reoptimizationResult == StructureReoptimizer::ReoptimizationResult::isotypic)
			totalReoptimizationResult = reoptimizationResult;
	}

	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::same_symmetry)
	{
		if ((reoptimizationResult == StructureReoptimizer::ReoptimizationResult::isotypic) || (reoptimizationResult == StructureReoptimizer::ReoptimizationResult::homeotypic))
			totalReoptimizationResult = reoptimizationResult;
	}

	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::same_space_group_type)
	{
		if ((reoptimizationResult != StructureReoptimizer::ReoptimizationResult::optimal) && (reoptimizationResult != StructureReoptimizer::ReoptimizationResult::infeasible))
			totalReoptimizationResult = reoptimizationResult;
	}

	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::optimal)
	{
		if (reoptimizationResult != StructureReoptimizer::ReoptimizationResult::infeasible)
			totalReoptimizationResult = reoptimizationResult;
	}

	if (totalReoptimizationResult == StructureReoptimizer::ReoptimizationResult::infeasible)
		totalReoptimizationResult = reoptimizationResult;
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
