#include "StructuralOptimizationParameters.h"

#include <limits>

#include "LengthCasting.h"
#include "FileReader.h"

#include "ModelAtomTracker.h"
#include "StructureCustomizerBase.h"

using namespace MathematicalCrystalChemistry::Design::Optimization;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

StructuralOptimizationParameters::StructuralOptimizationParameters()
	: m_maxTotalStructureOptimizing{ 0 }
	, m_atomPairTrackersLifetime{ std::numeric_limits<size_type>::max() }
	, m_constraintsLifetime{ std::numeric_limits<size_type>::max() }
	, m_minMaxConstraintsLifetime{ std::numeric_limits<size_type>::max() }
	, m_unitCellLifetime{ std::numeric_limits<size_type>::max() }
	, m_atomPairTrackingCutoffRatio{ MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::defaultAtomPairTrackingCutoffRatio() }
	, m_pressure{ MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase::defaultCommonPressure() }
	, m_attractiveForceConstant{ MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase::defaultCommonAttractiveForceConstant() }
	, m_repulsiveForceConstant{ MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase::defaultCommonRepulsiveForceConstant() }
	, m_globalRelaxationParameters{}
	, m_localRelaxationParameters{}
	, m_preciseRelaxationParameters{}
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void StructuralOptimizationParameters::initialize() noexcept
{
	m_maxTotalStructureOptimizing = 0;
	m_atomPairTrackersLifetime = std::numeric_limits<size_type>::max();
	m_constraintsLifetime = std::numeric_limits<size_type>::max();
	m_minMaxConstraintsLifetime = std::numeric_limits<size_type>::max();
	m_unitCellLifetime = std::numeric_limits<size_type>::max();

	m_atomPairTrackingCutoffRatio = MathematicalCrystalChemistry::StructureModel::Components::Internal::ModelAtomTracker::defaultAtomPairTrackingCutoffRatio();
	m_pressure = MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase::defaultCommonPressure();
	m_attractiveForceConstant = MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase::defaultCommonAttractiveForceConstant();
	m_repulsiveForceConstant = MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase::defaultCommonRepulsiveForceConstant();

	m_globalRelaxationParameters.initialize();
	m_localRelaxationParameters.initialize();
	m_preciseRelaxationParameters.initialize();
}

void StructuralOptimizationParameters::initializeFromInputFileTexts(const std::string& inputTexts)
{
	initialize();
	using OperatingSystem::IO::FileReader;

	FileReader::tryReadData(inputTexts, "Maximum.Number.of.Total.Structural.Optimization", m_maxTotalStructureOptimizing);
	FileReader::tryReadData(inputTexts, "Atom.Pair.Trackers.Lifetime", m_atomPairTrackersLifetime);
	FileReader::tryReadData(inputTexts, "Constraints.Lifetime", m_constraintsLifetime);
	FileReader::tryReadData(inputTexts, "MinMax.Constraints.Lifetime", m_minMaxConstraintsLifetime);
	FileReader::tryReadData(inputTexts, "Unit.Cell.Lifetime", m_unitCellLifetime);

	FileReader::tryReadData(inputTexts, "Atom.Pair.Tracking.Cutoff.Ratio", m_atomPairTrackingCutoffRatio);
	FileReader::tryReadData(inputTexts, "Pressure", m_pressure);
	FileReader::tryReadData(inputTexts, "Attractive.Force.Constant", m_attractiveForceConstant);
	FileReader::tryReadData(inputTexts, "Repulsive.Force.Constant", m_repulsiveForceConstant);

	initializeGlobalStructuralRelaxationParameters(inputTexts);
	initializeLocalStructuralRelaxationParameters(inputTexts);
	initializePreciseStructuralRelaxationParameters(inputTexts);

	validateInitializedValues();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void StructuralOptimizationParameters::initializeGlobalStructuralRelaxationParameters(const std::string& inputTexts)
{
	std::string parameterTexts;
	{
		for (const auto& lineData : OperatingSystem::IO::FileReader::readDataEnumeration(inputTexts, "Global.Structural.Relaxation.Parameters"))
		{
			std::string lineText;
			{
				for (const auto& text : lineData)
				{
					lineText += text;
					lineText += "  ";
				}

				lineText.pop_back();
				lineText.pop_back();
			}

			parameterTexts += lineText;
			parameterTexts += "\n";
		}
	}

	m_globalRelaxationParameters.initializeFromInputFileTexts(parameterTexts);
}

void StructuralOptimizationParameters::initializeLocalStructuralRelaxationParameters(const std::string& inputTexts)
{
	std::string parameterTexts;
	{
		for (const auto& lineData : OperatingSystem::IO::FileReader::readDataEnumeration(inputTexts, "Local.Structural.Relaxation.Parameters"))
		{
			std::string lineText;
			{
				for (const auto& text : lineData)
				{
					lineText += text;
					lineText += "  ";
				}

				lineText.pop_back();
				lineText.pop_back();
			}

			parameterTexts += lineText;
			parameterTexts += "\n";
		}
	}

	m_localRelaxationParameters.initializeFromInputFileTexts(parameterTexts);
}

void StructuralOptimizationParameters::initializePreciseStructuralRelaxationParameters(const std::string& inputTexts)
{
	std::string parameterTexts;
	{
		for (const auto& lineData : OperatingSystem::IO::FileReader::readDataEnumeration(inputTexts, "Precise.Structural.Relaxation.Parameters"))
		{
			std::string lineText;
			{
				for (const auto& text : lineData)
				{
					lineText += text;
					lineText += "  ";
				}

				lineText.pop_back();
				lineText.pop_back();
			}

			parameterTexts += lineText;
			parameterTexts += "\n";
		}
	}

	m_preciseRelaxationParameters.initializeFromInputFileTexts(parameterTexts);
}

void StructuralOptimizationParameters::validateInitializedValues() const
{
	if (m_pressure < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "\"m_pressure\" is less than zero."));
	}

	else
	{
		if (m_atomPairTrackingCutoffRatio < 0.0)
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "\"m_atomPairTrackingCutoffRatio\" is less than zero."));
		}

		else
		{
			if (m_attractiveForceConstant < 0.0)
			{
				OperatingSystem::Configuration::ReportException reportException;
				throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "\"m_attractiveForceConstant\" is less than zero."));
			}

			else
			{
				if (0.0 < m_repulsiveForceConstant)
				{
					OperatingSystem::Configuration::ReportException reportException;
					throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "\"m_repulsiveForceConstant\" is more than zero."));
				}
			}
		}
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
