#include "StructuralOptimizer.h"

#include "StructuralRelaxer.h"

using namespace MathematicalCrystalChemistry::Design::Optimization;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

StructuralOptimizer::StructuralOptimizer()
	: m_structuralOptimizationParameters{}
{
}

StructuralOptimizer::StructuralOptimizer(const StructuralOptimizationParameters& structuralOptimizationParameters)
	: m_structuralOptimizationParameters{ structuralOptimizationParameters }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void StructuralOptimizer::optimize(ModelCrystalStructure& modelCrystalStructure) const
{
	initializeTimers();
	modelCrystalStructure.normalizeFractionalCoordinates();
	modelCrystalStructure.createTrackingAtomPairs();
	modelCrystalStructure.createGeometricalConstraints();


	while (m_totalStructureOptimizing < m_structuralOptimizationParameters.maxTotalStructureOptimizing())
	{
		if (applyGlobalStructuralRelaxation(modelCrystalStructure))
		{
			if (applyLocalStructuralRelaxation(modelCrystalStructure))
			{
				if (applyPreciseStructuralRelaxation(modelCrystalStructure))
					break;

				else
				{
					manageUnitCell(modelCrystalStructure);
					manageGeometricalConstraints(modelCrystalStructure);
					modelCrystalStructure.substituteChemicalBonds(m_structuralOptimizationParameters.preciseRelaxationParameters().feasibleConstraintErrorRate());
				}
			}

			else
			{
				manageUnitCell(modelCrystalStructure);
				manageGeometricalConstraints(modelCrystalStructure);
				modelCrystalStructure.substituteChemicalBonds(m_structuralOptimizationParameters.localRelaxationParameters().feasibleConstraintErrorRate());
			}
		}

		else
		{
			manageUnitCell(modelCrystalStructure);
			manageGeometricalConstraints(modelCrystalStructure);
		}
	}


	initializeTimers();
}

void StructuralOptimizer::optimize(ModelCrystalStructure& modelCrystalStructure, StructureDesignRecorder& structureDesignRecorder) const
{
	initializeTimers();
	modelCrystalStructure.normalizeFractionalCoordinates();
	modelCrystalStructure.createTrackingAtomPairs();
	modelCrystalStructure.createGeometricalConstraints();


	while (m_totalStructureOptimizing < m_structuralOptimizationParameters.maxTotalStructureOptimizing())
	{
		if (applyGlobalStructuralRelaxation(modelCrystalStructure, structureDesignRecorder))
		{
			if (applyLocalStructuralRelaxation(modelCrystalStructure, structureDesignRecorder))
			{
				if (applyPreciseStructuralRelaxation(modelCrystalStructure, structureDesignRecorder))
					break;

				else
				{
					manageUnitCell(modelCrystalStructure);
					manageGeometricalConstraints(modelCrystalStructure);
					modelCrystalStructure.substituteChemicalBonds(m_structuralOptimizationParameters.preciseRelaxationParameters().feasibleConstraintErrorRate());
				}
			}

			else
			{
				manageUnitCell(modelCrystalStructure);
				manageGeometricalConstraints(modelCrystalStructure);
				modelCrystalStructure.substituteChemicalBonds(m_structuralOptimizationParameters.localRelaxationParameters().feasibleConstraintErrorRate());
			}
		}

		else
		{
			manageUnitCell(modelCrystalStructure);
			manageGeometricalConstraints(modelCrystalStructure);
		}

		structureDesignRecorder.addGeometricalConstraintsOptimizing();
	}


	initializeTimers();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

bool StructuralOptimizer::manageGeometricalConstraints(ModelCrystalStructure& modelCrystalStructure) const
{
	if (m_atomPairTrackersTimer < m_structuralOptimizationParameters.atomPairTrackersLifetime())
	{
		if (m_constraintsTimer < m_structuralOptimizationParameters.constraintsLifetime())
		{
			if (m_structuralOptimizationParameters.minMaxConstraintsLifetime() <= m_minMaxConstraintsTimer)
			{
				modelCrystalStructure.createChemicalBonds();
				m_minMaxConstraintsTimer = 0;

				return true;
			}
		}

		else
		{
			modelCrystalStructure.createGeometricalConstraints();
			m_constraintsTimer = 0;
			m_minMaxConstraintsTimer = 0;

			return true;
		}
	}

	else
	{
		modelCrystalStructure.normalizeFractionalCoordinates();
		modelCrystalStructure.createTrackingAtomPairs();
		modelCrystalStructure.createGeometricalConstraints();

		m_atomPairTrackersTimer = 0;
		m_constraintsTimer = 0;
		m_minMaxConstraintsTimer = 0;

		return true;
	}

	return false;
}

bool StructuralOptimizer::manageUnitCell(ModelCrystalStructure& modelCrystalStructure) const
{
	if (modelCrystalStructure.getLowerIonicPackingFraction() < 2.0)
	{
		if (m_structuralOptimizationParameters.unitCellLifetime() < m_unitCellTimer)
		{
			size_type numAtom = modelCrystalStructure.atomicArrangement().size();
			{
				modelCrystalStructure.refineStructure(0.00000000001);

				if (numAtom != modelCrystalStructure.atomicArrangement().size())
					throw std::runtime_error("Spglib reduces the number of atoms.\n");
			}

			modelCrystalStructure.createTrackingAtomPairs();
			modelCrystalStructure.createGeometricalConstraints();

			m_atomPairTrackersTimer = 0;
			m_constraintsTimer = 0;
			m_minMaxConstraintsTimer = 0;
			m_unitCellTimer = 0;

			return true;
		}

		return false;
	}

	else
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "manageUnitCell", "The volume of unit cell is too small."));
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
