#include "StructuralRelaxationParameters.h"

#include <limits>

#include "LengthCasting.h"
#include "FileReader.h"

using namespace MathematicalCrystalChemistry::Design::Optimization;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

StructuralRelaxationParameters::StructuralRelaxationParameters()
	: m_maxStructureOptimizing{ 0 }
	, m_initialMaxAtomicDisplacement{ 0.0 }
	, m_finalMaxAtomicDisplacement{ 0.0 }
	, m_maxUnitCellDisplacementFactor{ 0.0 }
	, m_feasibleConstraintErrorRate{ std::numeric_limits<double>::max() }
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void StructuralRelaxationParameters::initialize() noexcept
{
	m_maxStructureOptimizing = 0;

	m_initialMaxAtomicDisplacement = 0.0;
	m_finalMaxAtomicDisplacement = 0.0;
	m_maxUnitCellDisplacementFactor = 0.0;

	m_feasibleConstraintErrorRate = std::numeric_limits<double>::max();
}

void StructuralRelaxationParameters::initializeFromInputFileTexts(const std::string& inputTexts)
{
	initialize();
	using OperatingSystem::IO::FileReader;

	FileReader::tryReadData(inputTexts, "Maximum.Number.of.Structural.Optimization", m_maxStructureOptimizing);
	FileReader::tryReadData(inputTexts, "Feasible.Error.Rate.of.Geometrical.Constraints", m_feasibleConstraintErrorRate);
	{
		double angstromDisplacement = 0.0;
		FileReader::tryReadData(inputTexts, "Initial.Maximum.Atomic.Displacement", angstromDisplacement);
		m_initialMaxAtomicDisplacement = MathToolkit::UnitConversion::LengthCasting::cast<MathToolkit::UnitConversion::LengthCasting::Unit::Angstrom, MathToolkit::UnitConversion::LengthCasting::Unit::AtomicUnit>(angstromDisplacement);
	}
	{
		double angstromDisplacement = 0.0;
		FileReader::tryReadData(inputTexts, "Final.Maximum.Atomic.Displacement", angstromDisplacement);
		m_finalMaxAtomicDisplacement = MathToolkit::UnitConversion::LengthCasting::cast<MathToolkit::UnitConversion::LengthCasting::Unit::Angstrom, MathToolkit::UnitConversion::LengthCasting::Unit::AtomicUnit>(angstromDisplacement);
	}
	FileReader::tryReadData(inputTexts, "Maximum.Unit.Cell.Displacement.Factor", m_maxUnitCellDisplacementFactor);

	validateInitializedValues();
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void StructuralRelaxationParameters::validateInitializedValues() const
{
	if (m_feasibleConstraintErrorRate < 0.0)
	{
		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "Feasible error rate of geometrical constraints is less than zero."));
	}

	else
	{
		if (m_maxUnitCellDisplacementFactor < 0.0)
		{
			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "Displacement factor of unit cell is less than zero."));
		}

		else
		{
			if (m_initialMaxAtomicDisplacement < 0.0)
			{
				OperatingSystem::Configuration::ReportException reportException;
				throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "Initial maximum displacement of atoms is less than zero."));
			}

			else
			{
				if (m_finalMaxAtomicDisplacement < 0.0)
				{
					OperatingSystem::Configuration::ReportException reportException;
					throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "Final maximum displacement of atoms is less than zero."));
				}

				else
				{
					if (m_initialMaxAtomicDisplacement < m_finalMaxAtomicDisplacement)
					{
						OperatingSystem::Configuration::ReportException reportException;
						throw std::invalid_argument(reportException(typeid(*this), "validateInitializedValues", "Initial maximum displacement of atoms is less than the final maximum displacement."));
					}
				}
			}
		}
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
