#include "StructureClassifier.h"

#include <set>
#include <map>

#include "FileReader.h"
#include "FileWriter.h"

#include "CifFileReader.h"
#include "CifFileWriter.h"

using namespace MathematicalCrystalChemistry::IO;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

std::string StructureClassifier::m_fingerprintFilename{ "fingerprint.txt" };
std::mutex StructureClassifier::m_fingerprintMutex;
std::mutex StructureClassifier::m_directoryMutex;


StructureClassifier::StructureClassifier()
	: m_needConventionalStructure{ true }
	, m_outputDirectoryPath{}
	, m_structureReporter{}
{
}

StructureClassifier::StructureClassifier(const std::filesystem::path& outputDirectoryPath)
	: m_needConventionalStructure{ true }
	, m_outputDirectoryPath{ outputDirectoryPath }
	, m_structureReporter{}
{
	if (!(std::filesystem::is_directory(m_outputDirectoryPath)))
	{
		std::string errorMessage;
		{
			errorMessage += "\"";
			errorMessage += m_outputDirectoryPath.generic_string();
			errorMessage += "\" is not a path to a directory.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", errorMessage));
	}
}

StructureClassifier::StructureClassifier(const std::filesystem::path& outputDirectoryPath, const bool needConventionalStructure)
	: m_needConventionalStructure{ needConventionalStructure }
	, m_outputDirectoryPath{ outputDirectoryPath }
	, m_structureReporter{}
{
	if (!(std::filesystem::is_directory(m_outputDirectoryPath)))
	{
		std::string errorMessage;
		{
			errorMessage += "\"";
			errorMessage += m_outputDirectoryPath.generic_string();
			errorMessage += "\" is not a path to a directory.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "constructor", errorMessage));
	}
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

std::filesystem::path StructureClassifier::classify(const ModelCrystalStructure& modelCrystalStructure, const std::string& outputDirectoryName) const
{
	ModelCrystalStructure primitiveStructure = modelCrystalStructure;
	{
		primitiveStructure.refineStructure();
		primitiveStructure.updateSpaceGroupInfo();
		primitiveStructure.createTrackingAtomPairs();
		primitiveStructure.createGeometricalConstraints();
	}

	ModelCrystalStructure conventionalStructure = modelCrystalStructure;
	{
		conventionalStructure.conventionalizeStructure();
		conventionalStructure.createTrackingAtomPairs();
		conventionalStructure.createGeometricalConstraints();
	}

	StructureFingerprint structureFingerprint = primitiveStructure.getStructureFingerprint();
	std::filesystem::path structureDirectoryPath = getStructureDirectoryPath(primitiveStructure, structureFingerprint);
	{
		if (!(outputDirectoryName.empty()))
			structureDirectoryPath /= outputDirectoryName;
	}
	createDirectoryIfNecessary(structureDirectoryPath);


	m_structureReporter.report(primitiveStructure, structureDirectoryPath);
	std::string primitiveFilename = primitiveStructure.getChemicalComposition().name();
	primitiveFilename += ".cif";
	outputModelCrystalStructure(primitiveStructure, structureDirectoryPath, primitiveFilename);

	std::string conventionalFilename{ "conventional-" };
	conventionalFilename += primitiveFilename;
	outputModelCrystalStructure(conventionalStructure, structureDirectoryPath, conventionalFilename);


	return structureDirectoryPath;
}

std::filesystem::path StructureClassifier::classify(const std::filesystem::path& cifFilePath) const
{
	if (std::filesystem::is_regular_file(cifFilePath) && ChemToolkit::Crystallography::FileIO::CifFileReader::isCorrectExtension(cifFilePath))
	{
		ChemToolkit::Crystallography::FileIO::CifFileReader cifFileReader{ cifFilePath };
		ChemToolkit::Crystallography::SpaceGroups::CrystallographicInformation crystallographicInformation = cifFileReader.readStructure();

		ModelCrystalStructure primitiveStructure;
		{
			crystallographicInformation.toCrystalStructure(primitiveStructure);

			primitiveStructure.removeSiteDuplications();
			primitiveStructure.assignCoordinationFeaturesFromElementSymbols();
			primitiveStructure.optimizeOxidationStates();
			primitiveStructure.assignCoordinationFeaturesFromElementSymbols();

			primitiveStructure.refineStructure();
			primitiveStructure.updateSpaceGroupInfo();

			primitiveStructure.createTrackingAtomPairs();
			primitiveStructure.createGeometricalConstraints();
		}
		StructureFingerprint structureFingerprint = primitiveStructure.getStructureFingerprint();

		std::filesystem::path structureDirectoryPath = getStructureDirectoryPath(primitiveStructure, structureFingerprint);
		{
			structureDirectoryPath /= primitiveStructure.getChemicalComposition().name();
			createDirectoryIfNecessary(structureDirectoryPath);

			m_structureReporter.report(primitiveStructure, structureDirectoryPath);
			std::filesystem::path copyFilePath = structureDirectoryPath;
			copyFilePath /= cifFilePath.filename();
			cifFileReader.copyFile(copyFilePath, OperatingSystem::IO::FileCopyOptions::skip_existing);
		}


		if (m_needConventionalStructure)
		{
			ModelCrystalStructure conventionalStructure = primitiveStructure;
			{
				conventionalStructure.conventionalizeStructure();
				conventionalStructure.createTrackingAtomPairs();
				conventionalStructure.createGeometricalConstraints();
			}

			std::string conventionalFilename{ "conventional-" };
			conventionalFilename += cifFilePath.filename().string();
			outputModelCrystalStructure(conventionalStructure, structureDirectoryPath, conventionalFilename);
		}

		return structureDirectoryPath;
	}

	else
	{
		std::string errorMessage;
		{
			errorMessage += "The path of \"";
			errorMessage += cifFilePath.generic_string();
			errorMessage += "\" is not a path to a cif file.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::invalid_argument(reportException(typeid(*this), "classify", errorMessage));
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void StructureClassifier::outputModelCrystalStructure(const ModelCrystalStructure& modelCrystalStructure, const std::filesystem::path& directoryPath, const std::string filename) const
{
	std::filesystem::path cifFilePath = directoryPath;
	cifFilePath /= filename;

	ChemToolkit::Crystallography::FileIO::CifFileWriter cifFileWriter{ cifFilePath };
	cifFileWriter.createFile(modelCrystalStructure, OperatingSystem::IO::FileCreateOptions::skip_existing);
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

void StructureClassifier::createFingerprintFile(const StructureFingerprint& structureFingerprint, const std::filesystem::path& directoryPath) const
{
	std::filesystem::path filePath = directoryPath;
	filePath /= m_fingerprintFilename;
	OperatingSystem::IO::FileWriter fileWriter{ filePath };

	fileWriter.createFile(OperatingSystem::IO::FileCreateOptions::skip_existing, structureFingerprint.getSymbol());
}

std::string StructureClassifier::readFingerprintFile(const std::filesystem::path& directoryPath) const
{
	std::filesystem::path filePath = directoryPath;
	filePath /= m_fingerprintFilename;
	OperatingSystem::IO::FileReader fileReader{ filePath };

	if (fileReader.isValid())
	{
		std::string fingerprintSymbol = fileReader.readFile();
		fingerprintSymbol.pop_back();
		return fingerprintSymbol;
	}

	else
		return std::string{};
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
