#include "StructureCreationReporter.h"

#include "FileReader.h"
#include "FileWriter.h"
#include "MpiController.h"

#include "StructureClassifier.h"

#include "StructureDesignRecorder.h"
#include "CreateStructures.h"

using namespace MathematicalCrystalChemistry::Runtime::Creation::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

StructureCreationReporter::StructureCreationReporter()
	: m_designedStructuresDirectoryPath{}
{
}

StructureCreationReporter::StructureCreationReporter(const std::filesystem::path& designedStructuresDirectoryPath)
	: m_designedStructuresDirectoryPath{ designedStructuresDirectoryPath }
{
	if (!(std::filesystem::is_directory(m_designedStructuresDirectoryPath)))
	{
		std::string errorMessage;
		{
			errorMessage += "\"";
			errorMessage += m_designedStructuresDirectoryPath.generic_string();
			errorMessage += "\" is not a directory path.";
		}

		OperatingSystem::Configuration::ReportException reportException;
		throw std::runtime_error(reportException(typeid(*this), "constructor", errorMessage));
	}
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void StructureCreationReporter::reportOptimalStructures() const
{
	std::filesystem::path mergedOptimaDirectoryPath = m_designedStructuresDirectoryPath;
	mergedOptimaDirectoryPath /= StructureDesignRecorder::optimalStructuresDirectoryName();
	createDirectoryIfNecessary(mergedOptimaDirectoryPath);


	for (const auto& directoryEntry : std::filesystem::directory_iterator(m_designedStructuresDirectoryPath))
	{
		std::filesystem::path mpiOptimaDirectoryPath = directoryEntry.path();

		if (isMpiDirectory(mpiOptimaDirectoryPath))
		{
			mpiOptimaDirectoryPath /= StructureDesignRecorder::optimalStructuresDirectoryName();

			if (std::filesystem::is_directory(mpiOptimaDirectoryPath))
			{
				for (const auto& fileEntry : std::filesystem::recursive_directory_iterator(mpiOptimaDirectoryPath))
				{
					std::filesystem::path filePath = fileEntry.path();

					if (isFingerprintFilePath(filePath))
					{
						std::string fingerprintSymbol = readFingerprintFile(filePath);

						if (!(fingerprintSymbol.empty()))
						{
							std::filesystem::path mergedStructureDirectoryPath = findMergedStructureDirectoryPath(fingerprintSymbol, filePath, mpiOptimaDirectoryPath, mergedOptimaDirectoryPath);
							moveToMergedDirectory(filePath.parent_path(), mergedStructureDirectoryPath);
						}
					}
				}
			}
		}
	}
}

void StructureCreationReporter::reportInfeasibleStructures() const
{
	std::filesystem::path mergedInfeasibleDirectoryPath = m_designedStructuresDirectoryPath;
	mergedInfeasibleDirectoryPath /= StructureDesignRecorder::infeasibleStructuresDirectoryName();
	createDirectoryIfNecessary(mergedInfeasibleDirectoryPath);


	for (const auto& directoryEntry : std::filesystem::directory_iterator(m_designedStructuresDirectoryPath))
	{
		std::filesystem::path mpiInfeasibleDirectoryPath = directoryEntry.path();

		if (isMpiDirectory(mpiInfeasibleDirectoryPath))
		{
			std::string mpiDirectoryName = getDirectoryName(mpiInfeasibleDirectoryPath);
			mpiInfeasibleDirectoryPath /= StructureDesignRecorder::infeasibleStructuresDirectoryName();


			if (std::filesystem::is_directory(mpiInfeasibleDirectoryPath))
			{
				for (const auto& categoryEntry : std::filesystem::directory_iterator(mpiInfeasibleDirectoryPath))
				{
					if (categoryEntry.is_directory())
					{
						std::filesystem::path newInfeasibleDirectoryPath = mergedInfeasibleDirectoryPath;
						newInfeasibleDirectoryPath /= mpiDirectoryName;
						createDirectoryIfNecessary(newInfeasibleDirectoryPath);

						newInfeasibleDirectoryPath /= getDirectoryName(categoryEntry.path());
						{
							if (std::filesystem::is_directory(newInfeasibleDirectoryPath))
								std::filesystem::remove_all(newInfeasibleDirectoryPath);
						}

						std::filesystem::rename(categoryEntry.path(), newInfeasibleDirectoryPath);
					}
				}
			}
		}
	}
}

void StructureCreationReporter::reportExceptionalStructures() const
{
	std::filesystem::path mergedExceptionalDirectoryPath = m_designedStructuresDirectoryPath;
	mergedExceptionalDirectoryPath /= StructureDesignRecorder::exceptionalStructuresDirectoryName();
	createDirectoryIfNecessary(mergedExceptionalDirectoryPath);


	for (const auto& directoryEntry : std::filesystem::directory_iterator(m_designedStructuresDirectoryPath))
	{
		std::filesystem::path mpiExceptionalDirectoryPath = directoryEntry.path();

		if (isMpiDirectory(mpiExceptionalDirectoryPath))
		{
			std::string mpiDirectoryName = getDirectoryName(mpiExceptionalDirectoryPath);
			mpiExceptionalDirectoryPath /= StructureDesignRecorder::exceptionalStructuresDirectoryName();


			if (std::filesystem::is_directory(mpiExceptionalDirectoryPath))
			{
				for (const auto& categoryEntry : std::filesystem::directory_iterator(mpiExceptionalDirectoryPath))
				{
					if (categoryEntry.is_directory())
					{
						std::filesystem::path newExceptionalDirectoryPath = mergedExceptionalDirectoryPath;
						newExceptionalDirectoryPath /= mpiDirectoryName;
						createDirectoryIfNecessary(newExceptionalDirectoryPath);

						newExceptionalDirectoryPath /= getDirectoryName(categoryEntry.path());
						{
							if (std::filesystem::is_directory(newExceptionalDirectoryPath))
								std::filesystem::remove_all(newExceptionalDirectoryPath);
						}

						std::filesystem::rename(categoryEntry.path(), newExceptionalDirectoryPath);
					}
				}
			}
		}
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

bool StructureCreationReporter::isMpiDirectory(const std::filesystem::path& directoryPath) const
{
	if (std::filesystem::is_directory(directoryPath))
	{
		if (OperatingSystem::Configuration::MpiController::isMpiDirectoryName(getDirectoryName(directoryPath)))
			return true;
		else
			return false;
	}

	else
		return false;
}

bool StructureCreationReporter::isFingerprintFilePath(const std::filesystem::path& filePath) const
{
	if (std::filesystem::is_regular_file(filePath))
	{
		if (filePath.filename().string() == MathematicalCrystalChemistry::IO::StructureClassifier::fingerprintFilename())
			return true;
		else
			return false;
	}

	else
		return false;
}

std::filesystem::path StructureCreationReporter::findMergedStructureDirectoryPath(const std::string& fingerprintSymbol, const std::filesystem::path& fingerprintFilePath, const std::filesystem::path& mpiOptimaDirectoryPath, const std::filesystem::path& mergedOptimaDirectoryPath) const
{
	std::filesystem::path spaceGroupDirectoryPath = fingerprintFilePath.parent_path().parent_path();
	std::filesystem::path mergedSpaceGroupDirectoryPath = mergedOptimaDirectoryPath;
	mergedSpaceGroupDirectoryPath /= std::filesystem::relative(spaceGroupDirectoryPath, mpiOptimaDirectoryPath);
	{
		if (!(std::filesystem::is_directory(mergedSpaceGroupDirectoryPath)))
		{
			if (!(std::filesystem::create_directories(mergedSpaceGroupDirectoryPath)))
			{
				std::string errorMessage;
				{
					errorMessage += "Could not create the directories of \"";
					errorMessage += mergedSpaceGroupDirectoryPath.generic_string();
					errorMessage += "\"";
				}

				OperatingSystem::Configuration::ReportException reportException;
				throw std::runtime_error(reportException(typeid(*this), "findMergedStructureDirectoryPath", errorMessage));
			}
		}
	}


	size_type numDirectory = 0;
	{
		for (const auto& directoryEntry : std::filesystem::directory_iterator(mergedSpaceGroupDirectoryPath))
		{
			if (directoryEntry.is_directory())
			{
				std::filesystem::path mergedStructureDirectoryPath = directoryEntry.path();

				for (const auto& fileEntry : std::filesystem::directory_iterator(mergedStructureDirectoryPath))
				{
					std::filesystem::path filePath = fileEntry.path();

					if (isFingerprintFilePath(filePath))
					{
						std::string mergedFingerprintSymbol = readFingerprintFile(filePath);

						if (!(mergedFingerprintSymbol.empty()))
						{
							if (fingerprintSymbol == mergedFingerprintSymbol)
								return mergedStructureDirectoryPath;
							else
								++numDirectory;
						}
					}
				}
			}
		}
	}

	std::filesystem::path mergedStructureDirectoryPath = mergedSpaceGroupDirectoryPath;
	{
		std::string directoryName;
		{
			directoryName += "Structure_";
			directoryName += std::to_string(numDirectory);
		}
		mergedStructureDirectoryPath /= directoryName;

		std::filesystem::remove_all(mergedStructureDirectoryPath);
		createDirectoryIfNecessary(mergedStructureDirectoryPath);
	}

	return mergedStructureDirectoryPath;
}

void StructureCreationReporter::moveToMergedDirectory(const std::filesystem::path& structureDirectoryPath, const std::filesystem::path& mergedStructureDirectoryPath) const
{
	for (const auto& fileEntry : std::filesystem::recursive_directory_iterator(structureDirectoryPath))
	{
		if (fileEntry.is_regular_file())
		{
			std::filesystem::path filePath = fileEntry.path();
			
			std::filesystem::path mergedFilePath = mergedStructureDirectoryPath;
			mergedFilePath /= std::filesystem::relative(filePath, structureDirectoryPath);
			{
				std::filesystem::path mergedDirectoryPath = mergedFilePath.parent_path();

				if (!(std::filesystem::is_directory(mergedDirectoryPath)))
				{
					if (!(std::filesystem::create_directories(mergedDirectoryPath)))
					{
						std::string errorMessage;
						{
							errorMessage += "Could not create the directories of \"";
							errorMessage += mergedDirectoryPath.generic_string();
							errorMessage += "\"";
						}

						OperatingSystem::Configuration::ReportException reportException;
						throw std::runtime_error(reportException(typeid(*this), "moveToMergedDirectory", errorMessage));
					}
				}
			}


			if (filePath.filename().string() == StructureDesignRecorder::resultFilename())
				mergeResultFile(filePath, mergedFilePath);
			else
			{
				if (!(std::filesystem::is_regular_file(mergedFilePath)))
					std::filesystem::rename(filePath, mergedFilePath);
			}
		}
	}	
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

std::string StructureCreationReporter::readFingerprintFile(const std::filesystem::path& filePath) const
{
	OperatingSystem::IO::FileReader fileReader{ filePath };

	if (fileReader.isValid())
	{
		std::string fingerprintSymbol = fileReader.readFile();
		fingerprintSymbol.pop_back();
		return fingerprintSymbol;
	}

	else
		return std::string{};
}

void StructureCreationReporter::mergeResultFile(const std::filesystem::path& resultFilePath, const std::filesystem::path& mergedResultFilePath) const
{
	size_type mergedStructureDesigning = 0;
	{
		OperatingSystem::IO::FileReader fileReader{ mergedResultFilePath };

		if (fileReader.isValid())
		{
			std::string fileText = fileReader.readFile();
			OperatingSystem::IO::FileReader::readData(fileText, "Number.of.Generated.Structures", mergedStructureDesigning);
		}
	}

	size_type thisStructureDesigning = 0;
	{
		OperatingSystem::IO::FileReader fileReader{ resultFilePath };

		if (fileReader.isValid())
		{
			std::string fileText = fileReader.readFile();
			OperatingSystem::IO::FileReader::readData(fileText, "Number.of.Generated.Structures", thisStructureDesigning);

			fileReader.removeFile();
		}

		else
		{
			std::string errorMessage;
			{
				errorMessage += "Could not open the file of \"";
				errorMessage += resultFilePath.generic_string();
				errorMessage += "\".";
			}

			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "mergeResultFile", errorMessage));
		}
	}

	mergedStructureDesigning += thisStructureDesigning;


	OperatingSystem::IO::FileWriter fileWriter{ mergedResultFilePath };
	{
		std::string outputString;
		outputString += "Number.of.Generated.Structures     ";
		outputString += std::to_string(mergedStructureDesigning);

		fileWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing, outputString);
	}
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
