#include "StructureCreator.h"

#include <vector>
#include <thread>
#include <mpi.h>

#include "ThreadController.h"
#include "MpiController.h"

#include "InternationalTables.h"

#include "IonicBond.h"
#include "MetallicBond.h"
#include "AnionicConstraint.h"
#include "CationicConstraint.h"
#include "NonIonicBond.h"
#include "NonMetallicBond.h"

#include "CreateStructures.h"
#include "StructureCreationReporter.h"

#include "StructureCustomizerBase.h"

using namespace MathematicalCrystalChemistry::Runtime::Creation;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructors

StructureCreator::StructureCreator(const StructureCreationPlan& structureCreationPlan)
	: m_structureCreationPlan{ structureCreationPlan }
{
}

StructureCreator::StructureCreator(StructureCreationPlan&& structureCreationPlan)
	: m_structureCreationPlan{ std::move(structureCreationPlan) }
{
}

// Constructors
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void StructureCreator::execute() const
{
	initializeDirectories();
	initializeChemicalParameters();
	initializeGeometricalConstraintParameters();

	MathematicalCrystalChemistry::Design::Generation::InitialStructureGenerator initialStructureGenerator;
	MathematicalCrystalChemistry::Design::Optimization::StructuralOptimizer structuralOptimizer{ m_structureCreationPlan.structuralOptimizationParameters() };
	initialStructureGenerator.setInitialStructureGenerationOptions(m_structureCreationPlan.initialStructureGenerationOptions());


	for (const auto& modelAtomCollection : m_structureCreationPlan.targetModelAtomCompositions().getModelAtomCollections())
	{
		if (isValidModelAtomCollection(modelAtomCollection))
		{
			for (const auto& typeAndNum : m_structureCreationPlan.initialStructureGenerationOptions().structureGenerating())
			{
				Internal::CreateStructures::initializeStructureDesigning();
				Internal::CreateStructures::setMaxStructureDesigning(getMaxStructureGenerating(typeAndNum.second));

				initialStructureGenerator.setGenerationType(typeAndNum.first);


				std::vector<Internal::CreateStructures> createStructuresCollection;
				{
					for (std::size_t threadRank = 0; threadRank < OperatingSystem::Configuration::ThreadController::maxThreading(); ++threadRank)
					{
						Internal::CreateStructures createStructures;
						createStructures.setInitialStructureGenerator(initialStructureGenerator);
						createStructures.setStructuralOptimizer(structuralOptimizer);
						createStructures.setStructureCreationReportOptions(m_structureCreationPlan.structureCreationReportOptions());

						createStructuresCollection.push_back(std::move(createStructures));
					}
				}

				std::vector<std::thread> operatingSystems;
				{
					for (std::size_t threadRank = 0; threadRank < OperatingSystem::Configuration::ThreadController::maxThreading(); ++threadRank)
						operatingSystems.push_back(std::thread{ createStructuresCollection[threadRank], modelAtomCollection });
				}


				for (auto& operatingSystem : operatingSystems)
					operatingSystem.join();
			}
		}
	}


	MPI_Barrier(MPI_COMM_WORLD);

	if (OperatingSystem::Configuration::MpiController::mpiRank() == 0)
	{
		Internal::StructureCreationReporter structureCreationReporter{ m_structureCreationPlan.structureCreationReportOptions().designedStructuresDirectoryPath() };
		structureCreationReporter.reportOptimalStructures();
		structureCreationReporter.reportInfeasibleStructures();
		structureCreationReporter.reportExceptionalStructures();
		structureCreationReporter.eraseMpiDirectories();
	}
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void StructureCreator::initializeDirectories() const
{
	std::filesystem::path workDirectoryPath = m_structureCreationPlan.structureCreationReportOptions().workDirectoryPath();
	workDirectoryPath /= OperatingSystem::Configuration::MpiController::getMpiDirectoryName();
	{
		if (std::filesystem::is_directory(workDirectoryPath))
			std::filesystem::remove_all(workDirectoryPath);

		if (!(std::filesystem::create_directories(workDirectoryPath)))
		{
			std::string errorMessage;
			{
				errorMessage += "Could not create directories of the path \"";
				errorMessage += workDirectoryPath.generic_string();
				errorMessage += "\"";
			}

			OperatingSystem::Configuration::ReportException reportException;
			throw std::invalid_argument(reportException(typeid(*this), "initializeDirectories", errorMessage));
		}
	}

	std::filesystem::path designedStructuresDirectoryPath = m_structureCreationPlan.structureCreationReportOptions().designedStructuresDirectoryPath();
	{
		if (!(std::filesystem::is_directory(designedStructuresDirectoryPath)))
		{
			if (!(std::filesystem::create_directories(designedStructuresDirectoryPath)))
			{
				std::string errorMessage;
				{
					errorMessage += "Could not create directories of the path \"";
					errorMessage += designedStructuresDirectoryPath.generic_string();
					errorMessage += "\"";
				}

				OperatingSystem::Configuration::ReportException reportException;
				throw std::invalid_argument(reportException(typeid(*this), "initializeDirectories", errorMessage));
			}
		}
	}
}

void StructureCreator::initializeChemicalParameters() const
{
	ChemToolkit::Crystallography::SpaceGroups::InternationalTables::setSpaceGroupPrecision(m_structureCreationPlan.chemicalParameters().spaceGroupPrecision());
}

void StructureCreator::initializeGeometricalConstraintParameters() const
{
	using MathematicalCrystalChemistry::StructureModel::Components::ModelCrystalStructure;

	using MathematicalCrystalChemistry::Design::Optimization::Internal::StructureCustomizerBase;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::IonicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::MetallicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::AnionicConstraint;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::CationicConstraint;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonIonicBond;
	using MathematicalCrystalChemistry::StructureModel::GeometricalConstraints::NonMetallicBond;


	ModelCrystalStructure::setCommonAtomPairTrackingCutoffRatio(m_structureCreationPlan.structuralOptimizationParameters().atomPairTrackingCutoffRatio());
	StructureCustomizerBase::setCommonPressure(m_structureCreationPlan.structuralOptimizationParameters().pressure());
	StructureCustomizerBase::setCommonAttractiveForceConstant(m_structureCreationPlan.structuralOptimizationParameters().attractiveForceConstant());
	StructureCustomizerBase::setCommonRepulsiveForceConstant(m_structureCreationPlan.structuralOptimizationParameters().repulsiveForceConstant());

	IonicBond::setDefaultCutoffRatio(m_structureCreationPlan.geometricalConstraintParameters().ionicBondCutoffLengthRatio());
	MetallicBond::setDefaultCutoffRatio(m_structureCreationPlan.geometricalConstraintParameters().metallicBondCutoffLengthRatio());
	AnionicConstraint::setDefaultCutoffRatio(m_structureCreationPlan.geometricalConstraintParameters().anionicConstraintCutoffLengthRatio());
	NonIonicBond::setDefaultCutoffRatio(m_structureCreationPlan.geometricalConstraintParameters().nonIonicBondCutoffLengthRatio());
	NonMetallicBond::setDefaultCutoffRatio(m_structureCreationPlan.geometricalConstraintParameters().nonMetallicBondCutoffLengthRatio());

	NonIonicBond::setDefaultNonBondingRangeRatio(m_structureCreationPlan.geometricalConstraintParameters().nonIonicBondRangeRatio());
	NonMetallicBond::setDefaultNonBondingRangeRatio(m_structureCreationPlan.geometricalConstraintParameters().nonMetallicBondRangeRatio());
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

StructureCreator::size_type StructureCreator::getMaxStructureGenerating(const size_type totalStructureGenerating) const noexcept
{
	size_type structureGenerating = totalStructureGenerating / OperatingSystem::Configuration::MpiController::mpiProcessing();
	{
		size_type remainder = totalStructureGenerating % OperatingSystem::Configuration::MpiController::mpiProcessing();;

		if (OperatingSystem::Configuration::MpiController::mpiRank() < remainder)
			structureGenerating += 1;
	}

	return structureGenerating;
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
