#include "StructureDesignRecorder.h"

#include "FileReader.h"
#include "FileWriter.h"
#include "CifFileWriter.h"

#include "StructureClassifier.h"

using namespace MathematicalCrystalChemistry::Design::Recording;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

std::mutex StructureDesignRecorder::m_directoryMutex;
std::mutex StructureDesignRecorder::m_reportMutex;
std::string StructureDesignRecorder::m_basicRecordFilename{ "basic_record.txt" };
std::string StructureDesignRecorder::m_resultFilename{ "result.txt" };
std::string StructureDesignRecorder::m_optimalStructuresDirectoryName{ "Optima" };
std::string StructureDesignRecorder::m_infeasibleStructuresDirectoryName{ "Infeasibles" };
std::string StructureDesignRecorder::m_exceptionalStructuresDirectoryName{ "Exceptions" };


StructureDesignRecorder::StructureDesignRecorder()
	: m_initializedTimePoint{ std::chrono::system_clock::now() }
	, m_structureOptimizing{ 0 }
	, m_geometricalConstraintsOptimizing{ 0 }
	, m_chemicalBondRecorder{}
	, m_dynamicsRecorder{}
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

void StructureDesignRecorder::reportOptimalStructure(ModelCrystalStructure modelCrystalStructure, const std::filesystem::path& designedStructuresDirectoryPath)
{
	std::filesystem::path outputDirectoryPath = designedStructuresDirectoryPath;
	createDirectoryIfNecessary(outputDirectoryPath);
	outputDirectoryPath /= m_optimalStructuresDirectoryName;
	createDirectoryIfNecessary(outputDirectoryPath);

	MathematicalCrystalChemistry::IO::StructureClassifier structureClassifier{ outputDirectoryPath };
	outputDirectoryPath = structureClassifier.classify(modelCrystalStructure);


	createResultFile(modelCrystalStructure, outputDirectoryPath);
}

void StructureDesignRecorder::reportOptimalStructure(ModelCrystalStructure modelCrystalStructure, const std::filesystem::path& designedStructuresDirectoryPath, const std::filesystem::path workDirectoryPath)
{
	std::filesystem::path outputDirectoryPath = designedStructuresDirectoryPath;
	createDirectoryIfNecessary(outputDirectoryPath);
	outputDirectoryPath /= m_optimalStructuresDirectoryName;
	createDirectoryIfNecessary(outputDirectoryPath);

	MathematicalCrystalChemistry::IO::StructureClassifier structureClassifier{ outputDirectoryPath };
	outputDirectoryPath = structureClassifier.classify(modelCrystalStructure, getDirectoryName(workDirectoryPath));


	if (m_chemicalBondRecorder.needRecord())
		m_chemicalBondRecorder.moveFiles(outputDirectoryPath);

	if (m_dynamicsRecorder.needRecord())
		m_dynamicsRecorder.moveFile(outputDirectoryPath);


	createBasicRecordFile(outputDirectoryPath);
	createResultFile(modelCrystalStructure, outputDirectoryPath.parent_path());
}

void StructureDesignRecorder::reportInfeasibleStructure(ModelCrystalStructure modelCrystalStructure, const std::filesystem::path& designedStructuresDirectoryPath, const std::filesystem::path workDirectoryPath)
{
	std::filesystem::path outputDirectoryPath = designedStructuresDirectoryPath;
	createDirectoryIfNecessary(outputDirectoryPath);
	outputDirectoryPath /= m_infeasibleStructuresDirectoryName;
	createDirectoryIfNecessary(outputDirectoryPath);
	{
		MathematicalCrystalChemistry::StructureModel::Features::AtomCategoryComposition atomCategoryComposition = modelCrystalStructure.getAtomCategoryComposition();
		outputDirectoryPath /= atomCategoryComposition.name();
		createDirectoryIfNecessary(outputDirectoryPath);
	}
	outputDirectoryPath /= getDirectoryName(workDirectoryPath);
	createDirectoryIfNecessary(outputDirectoryPath);


	if (m_chemicalBondRecorder.needRecord())
		m_chemicalBondRecorder.moveFiles(outputDirectoryPath);

	if (m_dynamicsRecorder.needRecord())
		m_dynamicsRecorder.moveFile(outputDirectoryPath);


	std::filesystem::path cifFilePath = outputDirectoryPath;
	{
		std::string cifFilename = modelCrystalStructure.getChemicalComposition().name();
		cifFilename += ".cif";

		cifFilePath /= cifFilename;
	}
	ChemToolkit::Crystallography::FileIO::CifFileWriter cifFileWriter{ cifFilePath };
	cifFileWriter.createFile(modelCrystalStructure, OperatingSystem::IO::FileCreateOptions::skip_existing);
}

void StructureDesignRecorder::reportExceptionalStructure(ModelCrystalStructure modelCrystalStructure, const std::filesystem::path& designedStructuresDirectoryPath, const std::filesystem::path workDirectoryPath)
{
	std::filesystem::path outputDirectoryPath = designedStructuresDirectoryPath;
	createDirectoryIfNecessary(outputDirectoryPath);
	outputDirectoryPath /= m_exceptionalStructuresDirectoryName;
	createDirectoryIfNecessary(outputDirectoryPath);
	{
		MathematicalCrystalChemistry::StructureModel::Features::AtomCategoryComposition atomCategoryComposition = modelCrystalStructure.getAtomCategoryComposition();
		outputDirectoryPath /= atomCategoryComposition.name();
		createDirectoryIfNecessary(outputDirectoryPath);
	}
	outputDirectoryPath /= getDirectoryName(workDirectoryPath);
	createDirectoryIfNecessary(outputDirectoryPath);

	
	if (m_chemicalBondRecorder.needRecord())
		m_chemicalBondRecorder.moveFiles(outputDirectoryPath);

	if (m_dynamicsRecorder.needRecord())
		m_dynamicsRecorder.moveFile(outputDirectoryPath);


	std::filesystem::path cifFilePath = outputDirectoryPath;
	{
		std::string cifFilename = modelCrystalStructure.getChemicalComposition().name();
		cifFilename += ".cif";

		cifFilePath /= cifFilename;
	}
	ChemToolkit::Crystallography::FileIO::CifFileWriter cifFileWriter{ cifFilePath };
	cifFileWriter.createFile(modelCrystalStructure, OperatingSystem::IO::FileCreateOptions::skip_existing);
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

void StructureDesignRecorder::createBasicRecordFile(const std::filesystem::path& outputDirectoryPath) const
{
	std::string outputString;
	{
		outputString += "Elapsed time:  ";
		{
			std::chrono::milliseconds elapsedMilliSeconds = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now().time_since_epoch() - m_initializedTimePoint.time_since_epoch());
			outputString += std::to_string(elapsedMilliSeconds.count());
		}
		outputString += " milliseconds\n";
		outputString += "Number.of.Structure.Optimization:  ";
		outputString += std::to_string(m_structureOptimizing);
		outputString += "\n";
		outputString += "Number.of.Geometrical.Constraints.Optimization:  ";
		outputString += std::to_string(m_geometricalConstraintsOptimizing);
		outputString += "\n";
	}


	std::filesystem::path reportFilePath = outputDirectoryPath;
	reportFilePath /= m_basicRecordFilename;

	OperatingSystem::IO::FileWriter fileWriter{ reportFilePath };
	fileWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing, outputString);
}

void StructureDesignRecorder::createResultFile(const ModelCrystalStructure& modelCrystalStructure, const std::filesystem::path& outputDirectoryPath) const
{
	std::lock_guard<std::mutex> guard{ m_reportMutex };


	std::filesystem::path reportFilePath = outputDirectoryPath;
	reportFilePath /= m_resultFilename;

	size_type thisStructureDesigning = 0;
	double unitCellVolume = 0.0;
	{
		OperatingSystem::IO::FileReader fileReader{ reportFilePath };

		if (fileReader.isValid())
		{
			std::string fileText = fileReader.readFile();
			OperatingSystem::IO::FileReader::readData(fileText, "Number.of.Generated.Structures", thisStructureDesigning);

			OperatingSystem::IO::FileReader::readData(fileText, "Volume.of.Unit.Cell", unitCellVolume);
			{
				double thisUnitCellVolume = modelCrystalStructure.unitCell().getCellVolume();
				{
					if (thisUnitCellVolume < unitCellVolume)
						unitCellVolume = thisUnitCellVolume;
				}
			}
		}

		++thisStructureDesigning;
	}

	OperatingSystem::IO::FileWriter fileWriter{ reportFilePath };
	{
		std::string outputString;
		outputString += "Number.of.Generated.Structures     ";
		outputString += std::to_string(thisStructureDesigning);
		outputString += "\n";
		outputString += "Volume.of.Unit.Cell     ";
		outputString += OperatingSystem::IO::FileWriter::toString(unitCellVolume);
		outputString += "\n";

		fileWriter.createFile(OperatingSystem::IO::FileCreateOptions::overwrite_existing, outputString);
	}
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
