#include "StructureOptimalityAppraiser.h"

#include "LogWriter.h"
#include "FileReader.h"

using namespace MathematicalCrystalChemistry::Runtime::Analysis::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

StructureOptimalityAppraiser::StructureOptimalityAppraiser()
	: m_maxReoptimizing{ 0 }
	, m_softStructureReoptimizer{}
	, m_hardStructureReoptimizer{}
{
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

StructureReoptimizer::ReoptimizationResult StructureOptimalityAppraiser::appraise(const ModelCrystalStructure& modelCrystalStructure, const StructureFingerprint& structureFingerprint, const std::filesystem::path& workDirectoryPath) const
{
	std::filesystem::path structureWorkDirectoryPath = workDirectoryPath.parent_path();

	std::filesystem::path softReoptimizationDirectoryPath = getSoftReoptimizationDirectoryPath(workDirectoryPath);
	StructureReoptimizer::ReoptimizationResult softReoptimizationResult = m_softStructureReoptimizer.reoptimize(modelCrystalStructure, structureFingerprint, softReoptimizationDirectoryPath);


	if(isDiscoveredReoptimizationResult(structureWorkDirectoryPath, softReoptimizationResult))
		return StructureReoptimizer::ReoptimizationResult::infeasible;

	else
	{
		if (softReoptimizationResult != StructureReoptimizer::ReoptimizationResult::infeasible)
		{
			std::filesystem::remove_all(softReoptimizationDirectoryPath);

			for (size_type rep = 0; rep < m_maxReoptimizing; ++rep)
			{
				StructureReoptimizer::ReoptimizationResult reoptimizationResult = m_hardStructureReoptimizer.reoptimize(modelCrystalStructure, structureFingerprint, getHardReoptimizationDirectoryPath(workDirectoryPath, rep));

				if (isDiscoveredReoptimizationResult(structureWorkDirectoryPath, reoptimizationResult))
					return StructureReoptimizer::ReoptimizationResult::infeasible;
			}

			return analyzeReoptimizationResults(workDirectoryPath);
		}

		else
			return softReoptimizationResult;
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

StructureReoptimizer::ReoptimizationResult StructureOptimalityAppraiser::analyzeReoptimizationResults(const std::filesystem::path& featuredStructureDirectoryPath) const
{
	StructureReoptimizer::ReoptimizationResult totalReoptimizationResult = StructureReoptimizer::ReoptimizationResult::isotypic;
	{
		for (const std::filesystem::directory_entry& directoryEntry : std::filesystem::directory_iterator(featuredStructureDirectoryPath))
		{
			if (directoryEntry.is_directory())
			{
				std::filesystem::path reoptimiationResultFilePath = directoryEntry.path();
				reoptimiationResultFilePath /= "reoptimization_result.txt";
				OperatingSystem::IO::FileReader fileReader{ reoptimiationResultFilePath };

				if (fileReader.isValid())
				{
					std::string reoptimizationResultText;
					{
						std::string fileText = fileReader.readFile();
						OperatingSystem::IO::FileReader::readData(fileText, "Reoptimization.Result", reoptimizationResultText);
					}

					downgradeTotalReoptimizationResult(totalReoptimizationResult, toReoptimizationResult(reoptimizationResultText));
				}
			}
		}
	}


	return totalReoptimizationResult;
}

bool StructureOptimalityAppraiser::isDiscoveredReoptimizationResult(const std::filesystem::path& structureWorkDirectoryPath, const StructureReoptimizer::ReoptimizationResult& reoptimizationResult) const
{
	std::filesystem::path reoptimizationResultDirectoryPath = structureWorkDirectoryPath;
	{
		switch (reoptimizationResult)
		{
		case StructureReoptimizer::ReoptimizationResult::isotypic:
			reoptimizationResultDirectoryPath /= "Isotypic";
			break;

		case StructureReoptimizer::ReoptimizationResult::homeotypic:
			reoptimizationResultDirectoryPath /= "Homeotypic";
			break;

		case StructureReoptimizer::ReoptimizationResult::same_symmetry:
			reoptimizationResultDirectoryPath /= "Same_symmetry";
			break;

		case StructureReoptimizer::ReoptimizationResult::same_space_group_type:
			reoptimizationResultDirectoryPath /= "Same_space_group_type";
			break;

		case StructureReoptimizer::ReoptimizationResult::optimal:
			reoptimizationResultDirectoryPath /= "Optimal";
			break;

		case StructureReoptimizer::ReoptimizationResult::infeasible:
			reoptimizationResultDirectoryPath /= "Infeasibles";
			break;

		default:
			OperatingSystem::Configuration::ReportException reportException;
			throw std::runtime_error(reportException(typeid(*this), "isDiscoveredReoptimizationResult", "\"reoptimizationResult\" is out of definitions."));
		}
	}


	if (std::filesystem::is_directory(reoptimizationResultDirectoryPath))
		return true;
	else
		return false;
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
