#include "StructureReoptimizer.h"

#include "MathematicalConstants.h"
#include "FileWriter.h"
#include "CifFileWriter.h"

using namespace MathematicalCrystalChemistry::Runtime::Analysis::Internal;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Constructor

StructureReoptimizer::StructureReoptimizer()
	: m_randomValueGenerating{ 0 }
	, m_engine{}
	, m_maxAtomFractuationRatio{ 0.0 }
	, m_structureDesignRecordOptions{}
	, m_structuralOptimizer{}
{
	std::random_device seed_gen;
	m_engine = std::mt19937(seed_gen());
}

// Constructor
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Public methods

StructureReoptimizer::ReoptimizationResult StructureReoptimizer::reoptimize(const ModelCrystalStructure& inputCrystalStructure, const StructureFingerprint& inputStructureFingerprint, const std::filesystem::path& workDirectoryPath) const
{
	ModelCrystalStructure modelCrystalStructure = inputCrystalStructure;
	applyFractuation(modelCrystalStructure);
	{
		if (m_structureDesignRecordOptions.needChemicalBondsRecord() || m_structureDesignRecordOptions.needMdRecord())
		{
			MathematicalCrystalChemistry::Design::Recording::StructureDesignRecorder structureDesignRecorder;
			{
				if (m_structureDesignRecordOptions.needChemicalBondsRecord())
				{
					structureDesignRecorder.chemicalBondRecorder().setRecordFrequency(m_structureDesignRecordOptions.chemicalBondsRecordFrequency());
					structureDesignRecorder.chemicalBondRecorder().createDataFiles(workDirectoryPath);
				}

				if (m_structureDesignRecordOptions.needMdRecord())
				{
					structureDesignRecorder.dynamicsRecorder().setRecordFrequency(m_structureDesignRecordOptions.mdRecordFrequency());
					structureDesignRecorder.dynamicsRecorder().createFile(workDirectoryPath, inputCrystalStructure.getChemicalComposition().name());
				}
			}

			m_structuralOptimizer.optimize(modelCrystalStructure, structureDesignRecorder);
		}

		else
			m_structuralOptimizer.optimize(modelCrystalStructure);
	}
	

	if (modelCrystalStructure.isOptimalSolution(m_structuralOptimizer.structuralOptimizationParameters().preciseRelaxationParameters().feasibleConstraintErrorRate()))
	{
		modelCrystalStructure.updateSpaceGroupInfo();

		createPrimitiveCifFile(modelCrystalStructure, workDirectoryPath);
		createConventionalCifFile(modelCrystalStructure, workDirectoryPath);
		{
			std::filesystem::path structureInfoFilePath = workDirectoryPath;
			structureInfoFilePath /= "structure-info.txt";
			modelCrystalStructure.createStructureInformationFile(structureInfoFilePath);
		}

		auto structureFingerprint = modelCrystalStructure.getStructureFingerprint();
		{
			std::filesystem::path fingerprintFilePath = workDirectoryPath;
			fingerprintFilePath /= "fingerprint.txt";
			structureFingerprint.report(fingerprintFilePath);
		}


		if (inputStructureFingerprint == structureFingerprint)
		{
			createReoptimizationResultFile(modelCrystalStructure, ReoptimizationResult::isotypic, workDirectoryPath);
			return ReoptimizationResult::isotypic;
		}

		else
		{
			if (MathematicalCrystalChemistry::StructureModel::Components::isHomeotypicStructureFingerprint(inputStructureFingerprint, structureFingerprint))
			{
				createReoptimizationResultFile(modelCrystalStructure, ReoptimizationResult::homeotypic, workDirectoryPath);
				return ReoptimizationResult::homeotypic;
			}
			
			else
			{
				if (hasSameSymmetry(inputCrystalStructure, modelCrystalStructure))
				{
					createReoptimizationResultFile(modelCrystalStructure, ReoptimizationResult::same_symmetry, workDirectoryPath);
					return ReoptimizationResult::same_symmetry;
				}

				else
				{
					if (hasSameSpaceGroupNumber(inputCrystalStructure, modelCrystalStructure))
					{
						createReoptimizationResultFile(modelCrystalStructure, ReoptimizationResult::same_space_group_type, workDirectoryPath);
						return ReoptimizationResult::same_space_group_type;
					}

					else
					{
						createReoptimizationResultFile(modelCrystalStructure, ReoptimizationResult::optimal, workDirectoryPath);
						return ReoptimizationResult::optimal;
					}
				}
			}
		}
	}

	else
	{
		createPrimitiveCifFile(modelCrystalStructure, workDirectoryPath);

		createReoptimizationResultFile(modelCrystalStructure, ReoptimizationResult::infeasible, workDirectoryPath);
		return ReoptimizationResult::infeasible;
	}
}

// Public methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private utility

void StructureReoptimizer::applyFractuation(ModelCrystalStructure& modelCrystalStructure) const
{
	using NumericalVector = MathToolkit::Generic::NumericalVector<double>;

	std::uniform_real_distribution<> thetaDist(0.0, MathToolkit::Numerics::MathematicalConstants::getPi());
	std::uniform_real_distribution<> phiDist(0.0, (2.0 * MathToolkit::Numerics::MathematicalConstants::getPi()));


	for (auto& modelAtom : modelCrystalStructure.atomicArrangement())
	{
		NumericalVector displacement(3);
		{
			std::uniform_real_distribution<> displacementSizeDist(0.0, m_maxAtomFractuationRatio * modelAtom.coordinationFeature().upperIonicBondRadius());

			double displacementSize = displacementSizeDist(m_engine);
			++m_randomValueGenerating;
			double theta = thetaDist(m_engine);
			++m_randomValueGenerating;
			double phi = phiDist(m_engine);
			++m_randomValueGenerating;

			displacement[0] = displacementSize * std::sin(theta) * std::cos(phi);
			displacement[1] = displacementSize * std::sin(theta) * std::sin(phi);
			displacement[2] = displacementSize * std::cos(theta);
		}

		modelAtom.coordinate() += displacement;
	}


	if (500 < m_randomValueGenerating)
	{
		std::random_device seed_gen;
		m_engine = std::mt19937(seed_gen());
	}
}

void StructureReoptimizer::createReoptimizationResultFile(const ModelCrystalStructure& modelCrystalStructure, const ReoptimizationResult& reoptimizationResult, const std::filesystem::path& directoryPath) const
{
	std::string outputText;
	{
		outputText += "Reoptimization.Result     ";
		{
			switch (reoptimizationResult)
			{
			case ReoptimizationResult::isotypic:
				outputText += "isotypic";
				break;

			case ReoptimizationResult::homeotypic:
				outputText += "homeotypic";
				break;

			case ReoptimizationResult::same_symmetry:
				outputText += "same_symmetry";
				break;

			case ReoptimizationResult::same_space_group_type:
				outputText += "same_space_group_type";
				break;

			case ReoptimizationResult::optimal:
				outputText += "optimal";
				break;

			case ReoptimizationResult::infeasible:
				outputText += "infeasible";
				break;

			default:
				OperatingSystem::Configuration::ReportException reportException;
				throw std::runtime_error(reportException(typeid(*this), "createReoptimizationResultFile", "\"reoptimizationResult\" is out of definitions."));
			}
		}
		outputText += "\n";
	}


	std::filesystem::path filePath = directoryPath;
	filePath /= "reoptimization_result.txt";
	OperatingSystem::IO::FileWriter fileWriter{ filePath };
	fileWriter.createFile(OperatingSystem::IO::FileCreateOptions::skip_existing, outputText);
}

void StructureReoptimizer::createPrimitiveCifFile(const ModelCrystalStructure& modelCrystalStructure, const std::filesystem::path& directoryPath) const
{
	std::filesystem::path cifFilePath = directoryPath;
	cifFilePath /= modelCrystalStructure.getChemicalComposition().name();
	cifFilePath += ".cif";

	ChemToolkit::Crystallography::FileIO::CifFileWriter cifFileWriter{ cifFilePath };
	cifFileWriter.createFile(modelCrystalStructure, OperatingSystem::IO::FileCreateOptions::overwrite_existing);
}

void StructureReoptimizer::createConventionalCifFile(const ModelCrystalStructure& modelCrystalStructure, const std::filesystem::path& directoryPath) const
{
	ModelCrystalStructure conventionalStructure = modelCrystalStructure;
	{
		conventionalStructure.conventionalizeStructure();
		conventionalStructure.createTrackingAtomPairs();
		conventionalStructure.createGeometricalConstraints();
	}
	std::filesystem::path conventionalCifFilePath = directoryPath;
	conventionalCifFilePath /= "conventional-";
	conventionalCifFilePath += modelCrystalStructure.getChemicalComposition().name();
	conventionalCifFilePath += ".cif";

	ChemToolkit::Crystallography::FileIO::CifFileWriter conventionalCifFileWriter{ conventionalCifFilePath };
	conventionalCifFileWriter.createFile(conventionalStructure, OperatingSystem::IO::FileCreateOptions::overwrite_existing);
}

// Private utility
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
