#include "SymmetryOperationTable.h"

#include <regex>

#include "MathematicalConstants.h"

using namespace ChemToolkit::Crystallography::SpaceGroups;


// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// Private methods

std::vector<MathToolkit::Generic::AffineMapping<double>> SymmetryOperationTable::translateIntoSymmetryOperations(const std::string& symbol)
{
	{
		std::regex rotationPat("C([\\d]+)");
		std::smatch sm;

		if (std::regex_match(symbol, sm, rotationPat))
		{
			size_type multiplicity = static_cast<size_type>(std::stoi(sm.str(1)));
			return getRotationMappings(multiplicity);
		}
	}
	{
		if (symbol == "m")
			return getReflectionMappings();
	}
	{
		if (symbol == "i")
			return getInversionMappings();
	}
	{
		std::regex rotoinversionPat("S([\\d]+)");
		std::smatch sm;

		if (std::regex_match(symbol, sm, rotoinversionPat))
		{
			size_type multiplicity = static_cast<size_type>(std::stoi(sm.str(1)));
			return getRotoinversionMappings(multiplicity);
		}
	}

	return std::vector<AffineMapping>{};
}

std::vector<MathToolkit::Generic::AffineMapping<double>> SymmetryOperationTable::getRotationMappings(const size_type multiplicity)
{
	double rotationAngle = 2.0 * MathToolkit::Numerics::MathematicalConstants::getPi();
	rotationAngle /= static_cast<double>(multiplicity);

	std::vector<AffineMapping> affineMappings;
	{
		NumericalMatrix xOperationMatrix(3, 3);
		{
			xOperationMatrix(0, 0) = 1.0;
			xOperationMatrix(1, 1) = std::cos(rotationAngle);
			xOperationMatrix(2, 1) = std::sin(rotationAngle);
			xOperationMatrix(1, 2) = (-1.0) * std::sin(rotationAngle);
			xOperationMatrix(2, 2) = std::cos(rotationAngle);
		}
		NumericalMatrix yOperationMatrix(3, 3);
		{
			yOperationMatrix(0, 0) = std::cos(rotationAngle);
			yOperationMatrix(2, 0) = (-1.0) * std::sin(rotationAngle);
			yOperationMatrix(1, 1) = 1.0;
			yOperationMatrix(0, 2) = std::sin(rotationAngle);
			yOperationMatrix(2, 2) = std::cos(rotationAngle);
		}
		NumericalMatrix zOperationMatrix(3, 3);
		{
			zOperationMatrix(0, 0) = std::cos(rotationAngle);
			zOperationMatrix(1, 0) = std::sin(rotationAngle);
			zOperationMatrix(0, 1) = (-1.0) * std::sin(rotationAngle);
			zOperationMatrix(1, 1) = std::cos(rotationAngle);
			zOperationMatrix(2, 2) = 1.0;
		}

		affineMappings.push_back(AffineMapping{ std::move(xOperationMatrix) });
		affineMappings.push_back(AffineMapping{ std::move(yOperationMatrix) });
		affineMappings.push_back(AffineMapping{ std::move(zOperationMatrix) });
	}

	return affineMappings;
}

std::vector<MathToolkit::Generic::AffineMapping<double>> SymmetryOperationTable::getReflectionMappings()
{
	std::vector<AffineMapping> affineMappings;
	{
		NumericalMatrix xOperationMatrix(3, 3);
		{
			xOperationMatrix(0, 0) = -1.0;
			xOperationMatrix(1, 1) = 1.0;
			xOperationMatrix(2, 2) = 1.0;
		}
		NumericalMatrix yOperationMatrix(3, 3);
		{
			yOperationMatrix(0, 0) = 1.0;
			yOperationMatrix(1, 1) = -1.0;
			yOperationMatrix(2, 2) = 1.0;
		}
		NumericalMatrix zOperationMatrix(3, 3);
		{
			zOperationMatrix(0, 0) = 1.0;
			zOperationMatrix(1, 1) = 1.0;
			zOperationMatrix(2, 2) = -1.0;
		}

		affineMappings.push_back(AffineMapping{ std::move(xOperationMatrix) });
		affineMappings.push_back(AffineMapping{ std::move(yOperationMatrix) });
		affineMappings.push_back(AffineMapping{ std::move(zOperationMatrix) });
	}

	return affineMappings;
}

std::vector<MathToolkit::Generic::AffineMapping<double>> SymmetryOperationTable::getInversionMappings()
{
	std::vector<AffineMapping> affineMappings;
	{
		NumericalMatrix operationMatrix(3, 3);
		{
			operationMatrix(0, 0) = -1.0;
			operationMatrix(1, 1) = -1.0;
			operationMatrix(2, 2) = -1.0;
		}

		affineMappings.push_back(AffineMapping{ std::move(operationMatrix) });
	}

	return affineMappings;
}

std::vector<MathToolkit::Generic::AffineMapping<double>> SymmetryOperationTable::getRotoinversionMappings(const size_type multiplicity)
{
	double rotationAngle = 2.0 * MathToolkit::Numerics::MathematicalConstants::getPi();
	rotationAngle /= static_cast<double>(multiplicity);

	std::vector<AffineMapping> affineMappings;
	{
		NumericalMatrix xOperationMatrix(3, 3);
		{
			xOperationMatrix(0, 0) = -1.0;
			xOperationMatrix(1, 1) = std::cos(rotationAngle);
			xOperationMatrix(2, 1) = std::sin(rotationAngle);
			xOperationMatrix(1, 2) = (-1.0) * std::sin(rotationAngle);
			xOperationMatrix(2, 2) = std::cos(rotationAngle);
		}
		NumericalMatrix yOperationMatrix(3, 3);
		{
			yOperationMatrix(0, 0) = std::cos(rotationAngle);
			yOperationMatrix(2, 0) = (-1.0) * std::sin(rotationAngle);
			yOperationMatrix(1, 1) = -1.0;
			yOperationMatrix(0, 2) = std::sin(rotationAngle);
			yOperationMatrix(2, 2) = std::cos(rotationAngle);
		}
		NumericalMatrix zOperationMatrix(3, 3);
		{
			zOperationMatrix(0, 0) = std::cos(rotationAngle);
			zOperationMatrix(1, 0) = std::sin(rotationAngle);
			zOperationMatrix(0, 1) = (-1.0) * std::sin(rotationAngle);
			zOperationMatrix(1, 1) = std::cos(rotationAngle);
			zOperationMatrix(2, 2) = -1.0;
		}

		affineMappings.push_back(AffineMapping{ std::move(xOperationMatrix) });
		affineMappings.push_back(AffineMapping{ std::move(yOperationMatrix) });
		affineMappings.push_back(AffineMapping{ std::move(zOperationMatrix) });
	}

	return affineMappings;
}

// Private methods
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
// **********************************************************************************************************************************************************************************************************************************************************************************************
